unit PropEditor;

interface

uses
  Windows, SysUtils, Messages, Classes, Graphics, Controls, Forms, Dialogs,
  GlowControls, ExtCtrls, ExtDlgs, StdCtrls, PropList, JPEG, GIFImage,
  PngImage, ComCtrls, Menus;

type
  TfrmPicture = class(TForm)
  private
    fPicture  : TPicture;
    opdImage  : TOpenPictureDialog;
    svdImage  : TSavePictureDialog;
    btnOK     : TShButton;
    btnCancel : TShButton;
    btnLoad   : TShButton;
    btnSave   : TShButton;
    btnClear  : TShButton;
    pnlImage  : TShPanel;
    pntImage  : TPaintBox;
    procedure btnLoadClick(Sender: TObject);
    procedure btnSaveClick(Sender: TObject);
    procedure btnClearClick(Sender: TObject);
    procedure pntImagePaint(Sender: TObject);
  public
    property Picture: TPicture read FPicture;
  end;

  TfrmStrings = class(TForm)
  private
    lblCount  : TLabel;
    redStrings: TRichEdit;
    btnOK     : TButton;
    btnCancel : TButton;
    btnClear  : TButton;
    function  GetStrings: TStrings;
  public
    { Public declarations }
    property Strings: TStrings read GetStrings;
  end;

  TMenuPropertyEditor = class;
  TfrmMenuEditor = class(TForm)
    private
      FEditor: TMenuPropertyEditor;
      FItemsOwner: TComponent;
      trvItems: TTreeView;
      pnlButtons: TPanel;
      btnNewItem: TButton;
      btnNewSubitem: TButton;
      btnDeleteItem: TButton;
      procedure MenuToNode;
      procedure NodeToMenu;
      procedure WMGetMinMaxInfo(var Msg: TMessage); message WM_GETMINMAXINFO;
      procedure btnAddItemClick(Sender: TObject);
      procedure btnDeleteItemClick(Sender: TObject);
      procedure trvItemsChange(Sender: TObject; Node: TTreeNode);
      procedure FormDestroy(Sender: TObject);
    public

  end;

  TPropertyEditor = class
  private
    fProp: TProperty;
  public
    constructor Create(AProp: TProperty); virtual;
    function Execute: Boolean; virtual; abstract;
    property Prop: TProperty read FProp;
  end;

  TPropertyEditorClass = class of TPropertyEditor;

  TStringPropertyEditor = class(TPropertyEditor)
  public
    function Execute: Boolean; override;
  end;

  TFontPropertyEditor = class(TPropertyEditor)
  public
    function Execute: Boolean; override;
  end;

  TColorPropertyEditor = class(TPropertyEditor)
  public
    function Execute: Boolean; override;
  end;

  TPicturePropertyEditor = class(TPropertyEditor)
  private
    fForm: TfrmPicture;
  public
    function Execute: Boolean; override;
  end;

  TMenuPropertyEditor = class(TPropertyEditor)
  private
    fForm    : TfrmMenuEditor;
    function GetMenuItems: TMenuItem;
  public
    function Execute: Boolean; override;
    property MenuItems: TMenuItem read GetMenuItems;
  end;

  TStringsPropertyEditor = class(TPropertyEditor)
  private
    fForm    : TfrmStrings;
  public
    function Execute: Boolean; override;
  end;

implementation

uses
  InspectFrm;

function TfrmStrings.GetStrings: TStrings;
begin
  Result:=redStrings.Lines;
end;

constructor TPropertyEditor.Create(AProp: TProperty);
begin
  inherited Create;
  FProp:=AProp;
end;

function TStringPropertyEditor.Execute: Boolean;
var
  OldValue: string;
begin
  with Prop do
  begin
    OldValue:=AsString;
    AsString:=InputBox((Instance as TComponent).Name,Name,OldValue);
    Result:=AsString<>OldValue;
  end;
end;
 
function TFontPropertyEditor.Execute: Boolean;
begin
  with TFontDialog.Create(nil) do
  try
    Font.Assign(TFont(Prop.AsObject));
    Result:=Execute;
    if Result then TFont(Prop.AsObject).Assign(Font);
  finally
    Free;
  end;
end;

function TColorPropertyEditor.Execute: Boolean;
begin
  with TColorDialog.Create(nil) do
  try
    Color:=Prop.AsInteger;
    Result:=Execute;
    if Result then Prop.AsInteger:=Color;
  finally
    Free;
  end;
end;

function TPicturePropertyEditor.Execute: Boolean;
begin
  fForm:= TfrmPicture.CreateNew(Nil);
  with fForm do
    try
      BorderStyle       := bsDialog;
      Caption           := ' ';
      Width             := 355;
      Height            := 300;
      Position          := poMainFormCenter;

      pnlImage          := TShPanel.Create(fForm);
      pnlImage.Parent   := fForm;
      pnlImage.Width    := 265;
      pnlImage.Height   := 265;
      pnlImage.Top      := 3;
      pnlImage.Left     := 3;
      pnlImage.Gradient := gsLeftVertical;

      pntImage          := TPaintBox.Create(pnlImage);
      pntImage.Parent   := pnlImage;
      pntImage.Align    := alClient;
      pntImage.OnPaint  := pntImagePaint;
      pntImage.AlignWithMargins := true;
      pntImage.Margins.Left     := 5;
      pntImage.Margins.Top      := 5;
      pntImage.Margins.Right    := 5;
      pntImage.Margins.Bottom   := 5;
      
      fPicture          := TPicture.Create;
      opdImage          := TOpenPictureDialog.Create(fForm);
      svdImage          := TSavePictureDialog.Create(fForm);
      btnOK             := TShButton.Create(fForm);
      btnCancel         := TShButton.Create(fForm);
      btnLoad           := TShButton.Create(fForm);
      btnSave           := TShButton.Create(fForm);
      btnClear          := TShButton.Create(fForm);

      btnOK.Parent      := fForm;
      btnCancel.Parent  := fForm;
      btnLoad.Parent    := fForm;
      btnSave.Parent    := fForm;
      btnClear.Parent   := fForm;

      btnLoad.OnClick   := btnLoadClick;
      btnSave.OnClick   := btnSaveClick;
      btnClear.OnClick  := btnClearClick;

      btnLoad.Left      := 270;
      btnLoad.Caption   := '';
      btnLoad.Top       := 3;
      btnLoad.Height    := 22;

      btnSave.Left      := 270;
      btnSave.Caption   := '';
      btnSave.Top       := 27;
      btnSave.Height    := 22;

      btnClear.Left     := 270;
      btnClear.Caption  := '';
      btnClear.Top      := 51;
      btnClear.Height   := 22;

      btnOK.Left        := 270;
      btnOK.Caption     := 'OK';
      btnOK.Top         := 219;
      btnOK.Height      := 22;

      btnCancel.Left    := 270;
      btnCancel.Caption := '';
      btnCancel.Top     := 244;
      btnCancel.Height  := 22;

      btnOK.ModalResult     := mrOk;
      btnCancel.ModalResult := mrCancel;

      with Prop do
        if PropType=TypeInfo(TPicture) then
          fPicture.Assign(TPicture(Prop.AsObject))
        else if PropType=TypeInfo(TBitmap) then begin
          fPicture.Bitmap:=TBitmap(Prop.AsObject);
          opdImage.Filter:='Bitmaps (*.bmp)|*.bmp';
          svdImage.Filter:=opdImage.Filter;
        end else if PropType=TypeInfo(TIcon) then begin
          fPicture.Icon:=TIcon(Prop.AsObject);
          opdImage.Filter:='Icon files (*.ico)|*.ico';
          svdImage.Filter:=opdImage.Filter;
        end else
          fPicture.Graphic:=TGraphic(Prop.AsObject);
      Result:=ShowModal=mrOk;
      if Result then
        with Prop do
          if PropType=TypeInfo(TPicture) then TPicture(Prop.AsObject).Assign(FPicture)
          else
            Prop.AsObject:=FPicture.Graphic;
    finally

      fPicture.Free;
      btnOK.Free;
      btnCancel.Free;
      opdImage.Free;
      svdImage.Free;
      btnLoad.Free;
      btnSave.Free;
      btnClear.Free;
      pntImage.Free;
      
      Free;
    end;
end;

procedure TfrmPicture.btnLoadClick(Sender: TObject);
begin
  with opdImage do
    if Execute then begin
      fPicture.LoadFromFile(FileName);
      pntImage.Repaint;
    end;
end;

procedure TfrmPicture.btnSaveClick(Sender: TObject);
begin
  with FPicture,svdImage do
    if not Graphic.Empty and Execute then SaveToFile(FileName);
end;

procedure TfrmPicture.btnClearClick(Sender: TObject);
begin
  fPicture.Graphic:=nil;
  pntImage.Repaint;
end;

procedure TfrmPicture.pntImagePaint(Sender: TObject);
var
  R: TRect;
begin
  if Assigned(FPicture.Graphic) and not FPicture.Graphic.Empty then
  begin
    with fPicture do R:=Rect(0,0,Width,Height);
    with pntImage,R do begin
      if (Right>ClientWidth) or (Bottom>ClientHeight) then
        if Right>Bottom then begin
          Right:=ClientWidth;
          Bottom:=Right*FPicture.Height div FPicture.Width;
        end else begin
          Bottom:=ClientHeight;
          Right:=Bottom*FPicture.Width div FPicture.Height;
        end;
      OffsetRect(R,(ClientWidth-Right) div 2,(ClientHeight-Bottom) div 2);
      Canvas.StretchDraw(R,FPicture.Graphic);
    end;
  end;
end;

//#####################################################################

const
  ScrollArea = 6;

function TMenuPropertyEditor.GetMenuItems: TMenuItem;
begin
  Result:=Prop.AsObject as TMenuItem;
end;

function TMenuPropertyEditor.Execute: Boolean;
begin
  fForm:= TfrmMenuEditor.CreateNew(Nil);
  with fForm do
  try
    FormStyle:=fsStayOnTop;
    OnDestroy:=FormDestroy;

    FEditor:=Self;

    trvItems:= TTreeView.Create(fForm);
    trvItems.Parent:=fForm;
    trvItems.Align:=alClient;
    trvItems.OnChange  := trvItemsChange;

    pnlButtons:= TPanel.Create(fForm);
    pnlButtons.Parent:=fForm;
    pnlButtons.Align:=alBottom;
    pnlButtons.Height:=30;

    btnNewItem:= TButton.Create(fForm);
    btnNewItem.Parent:=pnlButtons;
    btnNewItem.OnClick:=btnAddItemClick;
    btnNewItem.Tag:=1;

    btnNewSubitem:= TButton.Create(fForm);
    btnNewSubitem.Parent:=pnlButtons;
    btnNewSubitem.OnClick:=btnAddItemClick;
    btnNewSubitem.Left:=66;
    btnNewSubitem.Tag:=2;

    btnDeleteItem:= TButton.Create(fForm);
    btnDeleteItem.Parent:=pnlButtons;
    btnDeleteItem.OnClick:=btnDeleteItemClick;
    btnDeleteItem.Left:=122;

    FItemsOwner:=MenuItems.Owner;
    MenuToNode;

    with InspectPanel.ObjectInspector do begin
      if MenuItems.Count>0 then Instance:=MenuItems[0]
      else Instance:=nil;
    end;

    Result:=ShowModal=mrOK;
    if Result then begin
      with MenuItems do
        while Count>0 do MenuItems[0].Free;
      NodeToMenu;
    end;

  finally
    Free;
  end;
end;

procedure TfrmMenuEditor.FormDestroy(Sender: TObject);
begin
  trvItems.Free;
  btnNewItem.Free;
  btnNewSubitem.Free;
  btnDeleteItem.Free;
  pnlButtons.Free;
end;

procedure CopyMenuItem(Source,Dest: TMenuItem);
begin
  with Source do
  begin
    Dest.Break:=Break;
    Dest.Caption:=Caption;
    Dest.Checked:=Checked;
    Dest.Default:=Default;
    Dest.Enabled:=Enabled;
    Dest.GroupIndex:=GroupIndex;
    Dest.HelpContext:=HelpContext;
    Dest.Hint:=Hint;
    Dest.Name:=Name;
    Dest.RadioItem:=RadioItem;
    Dest.ShortCut:=ShortCut;
    Dest.Tag:=Tag;
    Dest.Visible:=Visible;
    Dest.OnClick:=OnClick;
  end;
end;

procedure TfrmMenuEditor.MenuToNode;
var
  i: Integer;

  procedure GetItems(Par: TTreeNode; MItem: TMenuItem);
  var
    i: Integer;
    TN: TTreeNode;
    MI: TMenuItem;
  begin
    MI:=TMenuItem.Create(nil);
    CopyMenuItem(MItem,MI);
    TN:=trvItems.Items.AddChildObject(Par,MI.Caption,MI);
    for i:=0 to Pred(MItem.Count) do GetItems(TN,MItem.Items[i]);
  end;

begin
  for i:=0 to Pred(FEditor.MenuItems.Count) do GetItems(nil,FEditor.MenuItems[i]);
end;

procedure TfrmMenuEditor.NodeToMenu;
var
  i: Integer;

  procedure GetItems(Par: TMenuItem; TN: TTreeNode);
  var
    i: Integer;
    MI: TMenuItem;
  begin
    MI:=TMenuItem.Create(FItemsOwner);
    CopyMenuItem(TMenuItem(TN.Data),MI);
    Par.Add(MI);
    for i:=0 to Pred(TN.Count) do GetItems(MI,TN[i]);
  end;

begin
  for i:=0 to trvItems.Items.Count-1 do
    if (trvItems.Items[i].Parent=nil) then
      GetItems(FEditor.MenuItems,trvItems.Items[i]);
end;

procedure TfrmMenuEditor.WMGetMinMaxInfo(var Msg: TMessage);
begin
  inherited;
  with PMinMaxInfo(Msg.LParam)^.ptMinTrackSize do begin
    X:=413;
    Y:=200;
  end;
end;

procedure TfrmMenuEditor.btnAddItemClick(Sender: TObject);
var
  NewItem: TMenuItem;
  i: Integer;
  iName: string;
begin
  NewItem:=TMenuItem.Create(Self);
  with NewItem do
  begin
    iName:='NewItem';
    i:=1;
    while Assigned(Self.FindComponent(iName+IntToStr(i))) do Inc(i);
    Name:=iName+IntToStr(i);
    Caption:=Name;
    with trvItems do
      if TControl(Sender).Tag=1 then
        Selected:=Items.AddObject(Selected,Caption,NewItem)
      else
        Selected:=Items.AddChildObject(Selected,Caption,NewItem);
  end;
end;

procedure TfrmMenuEditor.btnDeleteItemClick(Sender: TObject);
begin
  with trvItems do
  begin
    TMenuItem(Selected.Data).Free;
    Selected.Free;
  end;
end;

procedure TfrmMenuEditor.trvItemsChange(Sender: TObject; Node: TTreeNode);
begin
  InspectPanel.ObjectInspector.Instance:=Node.Data;
end;

//#####################################################################

function TStringsPropertyEditor.Execute: Boolean;
begin
  fForm:=TfrmStrings.CreateNew(nil);
  with fForm do
    try
    BorderStyle       := bsDialog;
    Caption           := ' .';
    Width             := 355;
    Height            := 300;
    Position          := poMainFormCenter;

    lblCount  := TLabel.Create(fForm);
    redStrings:= TRichEdit.Create(fForm);
    btnOK     := TButton.Create(fForm);
    btnCancel := TButton.Create(fForm);
    btnClear  := TButton.Create(fForm);

    lblCount.Parent:=fForm;
    redStrings.Parent:=fForm;
    btnCancel.Parent:=fForm;
    btnClear.Parent:=fForm;
    btnOK.Parent:=fForm;
    
    btnOK.ModalResult:=MrOK;
    btnOK.Caption:='OK';

    if TStrings(Prop.AsObject)<>NIL then
      redStrings.Lines.Assign(TStrings(Prop.AsObject));
    Result:=ShowModal=mrOk;
    if Result then TStrings(Prop.AsObject).Assign(redStrings.Lines);
  finally
    lblCount.Free;
    redStrings.Free;
    btnOK.Free;
    btnCancel.Free;
    btnClear.Free;
    Free;
  end;
end;

end.
