unit STreeView;
{$R-}

interface

uses

  SysUtils, Windows, Messages, Classes, Graphics, ImgList,
  Controls, Forms, Dialogs, ComCtrls, CommCtrl;

type
  TSTreeNode = class(TTreeNode)
  public
    Main: boolean;
  end;

  TSTreeView = class(TTreeView)
  private
    FSortType   : TSortType;
    procedure   SetSortType(Value: TSortType);
    procedure   TreeViewAdvancedCustomDrawItem(Sender: TCustomTreeView;Node: TTreeNode; State: TCustomDrawState; Stage: TCustomDrawStage;var PaintImages, DefaultDraw: Boolean);
  protected
    function    GetItemText(ANode: TSTreeNode): string;
  public
    fNodeRound  : TBitmap;
    DeltaX      : integer;
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    function    AlphaSort: Boolean;
    function    CustomSort(SortProc: TTVCompare; Data: Longint): Boolean;
    procedure   LoadFromFile(const AFileName: string);
    procedure   SaveToFile(const AFileName: string);
    procedure   GetItemList(AList: TStrings);
    procedure   SetItemList(AList: TStrings);
  published
    property    SortType: TSortType read FSortType write SetSortType default stNone;
  end;

implementation

{$R STreeView.res}

uses
  ManagerFrm;

function DefaultTreeViewSort(Node1, Node2: TSTreeNode; lParam: Integer): Integer; stdcall;
begin
  Result := lstrcmp(PChar(Node1.Text), PChar(Node2.Text));
end;

constructor TSTreeView.Create(AOwner: TComponent);
var
  bmp     : TBitmap;
  fImages : TImageList;
begin
  inherited Create(AOwner);
  DeltaX:=0;
  OnAdvancedCustomDrawItem:= TreeViewAdvancedCustomDrawItem;
  FSortType   := stNone;
  fImages     := TImageList.Create(Self);
  bmp := TBitmap.Create;
  bmp.LoadFromResourceName(HInstance,'TREEBMP');
  fImages.AddMasked(bmp,clFuchsia);
  bmp.LoadFromResourceName(HInstance,'TYPEBMP');
  fImages.AddMasked(bmp,clWhite);
  bmp.LoadFromResourceName(HInstance,'STRUCT');
  fImages.AddMasked(bmp,clWhite);
  bmp.LoadFromResourceName(HInstance,'MENUBMP');
  fImages.AddMasked(bmp,clFuchsia);
  fNodeRound  := TBitmap.Create;
  fNodeRound.LoadFromResourceName(HInstance,'NODEBMP');
  Images:=fImages;                          
  bmp.Free;
end;

destructor TSTreeView.Destroy;
begin
  fNodeRound.Free;
  inherited Destroy;
end;

procedure TSTreeView.TreeViewAdvancedCustomDrawItem(Sender: TCustomTreeView;Node: TTreeNode; State: TCustomDrawState; Stage: TCustomDrawStage;var PaintImages, DefaultDraw: Boolean);
var
  r1                : TRect;
  cx,cy             : integer;
  Node1             : TSTreeNode;
  Debug             : boolean;

  procedure LineX(EndX: integer);
  var
    i: integer;
  begin
    for i:=Canvas.PenPos.X to EndX do
      if (i mod 2)=0 then Canvas.Pixels[i,Canvas.PenPos.Y]:=Canvas.Pen.Color;
  end;

  procedure LineY(EndY: integer);
  var
    i: integer;
  begin
    for i:=Canvas.PenPos.Y to EndY do
      if (i mod 2)=0 then Canvas.Pixels[Canvas.PenPos.X,i]:=Canvas.Pen.Color;
  end;

begin
  Debug := false;
  DefaultDraw:=false;
  r1:=Node.DisplayRect(True);
  //Inc(r1.Right,Length(Node.Text));

  with Canvas do begin

    Pen.Style:=psSolid;
    Pen.Color:=$7dbdc6;

    Brush.Style:=bsClear;
    if cdsSelected in State then Brush.Color:=$00EFD3C6;


    if (Node.Level=0) then begin
      Font.Style:=[fsBold];
      r1.Right:=abs(r1.Left)+TextWidth(Node.Text)+5;
    end;

    FillRect(r1);

    if (Node.Level>=1) then begin
      if Node.Selected then begin
        if Focused then
          Font.Color := clWhite
        else begin
          Font.Color := clDefault;
          Brush.Color:=$00EEEEEE;
          FillRect(r1);
        end;
      end else
        Font.Color := FontColor[TFileType(FindFormat(ExtractFileExt(Node.Text)))];
        Font.Style:=[];
        if TSTreeNode(Node).Main then Font.Color:=clBlue;
    end;

    if cdsSelected in State then begin
      Brush.Color:=$00D89574;
      FrameRect(r1);
    end;

    TextOut(r1.Left+3,r1.Top+1,Node.Text);

    if (Node.Level>0) then begin
      // 
      cx:=r1.Left - (Indent div 2)-DeltaX;
      cy:=r1.Top + (r1.Bottom - r1.Top) div 2;

      // - .
      if Debug then Pen.Color:=clLime;
      MoveTo(cx+5,cy);
      //LineTo(cx+9,cy);
      LineX(cx+9);

      // - .
      if (Node.Parent<>nil) or (Node.GetPrevSibling<>nil) then begin
        if Debug then Pen.Color:=clBlue;
        MoveTo(cx,cy-9);
        //LineTo(cx,cy);
        LineY(cy);
      end; 

      // -.
      if (Node.GetNextSibling <> nil) then begin
        if Debug then Pen.Color:=clBlack;
        MoveTo(cx,cy+5);
        //LineTo(cx,cy+8);
        LineY(cy+8);
      end;

      //  /
      if Node.HasChildren then begin
        Draw(cx-4,cy-4,fNodeRound);
        MoveTo(cx-2,cy);
        LineTo(cx+3,cy);
        if not Node.Expanded then begin
          MoveTo(cx,cy-2);
          LineTo(cx,cy+3);
        end;
      end else begin
        if Debug then Pen.Color:=clRed;
        MoveTo(cx,cy);
        //LineTo(cx+8,cy);
        LineX(cx+8);
        if (Node.GetNextSibling <> nil) then begin
          MoveTo(cx,cy);
          //LineTo(cx,cy+8);
          LineY(cy+8);
        end;
      end;

      //.  
      if Debug then Pen.Color:=clYellow;
      Node1:=TSTreeNode(Node.Parent);
      while Node1<>nil do begin
        cx:=cx-Indent;
        if Node1.GetNextSibling<>nil then begin
          MoveTo(cx,r1.Top-3);
          //LineTo(cx,r1.Bottom);
          LineY(r1.Bottom);
        end;
        Node1:=TSTreeNode(Node1.Parent);
      end;
    end;

  end;

 Images.Draw(Canvas,r1.Left-DeltaX,r1.Top,Node.ImageIndex,true);

end;

procedure TSTreeView.SetSortType(Value: TSortType);
begin
  if SortType <> Value then
  begin
    FSortType := Value;
    if ((SortType in [stData, stBoth]) and Assigned(OnCompare)) or
      (SortType in [stText, stBoth]) then
      AlphaSort;
  end;
end;

procedure TSTreeView.LoadFromFile(const AFileName: string);
var
  AList: TStringList;
begin

  AList := TStringList.Create;
  Items.BeginUpdate;
  try
    AList.LoadFromFile(AFileName);
    SetItemList(AList);
  finally
    Items.EndUpdate;
    AList.Free;
  end;
end;

procedure TSTreeView.SaveToFile(const AFileName: string);
var
  AList: TStringList;
begin

  AList := TStringList.Create;
  try
    GetItemList(AList);
    AList.SaveToFile(AFileName);
  finally
    AList.Free;
  end;
end;

procedure TSTreeView.SetItemList(AList: TStrings);
var
  ALevel, AOldLevel, i, Cnt: Integer;
  S: string;
  ANewStr: string;
  AParentNode: TSTreeNode;
  TmpSort: TSortType;

  function GetBufStart(Buffer: PChar; var ALevel: Integer): PChar;
  begin
    ALevel := 0;
    while Buffer^ in [' ', #9] do
    begin
      Inc(Buffer);
      Inc(ALevel);
    end;
    Result := Buffer;
  end;

begin
  SendMessage(handle, TVM_DELETEITEM, 0, Longint(TVI_ROOT));
  AOldLevel := 0;
  AParentNode := nil;
  TmpSort := SortType;
  SortType := stNone;
  try
    for Cnt := 0 to AList.Count - 1 do
    begin
      S := AList[Cnt];
      if (Length(S) = 1) and (S[1] = Chr($1A)) then
        Break;

      ANewStr := GetBufStart(PChar(S), ALevel);
      if (ALevel > AOldLevel) or (AParentNode = nil) then
      begin
        if ALevel - AOldLevel > 1 then
          raise Exception.Create('  TreeNode');
      end
      else
      begin
        for i := AOldLevel downto ALevel do
        begin
          AParentNode := TSTreeNode(AParentNode.Parent);
          if (AParentNode = nil) and (i - ALevel > 0) then
            raise Exception.Create('  TreeNode');
        end;
      end;
      AParentNode := TSTreeNode(Items.AddChild(AParentNode, ANewStr));
      AOldLevel := ALevel;
    end;
  finally
    SortType := TmpSort;
  end;
end;

procedure TSTreeView.GetItemList(AList: TStrings);
var
  i, Cnt: integer;
  ANode: TSTreeNode;
begin

  AList.Clear;
  Cnt := Items.Count - 1;
  ANode := TSTreeNode(Items.GetFirstNode);
  for i := 0 to Cnt do
  begin
    AList.Add(GetItemText(ANode));
    ANode := TSTreeNode(ANode.GetNext);
  end;
end;

function TSTreeView.GetItemText(ANode: TSTreeNode): string;
begin
  Result := StringOfChar(' ', ANode.Level) + ANode.Text;
end;

function TSTreeView.AlphaSort: Boolean;
begin
  if HandleAllocated then
  begin
    Result := CustomSort(nil, 0);
  end
  else
    Result := False;
end;

function TSTreeView.CustomSort(SortProc: TTVCompare; Data: Longint): Boolean;
var
  SortCB: TTVSortCB;
  Node: TSTreeNode;
begin
  Result := False;
  if HandleAllocated then
  begin
    with SortCB do
    begin
      if not Assigned(SortProc) then
        lpfnCompare := @DefaultTreeViewSort
      else
        lpfnCompare := SortProc;
      hParent := TVI_ROOT;
      lParam := Data;
      Result := TreeView_SortChildrenCB(Handle, SortCB, 0);
    end; 
    if Items.Count > 0 then
    begin
      Node := TSTreeNode(Items.GetFirstNode);
      while Node <> nil do
      begin
        if Node.HasChildren then
          Node.CustomSort(SortProc, Data);
        Node := TSTreeNode(Node.GetNext);
      end;
    end;
  end;
end;

end.