unit SynEdit;

interface

uses
    Windows, Messages, SysUtils, Classes, Graphics, Controls, StdCtrls,
    Contnrs, Clipbrd, Forms, Dialogs, MessagesFrm, DebugFrm;

const
  tokBlank        = 0;
  tokText         = 1;
  tokError        = 2;
  tokOK           = 3;
  tokUser         = 255;

  TABlen          = 4;

type
  TCustomSyntaxMemo = class;
  TSynMemoRange     = class;
  TSynMemoSection   = class;
  TSynMemoStrings   = class;
  TSyntaxParser     = class;
  TSyntaxAttributes = class;
  EMPSyntaxMemo       = class(Exception);

  TToken              = tokBlank..tokUser;
  PToken              = ^TToken;
  TTokenSet           = set of TToken;
  TCharSet            = set of Char;

  TCommentBlock     = class
    BlockStart    : String;
    BlockEnd      : String;
    BlockToken    : TToken;
  end;

  TUserTokenEvent     = procedure(
      Sender          : TObject;
      Word            : string;
      Pos, Line       : Integer;
      var Token       : TToken
  ) of object;

  TCharWidths         = array [0..3] of array [Char] of Byte;
  TTokenStyle         = record
                          tsForeground : TColor;          //  
                          tsBackground : TColor;          //  
                          tsStyle      : TFontStyles;     //  
                        end;

  TCustomScrollBar  = class(TPersistent)
    private
      fControl    : TCustomSyntaxMemo;
      fKind       : TScrollBarKind;
      fIncrement  : TScrollBarInc;
      fCalcRange  : Integer;
      fPosition   : integer;
      fRange      : integer;
      fSize       : integer;
      fTracking   : boolean;
      fVisible    : boolean;
      fEnabled    : boolean;
      constructor Create(AControl: TCustomSyntaxMemo; AKind: TScrollBarKind);
      procedure   Update(ControlSB, AssumeSB: Boolean);
      procedure   SetPosition(Value: Integer);
      procedure   SetRange(Value: Integer);
      procedure   SetSize(Value: Integer);
      procedure   SetVisible(Value: Boolean);
      procedure   SetEnabled(Value: Boolean);
      function    ControlSize(ControlSB, AssumeSB: Boolean): Integer;
      function    NeedsScrollBarVisible: Boolean;
      procedure   ScrollMessage(var Msg: TWMScroll);
      procedure   MouseMessage (var Msg: TMessage);
    published
      property    Position : Integer read fPosition write SetPosition default 0;
      property    Range    : Integer read fRange write SetRange default 0;
      property    Size     : Integer read fSize write SetSize default 0;
      property    Tracking : Boolean read fTracking write fTracking default False;
      property    Visible  : Boolean read fVisible write SetVisible default True;
      property    Enabled  : Boolean read fEnabled write SetEnabled default True;
      property    Increment: TScrollBarInc read fIncrement write fIncrement default 8;
  end;

    // ,       
  TSyntaxAttributes = class
    private
        fRichMemo       : TCustomSyntaxMemo;
        fTokenStyles    : array [TToken] of TTokenStyle;
        fOnUserToken    : TUserTokenEvent;
        function        GetColor(const Token: TToken; const Index: Integer): TColor;
        function        GetStyle(const Token: TToken): TFontStyles;
        procedure       SetColor(const Token: TToken; const Index: Integer; const Value: TColor);
        procedure       SetStyle(const Token: TToken; const Value: TFontStyles);
    public
        constructor     Create(Owner: TCustomSyntaxMemo);
        procedure       Assign(Friend: TSyntaxAttributes);
        function        Equals(const T1, T2: TToken): Boolean;
        procedure       CopyAttrs(const SrcToken: TToken; DstTokArray: array of TToken);
        property        OnUserToken: TUserTokenEvent read fOnUserToken write fOnUserToken;
        property        FontColor[const Token: TToken]: TColor index 0 read GetColor write SetColor;
        property        BackColor[const Token: TToken]: TColor index 1 read GetColor write SetColor;
        property        FontStyle[const Token: TToken]: TFontStyles read GetStyle write SetStyle;
    end;

  TSyntaxTokenStyle = ( stsInSelection,     //   
                            stsPressed          //      - reserved for future use
                          );
  TSyntaxTokenStyles= set of TSyntaxTokenStyle;
  TSyntaxToken      = class
    public
        stStart         : Word;                     //  , 
        stLength        : Word;                     //  , 
        stToken         : TToken;                   //  () 
        stStyle         : TSyntaxTokenStyles;     //  
    end;

  TSyntaxParser     = class (TObjectList)
    private
        fSection        : TSynMemoSection;
        fVisibleIndex   : Integer;
        fNeedReparse    : Boolean;
        function        GetToken(const TokIndex: Integer): TSyntaxToken;
        procedure       SetToken(const TokIndex: Integer; const Value: TSyntaxToken);
    public
        constructor     Create(const AsCloneOf: TSyntaxParser = nil);
        procedure       Assign(const Friend: TSyntaxParser);
        procedure       Clear; override;
        procedure       AddToken(const Beg, Len: Integer; Token: TToken);
        procedure       SplitTokens(const sx, ex: Integer);
        function        LastToken: TToken;
        function        FirstToken: TToken;
        function        Parse(Line: string; LineIndex: Integer; LastToken: TToken; PA: TSyntaxAttributes): TToken;
        property        Tokens[const TokIndex: Integer]: TSyntaxToken read GetToken write SetToken; default;
        property        NeedReparse: Boolean read fNeedReparse write fNeedReparse;
        property        Section: TSynMemoSection read fSection write fSection;
        property        VisibleIndex: Integer read fVisibleIndex write fVisibleIndex;
    end;


  TStringsStateItem   = (
                        ssTextChanged,      //   
                        ssSectionsChanged,  //    (New, Explode)
                        ssNeedReIndex,      //    (Expand, Collapse)
                        ssNeedReparseAll,   //    
                        ssUndoProcess       //     
    );
  TStringsState       = set of TStringsStateItem;

  TSynMemoStrings    = class(TStringList)
    private
        fRichMemo       : TCustomSyntaxMemo;              // 
        fState          : TStringsState;                    //  
        fModified       : Boolean;                          //  
        fDirectAccess   : Boolean;                          //    
        function        GetParser(const Row: Integer): TSyntaxParser;
        procedure       SetModified(const Value: Boolean);
    protected
        procedure       Changed; override;
        procedure       Put(Index: Integer; const s: string); override;
        procedure       SetUpdateState(Updating: Boolean); override;
        function        ParseLine(const Index: Integer; const TestNextLine: Boolean): Boolean; virtual;
        property        State: TStringsState read fState write fState;
    public
        constructor     Create(const Owner: TCustomSyntaxMemo);
        destructor      Destroy; override;
        procedure       Clear; override;
        function        PositionToRC(Value: Integer): TPoint;
        function        RCToPosition(Col, Row: Integer): Integer;
        procedure       Parse(const EntireText : Boolean; const NeedRepaint: Boolean = False);
        procedure       Delete(Index: Integer); override;
        procedure       InsertItem(Index: Integer; const s: string; AObject: TObject); override;
        procedure       LoadFromStream(Stream: TStream); override;
        procedure       SaveToStream(Stream: TStream); override;
        procedure       LoadFromFile(const FileName: string); override;
        procedure       LoadFromUFile(const FileName: string);
        procedure       SaveToFile(const NewFileName: string); override;
        procedure       New; virtual;
        function        IsValidLineIndex(const Row: Integer): Boolean;
        property        Parser[const Row: Integer]: TSyntaxParser read GetParser;
        property        Modified: Boolean read fModified write SetModified;
        property        DirectAccess: Boolean read fDirectAccess write fDirectAccess;
        property        UpdateCount;
    end;


  TSynMemoSection    = class(TObjectList)
    private
        fParent         : TSynMemoSection;    //  - 
        fRowBeg         : Integer;              //    
        fRowEnd         : Integer;              //    
        fLevel          : Integer;              //   
        fCollapsed      : Boolean;              //  ,   
        function        GetSections(const Idx: Integer): TSynMemoSection;
        procedure       SetLevel(const Value: Integer);
    public
        constructor     Create;
        property        RowBeg: Integer read fRowBeg write fRowBeg;
        property        RowEnd: Integer read fRowEnd write fRowEnd;
        property        Sections[const Idx: Integer]: TSynMemoSection read GetSections; default;
        property        Level: Integer read fLevel write SetLevel;
        property        Parent: TSynMemoSection read fParent write fParent;
        property        Collapsed: Boolean read fCollapsed write fCollapsed;
    end;

  TSMSectionClone   = class(TSynMemoSection)
    private
        fRefCount       : Integer;
    protected
        procedure       AddRef;
        procedure       Release;
    public
        property        RefCount: Integer read fRefCount;
    end;

  TSectionMark = (smNone, smExpanded, smCollapsed, smEnd);

  TSynMemoSections   = class(TObject)
    private
        fRichMemo       : TCustomSyntaxMemo;          // 
        fRoot           : TSMSectionClone;            //  
        fIndexes        : TList;                        //        
        fMaxLevel       : Integer;                      //    
        fMaxExpandLevel : Integer;                      //     
        fErrorLine      : Integer;                      //       
        fErrorString    : string;                       //   
        procedure       ReIndex;
        procedure       MakeUnique;
        procedure       SetRoot(const Value: TSMSectionClone);
        function        GetSection(const Row: Integer): TSynMemoSection;
        procedure       SetSection(const Row: Integer; Value: TSynMemoSection);
    protected
        procedure       Scan; virtual;
        procedure       FillOutput(const Sl: TStringList); virtual;
        procedure       DeleteRow(const Row: Integer); virtual;
        procedure       InsertRow(const Row: Integer); virtual;
    public
        class function  DetectSectionMark(const s: string): TSectionMark;
        constructor     Create(Owner: TCustomSyntaxMemo);
        destructor      Destroy; override;
        function        New(const Row1, Row2: Integer; const IsCollapsed: Boolean = False): TSynMemoSection;
        procedure       Explode(const Row: Integer; const Recursive: Boolean);
        procedure       Collapse(const Row: Integer; const Recursive, SafeSelf: Boolean);
        procedure       Expand(const Row: Integer; const Recursive, ParentRecursive: Boolean);
        function        SectionBorder(const Row: Integer): TSectionMark;
        function        AsText: string;
        function        Next(Sec: TSynMemoSection): TSynMemoSection;
        function        Prev(Sec: TSynMemoSection): TSynMemoSection;
        function        Visible(const Sec: TSynMemoSection): Boolean;
        property        Section[const Row: Integer]: TSynMemoSection read GetSection write SetSection;
        property        MaxLevel: Integer read fMaxLevel;
        property        ErrorLine: Integer read fErrorLine;
        property        ErrorString: string read fErrorString;
        property        EntireSection: TSMSectionClone read fRoot write SetRoot;
        property        Indexes: TList read fIndexes;
    end;


  TSynMemoUndoItem  = class
        uiCaretPos      : TPoint;
        uiSelStart      : TPoint;
        uiSelEnd        : TPoint;
        uiSealing       : Boolean;
        uiText          : string;
        uiSections      : TSMSectionClone;
        destructor      Destroy; override;
    end;

  TSynMemoRange = class(TObject)
    private
        fRichMemo       : TCustomSyntaxMemo;  // 
        fStart          : TPoint;               //   
        fEnd            : TPoint;               //   
        fPos            : TPoint;               //      [Row,Col]
        fSealing        : Boolean;              //  Collapsed (,  )
        fMaxUndoDepth   : Integer;              //    
        fUndoStack      : TObjectList;          //  
        function        GetLength: Integer;
        function        GetPosition: Integer;
        function        GetText: string;
        function        GetMarkedText: string;
        procedure       CutFinalSpaces(const Row: Integer);
        procedure       SetLength(const Value: Integer);
        procedure       SetMaxUndoDepth(const Value: Integer);
        procedure       SetPosition(const Value: Integer);
        procedure       SetPos(const NewPos: TPoint);
        procedure       SetRange(const Index, Value: Integer);
        procedure       SetText(const Value: string);
        procedure       SetMarkedText(Value: string);
        procedure       MakeIndent;
        procedure       MakeUnIndent;
    protected
        function        AddUndo(const UndoText: string = ''): TSynMemoUndoItem;
        property        StartX : Integer read fStart.X write fStart.X;
        property        StartY : Integer read fStart.Y write fStart.Y;
        property        EndX   : Integer read fEnd.X   write fEnd.X;
        property        EndY   : Integer read fEnd.Y   write fEnd.Y;
    public
        constructor     Create(Owner: TCustomSyntaxMemo);
        destructor      Destroy; override;
        procedure       Collapse;
        procedure       Enlarge(const Value: Integer; const EnlargeLine: Boolean = False; const VisiblesOnly: Boolean = False);
        procedure       Delete;
        {    }
        procedure       CreateSection;
        procedure       ExplodeSection(const Recursive: Boolean);
        procedure       ExpandSection(const Recursive: Boolean);
        procedure       CollapseSection(const Recursive: Boolean);
        procedure       GotoSection(const GoForward: Boolean);
        {    }
        procedure       CopyToClipboard;
        procedure       CutToClipBoard;
        procedure       PasteFromClipboard;
        procedure       CopyAsHTML;
        {   }
        procedure       DoUndo;
        procedure       ClearUndo;
        procedure       SelectAll;
        procedure       SelectFromStart;
        procedure       SelectToEnd;
        function        IsEmpty: Boolean;
        function        CanUndo: Boolean;
        property        UndoStack: TObjectList read fUndoStack;
        property        PosX   : Integer index 0 read fPos.X write SetRange;
        property        PosY   : Integer index 1 read fPos.Y write SetRange;
        property        Pos: TPoint read fPos write SetPos;
        property        MaxUndoDepth: Integer read fMaxUndoDepth write SetMaxUndoDepth default 100;
        property        Position:     Integer read GetPosition write SetPosition;
        property        SelLength:    Integer read GetLength write SetLength;
        property        Text: string read GetText write SetText;
        property        MarkedText: string read GetMarkedText write SetMarkedText;
    end;

    TWordInfoEvent      = procedure (Sender: TCustomSyntaxMemo; const X, Y, WordIndex, Row: Integer; Showing: Boolean) of object;
    TRowIndexConvertionDirection = (cdNeedReal, cdNeedScreen);
    TSynMemoOption = (  smoReadOnly,                    //    ( )
                        smoOverwrite,                   //   
                        smoSkipSectionsOnCopy,          //        
                        smoSkipSectionsOnPaste,         //         
                        smoAutoGutterWidth,             //      
                        smoWriteMarkersOnSave          //      
    );
    TSynMemoOptions   = set of TSynMemoOption;
    TLogEvent           = procedure (Sender: TObject; LogStr: string) of object;
    TChangedItem        = (ciText, ciSelection, ciSections, ciUndoStack, ciOptions);
    TChangedItems       = set of TChangedItem;
    TChangeEvent      = procedure (Sender: TObject; ChangedItems: TChangedItems) of object;


    //  
    TBookmarkIndex      = 0..9;
    TBookmarkManager    = class
    private
        fRichMemo       : TCustomSyntaxMemo;
        fBookMarks      : array [TBookmarkIndex] of Integer;
        fImages         : TBitmap;
        function        GetBookMarks(const Index: TBookmarkIndex): Integer;
        procedure       SetBookMarks(const Index: TBookmarkIndex; const Row: Integer);
    public
        constructor     Create(Owner: TCustomSyntaxMemo);
        destructor      Destroy; override;
        function        Find(const Row: Integer; var Index: TBookmarkIndex): Boolean;
        procedure       Clear;
        property        BookMarks[const Index: TBookmarkIndex]: Integer read GetBookMarks write SetBookMarks; default;
    end;

  TSCharHighlight = class(TPersistent)
    private
      fFontColor, fBackColor : TColor;
      fFontStyle : TFontStyles;
    published
      property    FontColor : TColor read fFontColor write fFontColor;
      property    BackColor : TColor read fBackColor write fBackColor;
      property    FontStyle : TFontStyles read fFontStyle write fFontStyle;
    public
      constructor Create; reintroduce;
      destructor  Destroy; override;
    end;

  TSCustomHighlight = class(TComponent)
    protected
      function    Parse(Sender: TSyntaxParser; Line: string; LineIndex: Integer; LastToken: TToken; PA: TSyntaxAttributes): TToken; virtual; abstract;
      procedure   Change(RichMemo : TCustomSyntaxMemo; ChangedItems: TChangedItems); virtual; abstract;
      procedure   SetMemoColor(RichMemo: TCustomSyntaxMemo; Tok : TToken; MPColor : TSCharHighlight);
      procedure   GetHighlitedParam(RichMemo : TCustomSyntaxMemo); virtual;
    public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
    published

    end;

    TCustomSyntaxMemo = class(TCustomControl)
    private
        fLines          : TSynMemoStrings;
        fRange          : TSynMemoRange;
        fSections       : TSynMemoSections;
        fOptions        : TSynMemoOptions;
        fCharHeight     : Integer;
        fCharWidths     : TCharWidths;
        fParseAttributes: TSyntaxAttributes;
        fBookMarks      : TBookmarkManager;
        fOffsets        : TPoint;
        fDown           : Boolean;
        fHinting        : Boolean;
        fInsertMode     : Boolean;
        fVScroll        : TCustomScrollBar;
        fHScroll        : TCustomScrollBar;
        fSelColor       : TColor;
        fDefBackColor   : TColor;
        fDefForeColor   : TColor;
        fSelWord        : TPoint;
        fCaretVisible   : Boolean;
        fScreenLines    : array of Boolean;
        fGutterWidth    : Integer;
        fSectionIndent  : Integer;
        fChangesSummator: TChangedItems;
        fOnChange       : TChangeEvent;
        fOnWordInfo     : TWordInfoEvent;
        fHighliter      : TSCustomHighlight;
        fErrorLine      : integer;
        fNaviLine       : Integer;
        function        GetUserTokenEvent: TUserTokenEvent;
        { Sets }
        procedure       SetDefColor(const Index: Integer; const Value: TColor);
        procedure       SetGutterWidth(const Value: Integer);
        procedure       SetOffset(const Index, Value: Integer);
        procedure       SetOffsets(NewOffsets: TPoint);
        procedure       SetOptions(const Value: TSynMemoOptions);
        procedure       SetSectionIndent(const Value: Integer);
        procedure       SetSelColor(const Value: TColor);
        procedure       SetSelectedWord(const Value: TPoint);
        procedure       SetUserTokenEvent(const Value: TUserTokenEvent);
        { Others }
        procedure       Reset; virtual;
        procedure       CalcScreenParams;
        procedure       UpdateScrollBars;
        procedure       ScrollBy(DeltaX, DeltaY: Integer);
        function        ClientLines: Integer;
        function        FindNextWord(var wx, wy: Integer): Boolean;
        function        FindPrevWord(var wx, wy: Integer): Boolean;
        procedure       PaintGutter(const ACanvas: TCanvas; const Row, ScreenRow: Integer);
        procedure       PaintSectionMarks(const ACanvas: TCanvas; const Row, ScreenRow: Integer);
        procedure       PaintDots(const ACanvas: TCanvas);
        procedure       PaintTokens(const ACanvas: TCanvas; s: string; Row: integer; Sp: TSyntaxParser; TextIndent, SelStart, SelEnd: Integer);
//        procedure       PaintLine(const ScreenRow, Row: Integer);
        procedure       PaintLine(const ScreenRow, Row: Integer);
        function        RowIndexConvert(const Index: Integer; const Direction: TRowIndexConvertionDirection): Integer;
        function        FindVisibleRow(const Row, Delta: Integer; const EnsureInRange: Boolean): Integer;
        function        RangeRowToScreenRow(const Row: Integer): Integer;
        { Repaint manager }
        procedure       ReDraw;
        procedure       NeedReDrawLE(const Row: Integer);
        procedure       NeedRedrawAll;
        {   }
        function        CharWidth(N: TFontStyles): byte;
        function        CharPosToPixOffset(const Col, Row: Integer): Integer; overload;
        function        CharPosToPixOffset(const Col: Integer; s: string; Sp: TSyntaxParser): Integer; overload;
        function        PixOffsetToCharPos(const Pix, Row: Integer; const WordIndex: PInteger = nil): Integer;
        procedure       WndOffsetToPixOffset(OfsPoint: TPoint; var CharPix, Row: Integer; const TextRow: Boolean);
        function        PixOffsetToWndOffsetEx(const CharPix, ScreenRow: Integer): TPoint;
        function        GetSectionButtonRect(const ScreenRow, ALevel: Integer): TRect;
        function        IsLineVisible(const Row: Integer; const PScreenRow: PInteger = nil): Boolean;
        function        GetWndRect(const ScreenRow, Index: Integer): TRect;
        procedure       SetHighliter(const Value: TSCustomHighlight); virtual;
    protected
        function        CanResize(var NewWidth, NewHeight: Integer): Boolean; override;
        procedure       CreateParams(var Params: TCreateParams); override;
        procedure       WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
        procedure       WMSize(var Message: TMessage); message WM_SIZE;
        procedure       WMKillFocus(var Msg: TWMKillFocus); message WM_KILLFOCUS;
        procedure       WMLButtonDblClk(var Message: TWMMouse); message WM_LBUTTONDBLCLK;
        procedure       WMSetFocus(var Msg: TWMSetFocus); message WM_SETFOCUS;
        procedure       WMMouseWheel(var Message: TMessage); message WM_MouseWheel;
        procedure       WMHScroll   (var Message: TWMHScroll);    message WM_HSCROLL;
        procedure       WMVScroll   (var Message: TWMVScroll);    message WM_VSCROLL;
        procedure       CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
        procedure       CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
        procedure       MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
        procedure       MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
        procedure       MouseMove(Shift: TShiftState; X, Y: Integer); override;
        procedure       KeyDown(var Key: Word; Shift: TShiftState); override;
        procedure       KeyPress(var Key: Char); override;
        procedure       KeyUp(var Key: Word; Shift: TShiftState); override;
        procedure       FontChange(Sender: TObject);
        procedure       Paint; override;
        procedure       HideCaret;
        procedure       ShowCaret;
        procedure       Change(const ChangedItems: TChangedItems); virtual;
                        //     
        property        EntireRowRect[const ScreenRow: Integer]: TRect      index 0 read GetWndRect;
        property        TextRowRect[const ScreenRow: Integer]: TRect        index 1 read GetWndRect;
        property        EntireGutterRect[const ScreenRow: Integer]: TRect   index 2 read GetWndRect;
        property        SymbolsGutterRect[const ScreenRow: Integer]: TRect  index 3 read GetWndRect;
    public
        Path,Name       : String;
        fWord           : String;
        constructor     Create(AOwner: TComponent); override;
        destructor      Destroy; override;
        function        CharPosToWordIndex(const Col, Row: Integer): Integer;
        function        GetWordAtPos(const X, Y: Integer; var WordIndex, Row: Integer): Boolean;
        procedure       ShowWord(const Row, WordIndex: Integer);
        procedure       MakeVisible(const Col, Row: Integer; const Length: Integer = 1);
        procedure       Navigate(const Col, Row: Integer);
        procedure       ErrorLine(const Col, Row: Integer);
        procedure       NaviLine(const Col, Row: Integer);
        procedure       NeedRedraw(const Row: Integer);
        { Properties }
        property        BookMarks: TBookmarkManager read fBookMarks;
        property        DefBackColor: TColor index 0 read fDefBackColor write SetDefColor default clWindow;
        property        DefForeColor: TColor index 1 read fDefForeColor write SetDefColor default clBlack;
        property        GutterWidth: Integer read fGutterWidth write SetGutterWidth default 32;
        property        Lines: TSynMemoStrings read fLines;
        property        OffsetXPix: Integer index 0 read fOffsets.X write SetOffset;
        property        OffsetY: Integer index 1 read fOffsets.Y write SetOffset;
        property        Offsets: TPoint read fOffsets write SetOffsets;
        property        Range: TSynMemoRange read fRange;
        property        SectionIndent: Integer read fSectionIndent write SetSectionIndent default 16;
        property        Sections: TSynMemoSections read fSections;
        property        SelColor: TColor read fSelColor write SetSelColor default clSkyBlue;
        property        SelectedWord: TPoint read fSelWord write SetSelectedWord;
        property        SyntaxAttributes: TSyntaxAttributes read fParseAttributes;
        property        Options: TSynMemoOptions read fOptions write SetOptions;
        property        OnChange: TChangeEvent read fOnChange write fOnChange;
//        property        OnDrawWord: TDrawWordEvent read fOnDrawWord write fOnDrawWord;
        property        OnWordInfo: TWordInfoEvent read fOnWordInfo write fOnWordInfo;
        property        OnParseWord: TUserTokenEvent read GetUserTokenEvent write SetUserTokenEvent;
        property        ParentColor default False;
        property        Highliter  : TSCustomHighlight read fHighliter write SetHighliter;
        procedure       FindWord(const Col,Row: integer; frDown,frMatchCase: boolean);
    end;


    TSyntaxMemo       = class(TCustomSyntaxMemo)
    published
        property        Anchors;
        property        Align;
        property        Color;
        property        DefBackColor;
        property        DefForeColor;
        property        GutterWidth default 32;
        property        Font;
        property        Options;
        property        SelColor default clHighlight;
        property        SectionIndent default 32;
        property        TabStop default True;
        property        OnKeyDown;
        property        OnKeyPress;
        property        OnMouseDown;
        property        OnMouseMove;
        property        OnMouseUp;
        property        OnClick;
        property        OnContextPopup;
        property        OnDblClick;
        property        OnDragDrop;
        property        OnChange;
//        property        OnDrawWord;
        property        OnWordInfo;
        property        OnParseWord;
        property        Highliter;
    end;

var
  TokenBlock: TTokenSet = [];

const
  SectionMarks      : array [TSectionMark] of string = ('', '{<+}', '{<-}', '{>>}');
  SECTION_HEADER_LENGTH = 4;

implementation

{$R SynEdit.res}

uses
  StrUtils, Math, Types, ComCtrls;

var
  CF_SYNTAX: THandle;
  GlobalUntitledIndex: Integer = 1;
const
  UNTITLEDFN = 'Untitled';
  ROW_ABOVE_SCREEN    = -1;
  ROW_HIDEN           = -2;
  ROW_BELOW_SCREEN    = -3;

{$Region ' Standart Procedure  '}
procedure Swap(var A, B: Integer);
var
  t: Integer;
begin
  t := A;
  A := B;
  B := t;
end;

procedure DrawGradient(Canvas: TCanvas; FromColor, ToColor: TColor; Steps: Integer; R: TRect; Direction: Boolean);
var
  diffr, startr, endr: Integer;
  diffg, startg, endg: Integer;
  diffb, startb, endb: Integer;
  rstepr, rstepg, rstepb, rstepw: Real;
  i, stepw: Word;

begin
  if Steps = 0 then
    Steps := 1;

  FromColor := ColorToRGB(FromColor);
  ToColor := ColorToRGB(ToColor);

  startr := (FromColor and $0000FF);
  startg := (FromColor and $00FF00) shr 8;
  startb := (FromColor and $FF0000) shr 16;
  endr := (ToColor and $0000FF);
  endg := (ToColor and $00FF00) shr 8;
  endb := (ToColor and $FF0000) shr 16;

  diffr := endr - startr;
  diffg := endg - startg;
  diffb := endb - startb;

  rstepr := diffr / steps;
  rstepg := diffg / steps;
  rstepb := diffb / steps;

  if Direction then
    rstepw := (R.Right - R.Left) / Steps
  else
    rstepw := (R.Bottom - R.Top) / Steps;

  with Canvas do
  begin
    for i := 0 to steps - 1 do
    begin
      endr := startr + Round(rstepr * i);
      endg := startg + Round(rstepg * i);
      endb := startb + Round(rstepb * i);
      stepw := Round(i * rstepw);
      Pen.Color := endr + (endg shl 8) + (endb shl 16);
      Brush.Color := Pen.Color;
      if Direction then
        Rectangle(R.Left + stepw, R.Top, R.Left + stepw + Round(rstepw) + 1, R.Bottom)
      else
        Rectangle(R.Left, R.Top + stepw, R.Right, R.Top + stepw + Round(rstepw) + 1);
    end;
  end;
end;
{$EndRegion}
//########################################################################
{$Region ' TCustomScrollBar    '}
constructor TCustomScrollBar.Create(AControl: TCustomSyntaxMemo; AKind: TScrollBarKind);
begin
  inherited Create;
  fControl    := AControl;
  fKind       := AKind;
  Tracking    := true;
  Visible     := true;
  Enabled     := true;
  fIncrement  := 8;
end;

function TCustomScrollBar.ControlSize(ControlSB, AssumeSB: Boolean): Integer;
var
  BorderAdjust: Integer;

  function ScrollBarVisible(Code: Word): Boolean;
  var
    Style: Longint;
  begin
    Style := WS_HSCROLL;
    if Code = SB_VERT then Style := WS_VSCROLL;
    Result := GetWindowLong(FControl.Handle, GWL_STYLE) and Style <> 0;
  end;

  function Adjustment(Code, Metric: Word): Integer;
  begin
    Result := 0;
    if not ControlSB then
      if AssumeSB and not ScrollBarVisible(Code) then
        Result := -(GetSystemMetrics(Metric) - BorderAdjust)
      else if not AssumeSB and ScrollBarVisible(Code) then
        Result := GetSystemMetrics(Metric) - BorderAdjust;
  end;

begin
  BorderAdjust := Integer(GetWindowLong(FControl.Handle, GWL_STYLE) and
    (WS_BORDER or WS_THICKFRAME) <> 0);
  if fKind = sbVertical then
    Result := FControl.ClientHeight + Adjustment(SB_HORZ, SM_CXHSCROLL) else
    Result := FControl.ClientWidth + Adjustment(SB_VERT, SM_CYVSCROLL);
end;

function TCustomScrollBar.NeedsScrollBarVisible: Boolean;
begin
  Result := FRange > ControlSize(False, False);
end;

procedure TCustomScrollBar.Update(ControlSB, AssumeSB: Boolean);
var
  ScrollInfo: TScrollInfo;
  Code      : Word;
begin
  fCalcRange            := 0;
  Code                  := SB_HORZ;
  if fKind = sbVertical then begin
    Code                := SB_VERT;
  end;
  if Visible then begin
    fCalcRange := Range - ControlSize(ControlSB, AssumeSB);
    if fCalcRange < 0 then fCalcRange := 0;
  end;
  ScrollInfo.cbSize     := SizeOf(ScrollInfo);
  ScrollInfo.fMask      := SIF_ALL;
  ScrollInfo.nMin       := 0;
  ScrollInfo.nMax       := Range;
  ScrollInfo.nPage      := (ControlSize(ControlSB, AssumeSB) + 1);
  ScrollInfo.nPos       := fPosition;
  ScrollInfo.nTrackPos  := fPosition;
  SetScrollInfo(fControl.Handle, Code, ScrollInfo, True);
  if not Visible then ShowScrollBar(fControl.Handle, Code, Visible);
  if Enabled then
    EnableScrollBar(fControl.Handle,Code,ESB_ENABLE_BOTH)
  else begin
    EnableScrollBar(fControl.Handle,Code,ESB_DISABLE_BOTH);
    Exit;
  end;
  SetPosition(fPosition);
end;

procedure TCustomScrollBar.ScrollMessage(var Msg: TWMScroll);

  function GetRealScrollPosition: Integer;
  var
    SI: TScrollInfo;
    Code: Integer;
  begin
    SI.cbSize := SizeOf(TScrollInfo);
    SI.fMask := SIF_TRACKPOS;
    Code := SB_HORZ;
    if fKind = sbVertical then Code := SB_VERT;
    Result := Msg.Pos;
    if GetScrollInfo(FControl.Handle, Code, SI) then
      Result := SI.nTrackPos;
  end;

begin
  with Msg do
    case ScrollCode of
      SB_LINEUP       : SetPosition(Position - fIncrement);
      SB_LINEDOWN     : SetPosition(Position + fIncrement);
      SB_PAGEUP       : SetPosition(Position - ControlSize(True, False));
      SB_PAGEDOWN     : SetPosition(Position + ControlSize(True, False));
      SB_TOP          : SetPosition(0);
      SB_BOTTOM       : SetPosition(fCalcRange);
      SB_THUMBPOSITION:
                        if fCalcRange > 32767 then
                          SetPosition(GetRealScrollPosition)
                        else
                          SetPosition((Pos div Increment)*Increment);
      SB_THUMBTRACK   :
                        if Tracking then
                          if fCalcRange > 32767 then
                            SetPosition(GetRealScrollPosition)
                          else
                            SetPosition((Pos div Increment)*Increment);
      SB_ENDSCROLL    : ;
    end;
end;

procedure TCustomScrollBar.MouseMessage(var Msg: TMessage);
begin
  if fKind=sbHorizontal then begin
    if Short(Msg.WParamHi) > 0 then
      SetPosition(Position - fIncrement)
    else
      SetPosition(Position + fIncrement);
  end else begin
    if Short(Msg.WParamHi) > 0 then
      SetPosition(Position - fIncrement*4)
    else
      SetPosition(Position + fIncrement*4);
  end;
end;

procedure TCustomScrollBar.SetPosition(Value: Integer);
var
  Code    : Word;
  OldPos  : Integer;
begin
  if csReading in fControl.ComponentState then begin
    fPosition := (Value div fIncrement)*fIncrement;
    Exit;
  end;

  if not fControl.HandleAllocated or (Value=fPosition) then Exit;

  if Value > fCalcRange then
    Value := fCalcRange
  else if Value < 0 then
    Value := 0;

  if fKind = sbHorizontal then
    Code := SB_HORZ
  else
    Code := SB_VERT;

  OldPos    := fPosition;
  fPosition := (Value div fIncrement)*fIncrement;
  if OldPos=fPosition then Exit;

  if fKind = sbHorizontal then
    fControl.ScrollBy(OldPos-Value,0)
  else
    fControl.ScrollBy(0,OldPos-Value);

  if GetScrollPos(fControl.Handle, Code) <> fPosition then
    SetScrollPos(fControl.Handle, Code, fPosition, True);

  //fControl.Invalidate;
end;

procedure TCustomScrollBar.SetRange(Value: Integer);
begin
  if Value <> Range then begin
    fRange:=Value;
    if fRange < 0 then fRange := 0;
    fControl.UpdateScrollBars;
  end;
end;

procedure TCustomScrollBar.SetSize(Value: Integer);
const
  SysConsts: array[TScrollBarKind] of Integer = (SM_CYHSCROLL, SM_CYVSCROLL);
var
  NewValue: Integer;
begin
  if Value <> Size then begin
    NewValue := Value;
    if NewValue = 0 then Value := GetSystemMetrics(SysConsts[fKind]);
    fSize := Value;
    fControl.UpdateScrollBars;
    if NewValue = 0 then FSize := 0;
  end;
end;

procedure TCustomScrollBar.SetVisible(Value: boolean);
var
  Code: word;
begin
  fVisible:=Value;
  Code:=SB_HORZ;
  if fKind=sbVertical then Code:=SB_VERT;
  if fVisible and fControl.HandleAllocated then
    ShowScrollBar(fControl.Handle, Code, Visible)
  else
    fControl.UpdateScrollBars;
end;

procedure TCustomScrollBar.SetEnabled(Value: boolean);
var
  Code: word;
begin
  fEnabled:=Value;
  Code:=SB_HORZ;
  if fKind=sbVertical then Code:=SB_VERT;
  if fControl.HandleAllocated then
    if fEnabled then
      EnableScrollBar(fControl.Handle,Code,ESB_ENABLE_BOTH)
    else
      EnableScrollBar(fControl.Handle,Code,ESB_DISABLE_BOTH);
end;
{$EndRegion}
//########################################################################
{$Region ' THighlighter        '}
constructor TSCharHighlight.Create;
begin
  fFontColor := clWindowText;
  fBackColor := clNone;
  fFontStyle := [];
end;

destructor TSCharHighlight.Destroy;
begin
  inherited;
end;

constructor TSCustomHighlight.Create(AOwner: TComponent);
begin
  inherited;
end;

destructor TSCustomHighlight.Destroy;
begin
  inherited;
end;

procedure TSCustomHighlight.GetHighlitedParam(RichMemo: TCustomSyntaxMemo);
begin
  inherited;
end;

procedure TSCustomHighlight.SetMemoColor(RichMemo: TCustomSyntaxMemo; Tok : TToken; MPColor : TSCharHighlight);
begin
  RichMemo.SyntaxAttributes.FontColor[Tok]    := MPColor.FontColor;
  IF MPColor.BackColor<>clNone THEN
    RichMemo.SyntaxAttributes.BackColor[Tok]  := MPColor.BackColor;
  RichMemo.SyntaxAttributes.FontStyle[Tok]    := MPColor.FontStyle;
end;
{$EndRegion}
//########################################################################
{$Region ' TSyntaxAttributes '}
constructor TSyntaxAttributes.Create(Owner: TCustomSyntaxMemo);
var
  T: TToken;
begin
  inherited Create;
  fRichMemo := Owner;
  for T := Low(TToken) to High(TToken) do
    with fTokenStyles[T] do begin
      tsForeground    := clDefault;
      tsBackground    := clDefault;
      tsStyle         := [];
    end;
end;

procedure TSyntaxAttributes.Assign(Friend: TSyntaxAttributes);
begin
  if (self = Friend) or (Friend = nil) then Exit;
  fRichMemo.Lines.BeginUpdate;
  Move(Friend.fTokenStyles, fTokenStyles, SizeOf(fTokenStyles));
  fRichMemo.Lines.State := fRichMemo.Lines.State + [ssNeedReparseAll];
  fRichMemo.Lines.EndUpdate;
end;

function TSyntaxAttributes.Equals(const T1, T2: TToken): Boolean;
begin
  Result := (fTokenStyles[T1].tsForeground = fTokenStyles[T2].tsForeground)
    and (fTokenStyles[T1].tsBackground = fTokenStyles[T2].tsBackground)
    and (fTokenStyles[T1].tsStyle = fTokenStyles[T2].tsStyle);
end;

function TSyntaxAttributes.GetColor(const Token: TToken; const Index: Integer): TColor;
begin
  if Index = 0
    then Result := fTokenStyles[Token].tsForeground
    else Result := fTokenStyles[Token].tsBackground;
end;

function TSyntaxAttributes.GetStyle(const Token: TToken): TFontStyles;
begin
  Result := fTokenStyles[Token].tsStyle;
end;

procedure TSyntaxAttributes.SetColor(const Token: TToken; const Index: Integer; const Value: TColor);
begin
  if Index = 0
    then fTokenStyles[Token].tsForeground := Value
    else fTokenStyles[Token].tsBackground := Value;
end;

procedure TSyntaxAttributes.SetStyle(const Token: TToken; const Value: TFontStyles);
begin
  fTokenStyles[Token].tsStyle := Value;
end;

procedure TSyntaxAttributes.CopyAttrs(const SrcToken: TToken; DstTokArray: array of TToken);
var
  T: TToken;
begin
  for T := Low(DstTokArray) to High(DstTokArray) do
    fTokenStyles[DstTokArray[T]] := fTokenStyles[SrcToken];
end;
{$EndRegion}
//########################################################################
{$Region ' TSyntaxParser     '}
constructor TSyntaxParser.Create(const AsCloneOf: TSyntaxParser = nil);
begin
  inherited Create(True);
  if Assigned(AsCloneOf) then Assign(AsCloneOf);
end;

procedure TSyntaxParser.AddToken(const Beg, Len: Integer; Token: TToken);
var
  W: TSyntaxToken;
begin
  W := TSyntaxToken.Create;
  W.stStart := Word(Beg - 1);
  W.stLength := Word(Len);
  W.stToken := Token;
  W.stStyle := [];
  Add(W);
end;

function TSyntaxParser.GetToken(const TokIndex: Integer): TSyntaxToken;
begin
  Result := TSyntaxToken( inherited Items[TokIndex] );
end;

procedure TSyntaxParser.SetToken(const TokIndex: Integer; const Value: TSyntaxToken);
begin
  Items[TokIndex] := Value;
end;

procedure TSyntaxParser.Assign(const Friend: TSyntaxParser);
var
  i: Integer;
  T, NewT: TSyntaxToken;
begin
  Clear;
  for i := 0 to Friend.Count - 1 do begin
    T := Friend.Tokens[i];
    NewT := TSyntaxToken.Create;
    NewT.stStart    := T.stStart;
    NewT.stLength   := T.stLength;
    NewT.stToken    := T.stToken;
    NewT.stStyle    := T.stStyle;
    ADd(NewT);
  end;
  fSection      := Friend.Section;
  fVisibleIndex := Friend.VisibleIndex;
  fNeedReparse  := Friend.NeedReparse;
end;

procedure TSyntaxParser.Clear;
begin
  inherited Clear;
  fNeedReparse := False;
end;

procedure TSyntaxParser.SplitTokens(const sx, ex: Integer);
var
  wi: Integer;
  T: TSyntaxToken;

  function TestSel(const r: Integer): Boolean;
  var
    TT: TSyntaxToken;
  begin
    with T do begin
      Result := InRange(r, stStart + 1, stStart + stLength - 1);
      if Result then begin
        TT := TSyntaxToken.Create;
        TT.stStart  := r;
        TT.stLength := stStart + stLength - r;
        TT.stToken  := stToken;
        stLength := r - stStart;
        Insert(wi + 1, TT);
      end
    end
  end;

begin
  wi := 0;
  while wi < Count do begin
    T := Tokens[wi];
    if not TestSel(sx) then TestSel(ex);
    if InRange(T.stStart, sx, ex - 1) then
      Include(T.stStyle, stsInSelection)
    else
      Exclude(T.stStyle, stsInSelection);
    Inc(wi);
  end;
end;

function TSyntaxParser.FirstToken: TToken;
begin
  if Count > 0 then
    Result := Tokens[0].stToken
  else
    Result := tokBlank;
end;

function TSyntaxParser.LastToken: TToken;
begin
  if Count > 0  then
    Result := Tokens[Count - 1].stToken
  else
    Result := tokBlank;
end;

function TSyntaxParser.Parse(Line: string; LineIndex: Integer; LastToken: TToken; PA: TSyntaxAttributes): TToken;
begin
  Result:=tokBlank;
  if Assigned(PA.fRichMemo.fHighliter) then
    Result:=PA.fRichMemo.fHighliter.Parse(Self,Line,LineIndex,LastToken,PA)
  else begin
    Clear;
    AddToken(1, Length(Line), tokText);
  end;
  fNeedReparse := False;
end;
{$EndRegion}
//########################################################################
{$Region ' TSynMemoStrings   '}
constructor TSynMemoStrings.Create(const Owner: TCustomSyntaxMemo);
begin
  inherited Create;
  fRichMemo   := Owner;
  fState      := [];
  Inc(GlobalUntitledIndex);
end;

destructor TSynMemoStrings.Destroy;
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do Objects[i].Free;
  inherited;
end;

procedure TSynMemoStrings.Clear;
var
  i: Integer;
  Da: Boolean;
begin
  BeginUpdate;
  Da := fDirectAccess;
  fDirectAccess := True;
  for i := Count - 1 downto 0 do Objects[i].Free;
  inherited Clear;
  fRichMemo.Sections.Scan;
  fRichMemo.Reset;
  fDirectAccess := Da;
  fState := fState + [ssNeedReparseAll, ssNeedReIndex];
  SetModified(True);
  EndUpdate;
end;

function TSynMemoStrings.PositionToRC(Value: Integer): TPoint;
var
  i: Integer;
begin
  for i := 0 to Count - 1 do
    if Value < Length(Get(i)) + 2 then begin
      Result := Point(Value, i);
      Exit;
    end else
      Dec(Value, Length(Get(i)) + 2);
  Result.Y := Count - 1;
  Result.X := Length(Get(Result.Y));
end;

function TSynMemoStrings.RCToPosition(Col, Row: Integer): Integer;
var
  i: Integer;
begin
  Result := 0;
  for i := 0 to Row - 1 do Inc(Result, Length(Get(i)) + 2);
  Inc(Result, Col);
end;

procedure TSynMemoStrings.Put(Index: Integer; const s: string);
begin
  if fDirectAccess then
    inherited Put(Index, s)
  else begin
    BeginUpdate;
    inherited Put(Index, s);
    Parser[Index].NeedReparse := True;
    Include(fState, ssTextChanged);
    EndUpdate;
  end;
end;

procedure TSynMemoStrings.InsertItem(Index: Integer; const s: string; AObject: TObject);
begin
  if fDirectAccess then
    inherited InsertItem(Index, s, TSyntaxParser.Create)
  else begin
    BeginUpdate;
    inherited InsertItem(Index, s, TSyntaxParser.Create);
    Parser[Index].NeedReparse := True;
    Include(fState, ssTextChanged);
    if not (ssUndoProcess in fState) then fRichMemo.Sections.InsertRow(Index);
    fRichMemo.NeedRedrawLE(Index);
    EndUpdate;
  end;
end;

procedure TSynMemoStrings.Delete(Index: Integer);
begin
  if DirectAccess then
    inherited Delete(Index)
  else begin
    BeginUpdate;
    if not (ssUndoProcess in fState) then fRichMemo.Sections.DeleteRow(Index);
    if Assigned(Objects[Index]) then Objects[Index].Free;
    inherited Delete(Index);
    Parser[Index].NeedReparse := True;
    Include(fState, ssTextChanged);
    fRichMemo.NeedReDrawLE(Index);
    EndUpdate;
  end;
end;

procedure TSynMemoStrings.Changed;
begin
end;

procedure TSynMemoStrings.SetUpdateState(Updating: Boolean);
begin
  inherited;
  if Updating then begin
    fRichMemo.HideCaret;
    fState := fState - [ssTextChanged, ssSectionsChanged, ssNeedReIndex, ssNeedReparseAll];
    fRichMemo.Change([]);
  end else begin
    if fState * [ssTextChanged, ssSectionsChanged] <> [] then SetModified(True);
    if ssNeedReIndex in fState then fRichMemo.Sections.ReIndex;
    if fState * [ssTextChanged, ssNeedReparseAll] <> [] then Parse(ssNeedReparseAll in fState, True);
    fRichMemo.ReDraw;
    if not (csDesigning in fRichMemo.ComponentState) then fRichMemo.UpdateScrollBars;
  end;
end;

procedure TSynMemoStrings.Parse(const EntireText : Boolean; const NeedRepaint: Boolean = False);
var
  Row: Integer;
  NeedNext: Boolean;
  Sp: TSyntaxParser;
begin
  NeedNext := False;
  Row := 0;
  while Row < Count do begin
    Sp := Parser[Row];
    if EntireText or Sp.NeedReparse then begin
      NeedNext := Self.ParseLine(Row, not EntireText);
      if NeedRepaint then fRichMemo.NeedRedraw(Row);
    end else if NeedNext and (Sp.Count <> 0) then begin
      NeedNext := Self.ParseLine(Row, True);
      if NeedRepaint then fRichMemo.NeedRedraw(Row);
    end;
    Inc(Row);
  end;
  fState := fState - [ssTextChanged, ssNeedReparseAll];
end;

function TSynMemoStrings.ParseLine(const Index: Integer; const TestNextLine: Boolean): Boolean;
var
  i  : integer;
  Key: TToken;
begin
  Result := False;
  if Index >= Count then Exit;
  Key := tokBlank;

  for i := Index - 1 downto 0 do
    with Parser[i] do
      if Count > 0 then begin
        if LastToken in TokenBlock then Key := LastToken;
        Break;
      end;

  Key := Parser[Index].Parse(Get(Index),Index,Key,fRichMemo.fParseAttributes);

  if TestNextLine then
    for i := Index + 1 to Count - 1 do
      with Parser[i] do
        if Count > 0 then begin
          Result := ((Key in TokenBlock)        and (Key<>FirstToken))
                 or ((FirstToken in TokenBlock) and (Key<>FirstToken))
                 ;//or (Key=FirstToken); //!!!!!!!
          Break;
        end;
end;

procedure TSynMemoStrings.LoadFromStream(Stream: TStream);
begin
  BeginUpdate;
  fDirectAccess := True;
  try
    inherited LoadFromStream(Stream);
  finally
    fDirectAccess := False;
    fRichMemo.Sections.Scan;
    fRichMemo.NeedRedrawAll;
    fState := [ssNeedReIndex, ssNeedReparseAll];
    EndUpdate;
  end;
  fRichMemo.Change([ciText, ciSelection, ciSections, ciUndoStack]);
end;

procedure TSynMemoStrings.SaveToStream(Stream: TStream);
var
  Sl: TStringList;
begin
  Sl := TStringList.Create;
  Sl.Assign(self);
  if smoWriteMarkersOnSave in fRichMemo.Options then fRichMemo.Sections.FillOutput(Sl);
  try
    Sl.SaveToStream(Stream);
  finally
    Sl.Free;
  end;
  SetModified(False);
  fRichMemo.Change([ciText]);
end;

procedure TSynMemoStrings.LoadFromFile(const FileName: string);
begin
  inherited;
  SetModified(False);
  InvalidateRect(fRichMemo.Handle,nil,false);
  fRichMemo.CalcScreenParams;
end;

procedure TSynMemoStrings.LoadFromUFile(const FileName: string);

var
  ms: TMemoryStream;

  procedure Convert(var Stream: TMemoryStream);
  var
    Buff: CHAR;
    S: TMemoryStream;
    C,OC: Char;
    i,pos: integer;
    ch: char;
    F: boolean;
  begin
    Stream.Position:=0;
    S:=TMemoryStream.Create;
    C:=#0;

    Stream.Read(C,1);

    if (C=#$F7) then begin
      Stream.Read(ch,1);
      i := 0;
      repeat
        Stream.Read(ch,1);
        INC(i)
      until ch = #0;
      Stream.Read(i,4);
      Stream.Read(i,4);
      Stream.Read(ch,1);
      if ch = #$F7 then begin
        Stream.Read(ch,1);
        if ch = #$08 then begin
          Stream.Read(i,4);
          Stream.Position:=(Stream.Position+i);
          Stream.Read(ch,1);
        end;
      end;
      pos:=Stream.Position;
      if (ch=#$F0) or (ch=#$01) then begin
        Stream.Read(ch,1);
        Stream.Read(i,4);
        Stream.Position:=pos-1+i-4;
        Stream.Read(i,4);
        Stream.Read(C,1);
      end;
    end;

    if C>=#32 then S.Write(C,1);
    F:=false;

    while Stream.Position<Stream.Size do begin
      OC:=C;
      Stream.Read(C,1);
      case C of
        #13:
            begin
              S.Write(C,1);
              F:=true;
            end;
        #10:
            begin
              if (OC<>#13) and not F then begin
                Buff:=#13;
                S.Write(Buff,1);
                F:=false;
              end;
              S.Write(C,1);
            end;
        #9:
            begin
              Buff:=#160;
              i:=0;
              while i<TABlen do begin
                S.Write(Buff,1);
                INC(i);
              end;
            end;
      else
        if C>=#32 then S.Write(C,1);
      end;

    end;
    S.Position:=0;
    Stream:=S;
    Stream.Position:=0; 
  end;

begin
  ms := TMemoryStream.Create;
  ms.LoadFromFile(filename);
  Convert(ms);
  self.LoadFromStream(ms);
  fRichMemo.fVScroll.Position:=0;
  fRichMemo.CalcScreenParams;
  SetModified(False);
end;

procedure TSynMemoStrings.SaveToFile(const NewFileName: string);
var
  ms: TMemoryStream;

  procedure Convert(var Stream: TMemoryStream);
  var
    Buff: CHAR;
    S: TMemoryStream;
    C: Char;
    i: integer;
  begin
    Stream.Position:=0;
    S:=TMemoryStream.Create;
    C:=#0;
    while Stream.Position<Stream.Size do begin
      Stream.Read(C,1);
      case C of
        #160:
            begin
              Buff:=#9;
              i:=1;
              while C=#160 do begin
                Stream.Read(C,1);
                INC(i);
                if (i mod TABlen)=0 then S.Write(Buff,1);
              end;
              S.Write(C,1);
            end;
      else
        S.Write(C,1);
      end;
    end;
    Stream:=S;
    Stream.Position:=0;
  end;

begin
  ms := TMemoryStream.Create;
  self.SaveToStream(ms);
  Convert(ms);
  ms.SaveToFile(NewFileName);
  //self.LoadFromStream(ms);
  //inherited;
  SetModified(False);
end;

procedure TSynMemoStrings.New;
begin
  Clear;
  fRichMemo.Reset;
  Inc(GlobalUntitledIndex);
  fState := [];
  fRichMemo.Change([ciText, ciSelection, ciSections, ciUndoStack]);
end;

function TSynMemoStrings.IsValidLineIndex(const Row: Integer): Boolean;
begin
  Result := InRange(Row, 0, Count - 1);
end;

function TSynMemoStrings.GetParser(const Row: Integer): TSyntaxParser;
begin
  Result := TSyntaxParser(Objects[Row]);
end;

procedure TSynMemoStrings.SetModified(const Value: Boolean);
begin
  if Value <> fModified then begin
    fModified := Value;
    fRichMemo.Change([ciText]);
  end;
end;
{$EndRegion}
//########################################################################
{$Region ' TSMSection        '}
procedure TSMSectionClone.AddRef;
begin
  Inc(fRefCount);
end;

procedure TSMSectionClone.Release;
begin
  Dec(fRefCount);
  if fRefCount <= 0 then Free;
end;

constructor TSynMemoSection.Create;
begin
  inherited Create(True);
end;

function TSynMemoSection.GetSections(const Idx: Integer): TSynMemoSection;
begin
  Assert(InRange(Idx, 0, Count - 1), 'Bad nested section index: ' + IntToStr(Idx));
  Result := TSynMemoSection(Items[Idx])
end;

procedure TSynMemoSection.SetLevel(const Value: Integer);
var
  i: Integer;
begin
  fLevel := Value;
  for i := 0 to Count - 1 do Sections[i].SetLevel(fLevel + 1);
end;

constructor TSynMemoSections.Create(Owner: TCustomSyntaxMemo);
begin
  inherited Create;
  fRichMemo := Owner;
  fRoot     := TSMSectionClone.Create;
  fIndexes  := TList.Create;
  fRoot.AddRef;
  Scan;
end;

destructor TSynMemoSections.Destroy;
begin
  fIndexes.Free;
  fRoot.Free;
  inherited;
end;

class function TSynMemoSections.DetectSectionMark(const s: string): TSectionMark;
begin
  if s = '' then Result := smNone else
  if PDWORD(s)^ = PDWORD(SectionMarks[smExpanded])^ then Result := smExpanded else
  if PDWORD(s)^ = PDWORD(SectionMarks[smCollapsed])^ then Result := smCollapsed else
  if PDWORD(s)^ = PDWORD(SectionMarks[smEnd])^ then Result := smEnd else Result := smNone;
end;

function TSynMemoSections.SectionBorder(const Row: Integer): TSectionMark;
begin
  with Section[Row] do
    if Row = RowBeg then
      if Collapsed
        then Result := smCollapsed
        else Result := smExpanded
    else
      if Row = RowEnd
        then Result := smEnd
        else Result := smNone;
end;

function TSynMemoSections.Next(Sec: TSynMemoSection): TSynMemoSection;

    function _next(Sec: TSynMemoSection): TSynMemoSection;
    var
      n: Integer;
    begin
      if Sec = fRoot then
        Result := nil
      else begin
        n := Sec.Parent.IndexOf(Sec);
        if n < Sec.Parent.Count - 1 then
          Result := Sec.Parent.Sections[n + 1]
        else
          Result := _next(Sec.Parent);
      end;
    end;

begin
  if Sec.Count > 0 then
    Result := TSynMemoSection(Sec.First)
  else
    Result := _next(Sec);
end;

function TSynMemoSections.Prev(Sec: TSynMemoSection): TSynMemoSection;

    function _last(Sec: TSynMemoSection): TSynMemoSection;
    begin
      Result := Sec;
      if Result.Count > 0 then Result := _last(TSynMemoSection(Result.Last));
    end;

var
  n: Integer;
begin
  if Sec = fRoot then
    Result := nil
  else begin
    n := Sec.Parent.IndexOf(Sec);
    if n > 0 then
      Result := _last(Sec.Parent.Sections[n - 1])
    else
      if Sec.Parent = fRoot then
        Result := nil
      else
        Result := Sec.Parent;
  end;
end;

function TSynMemoSections.GetSection(const Row: Integer): TSynMemoSection;
begin
  Result := fRichMemo.Lines.Parser[Row].Section;
end;

procedure TSynMemoSections.SetSection(const Row: Integer; Value: TSynMemoSection);
begin
  fRichMemo.Lines.Parser[Row].Section := Value;
end;

function TSynMemoSections.Visible(const Sec: TSynMemoSection): Boolean;
begin
  if ssNeedReIndex in fRichMemo.Lines.State then ReIndex;
  Result := fRichMemo.Lines.Parser[Sec.RowBeg].VisibleIndex >= 0;
end;

procedure TSynMemoSections.Collapse(const Row: Integer; const Recursive, SafeSelf: Boolean);
var
  Sec: TSynMemoSection;

  procedure CollapseChildren(Father: TSynMemoSection);
  var
    i: Integer;
  begin
    if Recursive then
      for i := Father.Count - 1 downto 0 do CollapseChildren(Father[i]);
    if (Father.Level = 0) or ((Father = Sec) and SafeSelf) then Exit;
    Father.Collapsed := True;
  end;

begin
  with fRichMemo.Lines do begin
    if ssNeedReindex in State then ReIndex;
    BeginUpdate;
    Sec := Section[Row];
    CollapseChildren(Sec);
    State := State + [ssNeedReIndex];
    fRichMemo.NeedRedrawAll;
    EndUpdate;
  end;
  fRichMemo.Change([ciSections]);
end;

procedure TSynMemoSections.Expand(const Row: Integer; const Recursive, ParentRecursive: Boolean);

  procedure ExpandChildren(Father: TSynMemoSection);
  var
    i: Integer;
  begin
      Father.Collapsed := False;
      if Recursive then
        for i := Father.Count - 1 downto 0 do ExpandChildren(Father[i]);
  end;

var
  Sec: TSynMemoSection;
begin
  if ssNeedReindex in fRichMemo.Lines.State then ReIndex;
  with fRichMemo.fLines do begin
    BeginUpdate;
    Sec := Self.Section[Row];
    ExpandChildren(Sec);
    if ParentRecursive then
      while Sec.Level > 1 do begin
        Sec := Sec.Parent;
        Sec.Collapsed := False;
      end;
    State := State + [ssNeedReIndex];
    fRichMemo.NeedRedrawAll;
    EndUpdate;
  end;
  fRichMemo.Change([ciSections]);
end;

procedure TSynMemoSections.Explode(const Row: Integer; const Recursive: Boolean);

    procedure ExplodeChildren(Father: TSynMemoSection);
    var
      i, n: Integer;
      Child: TSynMemoSection;
    begin
      if Father.Level > 0 then begin
        n := Father.Parent.IndexOf(Father);
        for i := Father.Count-1 downto 0 do begin
          Child := Father[i];
          Child.Level := Child.Level - 1;
          Child.Parent := Father.Parent;
          Father.Parent.Insert(n+1, Father.Extract(Child));
          if Recursive then ExplodeChildren(Child);
        end;
        for i := Father.RowBeg to Father.RowEnd do
          with fRichMemo.Lines.Parser[i] do
            if Section = Father then Section := Father.Parent;
        Father.Parent.Delete(n);
      end;
    end;

begin
  if ssNeedReindex in fRichMemo.Lines.State then ReIndex;
  MakeUnique;
  with fRichMemo.Lines do begin
    BeginUpdate;
    ExplodeChildren(Section[Row]);
    State := State + [ssSectionsChanged, ssNeedReIndex];
    fRichMemo.NeedRedrawAll;
    SetModified(True);
    EndUpdate;
  end;
  fRichMemo.Change([ciText, ciSelection, ciSections]);
end;

function TSynMemoSections.New(const Row1, Row2: Integer; const IsCollapsed: Boolean = False): TSynMemoSection;
var
  i: Integer;
  Father, iSec: TSynMemoSection;
begin
    if ssNeedReindex in fRichMemo.Lines.State then ReIndex;
    MakeUnique;
    if ((Section[Row2] = Section[Row1].Parent) and (SectionBorder(Row2) = smNone))
    or ((Section[Row1] = Section[Row2].Parent) and (SectionBorder(Row1) = smNone)) then
      Explode(Row1, False);
    if (Section[Row1] <> Section[Row2])
    or (SectionBorder(Row1) <> smNone)
    or (SectionBorder(Row2) <> smNone) then begin
      Result := nil;
      Exit;
    end;
    fRichMemo.Lines.BeginUpdate;
    Father := Section[Row1];
    Result := TSynMemoSection.Create;
    Result.Parent    := Father;
    Result.RowBeg    := Row1;
    Result.RowEnd    := Row2;
    Result.Level     := Father.Level + 1;
    Result.Collapsed := IsCollapsed;
    for i := Row1 to Row2 do begin
      iSec := Section[i];
      if iSec = Father then
        Section[i] := Result
      else if (i = iSec.RowBeg) and (iSec.Parent = Father) then begin
        iSec.Level := Result.Level + 1;
        iSec.Parent := Result;
        Result.Add(Father.Extract(iSec));
      end;
    end;

    i := Father.Count;
    while (i > 0) and (Father[i-1].RowBeg > Row1) do Dec(i);
    Father.Insert(i, Result);

    with fRichMemo.Lines do begin
      State := State + [ssSectionsChanged, ssNeedReIndex];
      fRichMemo.NeedRedrawAll;
      SetModified(True);
      EndUpdate;
    end;

    fRichMemo.Change([ciText, ciSelection, ciSections]);
end;

procedure TSynMemoSections.DeleteRow(const Row: Integer);

    procedure UpdateIndexes(Sec: TSynMemoSection);
    var
      i: Integer;
    begin
      if Sec.RowBeg > Row then Dec(Sec.fRowBeg);
      if Sec.RowEnd > Row then begin
        Dec(Sec.fRowEnd);
        for i := 0 to Sec.Count - 1 do UpdateIndexes( Sec.Sections[i] );
      end;
    end;

begin
  MakeUnique;
  with fRichMemo do
    if Sections.SectionBorder(Row) <> smNone then Explode(Row, False);
  UpdateIndexes(fRoot);
  with fRichMemo.Lines do
    State := State + [ssSectionsChanged, ssNeedReIndex];
end;

procedure TSynMemoSections.InsertRow(const Row: Integer);
var
  ParentSec: TSynMemoSection;

  procedure UpdateIndexes(Sec: TSynMemoSection);
  var
    i: Integer;
  begin
    if Row <= Sec.RowBeg then Inc(Sec.fRowBeg);
    if Row <= Sec.RowEnd then begin
      Inc(Sec.fRowEnd);
      if InRange(Row, Sec.RowBeg, Sec.RowEnd) and (Sec.Level > ParentSec.Level) then ParentSec := Sec;
      for i := 0 to Sec.Count - 1 do UpdateIndexes( Sec.Sections[i] );
    end;
  end;

begin
  MakeUnique;
  ParentSec := fRoot;
  UpdateIndexes(fRoot);
  Section[Row] := ParentSec;
  with fRichMemo.Lines do State := State + [ssSectionsChanged, ssNeedReIndex];
end;

procedure TSynMemoSections.FillOutput(const Sl: TStringList);
const
  pm: array[Boolean] of string[4] = ('{<+}', '{<-}');

    procedure MarkSection(Sec: TSynMemoSection);
    var
      i: Integer;
    begin
      if Sec <> fRoot then begin
        Sl[Sec.RowBeg] := pm[Sec.Collapsed] + Sl[Sec.RowBeg];
        Sl[Sec.RowEnd] := '{>>}' + Sl[Sec.RowEnd];
      end;
      for i := 0 to Sec.Count - 1 do MarkSection(Sec.Sections[i]);
    end;

begin
  MarkSection(fRoot);
end;

procedure TSynMemoSections.MakeUnique;

    function Clone(const Father, Sec: TSynMemoSection): TSynMemoSection;
    var
      i: Integer;
    begin
      Result := TSynMemoSection.Create;
      Result.fParent    := Father;
      Result.fRowBeg    := Sec.fRowBeg;
      Result.fRowEnd    := Sec.fRowEnd;
      Result.fLevel     := Sec.fLevel;
      Result.fCollapsed := Sec.fCollapsed;
      for i := 0 to Sec.Count - 1 do Result.Add( Clone(Result, Sec.Sections[i]) );
    end;

var
  NewRoot: TSMSectionClone;
  i: Integer;
begin
  if ssUndoProcess in fRichMemo.Lines.State then Exit;
  if fRoot.fRefCount > 1 then begin
    NewRoot := TSMSectionClone.Create;
    NewRoot.fParent    := nil;
    NewRoot.fRowBeg    := fRoot.fRowBeg;
    NewRoot.fRowEnd    := fRoot.fRowEnd;
    NewRoot.fLevel     := 0;
    NewRoot.fCollapsed := False;
    for i := 0 to fRoot.Count - 1 do NewRoot.Add( Clone(NewRoot, fRoot.Sections[i]) );
    SetRoot(NewRoot);
  end;
end;

function TSynMemoSections.AsText: string;
var
  sl: TStringList;

    procedure SecAsString(Sec: TSynMemoSection);
    var i: Integer;
    begin
      sl.Append(Format('%s%d..%d %s',
         [StringOfChar(' ', Sec.Level * 4),
          Sec.RowBeg,
          Sec.RowEnd,
          ifthen(Sec.Collapsed, 'Collapsed')]));
      for i := 0 to Sec.Count - 1 do SecAsString(Sec.Sections[i]);
    end;

begin
  sl := TStringList.Create;
  sl.Append('Sections');
  SecAsString(fRoot);
  sl.Append('End of sections');
  Result := sl.Text;
  sl.Free;
end;

procedure TSynMemoSections.ReIndex;
var
  Row: Integer;

    procedure ProcessSection(const Sec: TSynMemoSection; ParentOpen: Boolean);
    var
      ChildIndex: Integer;
    begin
      ChildIndex := 0;
      if Sec.Level > fMaxLevel then fMaxLevel := Sec.fLevel;
      if ParentOpen and (Sec.fLevel > fMaxExpandLevel) then fMaxExpandLevel := Sec.fLevel;
      while Row <= Sec.fRowEnd do begin
        if (ChildIndex >= Sec.Count)
        or (Row < Sec.Sections[ChildIndex].RowBeg)
        or (Row > Sec.Sections[ChildIndex].RowEnd) then begin
          with fRichMemo.Lines.Parser[Row] do begin
            Section := Sec;
            if ParentOpen and (not Sec.Collapsed or (Row = Sec.RowBeg)) then
              fVisibleIndex := fIndexes.Add(Pointer(Row))
            else
              fVisibleIndex := -1;
          end;
          Inc(Row);
        end else begin
          ProcessSection(Sec.Sections[ChildIndex], ParentOpen and not Sec.Collapsed);
          Inc(ChildIndex);
        end;
      end;
    end;

begin
  Dec(fRoot.fRowEnd);
  fMaxLevel       := 0;
  fMaxExpandLevel := 0;
  Row             := 0;
  fIndexes.Clear;
  ProcessSection(fRoot, True);
  Inc(fRoot.fRowEnd);
  fRichMemo.Lines.State := fRichMemo.Lines.State - [ssNeedReIndex];
  fRichMemo.Change([ciSections]);
end;

procedure TSynMemoSections.SetRoot(const Value: TSMSectionClone);
begin
  if fRoot <> Value then begin
    fRoot.Release;
    fRoot := Value;
    fRoot.AddRef;
  end;
  ReIndex;
end;

procedure TSynMemoSections.Scan;
var
  Row, i: Integer;
  ParentSec, Sec: TSynMemoSection;
  Stack: TObjectStack;
  Sm: TSectionMark;
begin

  with fRoot do begin
    Clear;
    fRowBeg     := -1;
    fRowEnd     := fRichMemo.fLines.Count;
    fLevel      := 0;
    fParent     := nil;
    fCollapsed  := False;
  end;

  Stack := TObjectStack.Create;

  {         }
  Row := 0;
  ParentSec := fRoot;
  repeat
      while Row < fRichMemo.Lines.Count do begin
          Sm := TSynMemoSections.DetectSectionMark(fRichMemo.Lines[Row]);
          case Sm of
            smExpanded,
            smCollapsed:
                          begin
                            Sec := TSynMemoSection.Create;
                            with Sec do begin
                              fRowBeg     := Row;
                              fRowEnd     := -1;
                              fLevel      := ParentSec.fLevel + 1;
                              fParent     := ParentSec;
                              fCollapsed  := Sm = smCollapsed;
                            end;
                            ParentSec.Add(Sec);
                            Stack.Push(Sec);
                            ParentSec := Sec;
                            fRichMemo.Sections.Section[Row] := ParentSec;
                            fRichMemo.Lines[Row] := StuffString(fRichMemo.Lines[Row], 1, SECTION_HEADER_LENGTH, '');
                          end;
            smEnd:
                          begin
                            if Stack.Count = 0 then begin
                              fRichMemo.fLines.Insert(Row, SectionMarks[smExpanded]);
                              Continue;
                            end;
                            Sec := TSynMemoSection(Stack.Pop);
                            Sec.fRowEnd := Row;
                            fRichMemo.Sections.Section[Row] := Sec;
                            ParentSec := Sec.fParent;
                            fRichMemo.Lines[Row] := StuffString(fRichMemo.Lines[Row], 1, SECTION_HEADER_LENGTH, '');
                          end;
          else
            fRichMemo.Sections.Section[Row] := ParentSec;
          end;
          Inc(Row);
      end;

      if Stack.Count > 0 then
        for i := 0 to Stack.Count - 1 do
          fRichMemo.Lines.Append(SectionMarks[smEnd]);
  until Stack.Count = 0;

  Stack.Free;
end;
{$EndRegion}
//########################################################################
{$Region ' TSynMemoRange     '}
destructor TSynMemoUndoItem.Destroy;
begin
  uiSections.Release;
  inherited;
end;

constructor TSynMemoRange.Create(Owner: TCustomSyntaxMemo);
begin
  inherited Create;
  fRichMemo     := Owner;
  fSealing      := True;
  fMaxUndoDepth := 100;
  fUndoStack := TObjectList.Create(True);
  fUndoStack.Capacity := 100;
end;

destructor TSynMemoRange.Destroy;
begin
  fUndoStack.Free;
  inherited;
end;

//     ***SAT
procedure TSynMemoRange.MakeIndent;
var
  i,j   : Integer;
  S     : String;
begin
  fRichMemo.Lines.BeginUpdate;
  for i := StartY to EndY do begin
    j:=1;
    S:=fRichMemo.Lines[i];
    if S<>'' then
      while ((S[j]<#32) or (S[j]=#160)) and (j<Length(S)) do begin
        S[j]:=#160;
        Inc(j);
      end;
    fRichMemo.Lines[i]:=StringOfChar(#160,TABlen)+S;
  end;
  EndX:=EndX+TABlen;
  StartX:=StartX+TABlen;
  PosX:=PosX+TABlen;
  if StartX<0 then StartX:=0;
  if EndX<0 then EndX:=0;
  fRichMemo.Lines.EndUpdate;
end;

//      ***SAT
procedure TSynMemoRange.MakeUnIndent;
var
  i,k: Integer;

  procedure CheckPos(Flag: boolean);
  begin
    if Flag then begin
      if i=StartY then Inc(k,TABlen);
    end else begin
      if i=EndY then Inc(k,TABlen);
    end;
  end;

begin
  fRichMemo.Lines.BeginUpdate;
  k:=0;
  for i:=StartY to EndY do begin
    if fRichMemo.Lines[i]='' then Continue;
    if copy(fRichMemo.Lines[i],1,TABlen)=StringOfChar(#160,TABlen) then begin
      fRichMemo.Lines[i] := copy(fRichMemo.Lines[i],TABlen+1,length(fRichMemo.Lines[i])-TABlen);
      CheckPos(StartY<EndY);
    end else if (fRichMemo.Lines[i][1]=' ') or (fRichMemo.Lines[i][1]=#160) then begin
      fRichMemo.Lines[i] := copy(fRichMemo.Lines[i],2,length(fRichMemo.Lines[i])-1);
      CheckPos(StartY<EndY);
    end;
  end;
  EndX:=EndX-k;
  StartX:=StartX-k;
  PosX:=PosX-k;
  if StartY>EndY then EndX:=PosX;
  if StartX<0 then StartX:=0;
  if EndX<0 then EndX:=0;
  fRichMemo.Lines.EndUpdate;
end;

procedure TSynMemoRange.Collapse;
var
  yb, ye: Integer;
  e: Boolean;
begin
  yb       := fStart.Y;
  ye       := fEnd.Y;
  e        := IsEmpty();
  fSealing := True;
  fStart   := fPos;
  fEnd     := fPos;
  if not e then
    with fRichMemo do begin
      Lines.BeginUpdate;
      repeat
        NeedRedraw(yb);
        Inc(yb);
      until yb > ye;
      Lines.EndUpdate;
    end;
  fRichMemo.Change([ciSelection]);
end;

//  !!!
procedure TSynMemoRange.Enlarge(const Value: Integer; const EnlargeLine: Boolean = False; const VisiblesOnly: Boolean = False);
var
  n, PrevY: Integer;
begin
  fRichMemo.Lines.BeginUpdate;
  fSealing := False;
  if not EnlargeLine then begin
    n := PosX + Value;
    if n < 0 then n := -PosX else
      if n > Length(fRichMemo.Lines[PosY]) then n := Length(fRichMemo.Lines[PosY]) - PosX else n := Value;
    if n <> 0 then begin
      if (PosX = StartX) and (PosY = StartY) then Inc(fStart.X, n) else Inc(fEnd.X, n);
      if (StartY = EndY) and (StartX > EndX) then Swap(fStart.X, fEnd.X);
      PosX := PosX + n;
      fRichMemo.NeedRedraw(PosY);
    end
  end else begin
    if VisiblesOnly then
      n := fRichMemo.FindVisibleRow(PosY, Value, True)
    else
      n := EnsureRange(PosY + Value, 0, fRichMemo.Lines.Count - 1);
    if n <> PosY then begin
      PrevY := PosY;
      if PointsEqual(fPos, fStart) then
        fStart.Y := n
      else
        fEnd.Y := n;
      if StartY > EndY then begin
        Swap(fStart.Y, fEnd.Y);
        Swap(fStart.X, fEnd.X);
      end;
      PosY := n;
      fRichMemo.fVScroll.Position:=fRichMemo.OffsetY*fRichMemo.fCharHeight;
      for n := Min(PrevY, PosY) to Max(PrevY, PosY) do fRichMemo.NeedRedraw(n);
    end;
  end;
  fRichMemo.Lines.EndUpdate;
  fRichMemo.Change([ciSelection]);
end;

procedure TSynMemoRange.Delete;
begin
  if not IsEmpty then
    SetText('')
  else if PosX < Length(fRichMemo.fLines[PosY]) then begin
    EndX := StartX + 1;
    SetText('');
  end else if PosY < fRichMemo.fLines.Count - 1 then begin
    EndX := 0;
    EndY := StartY + 1; //***
    SetText('');
  end;
end;

procedure TSynMemoRange.CollapseSection(const Recursive: Boolean);
begin
  with fRichMemo do begin
    if Sections.SectionBorder(PosY) in [smNone, smEnd] then SetPos( Point(0, Sections.Section[PosY].RowBeg) );
    Sections.Collapse(PosY, Recursive, Recursive);
  end;
end;

procedure TSynMemoRange.ExpandSection(const Recursive: Boolean);
begin
  fRichMemo.Sections.Expand(PosY, Recursive, False);
end;

procedure TSynMemoRange.ExplodeSection(const Recursive: Boolean);
begin
  Collapse;
  AddUndo;
  fRichMemo.Sections.Explode(PosY, Recursive);
end;

procedure TSynMemoRange.CreateSection;
begin
  with fRichMemo do begin
    Lines.BeginUpdate;
    if (StartY = EndY)
    and (Sections.SectionBorder(StartY) = smNone) then begin
      Collapse;
      with AddUndo() do begin
        uiSelStart := Point(0, StartY + 1);
        uiSelEnd   := Point(0, StartY + 2);
        uiSealing  := False;
      end;
      Lines.Insert(StartY + 1, '');
      Sections.New(StartY, StartY + 1);
    end else
    if (StartY <> EndY)
    and (Sections.SectionBorder(StartY) = smNone)
    and (Sections.SectionBorder(EndY) = smNone) then begin
      with AddUndo() do begin
        uiSelStart := fPos;
        uiSelEnd   := fPos;
        uiSealing  := True;
      end;
      Sections.New(fStart.Y, fEnd.Y);
      Collapse;
    end;
    SetPos(Point(0, fStart.Y));
    Lines.EndUpdate;
  end;
end;

procedure TSynMemoRange.GotoSection(const GoForward: Boolean);
var
  Sec: TSynMemoSection;
begin
    Sec := fRichMemo.Sections.Section[PosY];
    repeat
      if GoForward then
        Sec := fRichMemo.Sections.Next(Sec)
      else
        Sec := fRichMemo.Sections.Prev(Sec);
    until (Sec = nil) or fRichMemo.Sections.Visible(Sec);
    if Sec <> nil then SetPos( Point(0, Sec.RowBeg) );
end;

procedure TSynMemoRange.CopyToClipboard;
var
  Data: THandle;
  DataPtr: Pointer;
  s: string;

  function Convert(S: String): String;
  var
    i: integer;
  begin
    i:=1;
    Result:='';
    while i<=Length(S) do begin
      while S[i]=#160 do begin
        INC(i);
        if (i mod TABlen)=0 then Result:=Result+#9;
      end;
      Result:=Result+S[i];
      INC(i);
    end;
  end;

begin
  if IsEmpty then Exit;
  Clipboard.AsText := Convert(self.GetText);
  if not (smoSkipSectionsOnCopy in fRichMemo.Options)
  and (fStart.Y <> fEnd.Y) then begin
    s := Convert(GetMarkedText);
    OpenClipboard(Application.Handle);
    try
      Data := GlobalAlloc(GMEM_MOVEABLE + GMEM_DDESHARE, Length(s) + 1);
      try
        DataPtr := GlobalLock(Data);
        try
          Move(PChar(s)^, DataPtr^, Length(s) + 1);
          SetClipboardData(CF_SYNTAX, Data);
        finally
          GlobalUnlock(Data);
        end;
      except
        GlobalFree(Data);
        raise;
      end;
    finally
      CloseClipboard;
    end;
  end;
end;

procedure TSynMemoRange.PasteFromClipboard;
var
  Data: THandle;

  function Convert(S: String): String;
  var
    i: integer;
  begin
    i:=1;
    Result:='';
    while i<=Length(S) do begin
      if S[i]=#9 then
        Result:=Result+StringOfChar(#160,TABlen)
      else
        Result:=Result+S[i];
      INC(i);
    end;
  end;

begin
  if smoReadOnly in fRichMemo.fOptions then Exit;
  fRichMemo.Lines.BeginUpdate;
  if not (smoSkipSectionsOnPaste in fRichMemo.Options) and Clipboard.HasFormat(CF_SYNTAX) then begin
    OpenClipboard(Application.Handle);
    Data := GetClipboardData(CF_SYNTAX);
    try
      SetMarkedText(Convert(PChar(GlobalLock(Data))));
    finally
      GlobalUnlock(Data);
      CloseClipboard;
    end;
  end else if Clipboard.HasFormat(CF_TEXT) then
    SetText(Convert(Clipboard.AsText));
  fRichMemo.Lines.EndUpdate;
end;

procedure TSynMemoRange.CutToClipBoard;
begin
  if not IsEmpty and not (smoReadOnly in fRichMemo.fOptions) then begin
    CopyToClipboard;
    self.Text := '';
  end;
end;

//    HTML
procedure TSynMemoRange.CopyAsHTML;
var
  i,w,s,e : Integer;
  Sp      : TSyntaxParser;
  tok     : TToken;
  Txt,Word: String;
  St      : String;
  Color   : TColor;
  Style   : TFontStyles;
  Data    : THandle;
  DataPtr : Pointer;

  function TimesString(W: String; Len: integer): String;
  var
    i: integer;
  begin
    Result:='';
    i:=0;
    while i<Len do begin
      Result:=Result+W;
      INC(i);
    end;
  end;

  function RedOfColor(cl:TColor):Byte;
  begin
    Result:=cl mod 256;
  end;

  function GreenOfColor(cl:TColor):Byte;
  begin
    Result:=(cl mod 65536) div 256;
  end;

  function BlueOfColor(cl:TColor):Byte;
  begin
    Result:=cl div 65536;
  end;

  function ColorToStringRrGgBb(Color: TColor): string;
  var
    cl:TColor;
  begin
      cl:=RGB(GetBValue(Color),GetGValue(Color),GetRValue(Color)); //  red  blue
      FmtStr(Result, '%s%.6x', ['#', Cl]); //  #RrGgBb
  end;

begin

  if fRichMemo = nil then Exit;
  if fRichMemo.Range.IsEmpty then Exit;

  s:=fRichMemo.Range.StartY;
  e:=fRichMemo.Range.EndY;

  if fRichMemo.Range.StartY>fRichMemo.Range.EndY then begin
    s:=fRichMemo.Range.EndY;
    e:=fRichMemo.Range.StartY;
  end;

  Txt:='<span style="font-family: '''+fRichMemo.Font.Name+'''; font-size: '+IntToStr(fRichMemo.Font.Size)+'pt;"><br />'#13#10;
  Word:='';

  for i:=s to e do begin
    Sp := fRichMemo.Lines.Parser[i];
    for w := 0 to Sp.Count - 1 do begin
      with Sp[w] do begin
        tok := stToken;
        Word:=Copy(fRichMemo.fLines[i], stStart + 1, stLength);

        if not (fRichMemo.SyntaxAttributes.Equals(tokText, tok))
        and not (fRichMemo.SyntaxAttributes.Equals(tokBlank, tok)) then begin
          Color := fRichMemo.SyntaxAttributes.FontColor[tok];
          Style := fRichMemo.SyntaxAttributes.FontStyle[tok];
          St:='';
          if fsBold       in Style then St:=St+'font-weight: bold;';
          if fsItalic     in Style then St:=St+'font-style: italic;';
          if fsUnderline  in Style then St:=St+'text-decoration: underline;';

          Txt:=Txt+'<span style="color: '+ColorToStringRrGgBb(Color)+';'+St+'">'+Word+'</span>';
        end else begin
          if tok=tokBlank then
            Txt:=Txt+TimesString('&#160;',stLength)
          else
            Txt:=Txt+Word;
        end;

      end;
    end;
    Txt:=Txt+'<br />'#13#10;
  end;
  Txt:=Txt+'</span><br />'#13#10;

  Clipboard.AsText := Txt;
  OpenClipboard(Application.Handle);
  try
    Data := GlobalAlloc(GMEM_MOVEABLE + GMEM_DDESHARE, Length(Txt) + 1);
    try
      DataPtr := GlobalLock(Data);
      try
        Move(PChar(Txt)^, DataPtr^, Length(Txt) + 1);
        SetClipboardData(CF_SYNTAX, Data);
      finally
        GlobalUnlock(Data);
      end;
    except
      GlobalFree(Data);
      raise;
    end;
  finally
    CloseClipboard;
  end;

end;

function TSynMemoRange.IsEmpty: Boolean;
begin
  Result := (StartX = EndX) and (StartY = EndY);
end;

procedure TSynMemoRange.SetPos(const NewPos: TPoint);
begin
  if PointsEqual(fPos, NewPos) then Exit;
  with fRichMemo do begin
    Lines.BeginUpdate;
    if NewPos.Y <> fPos.Y then CutFinalSpaces(fPos.Y);
    fPos.Y := EnsureRange(NewPos.Y, 0, Lines.Count);
    if fPos.Y = Lines.Count then
      if (Lines.Count = 0)
      or ((Lines.Count > 0) and (Lines[Lines.Count-1] <> '')) then begin
        Lines.Add('');
        SetPos(Point(0, Lines.Count-1));
      end;
    fPos.X := Max(0, NewPos.X);
    if fSealing then begin
      fStart := fPos;
      fEnd   := fPos;
    end;
    MakeVisible(fPos.X, fPos.Y);
    Lines.EndUpdate;
    Change([ciSelection]);
  end;
end;

procedure TSynMemoRange.SetRange(const Index, Value: Integer);
begin
  case Index of
    0:  SetPos( Point(Value, fPos.Y) );
    1:  SetPos( Point(fPos.X, Value) );
  end;
end;

procedure TSynMemoRange.CutFinalSpaces(const Row: Integer);
var
  s: string;
  Da: Boolean;
begin
  with fRichMemo.Lines do
    if IsValidLineIndex(fPos.Y) then begin
      s := TrimRight(Strings[fPos.Y]);
      if Length(s) <> Length(Strings[fPos.Y]) then begin
        Da := fDirectAccess;
        fDirectAccess := True;
        Strings[fPos.Y] := s;
        fDirectAccess := Da;
      end;
    end;
end;

function TSynMemoRange.GetText: string;
var
  Row: Integer;
begin
  if IsEmpty then
    Result := ''
  else
  if fStart.Y = fEnd.Y then
    Result := Copy(fRichMemo.Lines[fStart.Y], fStart.X + 1, fEnd.X - fStart.X)
  else begin
    Result := Copy(fRichMemo.Lines[fStart.Y], fStart.X + 1, MAXINT);
    for Row := StartY + 1 to EndY - 1 do
      Result := Result + #13#10 + fRichMemo.Lines[Row];
    Result := Result + #13#10 + LeftStr(fRichMemo.Lines[fEnd.Y], fEnd.X);
  end;
end;

function TSynMemoRange.GetMarkedText: string;
var
  Row, Row1, Row2: Integer;
begin
  with fRichMemo do begin
    Row1 := fStart.Y;
    if fStart.X > 0 then Inc(Row1);
    Row2 := fEnd.Y;
    if fEnd.X < Length(Lines[fEnd.Y]) then Dec(Row2);
    if Row1 > fStart.Y
        then Result := Copy(Lines[fStart.Y], fStart.X, MAXINT) + #13#10
        else Result := '';
    for Row := Row1 to Row2 do
      with Sections.Section[Row] do
        if (Row = RowBeg) and (RowEnd <= Row2) then
          if Collapsed then
            Result := Result + SectionMarks[smCollapsed] + Lines[Row] + #13#10
          else
            Result := Result + SectionMarks[smExpanded] + Lines[Row] + #13#10
        else if (Row = RowEnd) and (RowBeg >= Row1) then
          Result := Result + SectionMarks[smEnd] + Lines[Row] + #13#10
        else
          Result := Result + Lines[Row] + #13#10;
    if Row2 < fEnd.Y then
      Result := Result + LeftStr(Lines[fEnd.Y], fEnd.X)
    else
      System.SetLength(Result, Length(Result) - 2);
  end;
end;

procedure TSynMemoRange.DoUndo;
begin
  if CanUndo then
    with fRichMemo, TSynMemoUndoItem(fUndoStack.Last) do begin
      Lines.BeginUpdate;
      fStart :=   uiSelStart;
      fEnd :=     uiSelEnd;
      fSealing := uiSealing;
      Lines.State := Lines.State + [ssUndoProcess];
      Self.SetText(uiText);
      Lines.State := Lines.State - [ssUndoProcess];
      Sections.EntireSection := uiSections;
      SetPos( uiCaretPos );
      with fUndoStack do Delete(Count - 1);
      Change([ciUndoStack]);
      Lines.State := [ssNeedReIndex, ssNeedReparseAll];
      fRichMemo.NeedRedrawAll;
      Lines.EndUpdate;
    end;
end;

function TSynMemoRange.AddUndo(const UndoText: string = ''): TSynMemoUndoItem;
begin
  Result := TSynMemoUndoItem.Create;
  Result.uiCaretPos := fPos;
  Result.uiSelStart := fStart;
  Result.uiSelEnd   := fEnd;
  Result.uiSealing  := fSealing;
  Result.uiSections := fRichMemo.Sections.EntireSection;
  Result.uiText     := UndoText;
  Result.uiSections.AddRef;
  if fUndoStack.Count >= fMaxUndoDepth then fUndoStack.Delete(0);
  fUndoStack.Add(Result);
  fRichMemo.Change([ciUndoStack]);
end;

procedure TSynMemoRange.ClearUndo;
begin
  fUndoStack.Clear;
  Assert(fRichMemo.Sections.EntireSection.RefCount = 1,
      ' ClearUndo   fRichMemo.Sections.EntireSection.RefCount == 1');
  fRichMemo.Change([ciUndoStack]);
end;

procedure TSynMemoRange.SetMarkedText(Value: string);
type
  TIntArray = array of Integer;

  procedure _Scan(var s: string; var secs: TIntArray);
  const
    signs: array [Boolean] of Integer = (-1, +1);
  var
    Sl: TStringList;
    i: Integer;
    Stack: TStack;
    Sm: TSectionMark;
  begin
    Sl := TStringList.Create;
    Sl.Text := s;
    secs := nil;
    Stack := TStack.Create;
    for i := 0 to Sl.Count - 1 do begin
      Sm := TSynMemoSections.DetectSectionMark(Sl[i]);
      case Sm of
        smExpanded,
        smCollapsed:
                    begin
                      Stack.Push(Pointer((i + 1) * signs[Sm = smExpanded]));
                      Sl[i] := Copy(Sl[i], SECTION_HEADER_LENGTH + 1, MAXINT);
                    end;
        smEnd:
                    begin
                      if Stack.AtLeast(1) then begin
                        System.SetLength(secs, Length(secs) + 2);
                        secs[High(secs)-1] := Integer(Stack.Pop);
                        secs[High(secs)]   := i;
                      end;
                      Sl[i] := Copy(Sl[i], SECTION_HEADER_LENGTH + 1, MAXINT);
                    end;
      end;
    end;
    Stack.Free;
    s := Sl.Text;
    Sl.Free;
  end;

var
  i, Row1, Row2: Integer;
  SafeStart: TPoint;
  SecIndexes: TIntArray;
begin
  with fRichMemo do begin
    Lines.BeginUpdate;
    SafeStart := fStart;
    if (SafeStart.X > 0)
    and (TSynMemoSections.DetectSectionMark(Value) <> smNone) then
        System.Delete(Value, 1, SECTION_HEADER_LENGTH);
    _Scan(Value, SecIndexes);
    SetText(Value);
    Sections.ReIndex;
    for i := 0 to Length(SecIndexes) shr 1 - 1 do begin
      Row1 := SafeStart.Y + SecIndexes[i*2] * sign(SecIndexes[i*2]) - 1;
      Row2 := SafeStart.Y + SecIndexes[i*2 + 1];
      Sections.New(Row1, Row2, SecIndexes[i*2] < 0);
    end;
    SecIndexes := nil;
    Lines.EndUpdate;
  end;
end;

procedure TSynMemoRange.SetText(const Value: string);
var
  Sl: TStringList;
  n, n1, n2: Integer;
  f: boolean;
  tmp, text: String;
  NewSelEnd: Tpoint;
begin
  (**
  with fRichMemo do begin

    if IsEmpty and (PosX > Length(Lines[PosY])) then
      Lines[PosY] := Lines[PosY] + StringOfChar(' ', PosX - Length(Lines[PosY]));
      
    Sl := TStringList.Create;
    Sl.Text := LeftStr(Lines[StartY],StartX)+Value;
    text:=self.GetText;

    n1:=0;
    n2:=StartY;
    tmp:=RightStr(Lines[EndY],Length(Lines[EndY])-EndX);
    if Sl.Count>0 then begin
      n1:=length(Sl.Strings[Sl.Count-1]);
      Sl.Strings[Sl.Count-1] := Sl.Strings[Sl.Count-1]+tmp;
    end else begin
      n2:=StartY+1;
    end;

    fRichMemo.Lines.BeginUpdate;

    f:=false;
    if EndY=fLines.Count-1 then f:=true;
    for n:=StartY to EndY-1 do fLines.Delete(StartY);

    if f then
      fLines.Strings[fLines.Count-1]:=tmp
    else if (tmp='') or (StartX<>0) then
      fLines.Delete(StartY)
    else if StartX=0 then
      fLines.Strings[StartY]:=tmp;

    for n:=0 to Sl.Count-1 do
      fLines.Insert(n+StartY,Sl.Strings[n]);
      
    if (Sl.Count=0) and (tmp='') then begin
      fLines.Insert(StartY,'');
      //n1:=2;  //-->>>>>
    end;

    StartX:=n1;
    StartY:=n2+Sl.Count-1;

    if Value=#13#10 then begin
      StartY:=StartY+1; StartX:=0;
      fLines.Insert(StartY,'');
    end;

    EndX:=StartX;
    EndY:=StartY;

    fEnd := point(EndX,EndY);
    fStart := point(EndX,EndY);
    fSealing := True;

    SetPos(fEnd);

    Lines.EndUpdate;
    //!!!
    if not (ssUndoProcess in Lines.State) then
      with AddUndo do begin
        uiSelEnd   := point(StartX,StartY);
        uiText     := text;
      end;
    
    fVScroll.Range:=fLines.Count*fCharHeight+fCharHeight;

    if ((fVScroll.Position div fCharHeight)>StartY)
    or (StartY>(fVScroll.Position div fCharHeight)+ClientLines-2) then
      fVScroll.Position:=(StartY-(ClientLines div 2))*fCharHeight;
    UpdateScrollBars;

    Sl.Free;
    Invalidate;
    InvalidateRect(Handle,nil,false);

  end;
  (**)
  (**)
  with fRichMemo do
    if IsEmpty and (PosX > Length(Lines[PosY])) then
      Lines[PosY] := Lines[PosY] + StringOfChar(' ', PosX - Length(Lines[PosY]));
  Sl := TStringList.Create;
  Sl.Text := Value + #13#10;
  with fRichMemo do begin
    n1 := Sl.Count;
    n2 := EndY - StartY + 1;
    NewSelEnd.Y := fStart.Y + n1 - 1;
    Sl[0] := LeftStr(fLines[fStart.Y], fStart.X) + Sl[0];
    NewSelEnd.X := Length(Sl[n1-1]);
    Sl[n1-1] := Sl[n1-1] + RightStr(fLines[fEnd.Y], Length(fLines[fEnd.Y]) - fEnd.X);
    if not (ssUndoProcess in Lines.State) then
      with AddUndo do begin
        uiSelEnd   := NewSelEnd;
        uiText := Copy(fLines[fStart.Y], fStart.X + 1, MAXINT);
        for n := fStart.Y + 1 to fEnd.Y do uiText := uiText + #13#10 + fLines[n];
        System.SetLength(uiText, Length(uiText) - Length(fLines[fEnd.Y]) + fEnd.X);
        //ShowMessage(uiText+' >>>'+IntToStr(uiSelEnd.Y)+' >>>'+IntToStr(uiSelEnd.X)
                          //+' >>>'+IntToStr(PosY)+' >>>'+IntToStr(PosX) );
      end;
    Lines.BeginUpdate;
    for n := n2 - n1 downto 1 do Lines.Delete(fStart.Y);
    Lines[fStart.Y] := Sl[0];
    for n := 0 to n1-2 do begin
      Lines.Insert(fStart.Y+n,Sl[n]);
      //if fHScroll.Range<Length(Sl[n])*8 then fHScroll.Range:=Length(Sl[n])*8; !!!
    end;
    Lines[fStart.Y+n1-1] := Sl[n1-1];
    fEnd := NewSelEnd;
    fStart := fEnd;
    fSealing := True;
    SetPos( fEnd );

    Lines.EndUpdate;
    fVScroll.Range:=fLines.Count*fCharHeight+fCharHeight;
  end;
  Sl.Free;
  (**)
end;

function TSynMemoRange.GetLength: Integer;
var
  i: Integer;
begin
  Result := EndX - StartX;
  for i := StartY to EndY - 1 do Inc(Result, Length(fRichMemo.Lines[i]) + 2);
end;

function TSynMemoRange.GetPosition: Integer;
begin
  Result := fRichMemo.Lines.RCToPosition(StartX, StartY);
end;

procedure TSynMemoRange.SetLength(const Value: Integer);
var
  x0, dy, i, n: Integer;
begin
  with fRichMemo do begin
    dy := 0;
    x0 := StartX;
    n := Value;
    for i := StartY to fLines.Count - 1 do begin
      Dec(n, Length(fLines[i]) + 2 - x0);
      x0 := 0;
      if n <= 0 then begin
        Enlarge(dy, True);
        Enlarge(Length(fLines[i]) + 2 + n - StartX);
        Exit;
      end else
        Inc(dy);
    end;
  end;
end;

procedure TSynMemoRange.SetPosition(const Value: Integer);
begin
  Collapse;
  SetPos( fRichMemo.Lines.PositionToRC(Value) );
end;

function TSynMemoRange.CanUndo: Boolean;
begin
  Result := fUndoStack.Count > 0;
end;

procedure TSynMemoRange.SetMaxUndoDepth(const Value: Integer);
begin
  while fMaxUndoDepth > Value do begin
    fUndoStack.Delete(0);
    Dec(fMaxUndoDepth);
  end;
  fMaxUndoDepth := Value;
end;

procedure TSynMemoRange.SelectAll;
begin
  SetPosition(0);
  SetLength(Length(fRichMemo.Lines.Text));
end;

procedure TSynMemoRange.SelectFromStart;
var
  L: Integer;
begin
  L := fRichMemo.Lines.RCToPosition(fPos.X, fPos.Y);
  SetPosition(0);
  SetLength(L);
end;

procedure TSynMemoRange.SelectToEnd;
var
  L: Integer;
begin
  L := fRichMemo.Lines.RCToPosition(fPos.X, fPos.Y);
  SetPosition(L);
  SetLength(Length(fRichMemo.Lines.Text) - L);
end;
{$EndRegion}
//########################################################################
{$Region ' TBookmarkManager  '}
constructor TBookmarkManager.Create(Owner: TCustomSyntaxMemo);
begin
  inherited Create;
  fRichMemo := Owner;
  fImages := TBitmap.Create;
  fImages.LoadFromResourceName(HInstance, 'BOOKMARKS');
  Clear;
end;

destructor TBookmarkManager.Destroy;
begin
  fImages.Free;
  inherited;
end;

procedure TBookmarkManager.Clear;
var
  i: TBookmarkIndex;
begin
  for i := Low(TBookmarkIndex) to High(TBookmarkIndex) do fBookMarks[i] := -1;
end;

function TBookmarkManager.Find(const Row: Integer; var Index: TBookmarkIndex): Boolean;
var
  i: TBookmarkIndex;
begin
  for i := Low(TBookmarkIndex) to High(TBookmarkIndex) do
    if fBookMarks[i] = Row then begin
      Result := True;
      Index := i;
      Exit;
    end;
  Result := False;
end;

function TBookmarkManager.GetBookMarks(const Index: TBookmarkIndex): Integer;
begin
  Result := fBookMarks[Index];
end;

procedure TBookmarkManager.SetBookMarks(const Index: TBookmarkIndex; const Row: Integer);

    procedure SetBookMarkInt;
    var
      i: TBookmarkIndex;
      n: Integer;
    begin
        if Find(Row, i) then begin
          fBookMarks[i] := -1;
          if i = Index then Exit;
        end;
        if fBookMarks[Index] >= 0 then begin
          n := fBookMarks[Index];
          fBookMarks[Index] := -1;
          fRichMemo.NeedRedraw(n);
        end;
        fBookMarks[Index] := Row;
    end;

begin
  fRichMemo.fLines.BeginUpdate;
  SetBookMarkInt;
  fRichMemo.NeedRedraw(Row);
  fRichMemo.fLines.EndUpdate;
end;
{$EndRegion}
//########################################################################

constructor TCustomSyntaxMemo.Create(AOwner: TComponent);
var
  F: TFont;
begin
  inherited Create(AOwner);
  DoubleBuffered  := True;
  fLines          := TSynMemoStrings.Create(self);
  fRange          := TSynMemoRange.Create(self);
  fSections       := TSynMemoSections.Create(self);
  fParseAttributes:= TSyntaxAttributes.Create(self);
  fBookMarks      := TBookmarkManager.Create(self);
  Width           := 185;
  Height          := 89;
  Color           := clWindow;
  Cursor          := crIBeam;
  fSelWord        := Point(-1, -1);
  fDefBackColor   := clWindow;
  fDefForeColor   := clBlack;
  fSelColor       := $a56d53;
  fGutterWidth    := 32;
  fInsertMode     := True;
  ParentColor     := False;
  fErrorLine      :=-1;
  fNaviLine       :=-1;

  fVScroll:= TCustomScrollBar.Create(Self, sbVertical);
  fHScroll:= TCustomScrollBar.Create(Self, sbHorizontal);
  fHScroll.Range:=4096*8;

  F := TFont.Create;
  F.Name := 'Courier New';
  F.Size := 10;
  F.Style := [];
  F.Color := clBlack;
  Font.Assign(F);
  F.Free;

  TabStop         := True;
  Font.OnChange   := FontChange;
  fSectionIndent  := 16;
  fLines.SetModified(false);

  Path:='';
  Name:='';
end;

destructor TCustomSyntaxMemo.Destroy;
begin
  fScreenLines := nil;
  fBookMarks.Free;
  fSections.Free;
  fParseAttributes.Free;
  fRange.Free;
  fLines.Free;
  fVScroll.Free;
  fHScroll.Free;
  inherited;
end;

procedure TCustomSyntaxMemo.SetHighliter(const Value: TSCustomHighlight);
var
  i: integer;
begin
  fHighliter := Value;
  if Assigned(fHighliter) then begin
    fHighliter.GetHighlitedParam(Self);
  end else begin
    with fParseAttributes do begin
      for i:=tokBlank to tokUser do begin
        fTokenStyles[i].tsForeground        := clNone;
        fTokenStyles[i].tsStyle             := [];
      end;
    end;
  end;
end;

procedure TCustomSyntaxMemo.SetSelColor(const Value: TColor);
begin
  if Value <> fSelColor then begin
    fSelColor := Value;
    if not fRange.IsEmpty() then InvalidateRect(Handle,nil,false);
  end;
end;

function TCustomSyntaxMemo.ClientLines: Integer;
begin
  with TextRowRect[-1] do
    Result := (Bottom - Top {+ fCharHeight - 1}) div fCharHeight+1; //***SAT
end;

procedure TCustomSyntaxMemo.WMSIZE(var Message : TMessage);
begin
  InvalidateRect(Handle,nil,false);
  CalcScreenParams;
  if HandleAllocated and (fLines.Text='') then fLines.Add('');
  if not (csDesigning in ComponentState) then UpdateScrollBars;
end;

function TCustomSyntaxMemo.CanResize(var NewWidth, NewHeight: Integer): Boolean;
begin
  Result := Inherited CanResize(NewWidth, NewHeight);
end;

procedure TCustomSyntaxMemo.WMLButtonDblClk(var Message : TWMMouse);
var
  Row, WIndex: Integer;
  Sec: TSynMemoSection;
  T: TSyntaxToken;
begin
  inherited;
  if (GetKeyState(VK_CONTROL) and $8000) <> 0 then begin
    WndOffsetToPixOffset(Point(Message.XPos, Message.YPos), WIndex, Row, True);
    if Lines.IsValidLineIndex(Row) then begin
      Sec := Sections.Section[Row];
      Sections.Expand(Row, False, True);
      Range.Collapse;
      {       -1,  .  }
      Range.Pos := Point(0, EnsureRange(Sec.RowBeg, 0, Lines.Count - 1));
      Range.Enlarge(Sec.RowEnd - Sec.RowBeg + 1, True);
    end
  end else if GetWordAtPos(Message.XPos, Message.YPos, WIndex, Row) then
    if InRange(WIndex, 0, Lines.Parser[Row].Count - 1) then begin
      T := Lines.Parser[Row].Tokens[WIndex];
      Range.Collapse;
      Range.Pos := Point(T.stStart, Row);
      Range.Enlarge(T.stLength);
    end;
end;

procedure TCustomSyntaxMemo.CalcScreenParams;
var
  c: Char;
begin
  inherited;
  with Canvas do begin
    Font.Assign(Self.Font);
    fCharHeight := (Font.Size div 2)+Font.Size+2;
    Font.Style  := [];
    for c := Low(fCharWidths[0]) to High(fCharWidths[0]) do
      fCharWidths[0][c] := Byte(TextWidth(c));
    Font.Style  := [fsBold];
    for c := Low(fCharWidths[1]) to High(fCharWidths[1]) do
      fCharWidths[1][c] := Byte(TextWidth(c));
    Font.Style  := [fsItalic];
    for c := Low(fCharWidths[2]) to High(fCharWidths[2]) do
      fCharWidths[2][c] := Byte(TextWidth(c));
    Font.Style  := [fsBold,fsItalic];
    for c := Low(fCharWidths[3]) to High(fCharWidths[3]) do
      fCharWidths[3][c] := Byte(TextWidth(c));
    Font.Assign(Self.Font);
    Font.Style:=Self.Font.Style;
  end;
  fParseAttributes.fTokenStyles[tokText].tsForeground:=Font.Color;
  fParseAttributes.fTokenStyles[tokText].tsStyle     :=Font.Style;
  SetLength(fScreenLines, ClientLines);
  fVScroll.Range:=fLines.Count*fCharHeight+fCharHeight;
  fVScroll.fIncrement:=fCharHeight;
  fHScroll.fIncrement:=fCharHeight;
end;

procedure TCustomSyntaxMemo.WndOffsetToPixOffset(OfsPoint: TPoint; var CharPix, Row: Integer; const TextRow: Boolean);
var
  R: TRect;
begin
  R := TextRowRect[0];
  CharPix := OfsPoint.X - R.Left + OffsetXPix;
  Row := (OfsPoint.Y - R.Top) div fCharHeight;
  if TextRow then Row := FindVisibleRow(OffsetY, Row, False);
end;

function TCustomSyntaxMemo.PixOffsetToWndOffsetEx(const CharPix, ScreenRow: Integer): TPoint;
begin
  with TextRowRect[ScreenRow] do begin
    Inc(Left,CharPix-OffsetXPix);
    Result := TopLeft;
  end;
end;

function TCustomSyntaxMemo.CharWidth(N: TFontStyles): byte;
begin
  Result:=0;
  if fsBold   in N then Inc(Result,1);
  if fsItalic in N then Inc(Result,2);
end;

function TCustomSyntaxMemo.CharPosToPixOffset(const Col: Integer; s: string; Sp: TSyntaxParser): Integer;
var
  T: TSyntaxToken;
  i, j, WordIndex: Integer;
  Bold: byte;
begin
  Result := 0;
  WordIndex := 0;
  i := 0;
  while WordIndex < Sp.Count do begin
    T := Sp[WordIndex];
    j := Min(Col, T.stStart);
    Inc(Result, (j - i) * fCharWidths[0][' ']);
    i := j;
    if i = Col then Break;
    Bold := CharWidth(fParseAttributes.FontStyle[T.stToken]);
    j := Min(Col, T.stStart + T.stLength);
    if s<>'' then //***
      while i < j do begin
        Inc(Result, fCharWidths[Bold][s[i+1]]);
        Inc(i);
      end;
    if i = Col then Break;
    Inc(WordIndex);
  end;
  if i<Col then Inc(Result,(Col-i)*fCharWidths[0][' ']);
end;

function TCustomSyntaxMemo.CharPosToPixOffset(const Col, Row: Integer): Integer;
begin
  if fLines.IsValidLineIndex(Row) then
    Result := CharPosToPixOffset(Col, fLines[Row], fLines.Parser[Row])
  else
    Result := 0;
end;

function TCustomSyntaxMemo.PixOffsetToCharPos(const Pix, Row: Integer; const WordIndex: PInteger = nil): Integer;
var
  Sp: TSyntaxParser;
  s: string;
  Pos, WIndex: Integer;
  Bold: byte;
  T: TSyntaxToken;
begin
  Result := 0;
  if not InRange(Row, 0, fLines.Count - 1) then Exit;
  Sp := TSyntaxParser(fLines.Objects[Row]);
  s  := fLines[Row];
  WIndex := 0;
  Pos := 0;
  if Sp <> nil then
    while WIndex < sp.Count do begin
      T := Sp[WIndex];
      Bold := CharWidth(fParseAttributes.FontStyle[T.stToken]);
      while Result < T.stStart do begin
        Inc(Pos, fCharWidths[0][' ']);
        if Pos>Pix then begin
          if Assigned(WordIndex) then WordIndex^ := -WIndex;
          Exit;
        end;
        Inc(Result);
      end;
      if s<>'' then //***
        while Result < T.stStart + T.stLength do begin
          Inc(Pos,fCharWidths[Bold][s[Result+1]]);
          if Pos>Pix+(fCharWidths[Bold][s[Result]] div 2) then begin
            if Assigned(WordIndex) then WordIndex^ := WIndex;
            Exit;
          end;
          Inc(Result);
        end;
      Inc(WIndex);
    end;
  if Pos<Pix then Inc(Result, (Pix - Pos) div fCharWidths[0][' ']);
  if Assigned(WordIndex) then WordIndex^ := MAXINT;
end;

procedure TCustomSyntaxMemo.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTARROWS    //        }
                 or DLGC_WANTALLKEYS   //     Enter }
                 or DLGC_WANTTAB;
end;

procedure TCustomSyntaxMemo.KeyDown(var Key: Word; Shift: TShiftState);
var
  xn, yn, x: Integer;
  Flag: boolean;
begin
    inherited;
    Flag:=Lines.Modified;
    Lines.Modified:=true;
    with Range do
        case Key of
        VK_CONTROL:
            if (Shift = [ssCtrl]) and not fHinting and Assigned(fHighliter) then begin
                Cursor := crHandPoint;
                Hint := '';
                SelectedWord := Point(-1, -1);
                ShowHint := True;
                fHinting := True;
                Lines.Modified:=Flag;
                if not Flag then Lines.fState:=[];
            end;

        VK_RIGHT:
            if not (ssShift in Shift) then begin
                Collapse;
                if not (ssCtrl in Shift) then
                    PosX := PosX + 1
                else
                if FindNextWord(xn, yn) then
                    Pos := Point(xn, yn);
            end else
                if not (ssCtrl in Shift)
                    then Enlarge(1)
                    else begin
                        if FindNextWord(xn, yn) then begin
                            Enlarge(yn - PosY, True);
                            Enlarge(xn - PosX);
                        end;
                    end;

        VK_LEFT:
            if not (ssShift in Shift) then begin
                Collapse;
                if not (ssCtrl in Shift) then
                    PosX := PosX - 1
                else
                if FindPrevWord(xn, yn) then
                    Pos := Point(xn, yn);
            end else
                if not (ssCtrl in Shift)
                    then Enlarge(-1)
                    else begin
                        if FindPrevWord(xn, yn) then begin
                            Enlarge(yn - PosY, True);
                            Enlarge(xn - PosX);
                        end;
                    end;

        VK_DOWN:
            if not (ssShift in Shift) then
                if not (ssCtrl in Shift)
                    then begin
                        Collapse;
                        PosY := FindVisibleRow(PosY, 1, True);
                    end else begin
                        OffsetY := FindVisibleRow(OffsetY, 1, True);
                        if StartY < OffsetY then begin
                            Collapse;
                            PosY := OffsetY;
                        end;
                    end
            else
                if not (ssCtrl in Shift)
                    then Enlarge(1, True, True)
                    else GotoSection(True);

        VK_UP:
            if not (ssShift in Shift) then
                if not (ssCtrl in Shift)
                    then begin
                        Collapse;
                        PosY := FindVisibleRow(PosY, -1, True);
                    end else begin
                        OffsetY := FindVisibleRow(OffsetY, -1, True);
                        if StartY >= OffsetY + ClientLines then begin
                            Collapse;
                            PosY := OffsetY + ClientLines - 1;
                        end;
                    end
            else
                if not (ssCtrl in Shift)
                    then Enlarge(-1, True, True)
                    else GotoSection(False);

        VK_HOME:
            if not (ssShift in Shift) then begin
                Collapse;
                if ssCtrl in Shift then PosY := 0;
                PosX := 0;
            end else begin
                if ssCtrl in Shift then Enlarge(-PosY, True);
                Enlarge(-PosX);
            end;

        VK_END:
            if not (ssShift in Shift) then begin
                Collapse;
                if ssCtrl in Shift then PosY := fLines.Count - 1;
                PosX := Length(fLines[PosY]);
            end else begin
                if ssCtrl in Shift then Enlarge(fLines.Count-1 - PosY, True);
                Enlarge(Length(fLines[PosY]) - PosX);
            end;

        VK_NEXT:
            if not (ssShift in Shift) then begin
                Collapse;
                if not (ssCtrl in Shift)
                    then PosY := FindVisibleRow(PosY, ClientLines - 1, True)
                    else PosY := FindVisibleRow(OffsetY, ClientLines - 1, True)
            end else
                if not (ssCtrl in Shift)
                    then Enlarge(ClientLines() - 1, True)
                    else Enlarge(OffsetY + (ClientLines - 1) - PosY, True);

        VK_PRIOR:
            if not (ssShift in Shift) then begin
                Collapse;
                if not (ssCtrl in Shift)
                    then PosY := FindVisibleRow(PosY, -(ClientLines - 1), True)
                    else PosY := OffsetY
            end else
                if not (ssCtrl in Shift)
                    then Enlarge( -(ClientLines - 1), True)
                    else Enlarge( -(OffsetY - PosY), True);

        VK_DELETE:
            if not (smoReadOnly in fOptions) then begin
                Range.Delete;
                IF PosY<OffsetY THEN
                  SetOffsets(Point(0, FindVisibleRow(PosY, -(ClientLines DIV 2), True)));//***;
            end;

        VK_TAB:
            if not (smoReadOnly in fOptions) then
              if StartY <> EndY then
                fRange.MakeIndent
              else
                Range.Text := StringOfChar(' ', 2 - PosX mod 2);

        VK_BACK:
            if not (smoReadOnly in fOptions) then begin
              if (PosX = 0) and (FindVisibleRow(Range.PosY, -1, True) = Range.PosY - 1) then
                with Range do begin
                  StartX := Length(fLines[PosY - 1]);
                  if Range.EndX=0 then begin
                    StartY := PosY - 1;
                    if (fLines[PosY-1]='') then StartX:=1;
                  end else begin
                    StartY := PosY;
                    StartX := 0;
                  end;
                end
              else if PosX > 0 then begin
                if Length(Trim(copy(Lines.Strings[PosY],1,PosX)))=0 then begin
                  x:=PosY;
                  yn:=PosX+1;
                  if (x<>0) and (PosX>1) then begin
                    while yn>PosX-1 do begin
                      dec(x);
                      yn:=1;
                      if Length(Trim(Lines.Strings[x]))=0 then begin
                        yn:=PosX+1;
                        Continue;
                      end;
                      while Lines.Strings[x][yn]<=#32 do Inc(yn);
                      Dec(yn);
                    end;
                    StartX:=yn;
                  end else
                    StartX:=0; 
                end else if Range.EndX=Range.StartX then
                  StartX:=PosX-1;
              end;
              if not ((Range.EndX=0) and (StartX=0) and (StartY=0)) then Range.Delete;
              if PosY<OffsetY then
                SetOffsets(Point(0, FindVisibleRow(PosY, -(ClientLines DIV 2), True)));
            end;

        VK_RETURN:
            if not (smoReadOnly in fOptions) then begin
              x:=PosY;
              while (Length(Trim(Lines.Strings[x]))=0) and (x<>0) do dec(x);
              Range.SetText(#13#10+copy(Lines.Strings[x],1,system.pos(Trim(Lines.Strings[x]),Lines.Strings[x])-1));
            end;

        VK_INSERT:
            if Shift=[] then begin
              if not (smoReadOnly in fOptions) then
                 if smoOverwrite in fOptions
                    then SetOptions(fOptions + [smoOverwrite])
                    else SetOptions(fOptions - [smoOverwrite]);
            end else begin
              if Shift=[ssCtrl] then fRange.CopyToClipboard;
              if Shift=[ssShift] then fRange.PasteFromClipboard;
            end;

        ORD('C'):
            if ssCtrl in Shift then
                fRange.CopyToClipboard;

        ORD('I'):
            if Shift=[ssCtrl, ssShift] then
              fRange.MakeIndent;

        ORD('U'):
            if Shift=[ssCtrl, ssShift] then
              fRange.MakeUnIndent;

        ORD('X'):
            if Shift = [ssCtrl] then
                fRange.CutToClipBoard;

        ORD('V'):
            if (Shift = [ssCtrl]) and not (smoReadOnly in fOptions) then
                fRange.PasteFromClipboard;

        ORD('Z'):
            if ssCtrl in Shift then
                if CanUndo then DoUndo;

        VK_ADD:
            if ssCtrl in Shift then
                fRange.ExpandSection(ssShift in Shift);

        VK_SUBTRACT:
            if ssCtrl in Shift then
                fRange.CollapseSection(ssShift in Shift);

        VK_F3:
              begin
                FindWord(0,-1,True,True);
                Lines.Modified:=Flag;
                if not Flag then Lines.fState:=[];
              end;
        VK_F5:
            fRange.CreateSection;

        VK_F6:
            fRange.ExplodeSection(ssCtrl in Shift);

        Ord('0')..Ord('9'):
            if ssCtrl in Shift then
                if ssShift in Shift
                    then BookMarks[Key - Ord('0')] := fRange.PosY
                    else Navigate(0, BookMarks[Key - Ord('0')]);

        VK_NUMPAD0..VK_NUMPAD9:
            if ssCtrl in Shift then
                if ssShift in Shift
                    then BookMarks[Key - VK_NUMPAD0] := fRange.PosY
                    else Navigate(0, BookMarks[Key - VK_NUMPAD0]);

        end;
  if Trim(Lines.Strings[Range.PosY])='' then Lines.Strings[Range.PosY]:=System.StringOfChar(' ',Range.PosX);
  fErrorLine:=-1;
  fNaviLine:=-1;
  fVScroll.Position:=OffsetY*fCharHeight;
  InvalidateRect(Handle,nil,false);
end;

procedure TCustomSyntaxMemo.KeyPress(var Key: Char);
begin
    inherited;
    case Key of
        #32..#255 : if not (smoReadOnly in fOptions) then
                        Range.Text := Key;
    end;
end;

procedure TCustomSyntaxMemo.KeyUp(var Key: Word; Shift: TShiftState);
begin
  Cursor := crIBeam;
  ShowHint := False;
  SelectedWord := Point(-1, -1);
  if Assigned(fOnWordInfo) then
      fOnWordInfo(self, 0, 0, -1, -1, False);
  fHinting := False;
  inherited;
end;

procedure TCustomSyntaxMemo.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var Col, Row, SRow: Integer;
    R: TRect;
begin
  if not Focused then SetFocus;
  {    }
  if (Button = mbLeft) and (Shift = [ssLeft]) then
      if PtInRect(EntireGutterRect[-1], Point(X, Y)) then begin
          {     -   /   }
          WndOffsetToPixOffset(Point(X, Y), Col, SRow, False);
          Row := FindVisibleRow(OffsetY, SRow, False);
          if Row<0 then Exit;
          with fSections.Section[Row] do
              if (Level > 0) and (Row = RowBeg) then begin
                  R := GetSectionButtonRect(SRow, Level);
                  if PtInRect(R, Point(X, Y)) then
                      if Collapsed
                          then fSections.Expand(Row, False, False)
                          else fSections.Collapse(Row, False, False);
              end;
      end else begin
          fDown := True;
          {   }
          Range.Collapse;
          WndOffsetToPixOffset(Point(X, Y), Col, Row, True);
          //WndOffsetToPixOffset(Point((X div fCharWidths[0]['W'])*fCharWidths[0]['W']+fCharWidths[0]['W'], Y), Col, Row, True);
          if fLines.IsValidLineIndex(Row) then begin
              Col := PixOffsetToCharPos(Col, Row);
              Range.Pos := Point(Col, Row);
          end;
      end
  else if (Button = mbLeft) and (ssShift in Shift) then begin
    //    Shift+Left_Mouse_Button;
    WndOffsetToPixOffset(Point(X, Y), Col, Row, False);
    Row := FindVisibleRow(OffsetY, Row, True);
    Col := PixOffsetToCharPos(Col, Row);
    if (Col <> Range.PosX) or (Row <> Range.PosY) then begin
      Range.Enlarge(Row - Range.PosY, True);
      Range.Enlarge(Col - Range.PosX);
    end;
  end;

  inherited;
  if Trim(Lines.Strings[Range.PosY])='' then Lines.Strings[Range.PosY]:=System.StringOfChar(' ',Range.PosX);
  fErrorLine:=-1;
  fNaviLine:=-1;
  InvalidateRect(Handle,nil,false);
end;

procedure TCustomSyntaxMemo.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  Col, Row: Integer;
begin
    //  
    if fDown and (Shift = [ssLeft]) then begin
        WndOffsetToPixOffset(Point(X, Y), Col, Row, False);
        Row := FindVisibleRow(OffsetY, Row, True);
        Col := PixOffsetToCharPos(Col, Row);
        if (Col <> Range.PosX) or (Row <> Range.PosY) then begin
            Range.Enlarge(Row - Range.PosY, True);
            Range.Enlarge(Col - Range.PosX);
        end;
    end else if fHinting then begin
        WndOffsetToPixOffset(Point(X, Y), Col, Row, True);
        PixOffsetToCharPos(Col, Row, @Col);
        if Row<0 then Exit;
        if InRange(Col, 0, fLines.Parser[Row].Count - 1) then begin
            SelectedWord := Point(Col, Row);
            if Assigned(fOnWordInfo) then
                fOnWordInfo(self, X, Y, Col, Row, True)
            else
                with fLines.Parser[Row].Tokens[Col] do
                    Hint := Format('Word: "%s"'#13#10'Start: %d'#13#10'Length: %d'#13#10'As token #%d', [Copy(fLines[Row], stStart + 1, stLength), stStart, stLength, stToken]);
        end;
    end else if Shift = [] then begin
        if PtInRect(EntireGutterRect[-1], Point(X, Y))
            then Cursor := crDefault
            else Cursor := crIBeam;
    end;
    inherited;
end;

procedure TCustomSyntaxMemo.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
    inherited;
    fDown := False;
end;

procedure TCustomSyntaxMemo.CreateParams(var Params: TCreateParams);
begin
    inherited;
    with Params do begin
      Style   := Style and not WS_TABSTOP;
    end;
end;

procedure TCustomSyntaxMemo.WMHScroll(var Message: TWMHScroll);
begin
  if (Message.ScrollBar = 0) and fHScroll.Visible then
    fHScroll.ScrollMessage(Message)
  else
    inherited;
end;

procedure TCustomSyntaxMemo.WMVScroll(var Message: TWMVScroll);
begin
  if (Message.ScrollBar = 0) and fVScroll.Visible then
    fVScroll.ScrollMessage(Message)
  else
    inherited;
end;

procedure TCustomSyntaxMemo.WMMouseWheel(var Message: TMessage);
begin
    inherited;
    if GetKeyState(VK_CONTROL) and $8000 <> 0 then
      fHScroll.MouseMessage(Message)
    else
      fVScroll.MouseMessage(Message);
    InvalidateRect(Handle,nil,false);
end;

procedure TCustomSyntaxMemo.CMMouseEnter(var Message: TMessage);
begin
  SelectedWord := Point(-1, -1);
  fHinting := FALSE;
end;

procedure TCustomSyntaxMemo.CMMouseLeave(var Message: TMessage);
begin
  SelectedWord := Point(-1, -1);
  fHinting := FALSE;
end;

procedure TCustomSyntaxMemo.UpdateScrollBars;
begin
  if not HandleAllocated then Exit;
  if fVScroll.NeedsScrollBarVisible then begin
    fHScroll.Update(False, True);
    fVScroll.Update(True, False);
  end else if fHScroll.NeedsScrollBarVisible then begin
    fVScroll.Update(False, True);
    fHScroll.Update(True, False);
  end else begin
    fVScroll.Update(False, False);
    fHScroll.Update(True, False);
  end;
end;

procedure TCustomSyntaxMemo.ScrollBy(DeltaX, DeltaY: Integer);
begin
  if (Handle=0) and not IsWindowVisible(Handle) then Exit;
  HideCaret;
  OffsetY     := RowIndexConvert(fVScroll.Position div fCharHeight, cdNeedReal);
  OffsetXPix  := fHScroll.Position;
  InvalidateRect(Handle,nil,false);
  ShowCaret;
end;

procedure TCustomSyntaxMemo.WMKillFocus(var Msg: TWMKillFocus);
begin
    inherited;
    HideCaret;
    Windows.DestroyCaret;
end;

procedure TCustomSyntaxMemo.WMSetFocus(var Msg: TWMSetFocus);
begin
    inherited;
    if fLines.Text='' then fLines.Add('');
    Windows.CreateCaret(Handle, 0, 2, FCharHeight-2);
    ShowCaret;
end;

function TCustomSyntaxMemo.FindNextWord(var wx, wy: Integer): Boolean;
var Sp: TSyntaxParser;
begin
    Result := True;
    wx := CharPosToWordIndex(Range.PosX, Range.PosY);
    wy := Range.PosY;
    Sp := fLines.Parser[wy];
    {       }
    if wx < 0 then begin
        wy := Range.PosY;
        wx := Sp[-wx].stStart;
    end else
    {       }
    if wx = MAXINT then begin
        wy := FindVisibleRow(wy, 1, False);
        Result := wy >= 0;
        if Result then begin
            Sp := fLines.Parser[wy];
            if Sp.Count > 0
                then wx := Sp[0].stStart
                else wx := 0;
        end
    end else
    {       }
    if wx = Sp.Count - 1 then
        with Sp[wx] do
            wx := stStart + stLength
    {        }
    else
        wx := Sp[wx + 1].stStart;
end;

function TCustomSyntaxMemo.FindPrevWord(var wx, wy: Integer): Boolean;
var Sp: TSyntaxParser;
    function FindLastWordOfPrevRow: Boolean;
    begin
        wy := FindVisibleRow(wy, -1, False);
        Result := wy >= 0;
        if Result then begin
            Sp := fLines.Parser[wy];
            if Sp.Count > 0
                then
                    with Sp[Sp.Count - 1] do
                        wx := stStart + stLength
                else wx := 0;
        end;
    end;
begin
    Result := True;
    wx := CharPosToWordIndex(Range.PosX, Range.PosY);
    wy := Range.PosY;
    Sp := fLines.Parser[wy];
    {       }
    if wx < 0 then
        {    -    }
        if wx = -1
            then Result := FindLastWordOfPrevRow
            else wx := Sp[-wx - 1].stStart
    else
    {      }
    if wx = MAXINT then
        if Sp.Count = 0
            then Result := FindLastWordOfPrevRow
            else wx := Sp[Sp.Count-1].stStart
    else
    {      }
    if Sp[wx].stStart = Range.PosX then
        {    }
        if wx = 0
            then Result := FindLastWordOfPrevRow
            else wx := Sp[wx - 1].stStart
    else
        {    }
        wx := Sp[wx].stStart;
end;

procedure TCustomSyntaxMemo.FontChange(Sender: TObject);
begin
    fLines.BeginUpdate;
    CalcScreenParams;
    NeedRedrawAll;
    fLines.EndUpdate;
end;

procedure TCustomSyntaxMemo.SetSelectedWord(const Value: TPoint);
begin
    fLines.BeginUpdate;
    if fLines.IsValidLineIndex( fSelWord.Y ) then NeedRedraw(fSelWord.Y);
    fSelWord := Point(-1, -1);
    if fLines.IsValidLineIndex(Value.Y) and Assigned(fHighliter) then begin
        ShowWord(Value.Y, Value.X);
        fSelWord := Value;
        NeedRedraw(fSelWord.Y);
    end;
    fLines.EndUpdate;
end;

procedure TCustomSyntaxMemo.ShowWord(const Row, WordIndex: Integer);
begin
    if fLines.IsValidLineIndex(Row) then
        with fLines.Parser[Row] do
            if InRange(WordIndex, 0, Count - 1) then
                with Tokens[WordIndex] do
                    MakeVisible(stStart, Row, stLength);
end;

function TCustomSyntaxMemo.GetWordAtPos(const X, Y: Integer; var WordIndex, Row: Integer): Boolean;
var n, Row1: Integer;
begin
    WndOffsetToPixOffset(Point(X, Y), n, Row1, True);
    Result := fLines.IsValidLineIndex(Row1);
    if Result then begin
        Row := Row1;
        PixOffsetToCharPos(n, Row1, @WordIndex);
    end;
end;

procedure TCustomSyntaxMemo.Navigate(const Col, Row: Integer);
begin
  if not fLines.IsValidLineIndex(Row-1) then Exit;
  fLines.BeginUpdate;
  fSections.Expand(Row-1, False, True);
  SetOffsets(Point(0, FindVisibleRow(Row-1, -(ClientLines div 2), True)));
  //SetOffsets(Point(0, FindVisibleRow(0, 0, True)));
  fRange.Pos := Point(Col, Row-1);
  fVScroll.Position:=(Row-1-(ClientLines div 2))*fCharHeight;
  //fHScroll.Position:=(Col*8)-20*8; //!!!!!***
  //fLines[Row-1];
  fLines.EndUpdate;
end;

procedure TCustomSyntaxMemo.ErrorLine(const Col, Row: Integer);
begin
  fErrorLine:=Row-1;
  Navigate(Col, Row);
end;

procedure TCustomSyntaxMemo.NaviLine(const Col, Row: Integer);
begin
  fNaviLine:=Row-1;
  Navigate(Col, Row);
end;

procedure TCustomSyntaxMemo.FindWord(const Col,Row: integer; frDown,frMatchCase: boolean);
var
  SelPos,X,i:integer;
  s,textE:string;
begin

  if Row<>-1 then
    fRange.Position:=fLines.RCToPosition(Col, Row); //   
   
  if frDown then
    fRange.Position:=fRange.Position+fRange.SelLength ; //   " "

  if frMatchCase then
    textE:=fRange.fRichMemo.fLines.Text
  else
    textE:=UpperCase(fRange.fRichMemo.fLines.Text);

  X:=fRange.Position+1;

  if frDown then
    textE:=Copy(textE,X,Length(textE)-X+1)
  else
    textE:=Copy(textE,0,X-1);

  if frMatchCase then
    s:=fWord
  else
    s:=UpperCase(fWord);

  fLines.BeginUpdate;

  if frDown then begin
    SelPos := Pos(s,textE);
    if SelPos > 0 then begin
      fRange.Position := SelPos + X - 2;
      fRange.SelLength:=length(s);
      //SetOffsets(Point(0, FindVisibleRow(fRange.PosY, -(ClientLines DIV 2), True)));
      fVScroll.Position:=(fRange.PosY-(ClientLines div 2))*fCharHeight;
    end else
      MessagesFrm.MsgBox.MessageForm('Search string "'+fWord+'" not found.',
                                         'Warning !!!',MB_ICONEXCLAMATION,MB_OK);
  end else begin
    for i:=length(textE)-length(s) downto 1 do
      if copy(textE,i,length(s))=s then begin
        fRange.Position := i-1;
        fRange.SelLength:=Length(s);
        //SetOffsets(Point(0, FindVisibleRow(fRange.PosY, -(ClientLines DIV 2), True)));
        fVScroll.Position:=(fRange.PosY-(ClientLines div 2))*fCharHeight;
        Exit;
      end;
     MessagesFrm.MsgBox.MessageForm('Search string "'+fWord+'" not found.',
                                         'Warning !!!',MB_ICONEXCLAMATION,MB_OK);
  end;

  fLines.EndUpdate;

  InvalidateRect(Self.Handle,nil,false);
end;

//     PosX   PosY
procedure TCustomSyntaxMemo.MakeVisible(const Col, Row: Integer; const Length: Integer = 1);
var RowPix, CharPix, CharPixLen: Integer;
    NewOffsets: TPoint;
    R: TRect;
begin
    // 
    RowPix := RangeRowToScreenRow(Row);
    if RowPix = ROW_HIDEN then Exit else
    if RowPix = ROW_ABOVE_SCREEN then NewOffsets.Y := Row else
    if RowPix = ROW_BELOW_SCREEN then
        NewOffsets.Y := FindVisibleRow(Row, -(ClientLines - 1), True)
    else
        NewOffsets.Y := OffsetY;

    // 
    R := TextRowRect[0];
    CharPixLen := Length * 10;
    CharPix := CharPosToPixOffset(Col, Row);
    if CharPix < OffsetXPix then NewOffsets.X := CharPix else
    if CharPix + CharPixLen - OffsetXPix > R.Right - R.Left then
        NewOffsets.X := CharPix + CharPixLen - R.Right + R.Left
    else
        NewOffsets.X := OffsetXPix;

    //  
    if (NewOffsets.X <> OffsetXPix) or (NewOffsets.Y <> OffsetY) then
        SetOffsets(NewOffsets);
end;

function TCustomSyntaxMemo.CharPosToWordIndex(const Col, Row: Integer): Integer;
var
  Sp: TSyntaxParser;
  i, WBeg: Integer;
begin
  Result := MAXINT;
  if not InRange(Row, 0, fLines.Count - 1) then Exit;
  Sp := TSyntaxParser(fLines.Objects[Row]);
  if Sp <> nil then
    for i := 0 to Sp.Count - 1 do begin
      WBeg := Sp[i].stStart;
      if Col < WBeg then begin
        Result := -i;
        Exit;
      end else
      if Col < WBeg + Sp[i].stLength then begin
        Result := i;
        Exit;
      end;
    end;
end;

function TCustomSyntaxMemo.GetUserTokenEvent: TUserTokenEvent;
begin
  if Assigned(fParseAttributes) then
    Result := fParseAttributes.OnUserToken
  else
    Result := nil;
end;

procedure TCustomSyntaxMemo.SetUserTokenEvent(const Value: TUserTokenEvent);
begin
  if Assigned(fParseAttributes) then fParseAttributes.OnUserToken := Value;
end;

procedure TCustomSyntaxMemo.SetGutterWidth(const Value: Integer);
begin
  if fGutterWidth <> Value then begin
    fGutterWidth := Value;
    Repaint;
  end;
end;

function TCustomSyntaxMemo.IsLineVisible(const Row: Integer; const PScreenRow: PInteger = nil): Boolean;
var
  sr: Integer;
begin
  sr := RangeRowToScreenRow(Row);
  Result := InRange(sr, 0, ClientLines - 1);
  if Assigned(PScreenRow) then PScreenRow^ := sr;
end;

procedure TCustomSyntaxMemo.SetSectionIndent(const Value: Integer);
begin
  fSectionIndent := Value;
  NeedRedrawAll;
end;

function TCustomSyntaxMemo.GetWndRect(const ScreenRow, Index: Integer): TRect;
begin
  case Index of
    0:  begin
            if ScreenRow = -1 then begin
              Result := ClientRect;
            end else
              Result := Bounds(   0,
                                    ScreenRow * fCharHeight,
                                    ClientWidth,
                                    Min(fCharHeight, ClientHeight - ScreenRow * fCharHeight)  );
            Dec(Result.Right);
        end;

    1:  begin
            Result := EntireRowRect[ScreenRow];
            Result.Left := EntireGutterRect[ScreenRow].Right;
        end;

    2:  begin
            Result := SymbolsGutterRect[ScreenRow];
            Inc(Result.Right, fSectionIndent + 2);
        end;

    3:  begin
            Result := EntireRowRect[ScreenRow];
            Result.Right := Result.Left + fGutterWidth;
        end;
  end;
end;

function TCustomSyntaxMemo.GetSectionButtonRect(const ScreenRow, ALevel: Integer): TRect;
begin
  Result := SymbolsGutterRect[ScreenRow];
  Result.Left := Result.Right + 2;
  Result.Right  := Result.Left + 9;
  Result.Top    := (Result.Top + Result.Bottom) shr 1 - 4;
  Result.Bottom := Result.Top + 9;
end;

procedure TCustomSyntaxMemo.HideCaret;
begin
  if fCaretVisible and Assigned(Parent) then begin
    Windows.HideCaret(Handle);
    fCaretVisible := False;
  end;
end;

procedure TCustomSyntaxMemo.ShowCaret;
var
  n, ScreenRow: Integer;
  Cp: TPoint;
begin
  if  (Lines.UpdateCount = 0)
  and Lines.IsValidLineIndex(Range.PosY)
  and IsLineVisible(Range.PosY, @ScreenRow) then begin
    n := CharPosToPixOffset(Range.PosX, Range.PosY);
    Cp := PixOffsetToWndOffsetEx(n, ScreenRow);
    with TextRowRect[ScreenRow] do  // new
      if InRange(Cp.X, Left, Right - 1) then begin  // new
        Windows.SetCaretPos(Cp.X-1, Cp.Y+1);
        Windows.ShowCaret(Handle);
        fCaretVisible := True;
      end;
  end;
end;

function TCustomSyntaxMemo.RangeRowToScreenRow(const Row: Integer): Integer;
begin
  if Row < OffsetY then
    Result := ROW_ABOVE_SCREEN
  else
  if fLines.Parser[Row].VisibleIndex < 0 then
    Result := ROW_HIDEN
  else begin
    Result := RowIndexConvert(Row, cdNeedScreen) - RowIndexConvert(OffsetY, cdNeedScreen);
    if Result >= ClientLines then Result := ROW_BELOW_SCREEN;
  end;
end;

function TCustomSyntaxMemo.FindVisibleRow(const Row, Delta: Integer; const EnsureInRange: Boolean): Integer;
var
  n: Integer;
begin
  n := RowIndexConvert(Row, cdNeedScreen) + Delta;
  if EnsureInRange then n := EnsureRange(n, 0, fSections.Indexes.Count - 1);
  Result := RowIndexConvert(n, cdNeedReal);
end;

function TCustomSyntaxMemo.RowIndexConvert(const Index: Integer; const Direction: TRowIndexConvertionDirection): Integer;
begin
  Result := -1;
  case Direction of
  cdNeedReal:     if InRange(Index, 0, fSections.Indexes.Count - 1)
                      then Result := Integer(fSections.Indexes[Index]);
  cdNeedScreen:   if fLines.IsValidLineIndex(Index)
                      then Result := fLines.Parser[Index].VisibleIndex;
  end;
end;

procedure TCustomSyntaxMemo.SetOffsets(NewOffsets: TPoint);
begin
  if NewOffsets.X <> fOffsets.X then
    NewOffsets.X := EnsureRange(NewOffsets.X, 0, fHScroll.Range);
  if NewOffsets.Y <> fOffsets.Y then begin
    NewOffsets.Y := EnsureRange(RowIndexConvert(NewOffsets.Y, cdNeedScreen), 0, fVScroll.Range);
    NewOffsets.Y := RowIndexConvert(NewOffsets.Y, cdNeedReal);
  end;
  if (NewOffsets.X <> fOffsets.X) or (NewOffsets.Y <> fOffsets.Y) then begin
    fOffsets := NewOffsets;
    NeedRedrawAll;
    UpdateScrollBars;
  end;
end;

procedure TCustomSyntaxMemo.SetOffset(const Index, Value: Integer);
begin
  case Index of
    0:  SetOffsets(Point(Value, fOffsets.Y));
    1:  SetOffsets(Point(fOffsets.X, Value));
  end;
end;

procedure TCustomSyntaxMemo.Reset;
begin
  fOffsets        := Point(0, 0);
  fDown           := False;
  fOptions        := fOptions - [smoOverwrite, smoReadOnly];
  fBookMarks.Clear;
  fRange.fPos     := fOffsets;
  fRange.fStart   := fOffsets;
  fRange.fEnd     := fOffsets;
  fRange.fSealing := True;
  fRange.fUndoStack.Clear;
end;

procedure TCustomSyntaxMemo.SetOptions(const Value: TSynMemoOptions);
var
  oi: TSynMemoOption;
  os: TSynMemoOptions;
begin
  if fOptions <> Value then begin
    os := fOptions;
    fOptions := Value;
    for oi := Low(TSynMemoOption) to High(TSynMemoOption) do
      if [oi]*os <> [oi]*Value then begin
        case oi of
          smoAutoGutterWidth: NeedRedrawAll;
        end;
      end;
    Change([ciOptions]);
  end;
end;

procedure TCustomSyntaxMemo.Change(const ChangedItems: TChangedItems);
begin
  if ChangedItems = [] then
    fChangesSummator := []
  else begin
    fChangesSummator := fChangesSummator + ChangedItems;
    if fLines.UpdateCount = 0 then begin
      if Assigned(fOnChange) then fOnChange(self, fChangesSummator);
      fChangesSummator := [];
    end;
  end;
end;

procedure TCustomSyntaxMemo.Paint;
begin
  if Parent = nil then Exit;
  NeedRedrawAll;
end;

procedure TCustomSyntaxMemo.NeedRedraw(const Row: Integer);
var
  Index: Integer;
begin
  if Parent = nil then Exit;
  if IsLineVisible(Row, @Index) then fScreenLines[Index] := True;
  ReDraw;
end;

procedure TCustomSyntaxMemo.NeedReDrawLE(const Row: Integer);
var Index: Integer;
begin
  if Parent = nil then Exit;
  if IsLineVisible(Row, @Index) or (Index<0) then
  begin
    if Index<0 then Index := 0;
    while Index <= High(fScreenLines) do begin
      fScreenLines[Index] := True;
      Inc(Index);
    end;
  end;
  ReDraw;
end;

procedure TCustomSyntaxMemo.NeedRedrawAll;
var
  i: Integer;
begin
  for i := Low(fScreenLines) to High(fScreenLines) do fScreenLines[i] := True;
  ReDraw;
end;

procedure TCustomSyntaxMemo.ReDraw;
var
  i: Integer;
begin
  if (Parent = nil) or (Lines.UpdateCount > 0) then Exit;
  HideCaret;
  for i := Low(fScreenLines) to High(fScreenLines) do
    if fScreenLines[i] then begin
      PaintLine(i, FindVisibleRow(OffsetY, i, False));
      fScreenLines[i] := False;
    end;
  ShowCaret;
end;

procedure TCustomSyntaxMemo.SetDefColor(const Index: Integer; const Value: TColor);
begin
    case Index of
    0: fDefBackColor := Value;
    1: fDefForeColor := Value;
    end;
    InvalidateRect(Handle,nil,false);
end;

procedure TCustomSyntaxMemo.PaintLine(const ScreenRow, Row: Integer);
var
  Sp                          : TSyntaxParser;
  TextIndent,SelStart,SelEnd  : Integer;
  ClipRgn                     : HRGN;
  R                           : TRect;
  fBuffer                     : TBitmap;
begin

  if not Visible then Exit;

  with EntireRowRect[0] do begin
    fBuffer        := TBitmap.Create;
    fBuffer.Height := Bottom - Top;
    fBuffer.Width  := Right  - Left;
  end;

  PaintGutter(fBuffer.Canvas, Row, ScreenRow);

  if Lines.IsValidLineIndex(Row) then begin

    PaintSectionMarks(fBuffer.Canvas, Row, ScreenRow );
    TextIndent := PixOffsetToWndOffsetEx(0, ScreenRow).X;
    Sp := TSyntaxParser.Create( Lines.Parser[Row] );
    if not fRange.IsEmpty()
    and InRange(Row, fRange.StartY, fRange.EndY) then begin
      SelStart := IfThen(Row = fRange.StartY, fRange.StartX, -1);
      SelEnd   := IfThen(Row = fRange.EndY, fRange.EndX, MAXINT);
      Sp.SplitTokens(SelStart, SelEnd);
    end else begin
      SelStart := 0;
      SelEnd := 0;
    end;

    if (Row=fErrorLine) and (fErrorLine<>-1) then
      with fBuffer.Canvas do begin
        Brush.Color:=$0000ff;
        FillRect(TextRowRect[0]);
      end
    else if (Row=fNaviLine) and (fNaviLine<>-1) then
      with fBuffer.Canvas do begin
        Brush.Color:=$ffDDFF;
        FillRect(TextRowRect[0]);
      end
    else if (SelStart=SelEnd) and (Row=Range.PosY) then
      with fBuffer.Canvas do begin
        Brush.Color:=$ddffff;
        FillRect(TextRowRect[0]);
      end;

    with TextRowRect[ScreenRow] do ClipRgn := CreateRectRgn(Left, 0, Right, Bottom - Top);
    SelectClipRgn(fBuffer.Canvas.Handle, ClipRgn);

    PaintTokens( fBuffer.Canvas, Lines[Row], Row, Sp, TextIndent, SelStart, SelEnd );
    if fSections.Section[Row].Collapsed then PaintDots(fBuffer.Canvas);

    if (fSelWord.Y = Row) and Assigned(fHighliter) then
      with fLines.Parser[Row].Tokens[fSelWord.X] do begin
        if stToken<>tokBlank then begin
          R.Left  := TextIndent + CharPosToPixOffset(stStart, Row);
          R.Right := TextIndent + CharPosToPixOffset(stStart+stLength, Row);
          R.Top := 0;
          R.Bottom := fCharHeight - 1;
          fBuffer.Canvas.Pen.Color := clBlue;
          fBuffer.Canvas.MoveTo(R.Left,R.Bottom-2);
          fBuffer.Canvas.LineTo(R.Right,R.Bottom-2);
        end;
      end;

    SelectClipRgn(fBuffer.Canvas.Handle, 0);
    DeleteObject(ClipRgn);
    Sp.Free;

  end;

  with EntireRowRect[ScreenRow] do
      BitBlt(Canvas.Handle,0,Top,Right-Left,Bottom-Top,fBuffer.Canvas.Handle,0,0,SRCCOPY);
  fBuffer.Free;
end;

//     
procedure TCustomSyntaxMemo.PaintGutter(const ACanvas: TCanvas; const Row, ScreenRow: Integer);
var
  R: TRect;
  i: TBookmarkIndex;
  N: String;
  SelStart,SelEnd  : Integer;
begin
  N:=IntToStr(Row+1);
  R := EntireGutterRect[ScreenRow];
  Dec(R.Bottom,R.Top);
  R.Top := 0;
  with ACanvas do begin
    Brush.Style := bsSolid;
    Brush.Color := clBtnFace;
    DrawGradient(ACanvas,$88ccff,$eeffff,64,R,true);
    Pen.Color:=$88ccff;
    MoveTo(R.Right-2,R.Top);
    LineTo(R.Right-2,R.Bottom);
    Pen.Color:=$ffffff;
    MoveTo(R.Right-3,R.Top);
    LineTo(R.Right-3,R.Bottom);
    if not Lines.IsValidLineIndex(Row) then Exit;
    Brush.Style:=bsClear;

    if not fRange.IsEmpty() and InRange(Row, fRange.StartY, fRange.EndY) then begin
      SelStart := IfThen(Row = fRange.StartY, fRange.StartX, -1);
      SelEnd   := IfThen(Row = fRange.EndY, fRange.EndX, MAXINT);
    end else begin
      SelStart := 0;
      SelEnd := 0;
    end;

    if (fErrorLine<>-1) and (Row=Range.PosY) then begin
      Font.Color:=$0000ff;
      Font.Style:=[fsBold];
    end else if (fNaviLine<>-1) and (Row=Range.PosY) then begin
      Font.Color:=$0000ff;
      Font.Style:=[fsBold];
    end else if (SelStart=SelEnd) and (Row=Range.PosY) then
      Font.Color:=$008800
    else
      Font.Color:=$0088ff;

    Dec(R.Right,16);
    DrawText(ACanvas.Handle,PChar(N),Length(N),R,DT_VCENTER or DT_RIGHT or DT_SINGLELINE);
    Inc(R.Right,16);

    if fBookMarks.Find(Row, i) then
       BitBlt(ACanvas.Handle, R.Left + 2, R.Top + 2,11,11,
       fBookMarks.fImages.Canvas.Handle,i * 11,0,SRCCOPY );
  end;
end;

//   
procedure TCustomSyntaxMemo.PaintSectionMarks(const ACanvas: TCanvas; const Row, ScreenRow: Integer);
var
  SecPnt    : TPoint;
  Sec       : TSynMemoSection;
  MR, RR, GR: TRect;
begin
  RR := TextRowRect[ScreenRow];
  Dec(RR.Bottom, RR.Top);
  RR.Top := 0;

  GR := EntireGutterRect[ScreenRow];
  Dec(GR.Bottom, GR.Top);
  GR.Top := 0;

  Sec      := fSections.Section[Row];
  MR       := GetSectionButtonRect(ScreenRow, Sec.Level);
  Inc(MR.Left,2);
  Inc(MR.Right,2);
  Dec(MR.Top, EntireRowRect[ScreenRow].Top);
  Dec(MR.Bottom, EntireRowRect[ScreenRow].Top);
  SecPnt   := CenterPoint(MR);

  with ACanvas do begin
    Pen.Color   := $aaaaff;
    Brush.Color := $aaaaff;
    if Sec.RowBeg = Row then begin          //   -  
      FrameRect(MR);
      with MR do begin
        MoveTo(Left + 2, SecPnt.Y);         // -> -
        LineTo(Right - 2, SecPnt.Y);
        if Sec.Collapsed then begin         // -> +
          MoveTo(SecPnt.X, Top + 2);
          LineTo(SecPnt.X, Bottom - 2);
        end;
        MoveTo(Right, SecPnt.Y);            //    
        LineTo(GR.Right - 2, SecPnt.Y);
        if not Sec.Collapsed then begin     //   
          MoveTo(SecPnt.X, Bottom);
          LineTo(SecPnt.X, GR.Bottom);
        end;
      end;
    end else if Sec.RowEnd = Row then begin //   -  . 
      MoveTo(SecPnt.X, GR.Top);
      LineTo(SecPnt.X, SecPnt.Y);
      LineTo(GR.Right - 2, SecPnt.Y);
    end else if Sec.Level > 0 then begin    //  ,  - 
      MoveTo(SecPnt.X, GR.Top);
      LineTo(SecPnt.X, GR.Bottom);
    end;
  end;

end;

procedure TCustomSyntaxMemo.PaintDots(const ACanvas: TCanvas);
var
  R: TRect;
begin
  with ACanvas do begin
    R := ClipRect;
    R.Left := R.Right - 33;
    Brush.Style := bsSolid;
    Brush.Color := self.Color;
    FillRect( R );
    with ClipRect do R := Rect(Right - 32, Top + 1, Right - 10, Bottom - 1);
    Brush.Color := clBlue;
    FrameRect( R );
    R := Bounds(R.Left + 5, R.Top + 8, 2, 2);
    FillRect( R );
    OffsetRect(R, 5, 0);
    FillRect( R );
    OffsetRect(R, 5, 0);
    FillRect( R );
  end;
end;

procedure TCustomSyntaxMemo.PaintTokens(const ACanvas: TCanvas; s: string; Row: integer; Sp: TSyntaxParser; TextIndent, SelStart, SelEnd: Integer);
var
  wi,i,CharPos : Integer;
  col,iColor   : TColor;
  R            : TRect;

  function ReplaceStr(const S, Srch, Replace: string): string;
  {   }
  var
    I:Integer;
    Source:string;
  begin
    Source:= S;
    Result:= '';
    repeat
      I:=Pos(Srch, Source);
      if I > 0 then begin
        Result:=Result+Copy(Source,1,I-1)+Replace;
        Source:=Copy(Source,I+Length(Srch),MaxInt);
      end else
        Result:=Result+Source;
    until I<=0;
  end;
  
begin
    with ACanvas do begin
        R := ClipRect;
        //  
        Brush.Style := bsSolid;
        for wi := 0 to Sp.Count - 1 do
            with Sp[wi] do begin
                col := fParseAttributes.BackColor[stToken];
                if (col <> clDefault) and not (stsInSelection in stStyle) then begin
                    R.Left := TextIndent + CharPosToPixOffset(stStart, s, Sp);
                    R.Right := TextIndent + CharPosToPixOffset(stStart + stLength, s, Sp);
                    Brush.Color := col;
                    FillRect( R );
                end;
            end;
        //   
        if SelStart <> SelEnd then begin
            R.Left := TextIndent;
            if SelStart > 0 then
                Inc(R.Left, CharPosToPixOffset(fRange.StartX, s, Sp));
            if SelEnd < MAXINT
                then R.Right := TextIndent + CharPosToPixOffset(fRange.EndX, s, Sp)
                else R.Right := ClipRect.Right;
            Brush.Color := fSelColor;
            FillRect( R );
        end;
        //     
        Brush.Style := bsClear;
        Font.Assign(Self.Font);
        PenPos := Point(TextIndent, 0);
        CharPos := 0;
        for wi := 0 to Sp.Count - 1 do
            with Sp[wi] do begin
                with fParseAttributes.fTokenStyles[stToken] do begin
                    Font.Style := tsStyle;
                    if stsInSelection in stStyle then
                      Font.Color := clWhite
                    else if tsForeground = clDefault then
                      Font.Color := fDefForeColor
                    else
                      Font.Color := tsForeground;
                end;

                if (Row=fErrorLine) and (fErrorLine<>-1) then Font.Color := clWhite;

                R.TopLeft := PenPos;
                Inc(R.Left, (stStart-CharPos) * fCharWidths[0][' ']);

                TextOut(R.Left, -1, Copy(s, stStart + 1, stLength));

                if (stToken=tokError) or (stToken=tokOk) then begin
                  iColor:=clRed;
                  if stToken in [tokOk] then iColor:=clGreen;
                  R.Left  := TextIndent + CharPosToPixOffset(stStart, s, Sp);
                  R.Right := TextIndent + CharPosToPixOffset(stStart + stLength, s, Sp);
                  for i:=R.Left to R.Right do
                    case (i mod 4) of
                       0,2: Pixels[i,R.Bottom-2]:=iColor;
                       1  : Pixels[i,R.Bottom-3]:=iColor;
                       3  : Pixels[i,R.Bottom-1]:=iColor;
                    end;
                end;

                CharPos := stStart + stLength;
            end;
    end;
end;

initialization
  CF_SYNTAX := RegisterClipboardFormat('Syntax Memo Sections');

end.


