MODULE Forms;

IMPORT
  SYSTEM, Windows, Messages, Controls, SysUtils, Dialogs;

VAR
  defaultBrush: Windows.HBrush;
  Application*: TApplication;

PROCEDURE [WINAPI] WndProc* (wnd: Windows.HWND; msg: LONGINT; wParam: Windows.WParam; lParam: Windows.LParam): Windows.LResult;
VAR
  Form : SYSTEM.PTR;
BEGIN

  Form:= Application.List.Find(wnd);
  IF Form#NIL THEN Application.Exec(Form,msg,wParam,lParam); END;

  CASE msg OF
      Messages.WM_CLOSE:
                              IF wnd=Application.MainForm(TForm).Handle THEN
                                Application.fTerminated:=TRUE;
                                Windows.DeleteObject(defaultBrush);
                                Windows.PostQuitMessage(0);
                              ELSE
                                Windows.ShowWindow(wnd, Windows.SW_HIDE);
                              END;
                              RETURN 0;
  ELSE
    RETURN Windows.DefWindowProc(wnd, msg, wParam, lParam)
  END;
  RETURN 0

END WndProc;

TYPE
  TList*       = POINTER TO RECORD
    Next     : TList;
    Pointer* : SYSTEM.PTR;
    Handle*  : Windows.HWND;

    PROCEDURE Find*(H : Windows.HWND): SYSTEM.PTR;
    VAR
      Result: SYSTEM.PTR;
      List  : TList;
    BEGIN
      Result:=NIL;
      List:=Next;
      WHILE (List#NIL) & (Result=NIL) DO
        IF List.Handle=H THEN
          Result:=List.Pointer;
        END;
        List:=List.Next;
      END;
      RETURN Result;
    END Find;

    PROCEDURE Add*(P : SYSTEM.PTR; H : Windows.HWND);
    VAR
      New   : TList;
    BEGIN
      IF Find(H)#NIL THEN RETURN END;
      NEW(New);
      New.Pointer:=P;
      New.Next:=Next;
      New.Handle:=H;
      Next:=New;
    END Add;

  END;

  TForm*       = POINTER TO RECORD (Controls.TWinControl);

    PROCEDURE Create*;
    BEGIN
      Left   := 100;
      Top    := 100;
      Width  := 320;
      Height := 240;
      Caption:='Form';
    END Create;

    PROCEDURE CreateForm*;
    VAR
      Style: SET;
      res  : LONGINT;
      rect : Windows.TRect;
    BEGIN
      Style:=Windows.WS_OVERLAPPEDWINDOW;
      Handle := Windows.CreateWindowEx(
         Windows.WS_EX_APPWINDOW,
         Application.ClassName, Caption, Style,
         Left, Top, Width, Height, 0, 0, Windows.hInstance, 0
      );
      Create;
      Windows.SetWindowPos(Handle, Windows.HWND_NOTOPMOST,Left,Top,Width,Height,Windows.SWP_HIDEWINDOW);
    END CreateForm;

    PROCEDURE WindowProc*(msg,wParam,lParam: LONGINT);
    BEGIN
    END WindowProc;

    PROCEDURE Show*;
    BEGIN
      Windows.ShowWindow(Handle, Windows.SW_SHOW);
      Windows.UpdateWindow(Handle);
    END Show;

    PROCEDURE Hide*;
    BEGIN
      Windows.ShowWindow(Handle, Windows.SW_HIDE);
      Windows.UpdateWindow(Handle);
    END Hide;

  END;

  TApplication = POINTER TO RECORD

    ClassName* : ARRAY 32 OF CHAR;
    MainForm*  : SYSTEM.PTR;
    List*  : TList;
    fTerminated: BOOLEAN;

    PROCEDURE Initialize*;
    VAR
      class     : Windows.WndClassEx;
    BEGIN
      NEW(List);
      List.Next:=NIL;
      ClassName           :='FORM';
      defaultBrush        := Windows.CreateSolidBrush(Windows.GetSysColor(Windows.COLOR_BTNFACE));
      class.cbSize        := SIZE(Windows.WndClassEx);
      class.style         := Windows.CS_VREDRAW + Windows.CS_HREDRAW;
      class.lpfnWndProc   := WndProc;
      class.cbClsExtra    := 0;
      class.cbWndExtra    := 0;
      class.hInstance     := Windows.hInstance;
      class.hIcon         := Windows.LoadIcon(Windows.hInstance,Windows.IDI_APPLICATION);
      class.hCursor       := Windows.LoadCursor(0,SYSTEM.VAL(Windows.LPSTR,Windows.IDC_ARROW));
      class.hbrBackground := defaultBrush;
      class.lpszMenuName  := Windows.NULL;
      class.lpszClassName := SYSTEM.ADR(Application.ClassName);
      class.hIconSm       := Windows.NULL;
      Windows.RegisterClassEx(class);
    END Initialize;

    PROCEDURE Exec*(Form: SYSTEM.PTR; msg: LONGINT; wParam: Windows.WParam; lParam: Windows.LParam);
    BEGIN
      IF MainForm=NIL THEN RETURN END;
      Form(TForm).WindowProc(msg,wParam,lParam);
    END Exec;

    PROCEDURE CreateForm*(Form: SYSTEM.PTR);
    BEGIN
      IF MainForm=NIL THEN MainForm:=Form; END;
      Form(TForm).CreateForm;
      List.Add(Form,Form(TForm).Handle);
    END CreateForm;

    PROCEDURE Run*;
    BEGIN
      IF MainForm=NIL THEN RETURN END;
      MainForm(TForm).Show;
      REPEAT
        ProcessMessages;
      UNTIL fTerminated;
      Windows.UnregisterClass('FORM',Windows.hInstance);
      Windows.Shutdown(0);
      Windows.ExitProcess(0);
    END Run;

    PROCEDURE ProcessMessages*;
    VAR
      msg : Messages.Msg;
    BEGIN
      Windows.GetMessage(msg,0,0,0);
      Windows.TranslateMessage(msg);
      Windows.DispatchMessage(msg);
    END ProcessMessages;

    PROCEDURE Terminate*;
    BEGIN
      fTerminated:=TRUE;
    END Terminate;

  END;

BEGIN
  NEW(Application);
END Forms.
