MODULE SysUtils;

IMPORT
  SYSTEM, Windows;

VAR
   DefaultFCR*: SET;
   tene    : ARRAY 23 OF LONGREAL; (* e = 0..22: exact values of 10^e *)
   ten     : ARRAY 27 OF LONGREAL;
   eq, gr  : ARRAY 20 OF SET;
  H, L    : INTEGER;

TYPE
  TRect* = Windows.TRect;

PROCEDURE Rect*(Left,Top,Width,Height: LONGINT): TRect;
VAR
  R: TRect;
BEGIN
    R.left:=Left;
    R.top:=Top;
    R.right:=Width;
    R.bottom:=Height;
    RETURN R;
END Rect;

PROCEDURE Length*(VAR str: ARRAY OF CHAR): LONGINT;
VAR
  i, l: LONGINT;
BEGIN
    l := LEN(str);
    i := 0;
    WHILE (i < l) & (str[i] # 0X) DO
       INC(i);
    END;
    RETURN i;
END Length;

PROCEDURE Expo* (x: REAL): LONGINT;
BEGIN
   RETURN ASH(SYSTEM.VAL(LONGINT, x), -23) MOD 256;
END Expo;

PROCEDURE ExpoL* (x: LONGREAL): LONGINT;
VAR
  i: LONGINT;
BEGIN
   SYSTEM.GET(SYSTEM.ADR(x) + H, i);
   RETURN ASH(i, -20) MOD 2048;
END ExpoL;

PROCEDURE SetExpo* (e: LONGINT; VAR x: REAL);
VAR
  i: LONGINT;
BEGIN
   SYSTEM.GET(SYSTEM.ADR(x), i);
   i := ASH(ASH(ASH(i, -31), 8) + e MOD 256, 23) + i MOD ASH(1, 23);
   SYSTEM.PUT(SYSTEM.ADR(x), i);
END SetExpo;

PROCEDURE SetExpoL* (e: LONGINT; VAR x: LONGREAL);
VAR
  i: LONGINT;
BEGIN
   SYSTEM.GET(SYSTEM.ADR(x) + H, i);
   i := ASH(ASH(ASH(i, -31), 11) + e MOD 2048, 20) + i MOD ASH(1, 20);
   SYSTEM.PUT(SYSTEM.ADR(x) + H, i);
END SetExpoL;

PROCEDURE Real* (h: LONGINT): REAL;
VAR
  x: REAL;
BEGIN
  SYSTEM.PUT(SYSTEM.ADR(x), h);
  RETURN x;
END Real;

PROCEDURE RealL* (h, l: LONGINT): LONGREAL;
VAR
  x: LONGREAL;
BEGIN
  SYSTEM.PUT(SYSTEM.ADR(x) + H, h);
  SYSTEM.PUT(SYSTEM.ADR(x) + L, l);
  RETURN x;
END RealL;

PROCEDURE Ten* (e: LONGINT): LONGREAL;
VAR
  E: LONGINT;
  r: LONGREAL;
BEGIN
   IF e < -307 THEN
     RETURN 0
  ELSIF 308 < e THEN
    RETURN RealL(2146435072, 0)
  END;
   INC(e, 307);
   r:= ten[e DIV 23] * tene[e MOD 23];
   IF e MOD 32 IN eq[e DIV 32] THEN
     RETURN r
   ELSE
      E:= ExpoL(r);
      SetExpoL(1023+52, r);
      IF e MOD 32 IN gr[e DIV 32] THEN
        r:= r-1
      ELSE
        r:= r+1
      END;
      SetExpoL(E, r);
      RETURN r
   END
END Ten;

PROCEDURE IntToStr*(val: LONGINT): ARRAY OF CHAR;
VAR
  i, j  : LONGINT;
  str   : ARRAY 16 OF CHAR;
  digits: ARRAY 16 OF LONGINT;
BEGIN
    IF val = MIN(LONGINT) THEN
       COPY("-2147483648", str);
       RETURN str
    END;
    IF val < 0 THEN
       val := -val;
       str[0] := "-";
       j := 1
    ELSE
       j := 0
    END;
    i := 0;
    REPEAT
       digits[i] := val MOD 10;
       INC(i);
       val := val DIV 10
    UNTIL val = 0;
    DEC(i);
    WHILE i >= 0 DO
       str[j] := CHR(digits[i]+ORD("0"));
       INC(j);
       DEC(i)
    END;
    str[j] := 0X;
    RETURN str;
END IntToStr;

PROCEDURE StrToInt*(str: ARRAY OF CHAR): LONGINT;
VAR
  i, d: LONGINT;
  ch  : CHAR;
  neg : BOOLEAN;
  val : LONGINT;
BEGIN
    i := 0;
    ch := str[0];
    WHILE (ch # 0X) & (ch <= " ") DO
       INC(i);
       ch := str[i]
    END;
    neg := FALSE;
    IF ch = "+" THEN
      INC(i);
      ch := str[i];
    END;
    IF ch = "-" THEN
      neg := TRUE;
      INC(i);
      ch := str[i]
  END;
    WHILE (ch # 0X) & (ch <= " ") DO
       INC(i);
       ch := str[i]
    END;
    val := 0;
    WHILE (ch >= "0") & (ch <= "9") DO
       d := ORD(ch)-ORD("0");
       INC(i);
       ch := str[i];
       IF val <= ((MAX(LONGINT)-d) DIV 10) THEN
          val := 10*val+d
       ELSIF neg & (val = 214748364) & (d = 8) & ((ch < "0") OR (ch > "9")) THEN
          val := MIN(LONGINT);
          neg := FALSE
       ELSE
          HALT(99)
      END
    END;
    IF neg THEN val := -val END;
    RETURN val;
END StrToInt;

PROCEDURE RealToStr*(x: LONGREAL): ARRAY OF CHAR;
VAR
  e, h, l, n, len: LONGINT;
  i, j, pos: LONGINT;
  z: LONGREAL;
  d: ARRAY 16 OF CHAR;
  s: ARRAY 16 OF CHAR;

   PROCEDURE Wr(ch: CHAR);
   BEGIN
      IF (ch = 0X) OR (pos>16) THEN HALT(42) END;
      IF pos < len THEN
        s[pos] := ch;
        INC(pos)
    END;
   END Wr;

BEGIN
   len := LEN(s)-1; pos := 0;
   e:= ExpoL(x);
   IF e = 2047 THEN
      Wr("N"); Wr("a"); Wr("N")
  ELSE
      n := 14;
      IF (x < 0) & (e # 0) THEN
        Wr("-");
        x:= - x
    END;
      IF e = 0 THEN
         h:= 0;
         l:= 0 (* no denormals *)
      ELSE
         e:= (e - 1023) * 301029 DIV 1000000; (* ln(2)/ln(10) = 0.301029996 *)
         z:= Ten(e+1);
         IF x >= z THEN x:= x/z; INC(e) ELSE x:= x * Ten(-e) END;
         IF x >= 10 THEN
            x:= x * Ten(-1) + 0.5D0 / Ten(n); INC(e)
         ELSE
            x:= x + 0.5D0 / Ten(n);
            IF x >= 10 THEN x:= x * Ten(-1); INC(e) END
         END;
         x:= x * Ten(7); h:= ENTIER(x); x:= (x-h) * Ten(8); l:= ENTIER(x)
      END;
      i := 15; WHILE i > 7 DO d[i]:= CHR(l MOD 10 + ORD("0")); l:= l DIV 10; DEC(i) END;
      WHILE i >= 0 DO d[i]:= CHR(h MOD 10 + ORD("0")); h:= h DIV 10; DEC(i) END;
      IF ABS(e) > 8 THEN (* scientific notation *)
         j := 15; WHILE (j > 0) & (d[j] = "0") DO DEC(j) END;
         Wr(d[0]); IF j # 0 THEN Wr(".") END; i := 1; WHILE i <= j DO Wr(d[i]); INC(i) END;
         IF e < 0 THEN Wr("D"); Wr("-"); e:= - e ELSE Wr("D"); Wr("+") END;
         Wr(CHR(e DIV 100 + ORD("0"))); e:= e MOD 100;
         Wr(CHR(e DIV 10 + ORD("0"))); Wr(CHR(e MOD 10 + ORD("0")))
      ELSE
         IF e < 0 THEN (* leading zeros *)
            j := (* !15*) 14; WHILE (j > 0) & (d[j] = "0") DO DEC(j) END;
            Wr("0"); Wr("."); INC(e);
            WHILE e < 0 DO Wr("0"); INC(e) END;
            i := 0; WHILE i <= j DO Wr(d[i]); INC(i) END
         ELSE
            i := 0; WHILE (e >= 0) & (i < 16 ) DO Wr(d[i]); INC(i); DEC(e) END;
            IF i < 16 THEN
               Wr(".");
               WHILE i < (*16*) 15 DO Wr(d[i]); INC(i); END;
               WHILE s[pos - 1] = "0" DO DEC(pos) END;
               IF s[pos - 1] = "." THEN DEC(pos) END;
            END
         END
      END
   END;
   s[pos] := 0X;
   RETURN s;
END RealToStr;

PROCEDURE StrToReal*(s: ARRAY OF CHAR): LONGREAL;
VAR
  p, e: LONGINT;
  y, g: LONGREAL;
  neg, negE: BOOLEAN;
  r   : LONGREAL;
BEGIN
  p := 0;
  WHILE (s[p] = " ") OR (s[p] = "0") DO INC(p) END;
  IF s[p] = "-" THEN neg := TRUE; INC(p) ELSE neg := FALSE END;
  WHILE (s[p] = " ") OR (s[p] = "0") DO INC(p) END;
  y := 0;
  WHILE ("0" <= s[p]) & (s[p] <= "9") DO
     y := y * 10 + (ORD(s[p]) - 30H);
     INC(p);
  END;
  IF s[p] = "." THEN
     INC(p); g := 1;
     WHILE ("0" <= s[p]) & (s[p] <= "9") DO
        g := g / 10; y := y + g * (ORD(s[p]) - 30H);
        INC(p);
     END;
  END;
  IF (s[p] = "D") OR (s[p] = "E") THEN
     INC(p); e := 0;
     IF s[p] = "-" THEN negE := TRUE; INC(p) ELSE negE := FALSE END;
     WHILE (s[p] = "0") DO INC(p) END;
     WHILE ("0" <= s[p]) & (s[p] <= "9") DO
        e := e * 10 + (ORD(s[p]) - 30H);
        INC(p);
     END;
     IF negE THEN
       y := y / Ten(e)
     ELSE
       y := y * Ten(e)
    END;
  END;
  IF neg THEN y := -y END;
  r := y;
  RETURN r;
END StrToReal;

PROCEDURE StrToBool*(str: ARRAY OF CHAR): BOOLEAN;
VAR
  i: LONGINT;
  b: BOOLEAN;
BEGIN
    i := 0;
    WHILE (str[i] # 0X) & (str[i] <= " ") DO
       INC(i)
    END;
    CASE CAP(str[i]) OF
       "Y", "T": b := TRUE
      |"O": b := CAP(str[i+1]) = "N"
    ELSE
       b := FALSE
    END;
    RETURN b;
END StrToBool;

PROCEDURE BoolToStr*(b: BOOLEAN): ARRAY OF CHAR;
VAR
  str: ARRAY 3 OF CHAR;
BEGIN
    IF b THEN
       COPY("Yes", str)
    ELSE
       COPY("No", str)
    END;
    RETURN str;
END BoolToStr;

PROCEDURE StrToSet*(str: ARRAY OF CHAR): SET;
VAR
  i, d, d1: LONGINT;
  ch: CHAR;
  dot: BOOLEAN;
  set: SET;
BEGIN
    set := {};
    dot := FALSE;
    i := 0; ch := str[i];
    WHILE (ch # 0X) & (ch # "}") DO
       WHILE (ch # 0X) & ((ch < "0") OR (ch > "9")) DO INC(i); ch := str[i] END;
       d := 0; WHILE (ch >= "0") & (ch <= "9") DO d := d*10 + ORD(ch) - 30H; INC(i); ch := str[i] END;
       IF d <= MAX(SET) THEN INCL(set, d) END;
       IF dot THEN
          d1 := 0;
          WHILE (d1 <= MAX(SET)) & (d1 < d) DO INCL(set, d1); INC(d1) END;
          dot := FALSE
       END;
       WHILE ch = " " DO INC(i); ch := str[i] END;
       IF ch = "." THEN d1 := d + 1; dot := TRUE END
    END;
    RETURN set;
END StrToSet;

PROCEDURE SetToStr*(set: SET): ARRAY OF CHAR;
VAR
  i, j, k: LONGINT;
  noFirst: BOOLEAN;
  str: ARRAY MAX(SET)*2+2 OF CHAR;
BEGIN
    str[0] := "{"; i := 0; k := 1; noFirst := FALSE;
    WHILE i <= MAX(SET) DO
       IF i IN set THEN
          IF noFirst THEN str[k] := ","; INC(k) ELSE noFirst := TRUE END;
          IF i >= 10 THEN str[k] := CHR(i DIV 10 + 30H); INC(k) END;
          str[k] := CHR(i MOD 10 + 30H); INC(k);
          j := i; INC(i);
          WHILE (i <= MAX(SET)) & (i IN set) DO INC(i) END;
          IF i-2 > j THEN
             str[k] := "."; str[k+1] := "."; INC(k, 2); j := i - 1;
             IF j >= 10 THEN str[k] := CHR(j DIV 10 + 30H); INC(k) END;
             str[k] := CHR(j MOD 10 + 30H); INC(k)
          ELSE
            i := j
          END
       END;
       INC(i)
    END;
    str[k] := "}";
    str[k+1] := 0X;
    RETURN str;
END SetToStr;

(************************************************************************)

PROCEDURE RealX (h, l: LONGINT; adr: LONGINT);
BEGIN
  SYSTEM.PUT(adr + H, h);
  SYSTEM.PUT(adr + L, l);
END RealX;

PROCEDURE fcr(): SET;
  CODE {SYSTEM.i386, SYSTEM.FPU}
   PUSH 0
   FSTCW [ESP]
   FWAIT
   POP EAX
END fcr;

PROCEDURE FCR*(): SET;
BEGIN
   RETURN fcr()
END FCR;

PROCEDURE setfcr(s: SET);
  CODE {SYSTEM.i386, SYSTEM.FPU}
   FLDCW s[EBP]
END setfcr;

PROCEDURE SetFCR*(s: SET);
BEGIN
   setfcr(s)
END SetFCR;

PROCEDURE InitHL;
VAR
  i: LONGINT;
  dmy: INTEGER;
  littleEndian: BOOLEAN;
BEGIN
   DefaultFCR := (FCR()-{0, 2, 3, 10, 11})+{0 .. 5, 8, 9};
   SetFCR(DefaultFCR);
   dmy := 1;
   i := SYSTEM.ADR(dmy);
   SYSTEM.GET(i, littleEndian);
   IF littleEndian THEN
     H := 4;
     L := 0
  ELSE
    H := 0;
    L := 4
  END;
END InitHL;

BEGIN
  InitHL;
  RealX(03FF00000H, 0, SYSTEM.ADR(tene[0]));
  RealX(040240000H, 0, SYSTEM.ADR(tene[1])); (* 1 *)
  RealX(040590000H, 0, SYSTEM.ADR(tene[2])); (* 2 *)
  RealX(0408F4000H, 0, SYSTEM.ADR(tene[3])); (* 3 *)
  RealX(040C38800H, 0, SYSTEM.ADR(tene[4])); (* 4 *)
  RealX(040F86A00H, 0, SYSTEM.ADR(tene[5])); (* 5 *)
  RealX(0412E8480H, 0, SYSTEM.ADR(tene[6])); (* 6 *)
  RealX(0416312D0H, 0, SYSTEM.ADR(tene[7])); (* 7 *)
  RealX(04197D784H, 0, SYSTEM.ADR(tene[8])); (* 8 *)
  RealX(041CDCD65H, 0, SYSTEM.ADR(tene[9])); (* 9 *)
  RealX(04202A05FH, 020000000H, SYSTEM.ADR(tene[10])); (* 10 *)
  RealX(042374876H, 0E8000000H, SYSTEM.ADR(tene[11])); (* 11 *)
  RealX(0426D1A94H, 0A2000000H, SYSTEM.ADR(tene[12])); (* 12 *)
  RealX(042A2309CH, 0E5400000H, SYSTEM.ADR(tene[13])); (* 13 *)
  RealX(042D6BCC4H, 01E900000H, SYSTEM.ADR(tene[14])); (* 14 *)
  RealX(0430C6BF5H, 026340000H, SYSTEM.ADR(tene[15])); (* 15 *)
  RealX(04341C379H, 037E08000H, SYSTEM.ADR(tene[16])); (* 16 *)
  RealX(043763457H, 085D8A000H, SYSTEM.ADR(tene[17])); (* 17 *)
  RealX(043ABC16DH, 0674EC800H, SYSTEM.ADR(tene[18])); (* 18 *)
  RealX(043E158E4H, 060913D00H, SYSTEM.ADR(tene[19])); (* 19 *)
  RealX(04415AF1DH, 078B58C40H, SYSTEM.ADR(tene[20])); (* 20 *)
  RealX(0444B1AE4H, 0D6E2EF50H, SYSTEM.ADR(tene[21])); (* 21 *)
  RealX(04480F0CFH, 064DD592H, SYSTEM.ADR(tene[22])); (* 22 *)

  RealX(031FA18H, 02C40C60DH, SYSTEM.ADR(ten[0])); (* -307 *)
  RealX(04F7CAD2H, 03DE82D7BH, SYSTEM.ADR(ten[1])); (* -284 *)
  RealX(09BF7D22H, 08322BAF5H, SYSTEM.ADR(ten[2])); (* -261 *)
  RealX(0E84D669H, 05B193BF8H, SYSTEM.ADR(ten[3])); (* -238 *)
  RealX(0134B9408H, 0EEFEA839H, SYSTEM.ADR(ten[4])); (* -215 *)
  RealX(018123FF0H, 06EEA847AH, SYSTEM.ADR(ten[5])); (* -192 *)
  RealX(01CD82742H, 091C6065BH, SYSTEM.ADR(ten[6])); (* -169 *)
  RealX(0219FF779H, 0FD329CB9H, SYSTEM.ADR(ten[7])); (* -146 *)
  RealX(02665275EH, 0D8D8F36CH, SYSTEM.ADR(ten[8])); (* -123 *)
  RealX(02B2BFF2EH, 0E48E0530H, SYSTEM.ADR(ten[9])); (* -100 *)
  RealX(02FF286D8H, 0EC190DCH, SYSTEM.ADR(ten[10])); (* -77 *)
  RealX(034B8851AH, 0B548EA4H, SYSTEM.ADR(ten[11])); (* -54 *)
  RealX(0398039D6H, 065896880H, SYSTEM.ADR(ten[12])); (* -31 *)
  RealX(03E45798EH, 0E2308C3AH, SYSTEM.ADR(ten[13])); (* -8 *)
  RealX(0430C6BF5H, 026340000H, SYSTEM.ADR(ten[14])); (* 15 *)
  RealX(047D2CED3H, 02A16A1B1H, SYSTEM.ADR(ten[15])); (* 38 *)
  RealX(04C98E45EH, 01DF3B015H, SYSTEM.ADR(ten[16])); (* 61 *)
  RealX(0516078E1H, 011C3556DH, SYSTEM.ADR(ten[17])); (* 84 *)
  RealX(05625CCFEH, 03D35D80EH, SYSTEM.ADR(ten[18])); (* 107 *)
  RealX(05AECDA62H, 055B2D9EH, SYSTEM.ADR(ten[19])); (* 130 *)
  RealX(05FB317E5H, 0EF3AB327H, SYSTEM.ADR(ten[20])); (* 153 *)
  RealX(064794514H, 05230B378H, SYSTEM.ADR(ten[21])); (* 176 *)
  RealX(06940B8E0H, 0ACAC4EAFH, SYSTEM.ADR(ten[22])); (* 199 *)
  RealX(06E0621B1H, 0C28AC20CH, SYSTEM.ADR(ten[23])); (* 222 *)
  RealX(072CD4A7BH, 0EBFA31ABH, SYSTEM.ADR(ten[24])); (* 245 *)
  RealX(077936214H, 09CBD3226H, SYSTEM.ADR(ten[25])); (* 268 *)
  RealX(07C59A742H, 0461887F6H, SYSTEM.ADR(ten[26])); (* 291 *)

  eq[0]:= {0, 3, 4, 5, 9, 16, 23, 25, 26, 28, 31};
  eq[1]:= {2, 5, 6, 8, 9, 10, 11, 12, 13, 14, 15, 17, 18, 19, 20, 21, 23, 24, 25, 27, 28, 29, 30, 31};
  eq[2]:= {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28};
  eq[3]:= {0, 1, 2, 3, 5, 6, 7, 8, 9, 11, 14, 15, 16, 17, 18, 19, 20, 22, 27, 28, 29, 30, 31};
  eq[4]:= {0, 6, 7, 10, 11, 12, 13, 14, 15, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31};
  eq[5]:= {0, 1, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31};
  eq[6]:= {0, 1, 4, 5, 7, 8, 10, 14, 15, 16, 18, 20, 21, 23, 24, 25, 26, 28, 29, 30, 31};
  eq[7]:= {0, 1, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 16, 17, 18, 19, 23, 24, 26, 28, 29, 30, 31};
  eq[8]:= {0, 1, 2, 3, 4, 5, 6, 8, 9, 10, 11, 14, 16, 17, 18, 19, 20, 21, 24, 25, 26, 29};
  eq[9]:= {1, 2, 4, 6, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31};
  eq[10]:= {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30};
  eq[11]:= {0, 1, 2, 3, 4, 5, 6, 7, 8, 10, 12, 13, 14, 15, 16, 19, 20, 21, 22, 23, 27, 28, 29, 30};
  eq[12]:= {0, 1, 2, 3, 4, 5, 7, 8, 9, 10, 12, 14, 15, 16, 17, 18, 19, 20, 21, 23, 26, 27, 29, 30, 31};
  eq[13]:= {0, 1, 2, 3, 4, 5, 6, 7, 9, 10, 11, 13, 14, 15, 16, 17, 18, 20, 21, 23, 24, 27, 28, 29, 30, 31};
  eq[14]:= {0, 1, 2, 3, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31};
  eq[15]:= {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 12, 13, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 28};
  eq[16]:= {1, 2, 4, 11, 13, 16, 17, 18, 19, 22, 24, 25, 26, 27, 28, 29, 30, 31};
  eq[17]:= {1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 14, 15, 18, 19, 20, 21, 23, 25, 26, 27, 28, 29, 31};
  eq[18]:= {0, 2, 4, 5, 6, 8, 9, 11, 12, 13, 14, 16, 17, 19, 20, 22, 23, 24, 26, 27, 28, 29};
  eq[19]:= {2, 3, 4, 5, 6, 7};

  gr[0]:= {24, 27, 29, 30};
  gr[1]:= {0, 1, 3, 4, 7};
  gr[2]:= {29, 30, 31};
  gr[3]:= {4, 10, 12, 13, 21, 23, 24, 25, 26};
  gr[4]:= {1, 2, 3, 4, 5, 8, 9, 16, 17};
  gr[5]:= {2, 3, 4, 18};
  gr[6]:= {2, 3, 6, 9, 11, 12, 13, 17, 19, 22, 27};
  gr[7]:= {2};
  gr[8]:= {7, 12, 13, 15, 22, 23, 27, 28, 30, 31};
  gr[9]:= {0, 3, 5, 7, 8};
  gr[10]:= {};
  gr[11]:= {};
  gr[12]:= {11, 13, 22, 24, 25, 28};
  gr[13]:= {22, 25, 26};
  gr[14]:= {4, 5};
  gr[15]:= {10, 14, 27, 29, 30, 31};
  gr[16]:= {0, 3, 5, 6, 7, 8, 9, 10, 12, 14, 15, 20, 21, 23};
  gr[17]:= {0, 10, 12, 13, 16, 17, 22, 24, 30};
  gr[18]:= {};
  gr[19]:= {}
END SysUtils.
  










  
  
