unit DebugFrm;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ImgList;

type
  TErrorPos = RECORD
    StrPos,CharPos,CodePos: INTEGER;
    Msg: String;
    Typ: integer;
  END;

  TErrorObj = CLASS
    StrPos,CharPos,CodePos: INTEGER;
    Name: String;
    Hint: Byte;
  END;

  TDebugPanel = class(TForm)
    ImageList1: TImageList;
    MessageList: TListBox;
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure MessageListDrawItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
    procedure FormDestroy(Sender: TObject);
    procedure MessageListDblClick(Sender: TObject);
  private
    { Private declarations }
    fAdded: boolean;
  public
    { Public declarations }
    ImageIndex: array of Byte;
    procedure   AddMsg(S: String; Hint: Byte; P: TErrorObj);
    procedure   AddErr(S,Name: String; Hint: Byte; P: TErrorPos);
    procedure   Clear;
  end;

var
  DebugPanel: TDebugPanel;
  FirstDebug: boolean;

implementation

uses
  ManagerFrm, MainFrm;

{$R *.dfm}

procedure TDebugPanel.FormDestroy(Sender: TObject);
var
  i: integer;
begin
  with MessageList.Items do
    for i := Count - 1 downto 0 do Objects[i].Free;
end;

procedure TDebugPanel.FormMouseUp(Sender: TObject; Button: TMouseButton;Shift: TShiftState; X, Y: Integer);
begin
  SetFocus;
end;

procedure TDebugPanel.MessageListDrawItem(Control: TWinControl; Index: Integer;Rect: TRect; State: TOwnerDrawState);
begin
  with MessageList do begin
    if odSelected in State then begin
      Canvas.Brush.Color:=$00D8D8D8;
      Canvas.Font.Color:=clBlack;
    end else
      Canvas.Brush.Color:=clWhite;

    if ImageIndex[Index]=4 then Canvas.Font.Color:=clRed;

    Canvas.FillRect(Rect);
    Canvas.Brush.Style := bsClear;
    ImageList1.Draw(Canvas, Rect.Left+1, Rect.Top+1, ImageIndex[Index]);
    Canvas.TextOut(Rect.Left + ImageList1.Width + 4, Rect.Top+1, Items[Index]);
  end;
end;

procedure TDebugPanel.AddMsg(S: String; Hint: Byte; P: TErrorObj);
var
  i: integer;
begin
  if not Showing then begin
    ManualDock(MainForm.BottomPanel);
    Show;
  end;
  i:= Length(ImageIndex);
  SetLength(ImageIndex,i+1);
  ImageIndex[i]:=Hint;
  MessageList.Items.AddObject(S,P);
  if not fAdded then
    MessageList.Selected[MessageList.Items.Count-1]:=true;
end;

procedure TDebugPanel.MessageListDblClick(Sender: TObject);
begin
  if MessageList.Items.Objects[MessageList.ItemIndex]<>NIL then
    with TErrorObj(MessageList.Items.Objects[MessageList.ItemIndex]) do begin
      MainFrm.MainForm.Memo.SetFocus;

      if ManagerPanel.TreeView.Items.Count>0 then
        MainFrm.MainForm.NewPage(Name,_ModFile,(QuickSearch(ManagerPanel.TreeView.Items.Item[0],Name)<>NIL))
      else
        MainFrm.MainForm.NewPage(Name,_ModFile,false);
        
      if Hint=4 then
        MainFrm.MainForm.Memo.ErrorLine(CharPos-2,StrPos)
      else
        MainFrm.MainForm.Memo.NaviLine(CharPos-2,StrPos);
    end;
end;

procedure TDebugPanel.AddErr(S,Name: String; Hint: Byte; P: TErrorPos);
var
  Obj: TErrorObj;
begin
  Obj:= TErrorObj.Create;   
  Obj.StrPos:=P.StrPos;
  Obj.CharPos:=P.CharPos;
  Obj.CodePos:=P.CodePos;
  Obj.Name:=Name;
  Obj.Hint:=Hint;
  AddMsg(S,Hint,Obj);
    if (Hint=4) and FirstDebug then begin

      if ManagerPanel.TreeView.Items.Count>0 then begin
        MainFrm.MainForm.NewPage(Name,_ModFile,(QuickSearch(ManagerPanel.TreeView.Items.Item[0],Name)<>NIL))
      end else
        MainFrm.MainForm.NewPage(Name,_ModFile,false);

      MainFrm.MainForm.Memo.ErrorLine(P.CharPos-2,P.StrPos);
      FirstDebug:=false;
      MessageList.Selected[MessageList.Items.Count-1]:=true;
      fAdded:=true;
    end;
end;

procedure TDebugPanel.Clear;
var
  i: integer;
begin
  with MessageList.Items do
    for i := Count - 1 downto 0 do Objects[i].Free;
  MessageList.Clear;
  SetLength(ImageIndex,0);
  fAdded:=false;
end;

end.
