unit ManagerFrm;

interface

uses
  Windows, Forms, Classes, SysUtils, STreeView, ToolFrm,
  Controls, ComCtrls, Graphics, SynEdit, SPageControl,
  Dialogs, Menus, ActnPopup, OBRParser, ActnMan;

type
  TManagerPanel = class(TForm)
    PrjFilePopup: TPopupActionBar;
    PrjAdd: TMenuItem;
    PrjRen: TMenuItem;
    N3: TMenuItem;
    N4: TMenuItem;
    N5: TMenuItem;
    N2: TMenuItem;
    ModItem: TMenuItem;
    ModFilePopup: TPopupActionBar;
    Setasmainmodule1: TMenuItem;
    Unsetmainmodule1: TMenuItem;
    N1: TMenuItem;
    Removefromproject1: TMenuItem;
    Moduleoptions1: TMenuItem;
    Projectoptions1: TMenuItem;
    N6: TMenuItem;
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure PrjAddClick(Sender: TObject);
    procedure RenClick(Sender: TObject);
    procedure PrjRenClick(Sender: TObject);
    procedure SaveClick(Sender: TObject);
    procedure ModItemClick(Sender: TObject);
    procedure TreeViewDblClick(Sender: TObject);
    procedure TreeViewEdited(Sender: TObject; Node: TTreeNode; var S: string);
    procedure FormResize(Sender: TObject);
    procedure SetAsMainModClick(Sender: TObject);
    procedure UnSetMainModClick(Sender: TObject);
    procedure Removefromproject1Click(Sender: TObject);
    procedure PrjFilePopupGetControlClass(Sender: TCustomActionBar;
      AnItem: TActionClient; var ControlClass: TCustomActionControlClass);
  private
    { Private declarations }
    PopupNode: TTreeNode;
    procedure TreeViewExpanding(Sender: TObject; Node: TTreeNode;var AllowExpansion: Boolean);
    procedure TreeViewCollapsing(Sender: TObject; Node: TTreeNode;var AllowCollapse: Boolean);
    procedure TreeViewContextPopup(Sender: TObject; MousePos: TPoint; var Handled: Boolean);
    procedure TreeViewChange(Sender: TObject; Node: TTreeNode);
    procedure SaveChange;
  public
    { Public declarations }
    TreeView  : TSTreeView;
    OldNode   : TTreeNode;
    NewCreated: boolean;
    procedure SetMain(Node: TTreeNode; Flag: boolean);
    procedure AddModFile(Name: String; AddInTab: boolean; n: integer);
    procedure SetModified(Flag: boolean);
    function  GetModified: boolean;
    procedure SetNewProject(Path: String);
  end;

  TFileType = (_Nil,_PrjFile,_ResFile,_AsmFile,_ModFile);
  TFileHead = class
    Types   : TFileType;
    Modified: boolean;
    Main    : boolean;
    Name    : String;
    Search  : String;
    Path    : String;
    Options : SET OF 0..31;
  end;

  const
    FileFormat: array [TFileType] of String[4] = (
      '','.OPR','.RES','.ASM','.MOD'
    );
    FontColor:  array [TFileType] of TColor = (
      $00000000,$00000000,$00BB7755,$00000000,$00008800
    );

  procedure FreeNode(Node: TTreeNode);
  procedure DeleteNode(Node: TTreeNode);
  function  FindFormat(S: String): integer;
  function  FindType(S: String): TFileType;
  function  QuickSearch(Noddy: TTreeNode; SearchTarget: string): TTreeNode;
  
var
  ManagerPanel  : TManagerPanel;
  OldData       : TMemoryStream;

implementation

uses
  MainFrm, DebugFrm, InspectFrm, StructFrm;

{$R *.dfm}

procedure FreeNode(Node: TTreeNode);
var
  i: integer;
begin
  with Node do
    for i:=Count-1 downto 0 do begin
      if Item[i].Count>0 then FreeNode(Item[i]);
      TFileHead(Item[i].Data).Main:=false;
      TSTreeNode(Item[i]).Main:=false;
      Item[i].DeleteChildren;
      TFileHead(Item[i].Data).Free;
    end;
end;

procedure DeleteNode(Node: TTreeNode);
var
  ParentNode: TTreeNode;
begin
  FreeNode(Node);
  TFileHead(Node.Data).Free;
  ParentNode:=Node.Parent;
  Node.Delete;
  if TFileHead(ParentNode.Data).Types=_NIL then begin
    TFileHead(ParentNode.Data).Free;
    ParentNode.Delete;
  end;
end;

procedure TManagerPanel.FormCreate(Sender: TObject);
begin
  DoubleBuffered             := true;
  TreeView                   := TSTreeView.Create(Self);
  TreeView.Parent            := Self;
  TreeView.DeltaX            := 17;
  TreeView.DoubleBuffered    := true;
  TreeView.BorderStyle       := bsNone;
  TreeView.SortType          := stNone;
  TreeView.OnCollapsing      := TreeViewCollapsing;
  TreeView.OnExpanding       := TreeViewExpanding;
  TreeView.OnContextPopup    := TreeViewContextPopup;
  TreeView.OnChange          := TreeViewChange;
  TreeView.OnDblClick        := TreeViewDblClick;
  TreeView.OnEdited          := TreeViewEdited;
  PrjFilePopup.Images        := TreeView.Images;
  MainForm.fPages.Images     := TreeView.Images;
  TreeView.Visible           := false;
end;

procedure TManagerPanel.FormDestroy(Sender: TObject);
begin
  if not Assigned(TreeView) or (TreeView.Items.Count=0) then Exit;
  FreeNode(TreeView.Items.Item[0]);
  TFileHead(TreeView.Items.Item[0].Data).Free;
  TreeView.Free;
end;

function FindFormat(S: String): integer;
var
  i: TFileType;
begin
  Result:=0;
  for i:=Low(TFileType) to High(TFileType) do
    if SameText(S,FileFormat[i]) then begin
      Result:=Ord(i); Exit;
    end;
end;

function FindType(S: String): TFileType;
var
  i: TFileType;
begin
  Result:=_Nil;
  for i:=Low(TFileType) to High(TFileType) do
    if SameText(S,FileFormat[i]) then begin
      Result:=i;
      Exit;
    end;
end;

procedure TManagerPanel.SetAsMainModClick(Sender: TObject);
begin
  if not Assigned(TreeView) then Exit;
  if TreeView.Items.Count=0 then Exit;
  SetMain(QuickSearch(TreeView.Items.Item[0],TreeView.Selected.Text),true);
  SetModified(true);
end;

procedure TManagerPanel.UnSetMainModClick(Sender: TObject);
begin
  if TreeView.Items.Count=0 then Exit;
  SetMain(QuickSearch(TreeView.Items.Item[0],TreeView.Selected.Text),false);
  SetModified(true);
end;

procedure TManagerPanel.SetMain(Node: TTreeNode; Flag: boolean);

  procedure ClearMain(Nodes: TTreeNode);
  var
    i: integer;
  begin
    with Nodes do
      for i:=Count-1 downto 0 do begin
        if Item[i].Count>0 then ClearMain(Item[i]);
        TFileHead(Item[i].Data).Main:=false;
        TSTreeNode(Item[i]).Main:=false;
      end;
  end;

begin
  ClearMain(TreeView.Items.Item[0]);
  TFileHead(Node.Data).Main:=Flag;
  TSTreeNode(Node).Main:=Flag;
  TreeView.Invalidate;
end;

procedure TManagerPanel.PrjAddClick(Sender: TObject);
begin
  MainForm.OpenFile(true);
end;

procedure TManagerPanel.PrjFilePopupGetControlClass(Sender: TCustomActionBar;
  AnItem: TActionClient; var ControlClass: TCustomActionControlClass);
begin
  if Assigned(PrjFilePopup.PopupMenu) then
    PrjFilePopup.PopupMenu.ColorMap:= MainForm.XPColorMap1;
  if Assigned(ModFilePopup.PopupMenu) then
    ModFilePopup.PopupMenu.ColorMap:= MainForm.XPColorMap1;
end;

procedure TManagerPanel.PrjRenClick(Sender: TObject);
begin
  TreeView.Selected.EditText;
end;

procedure TManagerPanel.Removefromproject1Click(Sender: TObject);
var
  Active   : TDTabSheet;
begin
  Active:=MainForm.FindPage(TreeView.Selected.Text);
  if Active<>NIL then begin
    MainForm.fPages.ActivePage:=Active;
    DeleteNode(TreeView.Selected);
    MainForm.CloseFile;
  end else
    DeleteNode(TreeView.Selected);
  SetModified(true);
end;

procedure TManagerPanel.RenClick(Sender: TObject);
begin
  TreeView.Selected.EditText;
end;

procedure TManagerPanel.FormMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  SetFocus;
end;

procedure TManagerPanel.FormResize(Sender: TObject);
begin
  TreeView.Top    := 0;
  TreeView.Left   := -TreeView.DeltaX;
  TreeView.Width  := Width+TreeView.DeltaX;
  TreeView.Height := Height;
end;

procedure TManagerPanel.SetModified(Flag: boolean);
begin
  if TreeView.Items.Count>0 then
    if TFileHead(TreeView.Items.Item[0].Data).Modified<>Flag then
      TFileHead(TreeView.Items.Item[0].Data).Modified:=Flag
end;

function TManagerPanel.GetModified: boolean;
begin
  Result:=false;
  if TreeView.Items.Count>0 then
    Result:=TFileHead(TreeView.Items.Item[0].Data).Modified;
end;

procedure TManagerPanel.SetNewProject(Path: String);
var
  Head: TFileHead;
begin
  if not TreeView.Visible then TreeView.Visible:=true;
  Head        := TFileHead.Create;
  Head.Types  := _PrjFile;
  Head.Name   := ExtractFileName(Path);
  Head.Path   := '';
  Head.Search := '';
  Head.Modified:=false;
  TreeView.Items.AddChildObject(NIL,Head.Name,Head);
  NewCreated:=true;
end;

procedure TManagerPanel.TreeViewContextPopup(Sender: TObject; MousePos: TPoint; var Handled: Boolean);
begin
  TreeView.PopupMenu   := NIL;
  PopupNode := (Sender as TTreeView).GetNodeAt(MousePos.X,MousePos.Y);
  if TFileHead(PopupNode.Data).Types=_Nil then Exit;
  if PopupNode<>nil then begin
    TTreeView(Sender).Selected:=PopupNode;
    if SameText(TObject(PopupNode.Data).ClassName,'TFileHead') then
      case TFileHead(PopupNode.Data).Types of
        _PrjFile: TreeView.PopupMenu   := PrjFilePopup;
        _ModFile:
                  begin
                    if TFileHead(PopupNode.Data).Main then begin
                      ModFilePopup.Items[0].ImageIndex:=8;
                      ModFilePopup.Items[0].Enabled:=false;
                      ModFilePopup.Items[1].Enabled:=true;
                    end else begin
                      ModFilePopup.Items[0].ImageIndex:=-1;
                      ModFilePopup.Items[0].Enabled:=true;
                      ModFilePopup.Items[1].Enabled:=false;
                    end;
                    TreeView.PopupMenu   := ModFilePopup;
                  end;
        _ResFile: ;
      else
        //ShowMessage(IntToStr(ord(TFileHead(PopupNode.Data).Types)));
      end;
  end;
end;

procedure TManagerPanel.TreeViewCollapsing(Sender: TObject; Node: TTreeNode;var AllowCollapse: Boolean);
begin
  if Node.ImageIndex=2 then begin
    Node.ImageIndex:=1;
    Node.SelectedIndex:=Node.ImageIndex;
  end;
end;

procedure TManagerPanel.TreeViewExpanding(Sender: TObject; Node: TTreeNode;var AllowExpansion: Boolean);
begin
  if Node.ImageIndex=1 then begin
    Node.ImageIndex:=2;
    Node.SelectedIndex:=Node.ImageIndex;
  end;
end;

procedure TManagerPanel.SaveChange;
begin

end;

procedure TManagerPanel.TreeViewEdited(Sender: TObject; Node: TTreeNode; var S: string);
var
  Tab: TDTabSheet;
  Memo:TCustomSyntaxMemo;
  oldName, newName: String;
begin
  if TFileHead(Node.Data).Types=_Nil then begin
    S:=Node.Text;
    Exit;
  end;
  oldName:=TFileHead(Node.Data).Path+TFileHead(Node.Data).Name;
  Tab:=MainForm.FindPage(Node.Text);
  Memo:=MainForm.FindMemo(Node.Text);
  if Tab<>NIL then begin
    Tab.Caption:=S;
    Memo.Name:=S;
    Memo.Path:=TFileHead(Node.Data).Path;
    Memo.Lines.Modified:=true;
    OBRManage(Tab.Caption,StructPanel.TreeView,Memo.Lines);
  end;
  newName:=TFileHead(Node.Data).Path+S;
  if FileExists(oldName) then RenameFile(oldName,newName);
  TFileHead(Node.Data).Search:=S;
  TFileHead(Node.Data).Name:=S;
  MainForm.fMemoInfo.Path:=TFileHead(Node.Data).Path;
  MainForm.fMemoInfo.Name:=TFileHead(Node.Data).Name;
  InspectPanel.ObjectInspector.Update;
end;

procedure TManagerPanel.TreeViewChange(Sender: TObject; Node: TTreeNode);
var
  ParentNode: TTreeNode;
begin
  if Node=NIL then Exit;
  if TFileHead(Node.Data).Types=_Nil then Exit;
  if Node.Level=0 then Exit;
  if TFileHead(Node.Data).Types=_ModFile then Exit;

  SaveChange;
  OldNode:=NIL;

  ParentNode:=Node.Parent;
  while ParentNode.Level>1 do ParentNode:=ParentNode.Parent;
  if ParentNode.Level=1 then MainForm.SetPage(ParentNode.Text,ParentNode) else
  if Node.Level=1 then MainForm.SetPage(Node.Text,Node);

  TreeView.ReadOnly:=true;

  if Node.Level>1 then begin
    ParentNode:=Node.Parent;
    while ParentNode.Level<>1 do ParentNode:=ParentNode.Parent;
    case TFileHead(ParentNode.Data).Types of
      _ResFile: ;
    end
  end else if Node.Level=1 then begin
    case TFileHead(Node.Data).Types of
      _ResFile: ;
      _ModFile: ;
    end;
  end;

  if MainForm.Memo.Visible then MainForm.Memo.Update;
  TDTabSheet(MainForm.fPages.ActivePage).SelectNode:=Node;
end;

procedure TManagerPanel.SaveClick(Sender: TObject);
begin
  with TFileHead(PopupNode.Data) do
    case Types of
      _ResFile: ;
      _ModFile: ;
    end;
end;

function QuickSearch(Noddy: TTreeNode; SearchTarget: string): TTreeNode;
var
  Searching: boolean;
begin
  Searching := true;
  while Searching and (Noddy<>nil) do begin
    if SameText(Trim(TFileHead(Noddy.Data).Search),Trim(SearchTarget)) then
      Searching := false
    else
      Noddy := Noddy.GetNext;
  end;
  Result:=Noddy;
end;

procedure TManagerPanel.TreeViewDblClick(Sender: TObject);
var
  Node: TTreeNode;
begin
  Node:=(Sender as TSTreeView).Selected;
  if (Node.Level>=1) and (TFileHead(Node.Data).Types<>_Nil) then MainForm.SetPage(Node.Text,Node);
end;

procedure TManagerPanel.AddModFile(Name: String; AddInTab: boolean; n: integer);
var
  Root,First      : TTreeNode;
  FN              : String;
  Head,Fold       : TFileHead;

begin
  TreeView.Items.BeginUpdate;
  FN:=ExtractFileName(Name);
  FN[1]:=UpCase(FN[1]);
  if TreeView.Items.Count=0 then Exit;
  Root := TreeView.Items.Item[0];
  if (QuickSearch(Root,FN)<>NIL) or (Root=NIL) then begin
    MainForm.NewPage(FN,_ModFile,true);
    TreeView.Items.EndUpdate;
    Exit;
  end;

  Head          := TFileHead.Create;
  Head.Types    := _ModFile;
  Head.Name     := FN;
  Head.Search   := FN;
  Head.Main     := false;
  if (n<>-1) and AddInTab then Head.Main     := true;
  Head.Path     := ExtractFilePath(Name);

  if Head.Path<>'' then begin
    First:=QuickSearch(Root,Head.Path);
    if First=NIL then begin
      Fold          := TFileHead.Create;
      Fold.Types    := _Nil;
      Fold.Name     := '';
      Fold.Path     := Head.Path;
      Fold.Search   := Head.Path;
      First:=TreeView.Items.AddChildObject(Root,copy(Fold.Path,1,10)+'..',Fold);
      First.ImageIndex:=1;
    end;
    First:=TreeView.Items.AddChildObject(First,FN,Head);
  end else
    First:=TreeView.Items.AddChildObject(Root,FN,Head);

  TSTreeNode(First).Main:=false;
  if (n<>-1) and AddInTab then TSTreeNode(First).Main:=true;

  Root.Expand(true);
  First.Expand(true);
  First.ImageIndex:=30+FindFormat(ExtractFileExt(First.Text));
  First.SelectedIndex:=First.ImageIndex;
  {
  for i:=0 to Root.Count-1 do
    with Root.Item[i] do
      for j:=0 to Count-1 do Item[j].Collapse(true);

  for i:=0 to First.Count-1 do
    with First.Item[i] do
      for j:=0 to Count-1 do
        Item[j].Parent.Collapse(true);
  }
  First.AlphaSort;
  Root.AlphaSort;
  TreeView.Items.EndUpdate;

  if AddInTab then MainForm.NewPage(Head.Path+Head.Name,_ModFile,true);
  OldNode:=NIL;
end;

procedure TManagerPanel.ModItemClick(Sender: TObject);
var
  Root: TTreeNode;
  i   : WORD;
begin
  IF TreeView.Items.Count=0 THEN SetNewProject('Project.opr');
  Root := TreeView.Items.Item[0];
  i:=1;
  while (i<$FFFF) and (QuickSearch(Root,'Module'+IntToStr(i)+'.mod')<>NIL) do inc(i);
  AddModFile('Module'+IntToStr(i)+'.mod',true,-1);
  SetModified(true);
end;

initialization
  OldData := TMemoryStream.Create;
finalization
  OldData.Free;
  
end.
