unit OptionsFrm;

interface

uses
  Forms, Controls, ComCtrls, Graphics, SPageControl,
  GlowControls, SysUtils, StdCtrls;

type
  TOptionsForm = class(TForm)
    procedure   FormCreate(Sender: TObject);
    procedure   CloseClick(Sender: TObject);
    procedure   SaveClick(Sender: TObject);
  private
    { Private declarations }
    fPageControl: TSPageControl;
  public
    { Public declarations }
    procedure   ProjectOpt(S: TObject; Flag: boolean);
    function    CreateButton(Sender: TForm; Capt: String; ALeft,ATop: integer): TShButton;
  end;

  TPrjOptionsForm = class(TForm)
    procedure   FormCreate(Sender: TObject);
    procedure   CloseClick(Sender: TObject);
    procedure   SaveClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  OptionsForm: TOptionsForm;
  PrjOptionsForm: TPrjOptionsForm;

  OpCheck : array [0..16] of string = (
    'Index check',
    'Overflow check',
    'Type checking',
    'Pointer initialization',
    'Assert evaluation',

    'Don''t create any files',
    'Display warnings',
    'Trace procedure usage',
    'Aos, different SymFile',
    'Checked external function',
    '',
    'Find text position of breakpc',
    'Prevent compilation of modules that import SYSTEM',
    'Generation of new symbol file allowed',
    'Generation of extended symbol file allowed',
    'Retain filename prefix on .Sym and .Obj file (Native)',
    'Zero all the stack (only if ptrinit)'
  );

implementation

{$R *.dfm}
{$R PrjOptionsForm.dfm}

procedure TPrjOptionsForm.CloseClick(Sender: TObject);
begin
  Close;
end;

procedure TPrjOptionsForm.SaveClick(Sender: TObject);
begin
  //
end;

procedure TPrjOptionsForm.FormCreate(Sender: TObject);
var
  Button  : TShButton;
begin
  Font.Color:=$00608E9C;
  OptionsForm.ProjectOpt(Self,false);
  Button:=OptionsForm.CreateButton(Self,'Save',Self.ClientWidth-150,Self.ClientHeight-35);
  Button.OnClick:=SaveClick;
  Button:=OptionsForm.CreateButton(Self,'Close',Self.ClientWidth-80,Self.ClientHeight-35);
  Button.OnClick:=CloseClick;
end;

//=================================================================

procedure TOptionsForm.CloseClick(Sender: TObject);
begin
  Close;
end;

procedure TOptionsForm.SaveClick(Sender: TObject);
begin
  //
end;

function TOptionsForm.CreateButton(Sender: TForm; Capt: String; ALeft,ATop: integer): TShButton;
var
  Button  : TShButton;
begin
  Button:= TShButton.Create(Sender);
  with Button do begin
    Parent        := Sender;
    Caption       := Capt;
    Font.Color    := $00608E9C;
    Width         := Button.Width-8;
    Height        := Button.Height-2;
    Left          := ALeft;
    Top           := ATop;
    ShowFocusRect := false;
    with Colors do begin
      DefaultFrom := clCream;
      DefaultTo   := $00A7CDD5;
      ClickedFrom := $00C4E2E7;
      ClickedTo   := clWhite;
      BorderColor := $00608E9C;
      OverFrom    := clWhite;
      OverTo      := $00C4E2E7;
    end;
  end;
  Result:=Button;
end;

procedure TOptionsForm.FormCreate(Sender: TObject);
const
  OpName : array [0..0] of string = (
    'Compiler options'
  );
var
  SubPage : TSTabSheet;
  i       : integer;
  Button  : TShButton;
begin

  fPageControl:= TSPageControl.Create(Self);
  with fPageControl do begin
    Parent:=Self;
    Align:= alClient;
    AlignWithMargins:=true;
    Margins.Left:=10;
    Margins.Top:=10;
    Margins.Right:=10;
    Margins.Bottom:=44;
    TabBackGroundColor:=Color;
    fPageControl.TabPosition:=tpLeft;
    TabBackGround.LoadFromResourceName(0,'LTABOFF');
    TabBackGroundActive.LoadFromResourceName(0,'LTABON');
    RoundEdges:=true;
    TabSheetBorderColor:=$00608E9C;
    TabMargin.LeftMargin:=10;
    TabWidth:=22;
    TabHeight:=159;
    ActiveFont.Color := clMaroon; 
  end;

  for i:=0 to Length(OpName)-1 do begin
    SubPage:= TSTabSheet.Create(fPageControl);
    with SubPage do begin
      Parent       := fPageControl;
      SPageControl := fPageControl;
      Caption      := OpName[i];
      Color        := clCream;
      TextColor    := $00608E9C;
      Tag          := i;
    end;

    case i of
      0: ProjectOpt(SubPage,true);
    end;
  end;

  Button:=CreateButton(Self,'Save',Self.ClientWidth-150,Self.ClientHeight-35);
  Button.OnClick:=SaveClick;
  Button:=CreateButton(Self,'Close',Self.ClientWidth-80,Self.ClientHeight-35);
  Button.OnClick:=CloseClick;

end;

procedure TOptionsForm.ProjectOpt(S: TObject; Flag: boolean);
var
  GroupBox1     : TGroupBox;
  CheckBox1     : TSCheckBox;
  RadioButton1  : TSRadioButton;
  i,w,GBHeight  : integer;
  TextColor     : TColor;
  Sender        : TWinControl;
begin
  if Flag then begin
    TextColor := TSTabSheet(S).TextColor;
    Sender    := TSTabSheet(S);
  end else begin
    TextColor := TForm(S).Font.Color;
    Sender    := TForm(S);
  end;

  GroupBox1:= TGroupBox.Create(Sender);
  GroupBox1.Parent:=Sender;
  GroupBox1.Caption:='Compiler options';
  GroupBox1.Left:=5;
  GroupBox1.Top:=0;
  GroupBox1.Width:=Sender.Width-10;
  for i:=0 to 11 do begin
    if i<>5 then begin
      CheckBox1 := TSCheckBox.Create(Sender);
      CheckBox1.Parent:=Sender;
      CheckBox1.Left:=10;
      if i<5 then
        w:=i
      else
        w:=i+5;
      CheckBox1.Top:=16*i+15;
      CheckBox1.Tag:=w;
      CheckBox1.Caption:=OpCheck[CheckBox1.Tag]+' '+IntToStr(CheckBox1.Tag);
      CheckBox1.Font.Color:= TextColor;
      if i<5 then
        w:=(GroupBox1.Width div 2)-7
      else
        w:=GroupBox1.Width-7;
      CheckBox1.Width:=w;
    end;

    if i<5 then begin
      if OpCheck[i+5]<>'' then begin
        CheckBox1 := TSCheckBox.Create(Sender);
        CheckBox1.Parent:=Sender;
        CheckBox1.Left:=7+(GroupBox1.Width div 2);
        CheckBox1.Top:=16*i+15;
        CheckBox1.Tag:=i+5;
        CheckBox1.Caption:=OpCheck[CheckBox1.Tag]+' '+IntToStr(CheckBox1.Tag);
        CheckBox1.Font.Color:= TextColor;
        CheckBox1.Width:=(GroupBox1.Width div 2)-7;
      end;
    end;
    GBHeight:=16*i+15;
  end;
  GroupBox1.Height:=GBHeight+25;

  GroupBox1:= TGroupBox.Create(Sender);
  GroupBox1.Parent:=Sender;
  GroupBox1.Caption:='Compiler mode';
  GroupBox1.Left:=5;
  GroupBox1.Top:=GbHeight+30;
  GroupBox1.Width:=(Sender.Width div 2)-5;

  RadioButton1 := TSRadioButton.Create(Sender);
  RadioButton1.Parent:=Sender;
  RadioButton1.Left:=10;
  RadioButton1.Top:=GbHeight+45;
  RadioButton1.Width:=GroupBox1.Width-10;
  RadioButton1.Font.Color:= TextColor;
  RadioButton1.Caption:='Oberon 1';

  RadioButton1 := TSRadioButton.Create(Sender);
  RadioButton1.Parent:=Sender;
  RadioButton1.Left:=10;
  RadioButton1.Top:=GbHeight+60;
  RadioButton1.Width:=GroupBox1.Width-10;
  RadioButton1.Font.Color:= TextColor;
  RadioButton1.Caption:='Oberon 2';

  RadioButton1 := TSRadioButton.Create(Sender);
  RadioButton1.Parent:=Sender;
  RadioButton1.Left:=10;
  RadioButton1.Top:=GbHeight+75;
  RadioButton1.Width:=GroupBox1.Width-10;
  RadioButton1.Font.Color:= TextColor;
  RadioButton1.Caption:='Active Oberon';
  RadioButton1.Checked:=true;

  GroupBox1.Height:=70;

  if Flag then begin
    GroupBox1:= TGroupBox.Create(Sender);
    GroupBox1.Parent:=Sender;
    GroupBox1.Caption:='Linking mode';
    GroupBox1.Left:=(Sender.Width div 2)+5;
    GroupBox1.Top:=GbHeight+30;
    GroupBox1.Width:=(Sender.Width div 2)-10;
    GroupBox1.Height:=70;

    CheckBox1 := TSCheckBox.Create(Sender);
    CheckBox1.Parent:=Sender;
    CheckBox1.Left:=GroupBox1.Left+5;
    CheckBox1.Top:=GbHeight+45;
    CheckBox1.Font.Color:= TextColor;
    CheckBox1.Width:=GroupBox1.Width-10;
    CheckBox1.Caption:='Linking with .Link file';

    CheckBox1 := TSCheckBox.Create(Sender);
    CheckBox1.Parent:=Sender;
    CheckBox1.Left:=GroupBox1.Left+5;
    CheckBox1.Top:=GbHeight+60;
    CheckBox1.Width:=GroupBox1.Width-10;
    CheckBox1.Font.Color:= TextColor;
    CheckBox1.Caption:='Create .Link file';
  end;

end;

end.
