UNIT OPOi386;

INTERFACE

USES
	OPT, OPM, _SYSTEM, SYSUTILS;

CONST

	noScale = 0; Scale1 = 0; Scale2 = 1; Scale4 = 2; Scale8 = 3;	(* scale factors *)
	RegReg 	= 0;	(* register to register *)
	RegMem 	= 1;	(* register to memory *)
	MemReg 	= 2;	(* memory to register *)
	ImmReg 	= 3;	(* immediate to register *)
	ImmMem 	= 4;	(* immediate to memory *)
	RegSt 	= 5; 	(* floating point st, reg *)
	StReg 	= 6; 	(* floating point reg, st *)
	StRegP 	= 7;	(* floating point ...p reg, st (with pop *)
	MemSt 	= 8; 	(* floating point st, mem *)

	Regs 	= RegReg;
	Mem 	= MemReg;         
	Imme 	= ImmReg;

	sReal 	= 0; 	(* short real 32 bit = Bit32 *)
	lRReal 	= 2; 	(* long real 64 bit *)
	eReal 	= 4;	(* extended real 80 bit, only valid in GenFLD, GenFSTP *)
	sIInt 	= 1;	(* short integer 32 bit *)

	CodeLength 	= 65000;	(* 64KByte Code Lenght per Module *)
	MaxConstLength	= 64*1024;	(* Max Const size allowed, limited by Object File Format*)

  EAX = 0; ECX = 1; EDX = 2; EBX = 3; ESP = 4; EBP = 5; ESI = 6; EDI = 7; (* 32 bit register *)
  AX = 8; CX = 9; DX = 10; BX = 11; SP =  12; BP = 13; SI = 14; DI = 15; (* 16 bit register *)
  AL = 16; CL = 17; DL = 18; BL = 19; AH = 20; CH = 21; DH = 22; BH = 23; (* 8 bit register *)

	(* register/memory size (8,16, 32 or 64 bit) *)
	Bit8 = AL; Bit16 = AX; Bit32 = EAX (* must be 0 *); Bit64 = 40;


	noDisp = 0; Disp8 = 1; Disp32 = 2; none = -1;
	noBase = none; noInx = none; noImm = 0;

	BUG = 22;

  gfCOMPP   = 0;
  gfTST     = 1;
  gfLDZ     = 2;
  gfLD1     = 3;
  gfABS     = 4;
  gfCHS     = 5;
  gfSTSW    = 6;
  gfINCSTP  = 7;
  gfDECSTP  = 8;

TYPE
  ConstArray = ARRAY OF CHAR;
		
VAR
	pcc, dsize: LONGINT; 		(* code size, data size *)
	csize: LONGINT;		(* constant size *)
	lastImmSize: SHORTINT;		(* last immediate Size *)
	constant: ConstArray;	(* constant area *)
	code: ARRAY [0..CodeLength] OF CHAR;	(* i386/i387 code area *)
	CodeErr: BOOLEAN;

  PROCEDURE GenMOV (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
  PROCEDURE GenLEA(reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenMOVSX (mode, s: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenMOVZX(mode, s: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenIN(size: INTEGER);
  PROCEDURE GenOUT(size: INTEGER);
  PROCEDURE GenPUSH (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
  PROCEDURE GenPOP (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenJMP (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenJcc (op: SHORTINT; disp: LONGINT);
  PROCEDURE GenNEG (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenNOT (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenTyp1 (op, mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
  PROCEDURE GenBTS (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
  PROCEDURE GenBTR (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
  PROCEDURE GenShiftRot (op, mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
  PROCEDURE GenTEST (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
  PROCEDURE GenBT (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
  PROCEDURE Prefix (reg: INTEGER; VAR w: SHORTINT);
  PROCEDURE PutByte (b: INTEGER);
  PROCEDURE GenIDIV (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenDEC (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenIMUL (mode: SHORTINT; shortform: BOOLEAN; reg, base, inx: INTEGER; scale: SHORTINT;disp, imm: LONGINT);
  PROCEDURE GenRepString (op, size: INTEGER);
  PROCEDURE GenINC (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE PutByteAt (pos: LONGINT; byte: INTEGER);
  PROCEDURE GenSetcc(op, mode: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GetDWord (pos: LONGINT; VAR dw: LONGINT);
  PROCEDURE PutDWordAt (pos, dw: LONGINT);
  PROCEDURE GenFFREE (freg: INTEGER);
  PROCEDURE GenFLD (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenFSTP (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenFop1 (op: INTEGER);
  PROCEDURE GenFSUBR (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenFMUL (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenFADD (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenFSUB (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenFDIVR (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenFDIV(mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenFCOMP (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenFSTCW (base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenFLDCW (base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenFST (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GetConsDWord (pos: LONGINT; VAR dw: LONGINT);
  PROCEDURE PutConsDWord (pos, dw: LONGINT);
  PROCEDURE InlineCode (VAR code: ARRAY OF CHAR; parSize: INTEGER);
  PROCEDURE GenCALL (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
  PROCEDURE GenRET (size: LONGINT);
  PROCEDURE Init;
  
IMPLEMENTATION

PROCEDURE PutByte (b: INTEGER);
BEGIN
  IF pcc >= CodeLength THEN BEGIN
    IF NOT CodeErr THEN BEGIN
      OPM.err (210);
      CodeErr:= TRUE
    END;
    pcc:= 0
  END ELSE BEGIN
    code[pcc]:= CHR(b);
    INC (pcc);
    IF pcc >= OPM.breakpc THEN BEGIN
      OPM.err(255);
      OPM.breakpc := MAX(_LONGINTS)
    END;
  END
END;
	
PROCEDURE PutWord (w: LONGINT);
BEGIN
  IF pcc >= CodeLength - 2 THEN BEGIN
    IF NOT CodeErr THEN BEGIN
      OPM.err (210);
      CodeErr:= TRUE
    END;
    pcc:= 0
  END ELSE BEGIN
    code[pcc]:= CHR (_SYSTEM._MOD(w,$100)); INC(pcc);
    code[pcc]:= CHR ( _SYSTEM._MOD(_SYSTEM._DIV(w,$100),$100));
    INC(pcc);
    IF pcc >= OPM.breakpc THEN BEGIN
      OPM.err(255);
      OPM.breakpc := MAX(_LONGINTS)
    END;
  END
END;

PROCEDURE PutDWord(dw: LONGINT);
BEGIN
  IF pcc >= CodeLength - 4 THEN BEGIN
    IF NOT CodeErr THEN BEGIN
      OPM.err (210);
      CodeErr:= TRUE
    END;
    pcc:= 0
  END ELSE BEGIN
    code[pcc]:= CHR(_SYSTEM._MOD(dw,$100));                      INC(pcc);
    code[pcc]:= CHR(_SYSTEM._MOD(_SYSTEM._DIV(dw,$100),$100));   INC(pcc);
    code[pcc]:= CHR(_SYSTEM._MOD(_SYSTEM._DIV(dw,$10000),$100)); INC(pcc);
    code[pcc]:= CHR(_SYSTEM._DIV(dw,$1000000));                  INC(pcc);
    IF pcc >= OPM.breakpc THEN BEGIN
      OPM.err(255);
      OPM.breakpc := MAX(_LONGINTS)
    END;
  END
END;

PROCEDURE PutByteAt (pos: LONGINT; byte: INTEGER);
VAR
  currentPC: LONGINT;
BEGIN
  currentPC:= pcc;
  pcc:= pos;
  PutByte (byte);
  pcc:= currentPC;
END;
	
PROCEDURE PutDWordAt (pos, dw: LONGINT);
VAR
  currentPC: LONGINT;
BEGIN
  currentPC:= pcc;
  pcc:= pos;
  PutDWord (dw);
  pcc:= currentPC
END;

PROCEDURE GetByte (pos: LONGINT; VAR byte: INTEGER);
BEGIN
  byte:= ORD (code [pos])
END;
	
PROCEDURE GetDWord (pos: LONGINT; VAR dw: LONGINT);
VAR
  byte: INTEGER;
BEGIN
  dw:= ORD(code[pos]) + ORD(code[pos+1])* $100 + ORD(code[pos+2])* $10000;
  byte:= ORD(code[pos+3]);
  IF byte >= 128 THEN byte:= byte - 256;
  dw:= (byte * $1000000) + dw;
END;
	
PROCEDURE PutConsDWord (pos, dw: LONGINT);
BEGIN
  constant [pos]  := CHR(_SYSTEM._MOD(dw,$100)); (* low byte first *)
  constant [pos+1]:= CHR(_SYSTEM._MOD(_SYSTEM._DIV(dw,$100),$100));
  constant [pos+2]:= CHR(_SYSTEM._MOD(_SYSTEM._DIV(dw,$10000),$100));
  constant [pos+3]:= CHR(_SYSTEM._DIV(dw,$1000000));
END;
	
PROCEDURE GetConsDWord (pos: LONGINT; VAR dw: LONGINT);
VAR
  byte: INTEGER;
BEGIN 
  dw:= ORD(constant[pos]) + ORD(constant[pos+1]) * $100 + ORD(constant[pos+2]) * $10000;
  byte:= ORD(constant[pos+3]);
  IF byte >= 128 THEN byte:= byte - 256;
  dw:= byte * $1000000 + dw
END;
		
PROCEDURE PutReg (reg1, reg2: INTEGER);
BEGIN
  reg1:= _SYSTEM._MOD(reg1,8);
  reg2:= _SYSTEM._MOD(reg2,8);
  PutByte (3 * $40 + reg1 * 8+ reg2)
END;

PROCEDURE PutRMInx (mode: SHORTINT; reg, base, inx, scale: INTEGER; disp: LONGINT);
BEGIN
  reg:= _SYSTEM._MOD(reg,8);
  inx:= _SYSTEM._MOD(inx,8);
  IF mode = omAbs THEN BEGIN
    PutByte (noDisp * $40 + reg * 8 + 4); (* escape to two bytes *)
    PutByte (scale * $40 + inx * 8+ 5);
    PutDWord (disp)
  END ELSE BEGIN(* RegRel *)
    base:= _SYSTEM._MOD(base,8);
    IF (disp = 0) AND (base <> EBP) THEN BEGIN(* no displacement *)
      PutByte (noDisp * $40 + reg * 8 + 4); (* escape to two bytes *)
      PutByte (scale * $40 + inx * 8 + base)
    END ELSE IF (disp <= 127) AND (disp >= -128) THEN BEGIN(* 8 bit displacement *)
      PutByte (Disp8 * $40 + reg * 8 + 4); (* escape to two bytes *)
      PutByte (scale * $40 + inx * 8 + base);
      PutByte (disp);
    END ELSE BEGIN
      PutByte (Disp32 * $40 + reg * 8 + 4); (* escape to two bytes *)
      PutByte (scale * $40 + inx * 8 + base);
      PutDWord (disp)
    END
  END
END;

PROCEDURE PutRegMem (mode: SHORTINT; reg, base: INTEGER; disp: LONGINT);
(* put register memory *)
BEGIN
  reg:= reg MOD 8;
  IF mode = omAbs THEN BEGIN
    PutByte (noDisp * $40 + reg * 8 + 5); (* Adr32 *) 
    PutDWord (disp)
  END ELSE BEGIN(* mode = RegRel *)
    base:= base MOD 8;
    IF base = ESP THEN (* 2 bytes address encoding necessary *)
      PutRMInx (mode, reg, base, 4, Scale1, disp) (* no index register: reg, disp[ESP] *)
    ELSE IF (disp = 0) AND (base <> EBP) THEN (* no displacement *)
      PutByte (noDisp * $40 + reg * 8 + base)
    ELSE IF (disp <= 127) AND (disp >= -128) THEN BEGIN(* 8 bit displacement *)
      PutByte (Disp8 * $40 + reg * 8 + base);
      PutByte (disp)
    END ELSE BEGIN(* 32 bit displacement *)
      PutByte (Disp32 * $40 + reg * 8 + base);
      PutDWord (disp)
    END
  END
END;
	
PROCEDURE PutMem (reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  IF base = noBase THEN BEGIN(* Abs *)
    IF inx = noInx THEN
      PutRegMem (omAbs, reg, noBase, disp)
    ELSE
      PutRMInx (omAbs, reg, noBase, inx, scale, disp);
  END ELSE IF inx = noInx THEN
    PutRegMem (omRegRel, reg, base, disp)
  ELSE
    PutRMInx (omRegRel, reg, base, inx, scale, disp);
END;
	
PROCEDURE PutDisp (disp: LONGINT; size: INTEGER);
BEGIN
  lastImmSize:= size;
  IF size >= Bit8 THEN PutByte (disp)
  ELSE IF size >= Bit16 THEN PutWord (disp)
  ELSE PutDWord (disp) (* size =  Bit32 *)
END;
	
PROCEDURE Prefix (reg: INTEGER; VAR w: SHORTINT);
BEGIN
  IF reg IN [AX..DI] THEN BEGIN
    PutByte ($66);
    w:= 1 (* Operand Size Prefix *)
  END ELSE IF reg IN [EAX..EDI] THEN
    w:= 1
  ELSE
    w:= 0;
END;
		
PROCEDURE GenMOV (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
VAR
  w: SHORTINT;
  r: INTEGER;
BEGIN
  Prefix (reg, w);
  CASE mode OF
      RegReg:
              BEGIN
                PutByte ($8A + w);
                PutReg (reg, base);
              END;
      RegMem:
              IF ( (reg = EAX) OR (reg = AX) OR (reg = AL) )
              AND (inx = noInx) AND (base = none) THEN BEGIN
                PutByte ($0A2 + w);
                PutDWord (disp)
              END ELSE BEGIN
                PutByte ($88 + w);
                PutMem (reg, base, inx, scale, disp)
              END;
      MemReg:
              BEGIN
                IF ( (reg = EAX) OR (reg = AX) OR (reg = AL) )
                AND (inx = noInx) AND (base = none) THEN BEGIN
                  PutByte ($0A0 + w);
                  PutDWord (disp);
                END ELSE BEGIN
                  PutByte ($8A + w);
                  PutMem (reg, base, inx, scale, disp);
                END;
              END;
      ImmReg:
              BEGIN
                r:= reg;
                reg:= reg MOD 8;
                PutByte ($0B0 + w * 8 + reg);
                PutDisp (imm, r)
              END;
      ImmMem:
              BEGIN
                PutByte ($0C6 + w);
                PutMem (reg, base, inx, scale, disp);
                PutDisp (imm, reg);
              END;
  ELSE
    HALT(BUG);
  END
END;

PROCEDURE GenMOVSX (mode, s: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
VAR
  w: SHORTINT; (* s = 0: 8 bit; s = 1: 16/32 bit *)
BEGIN
  Prefix (reg, w);
  PutByte ($0F);
  PutByte ($0BE + s);
  CASE mode OF
    RegReg: PutReg (reg, base);
    MemReg: PutMem (reg, base, inx, scale, disp);
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenMOVZX(mode, s: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
VAR
  w: SHORTINT; (* s = 0: 8 bit; s = 1: 16/32 bit *)
BEGIN
  Prefix (reg, w);
  PutByte ($0F);
  PutByte ($0B6+s);
  CASE mode OF
    RegReg: PutReg (reg, base);
    MemReg: PutMem (reg, base, inx, scale, disp);
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenIN(size: INTEGER);
BEGIN
  CASE size OF
    Bit32 : PutByte($0ED);		(* IN EAX, DX *)
    Bit16 : PutWord($0ED66);   (* IN AX, DX *)
    Bit8  : PutByte($0EC);	    (* IN AL, DX *)
   END
END;
	
PROCEDURE GenOUT(size: INTEGER);
BEGIN
  CASE size OF
    Bit32 : PutByte($0EF);		(* OUT DX, EAX *)
    Bit16 : PutWord($0EF66);   (* OUT DX, AX *)
    Bit8  : PutByte($0EE);	    (* OUT DX, AL *)
   END
END;

PROCEDURE GenPUSH (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
VAR
  w: SHORTINT;
BEGIN
  Prefix (reg, w);
  CASE mode OF
    Regs:
            BEGIN
              reg:= _SYSTEM._MOD(reg,8);
              PutByte ($50 + reg)
            END;
    Mem: (* push memory *)
            BEGIN
              reg:= ESI;
              PutByte ($0FF);
              PutMem (reg, base, inx, scale, disp)
            END;
    Imme: (* push immediate *)
            IF (imm <= 127) AND (imm >= -128) THEN BEGIN
              PutByte ($6A);
              PutByte (imm)
            END ELSE BEGIN
              PutByte ($68);
              PutDWord(imm)
            END
  ELSE
    HALT (BUG)
  END
END;
	
PROCEDURE GenPOP (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
VAR
  w: SHORTINT;
BEGIN
  Prefix (reg, w);
  CASE mode OF
    Regs:
            BEGIN
              reg:= reg MOD 8;
              PutByte ($58 + reg)
            END;
    Mem: (* pop memory *)
            BEGIN
              reg:= EAX;
              PutByte ($8F);
              PutMem (reg, base, inx, scale, disp)
            END;
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenXCHG (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
VAR
  w: SHORTINT;
BEGIN
  Prefix (reg, w);
  CASE mode OF
    RegReg:
              IF (reg = EAX) OR (reg = AX) THEN BEGIN
                base:= base MOD 8;
                PutByte ($90 + base);
              END ELSE BEGIN
                PutByte ($86 + w);
                PutReg (reg, base);
              END;
    RegMem:
              BEGIN
                PutByte ($86 + w);
                PutMem (reg, base, inx, scale, disp);
              END
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenLEA(reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  IF (reg = base) AND (inx = noInx) AND (disp = 0) THEN
    (* skip: LEA reg, 0[reg] has no effect *)
  ELSE BEGIN
    PutByte ($8D);
    PutMem (reg, base, inx, scale, disp)
  END
END;

PROCEDURE GenTyp1 (op, mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
(* general code generator procedure for ADD, ADC, AND, XOR, OR, SBB, SUB, CMP *)
VAR
  w, wImm: SHORTINT;
BEGIN
  Prefix (reg, w);
  CASE mode OF
    RegReg: (* reg = destination, base = source *)
              BEGIN
                PutByte (op + $02 + w);
                PutReg (reg, base);
              END;
    RegMem:
              BEGIN
                PutByte (op + w);
                PutMem (reg, base, inx, scale, disp)
              END;
    MemReg:
              BEGIN
                PutByte (op + $2 + w);
                PutMem (reg,base, inx, scale, disp)
              END;
    ImmReg:
              IF (reg = EAX) OR (reg = AX) OR (reg = AL) THEN BEGIN
                PutByte (op + $4 + w);
                PutDisp (imm, reg);
              END ELSE BEGIN
                IF reg >= AL THEN BEGIN(* 8 bit *)
                  PutByte ($80);
                  wImm:= Bit8;
                END ELSE IF (imm <= 127) AND (imm >= -128) THEN BEGIN(* sign extended *)
                  PutByte ($83);
                  wImm:= Bit8
                END ELSE BEGIN(* 16/32 bit immediate *)
                  PutByte ($81);
                  wImm:= reg (* 16/32 bit *)
                END;
                PutReg (op DIV 8, reg); (* op DIV 8 is code for op *)
                PutDisp (imm, wImm)
              END;
    ImmMem:
              BEGIN
                IF reg >= AL THEN BEGIN(* byte ptr *)
                  PutByte ($80);
                  wImm:= Bit8
                END ELSE IF (imm <= 127) AND (imm >= -128) THEN BEGIN (* sign extended *)
                  PutByte ($83);
                  wImm:= Bit8 (* 16/32 bit *)
                END ELSE BEGIN(* 16/32 bit immediate *)
                  PutByte ($81);
                  wImm:= reg (* 16/32 bit *)
                END;
                PutMem (op DIV 8, base, inx, scale, disp);
                PutDisp (imm, wImm)
              END;
  ELSE
    HALT (BUG)
  END
END;
		
PROCEDURE GenINC (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
VAR
  w: SHORTINT;
BEGIN
  Prefix (reg, w);
  IF mode = ImmReg THEN BEGIN
    IF w <> 0 THEN BEGIN
      reg:= reg MOD 8;
      PutByte ($40 + reg)
    END ELSE BEGIN
      PutByte ($0FE + w);
      PutReg (0, reg)
    END
  END ELSE BEGIN
    PutByte ($0FE + w);
    PutMem (0, base, inx, scale, disp)
  END
END;
	
PROCEDURE GenDEC (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
VAR
  w: SHORTINT;
BEGIN
  Prefix (reg, w);
  IF mode = ImmReg THEN BEGIN
    IF w <> 0 THEN BEGIN
      reg:= reg MOD 8;
      PutByte ($48 + reg)
    END ELSE BEGIN
      PutByte ($0FE + w);
      PutReg (1, reg)
    END
  END ELSE BEGIN
    PutByte ($0FE + w);
    PutMem (1, base, inx, scale, disp)
  END
END;

PROCEDURE GenNEG (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
VAR
  w: SHORTINT;
BEGIN
  Prefix (reg, w);
  PutByte ($0F6 + w);
  IF mode = Regs THEN
    PutReg (3, reg)
  ELSE
    PutMem (3, base, inx, scale, disp);
END;

PROCEDURE GenIMUL (mode: SHORTINT; shortform: BOOLEAN; reg, base, inx: INTEGER; scale: SHORTINT;disp, imm: LONGINT);
VAR
  w: SHORTINT;
BEGIN
  Prefix (reg, w);
  CASE mode OF
    RegReg: (* reg := reg * base *)
              IF shortform THEN BEGIN
                PutByte ($0F6 + w);
                PutReg (5, base)
              END ELSE BEGIN
                PutByte ($0F);
                PutByte ($0AF);
                PutReg (reg, base)
              END;
    MemReg: (* reg := reg * mem *)
              IF shortform THEN BEGIN
                PutByte ($0F6 + w);
                PutMem (5, base, inx, scale, disp)
              END ELSE BEGIN
                PutByte ($0F);
                PutByte ($0AF);
                PutMem (reg, base, inx, scale, disp)
              END;
    ImmReg: (* reg := base * imm *)
              IF (imm <= 127) AND (imm >= -128) THEN BEGIN
                PutByte ($6B);
                PutReg (reg, base);
                PutByte (imm)
              END ELSE BEGIN
                PutByte ($69);
                PutReg (reg, base);
                PutDisp (imm, base)
              END;
    ImmMem:
              IF (imm <= 127) AND (imm >= -128) THEN BEGIN
                PutByte ($6B);
                PutMem (reg, base, inx, scale, disp);
                PutByte (imm)
              END ELSE BEGIN
                PutByte ($69);
                PutMem (reg, base, inx, scale, disp);
                PutDisp (imm, base)
              END;
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenIDIV (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
VAR
  w: SHORTINT;
BEGIN
  Prefix (reg, w);
  PutByte ($0F6 + w);
  IF mode = RegReg THEN
    PutReg (7, reg)
  ELSE
    PutMem (7, base, inx, scale, disp);
END;

PROCEDURE GenShiftRot (op, mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
(* Generates code for ROL, ROR, SAL, SAR, SHL and SHR *)
VAR
  w: SHORTINT;
BEGIN
  Prefix (reg, w);
  CASE mode OF
    RegReg: (* reg by  base, base = CL *)
              BEGIN
                PutByte ($0D2 + w);
                PutReg (op, reg)
              END;
    RegMem: (* mem by reg; reg = CL *)
              BEGIN
                PutByte ($0D2 + w);
                PutMem (op, base, inx, scale, disp)
              END;
    ImmReg: (* register by immediate *)
              IF imm = 1 THEN BEGIN
                PutByte ($0D0 + w);
                PutReg (op, reg)
              END ELSE BEGIN
                PutByte ($0C0 + w);
                PutReg (op, reg);
                PutByte (imm);
              END;
    ImmMem: (* memory by immediate *)
              IF imm = 1 THEN BEGIN
                PutByte ($0D0 + w);
                PutMem (op, base, inx, scale, disp)
              END ELSE BEGIN
                PutByte ($0C0 + w);
                PutMem (op, base, inx, scale, disp);
                PutByte (imm)
              END
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenNOT (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
VAR
  w: SHORTINT;
BEGIN
  Prefix (reg, w);
  PutByte ($0F6 + w);
  IF mode = Regs THEN
    PutReg (2, reg)
  ELSE
    PutMem (2, base, inx, scale, disp);
END;

PROCEDURE GenString (op, size: INTEGER);
VAR
  w: SHORTINT;
BEGIN
  Prefix (size, w);
  PutByte (op * 2 + w)
END;

PROCEDURE GenRepString (op, size: INTEGER); (* without REPE CMPS, REPE SCAS *)
VAR
  w: SHORTINT;
BEGIN
  Prefix (size, w);
  PutByte ($0F3);
  PutByte (op * 2 + w)
END;
	
PROCEDURE GenRepCmpsScas(op, size: INTEGER);
(* REPE CMPS, REPE SCAS *)
VAR
  w: SHORTINT;
BEGIN
  Prefix (size, w);
  PutByte ($0F3);
  PutByte (op * 2 + w)
END;
	
PROCEDURE GenTEST (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
VAR
  w: SHORTINT;
BEGIN
  Prefix(reg, w);
  CASE mode OF
    RegReg:	(* TEST reg, base *)
              BEGIN
                IF w=0 THEN
                  PutByte($84)
                ELSE
                  PutByte($85);
                PutReg(base, reg)
              END;
    RegMem:	(* TEST mem, reg *)
              BEGIN
                IF w=0 THEN
                  PutByte($84)
                ELSE
                  PutByte($85);
                PutMem (reg, base, inx, scale, disp)
              END;
    ImmReg:	(* TEST reg, imm *)
              IF reg IN [EAX, AX, AL] THEN BEGIN
                IF w=0 THEN
                  PutByte($0A8)
                ELSE
                  PutByte($0A9);
                PutDisp(imm, (reg DIV 8) * 8)
              END ELSE BEGIN
                IF w=0 THEN
                  PutByte($0F6)
                ELSE
                  PutByte($0F7);
                PutReg(0, reg);
                PutDisp(imm, (reg DIV 8) * 8)
              END;
    ImmMem:	(* TEST mem, imm *)
              BEGIN
                IF w=0 THEN
                  PutByte($0F6)
                ELSE
                  PutByte($0F7);
                PutMem(0, base, inx, scale, disp);
                PutDisp(imm, (reg DIV 8) * 8)
              END;
  ELSE
    HALT (BUG)
  END
END;
	
PROCEDURE GenBT (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
VAR
  w: SHORTINT;
BEGIN
  Prefix (reg, w);
  PutByte ($0F);
  CASE mode OF
    RegReg: (* BT reg, base *)
              BEGIN
                PutByte ($0A3);
                PutReg (base, reg)
              END;
    RegMem: (* BT mem, reg *)
              BEGIN
                PutByte ($0A3);
                PutMem (reg, base, inx, scale, disp)
              END;
    ImmReg: (* BT reg, imm *)
              BEGIN
                PutByte ($0BA);
                PutReg (4, reg);
                PutByte (imm)
              END;
    ImmMem: (* BT mem, imm *)
              BEGIN
                PutByte ($0BA);
                PutMem (4, base, inx, scale, disp);
                PutByte (imm)
              END;
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenBTR (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
  VAR w: SHORTINT;
BEGIN
  Prefix (reg, w);
  PutByte ($0F);
  CASE mode OF
    RegReg: (* BTR reg, base *)
              BEGIN
                PutByte ($0B3);
                PutReg (base, reg)
              END;
    RegMem: (* BTR mem, reg *)
              BEGIN
                PutByte ($0B3);
                PutMem (reg, base, inx, scale, disp)
              END;
    ImmReg: (* BTR reg, imm *)
              BEGIN
                PutByte ($0BA);
                PutReg (6, reg);
                PutByte (imm)
              END;
    ImmMem: (* BTR mem, imm *)
              BEGIN
                PutByte ($0BA);
                PutMem (6, base, inx, scale, disp);
                PutByte (imm)
              END;
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenBTS (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp, imm: LONGINT);
VAR
  w: SHORTINT;
BEGIN
  Prefix (reg, w);
  PutByte ($0F);
  CASE mode OF
    RegReg: (* BTS reg, base *)
              BEGIN
                PutByte ($0AB);
                PutReg (base, reg) (* !! see also OPC.Include *)
              END;
    RegMem: (* BTS mem, reg *)
              BEGIN
                PutByte ($0AB);
                PutMem (reg, base, inx, scale, disp)
              END;
    ImmReg: (* BTS reg, imm *)
              BEGIN
                PutByte ($0BA);
                PutReg (5, reg);
                PutByte (imm)
              END;
    ImmMem: (* BTS mem, imm *)
              BEGIN
                PutByte ($0BA);
                PutMem (5, base, inx, scale, disp);
                PutByte (imm);
              END;
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenCALL (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
(* no intersegment call implemented yet *)
BEGIN
  IF mode = Imme THEN BEGIN (* direct call within segment *)
    PutByte ($0E8);
    PutDWord (disp)
  END ELSE BEGIN
    PutByte ($0FF);
    IF mode = Regs THEN
      PutReg (2, reg) (* call reg *)
    ELSE
      PutMem (2, base, inx, scale, disp); (* Memory call *)
  END
END;

PROCEDURE GenJMP (mode: SHORTINT; reg, base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  IF mode =  Imme THEN BEGIN(* direct jmp whithin segment *)
    IF (disp <= 127) AND (disp >= -128) THEN BEGIN(* short jump *)
      PutByte ($0EB);
      PutByte (disp)
    END ELSE BEGIN
      PutByte ($0E9);
      PutDWord (disp)
    END;
  END ELSE BEGIN
    PutByte ($0FF);
    IF mode = Regs THEN
      PutReg (4, reg) (* jmp reg *)
    ELSE
      PutMem (4, base, inx, scale, disp); (* jmp memory *)
  END;
END;
	
PROCEDURE GenRET (size: LONGINT);
BEGIN
  IF size = 0 THEN
    PutByte ($0C3)
  ELSE BEGIN
    PutByte ($0C2);
    PutWord (size)
  END;
END;
	
PROCEDURE GenJcc (op: SHORTINT; disp: LONGINT);
BEGIN
  IF (disp <= 127) AND (disp >= -128) THEN BEGIN(* short jmp *)
    PutByte ($70 + op);
    PutByte (disp)
  END ELSE BEGIN(* near jmp, always 32 bit wide *)
    PutByte ($0F);
    PutByte ($80 + op);
    PutDWord (disp)
  END
END;

PROCEDURE GenSetcc(op, mode: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  PutByte ($0F);
  PutByte ($90 + op);
  IF mode = Regs THEN
    PutReg (0, base) (* setcc reg *)
  ELSE
    PutMem (0, base, inx, scale, disp)
END;

PROCEDURE GenINT(intNumber: INTEGER);
(* int intNumber *)
BEGIN
  PutByte ($0CD);
  PutByte (intNumber)
END;

PROCEDURE InlineCode (VAR code: ARRAY OF CHAR; parSize: INTEGER);
VAR
  i, n: INTEGER;
BEGIN
  n := ORD(code[0]);
  i := 1;
  WHILE i <= n DO BEGIN
    PutByte(ORD(code[i]));
    INC(i)
  END;
END;

(* floating point encoding *)
	
PROCEDURE GenFLD (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  CASE mode OF
    Regs: (* fload st(0), st (i) ; st(i) = base *)
              BEGIN
                PutByte ($0D9);
                PutByte ($0C0 + base)
              END;
    Mem: (* fload st(0), mem *)
              IF size = eReal THEN BEGIN
                PutByte ($0DB);
                PutMem (5, base, inx, scale, disp)
              END ELSE BEGIN
                PutByte ($0D9 + size * 2);
                PutMem (0, base, inx, scale, disp)
              END
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenFST (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  CASE mode OF
    Regs: (* fstore st(0), st(i); st(i) = base *)
              BEGIN
                PutByte ($0DD);
                PutByte ($0D0 + base);
              END;
    RegMem: (* fstore mem, st(0) *)
              BEGIN
                PutByte ($0D9 + size * 2);
                PutMem (2, base, inx, scale, disp);
              END
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenFSTP (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  CASE mode OF
    Regs: (* fstore st(0), st(i); st(i) = base *)
              BEGIN
                PutByte ($0DD);
                PutByte ($0D8 + base)
              END;
    RegMem:
              IF size = eReal THEN BEGIN
                PutByte ($0DB);
                PutMem (7, base, inx, scale, disp)
              END ELSE BEGIN
                PutByte ($0D9 + size * 2);
                PutMem (3, base, inx, scale, disp)
              END
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenFCOM (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  CASE mode OF
    Regs: (* fcom st(0), st(i) *)
              BEGIN
                PutByte ($0D8);
                PutByte ($0D0 + base)
              END;
    Mem: (* fcom st(0), mem *)
              BEGIN
                PutByte ($0D8 + size * 2);
                PutMem (2, base, inx, scale, disp)
              END;
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenFCOMP (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  CASE mode OF
    Regs: (* fcomp st(0), st(i) *)
              BEGIN
                PutByte ($0D8);
                PutByte ($0D8+ base)
              END;
    Mem: (* fcomp st(0), mem *)
              BEGIN
                PutByte ($0D8 + size * 2);
                PutMem (3, base, inx, scale, disp)
              END;
  ELSE
    HALT (BUG)
  END
END;
	
PROCEDURE GenFtyp1 (op, mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  CASE mode OF
    RegSt: (* op st, base and pop not possible*)
              BEGIN
                PutByte ($0D8);
                PutByte ($0C0 + op * 8 + base)
              END;
    StReg: (* op base, st *)
              BEGIN
                PutByte ($0DC);
                PutByte ($0C0 + op * 8 + base)
              END;
    StRegP: (* op base, st *)
              BEGIN
                PutByte ($0DE);
                PutByte ($0C0 + op * 8 + base)
              END;
    MemSt:
              BEGIN
                PutByte ($0D8 + size * 2);
                PutMem (op, base, inx, scale, disp)
              END
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenFADD (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  GenFtyp1 (0, mode, size, base, inx, scale, disp)
END;

PROCEDURE GenFSUB (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  IF (mode = StReg) OR (mode = StRegP) THEN
    GenFtyp1 (5, mode, size, base, inx, scale, disp)
  ELSE
    GenFtyp1 (4, mode, size, base, inx, scale, disp)
END;
	
PROCEDURE GenFSUBR (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  IF (mode = StReg) OR (mode = StRegP) THEN
    GenFtyp1 (4, mode, size, base, inx, scale, disp)
  ELSE
    GenFtyp1 (5, mode, size, base, inx, scale, disp);
END;
	
PROCEDURE GenFMUL (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  GenFtyp1 (1, mode, size, base, inx, scale, disp)
END;
	
PROCEDURE GenFDIV(mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  IF (mode = StReg) OR (mode = StRegP) THEN
    GenFtyp1 (7, mode, size, base, inx, scale, disp)
  ELSE
    GenFtyp1 (6, mode, size, base, inx, scale, disp);
END;
	
PROCEDURE GenFDIVR (mode, size: SHORTINT; base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  IF (mode = StReg) OR (mode = StRegP) THEN
    GenFtyp1 (6, mode, size, base, inx, scale, disp)
  ELSE
    GenFtyp1 (7, mode, size, base, inx, scale, disp);
END;

PROCEDURE GenFFREE (freg: INTEGER);
BEGIN
  PutByte ($0DD);
  PutByte ($0C0 + freg)
END;
	
PROCEDURE GenFop1 (op: INTEGER);
BEGIN
  CASE op OF
    gfCOMPP  : BEGIN PutByte ($0DE); PutByte ($0D9); END;
    gfTST    : BEGIN PutByte ($0D9); PutByte ($0E4); END;
    gfLDZ    : BEGIN PutByte ($0D9); PutByte ($0EE); END;
    gfLD1    : BEGIN PutByte ($0D9); PutByte ($0E8); END;
    gfABS    : BEGIN PutByte ($0D9); PutByte ($0E1); END;
    gfCHS    : BEGIN PutByte ($0D9); PutByte ($0E0); END;
    gfSTSW   : BEGIN PutByte ($0DF); PutByte ($0E0); END;
    gfINCSTP : BEGIN PutByte ($0D9); PutByte ($0F7); END;
    gfDECSTP : BEGIN PutByte ($0D9); PutByte ($0F6); END;
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE GenFSTCW (base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  PutByte ($0D9);
  PutMem (7, base, inx, scale, disp)
END;

PROCEDURE GenFLDCW (base, inx: INTEGER; scale: SHORTINT; disp: LONGINT);
BEGIN
  PutByte ($0D9);
  PutMem (5, base, inx, scale, disp)
END;

PROCEDURE Init;
BEGIN
  SETLENGTH(constant, 1024);
END;

END.