unit Compiler;

interface

uses
  Windows, SysUtils, Classes, ShellApi, DebugFrm, ManagerFrm, LNTi386, LNKi386;

  procedure Compile(fFileName: STRING; AndLink,AndExec: BOOLEAN);
  
implementation

uses
  _SYSTEM, OPT, OPS, OPM, OPV, OPL, OPP, OPB, OPA,
  OPCi386, OPLi386, OPVi386, OPOi386;

var
  fStack    : TModList;

procedure Module(modName: String);
var
  OPS : TScaner;
  n   : integer;
  sym : TToken;
  impName: String;
begin
  if fStack.Find(modName,n) then Exit;
  OPS:= TScaner.Create(modName);
  OPS.Get(sym);
  IF sym=_module THEN OPS.Get(sym) ELSE err(16);
  IF sym = _ident THEN BEGIN
    OPS.Get(sym);
    OPS.CheckSym(sym,_semicolon);
    IF sym = _import THEN BEGIN
      OPS.Get(sym);
      REPEAT
        IF sym = _ident THEN BEGIN
          impName:=OPS.name;
          if (FileExists(impName+moduleext)) then Module(impName+moduleext);
        END;
        OPS.Get(sym);
        IF sym = _becomes THEN BEGIN
          OPS.Get(sym);
          IF sym = _ident THEN BEGIN
            impName:=OPS.Name;
            if (FileExists(impName+moduleext)) then begin
              Module(impName+moduleext);
              OPS.Get(sym);
            end else if impName='SYSTEM' then
              OPS.Get(sym)
            else
              err(152);
          END ELSE
            err(ORD(_ident));
        END ELSE IF sym = _period THEN
          err(200);
        IF      sym = _comma THEN OPS.Get(sym)
        ELSE IF sym = _ident THEN err(ORD(_comma))
        ELSE BREAK;
      UNTIL FALSE;
      OPS.CheckSym(sym,_semicolon);
    END;
  END ELSE
    err(ORD(_ident));
  fStack.Add(modName);
  OPS.Free;
end;

PROCEDURE BUILD(VAR p: OPT.Nodes; S: STRING);
VAR
  optn: TSET;
begin
  optn:=DefOpt;

  OPM.OPMBegin;
  OPT.OPTBegin;
  OPL.OPLBegin;

  OPLi386.OPLBegin;
  OPLi386.Init;

  OPV.OPVBegin;
  OPB.OPBBegin;
  OPA.OPABegin;

  breakpc:=0;
  OPM.Init(optn);
  OPV.Init(optn, 0);

  OnewSF := OPM.newsf IN optn;
  OextSF := OPM.extsf IN optn;
  OPP.Module(p,s,optn);
  
  IF OPM.noerr THEN BEGIN
    OPL.Init;
    OPM.errpos.CodePos := 0;
    OPM.FNBegin(S);
    IF OPM.noerr THEN BEGIN
      OPT.OPTExport(S, OnewSF, OextSF);
      OPV.AdrAndSize(OPT.topScope);
      IF OPM.noerr THEN BEGIN
					OPM.errpos.CodePos := 0;
          OPOi386.Init;
          OPCi386.OPCBegin;
					OPCi386.Init;
					OPVi386.Module(p);
					IF OPM.noerr THEN BEGIN
						OPLi386.OutCode(S);
						IF OPM.noerr THEN BEGIN
              DebugPanel.AddMsg('Size '''+S+'.Obj'' - '+IntToStr(OPOi386.pcc)+' bytes',1,NIL);
            END;
					END;
      END;

    END;
  END;
  
  OPL.Close;
  OPLi386.Close;
  OPT.CloseScope;       
  OPT.Close;
  //p:=NIL;
  //OPT.OBJDestroy;
END;

procedure Compile(fFileName: STRING; AndLink,AndExec: BOOLEAN);
var
  i,l     : integer;
  S,E     : STRING;
  Ticks   : Cardinal;
  Linker  : TLinker;
  p       : OPT.Nodes;
  Flag    : BOOLEAN;
begin

  DebugPanel.Clear;
  FirstDebug:=true;
  fStack:= TModList.Create;
  OPM.ConNoErr    := TRUE;
  Module(fFileName);

  Ticks:=Windows.GetTickCount; 

  i:=0;
  WHILE i<fStack.count DO BEGIN
    s:=fStack.FindNum(i);
    DebugPanel.AddMsg('Compile "'+ExtractFileName(S)+'"',5,NIL);
    LNTi386.Init();
    BUILD(p,s);
    LNTi386.Done;
    IF NOT OPM.noerr THEN BREAK;
    INC(i);
  END;

  IF AndLink THEN
    IF OPM.noerr THEN BEGIN

      //#################################################
      LNTi386.Init();
      Linker  := TLinker.Create;
      Flag    := FALSE;

      FOR i:=0 TO fStack.count-1 DO BEGIN
        S:=fStack.FindNum(i);
        E:=ExtractFileExt(S);
        S:=COPY(S,1,SYSTEM.Pos(E,S)-1);
        IF S='Kernel' THEN Flag:=TRUE;
        LNTi386.AddModule(s);
      END;

      IF ManagerPanel.TreeView.Items.Count>0 THEN BEGIN
        S:=ExtractFileName(ManagerPanel.TreeView.Items.Item[0].Text);
        E:=ExtractFileExt(S);
        S:=COPY(S,1,SYSTEM.Pos(E,S)-1);
      END;

      DebugPanel.AddMsg('Linkin "'+S+E+'"',5,NIL);

      IF Flag THEN BEGIN
        AddImport('Kernel','bootHeapAdr', '','HeapAdr',   0);
        AddImport('Kernel','bootHeapSize','','HeapSize',  0);
        AddImport('Kernel','hInstance',   '','hInstance', 0);
      END;

      l:=Linker.DoLink(GUI,FALSE,'Program1.Link',S);
      Linker.Free;
      LNTi386.Done;
      //#################################################

      DecimalSeparator := '.';
      DebugPanel.AddMsg('Create file "'+S+OPM.exeext+'" '+FORMATFLOAT('0.00',l/1000)+'Kb. in '+FloatToStr(Windows.GetTickCount-Ticks)+'ms ... OK!',1,NIL);

      IF AndExec THEN BEGIN
        ShellExecute(0, 'open', PChar(S+OPM.exeext), nil, nil, SW_SHOWNORMAL);
      END;
    
    END;

  fStack.Free;
  FirstDebug:=false;

end;

end.
