UNIT LNBi386;

INTERFACE

USES
  Windows, SysUtils, Classes, OPM, _SYSTEM, LNTi386, DebugFrm,
  DIALOGS;

CONST
	Boundary = 32;

	newSF       = 0;
  sysnewSF    = 1;
  newarrSF    = 2;
  StartSF     = 3;
  PassivateSF = 4;
  ActivateSF  = 5;
  LockSF      = 6;
	UnlockSF    = 7;
  copyarraySF = 10;
	CurProcSF   = 11;
  commandSF   = 12;
  listSF      = 13;
  modDescSF   = 14;
  expDescSF   = 15;
  objectSF    = 16;
  MaxSF       = 17;

	ProcType = 0;
  VarType = 1;
	
	OFVersion = #$AF;

	PaddingSize = 64;
	PageSize = 4096;
	hiddenTD = 20;

	done              = 0;
	fileNotFound      = 1;
	invalidObjFile    = 2;
	corruptedObjFile  = 4;
	cmdNotFound       = 5;
	moduleNotFound    = 6;
	notEnoughSpace    = 7;
	refCntNotZero     = 8;
	cyclicImport      = 9;
	incompImport      = 16;

TYPE
	ArrayDesc = RECORD
    a, b, c, len: LONGINT
  END;

CONST
	ArrayDescSize = 16;
	TagSize       = 4;

TYPE
	Name = STRING;
  ModuleName = Name;	(* alias for the lader/linker *)

	CommandDesc = RECORD
		name: Name;
		adr: LONGINT
	END;
	Commands = ARRAY OF CommandDesc;

	Command = PROCEDURE;

	NewMethod = RECORD
		mthNo, entryNo: INTEGER;
	END;
	
	ArrayPtr = ARRAY OF LONGINT;

	tType = ^TypeDesc;
	Module = ^ModuleDesc;
	
	ExportPtr = ^ExportDesc;
	ExportDesc = CLASS
		fp    : LONGINT;
    adr   : LONGINT;
		nofExp: INTEGER;
		dsc   : ARRAY OF ExportDesc;
		cAdr  : LONGINT;
    done  : BOOLEAN;
    types : tType;
	END;
	
	DumpExportDesc = RECORD
		fp    : LONGINT;
    adr   : LONGINT;
		nofExp: LONGINT;
		dsc   : LONGINT;
	END;

	TerminationHandler = PROCEDURE;
	
	TypeDesc = RECORD
		size, nofMethods, nofNewMethods,
    nofPtrs, tdAdr, tdEntry, tdSize,
    padSize, baseMod, baseEntry: LONGINT;
		name        : Name;
		newMethods  : ARRAY OF NewMethod;
		ptrOffset   : ARRAY OF LONGINT;
		baseType    : tType;
		extlev      : LONGINT;
		initialized : BOOLEAN;
		module      : Module;
	END;
	
	Bytes = ARRAY OF CHAR;

	ModuleDesc = RECORD
		link: Module;
		name: Name;
		init: BOOLEAN;
		refcnt, sb: LONGINT;
		entries: ARRAY OF LONGINT;
		cmds: ARRAY OF CommandDesc;
		ptrTab: ARRAY OF LONGINT;
		tdescs: ARRAY OF tType;
		imports: ARRAY OF	Module;
		data, code, refs: Bytes;
		publics, privates: LONGINT;
		nofimp: LONGINT; import: ArrayPtr;
		nofstrc: LONGINT; struct: ArrayPtr;
		nofreimp: LONGINT; reimp: ArrayPtr;
		export: ExportDesc;
		term: TerminationHandler;
		base, imageSize: LONGINT;
		codeBase, refBase: LONGINT;
		expAdr, expSize, expPadding: LONGINT;
		typeTableSize, typeTableAdr, typeTablePadding: LONGINT;
		modDescAdr: LONGINT;
		dataSize, conSize, codeSize, refSize: LONGINT;
		nofEntries, nofCmds, nofImps, nofTds, nofPtrs: LONGINT;
	END;

	DumpModule = ^DumpModuleDesc;
	DumpModuleDesc = RECORD
		link:	LONGINT;
		name: ARRAY [0..31] OF CHAR;
		init, trapped: BOOLEAN;
		refcnt, sb: LONGINT;
		entries, cmds, ptrTab,
    tdescs, imports, data,
    code, refs: LONGINT;
		publics, privates: LONGINT;
		nofimp: LONGINT;
    import: LONGINT;
		nofstrc: LONGINT;
    struct: LONGINT;
		nofreimp: LONGINT;
    reimp: LONGINT;
		export: DumpExportDesc;
		term: TerminationHandler;
	END;

	InitPoint = ^InitPointNode;
	InitPointNode = RECORD
		entryPoint: LONGINT;
		objects: Module;
		next: InitPoint
	END;

CONST
	EUEnd       = 0;
  EURecord    = 1;
  EUobjScope  = 0;
  EUrecScope  = 1;
  EUerrScope  = -1;
  EUProcFlag  = $80000000;

	ExtTabWordSize = 16;
	Tag0WordOffset = -2;
	Mth0WordOffset = Tag0WordOffset-ExtTabWordSize;

	DefMaxImport = 64;
  DefMaxStruct = 32;
  DefMaxReimp  = 32;

TYPE
	LinkTab = ARRAY [0..256] OF RECORD
		cmod, entry: CHAR;
		link: LONGINT
	END;

	DataLinkTab = ARRAY [0..128] OF RECORD
		cmod: CHAR;
		entry: WORD;
		nofFixups: WORD;
		offset: ARRAY OF LONGINT
	END;

	Cmd = CommandDesc;

VAR
	SysFix: ARRAY [0..MaxSF] OF RECORD
    name, module, command: Name;
    adr: LONGINT;
  END;
	protObj     : tType;
	objectList  : Module;
	includeRefs : BOOLEAN;
	imageBase, imageSize: LONGINT;
	moduleDescSize, mDescPadSize: LONGINT;
	moduleTag   : LONGINT;

	nofEntryPoints: LONGINT;
	initPointList: InitPoint;

	padding: ARRAY [0..PaddingSize] OF BYTE;

	logName: Name;
	res: INTEGER;
	AddReloc: PROCEDURE (offs, val: DWORD);

  FUNCTION  FindModule(name: STRING): Module;
  FUNCTION  Load(VAR m: Module; name: STRING; VAR base: LONGINT): BOOLEAN;
  PROCEDURE BeginLink(imgBase: LONGINT; VAR base: LONGINT);
  PROCEDURE EndLink(VAR Files: TMemoryStream; base: LONGINT; VAR initOffs: DWORD);
  FUNCTION  FindAdr(VAR Vmod: Module; VAR pat: Name; Vtype: LONGINT): LONGINT;
  PROCEDURE Close();

IMPLEMENTATION

PROCEDURE ReadNum(VAR F: TMemoryStream; VAR x: LONGINT);
VAR
  n: LONGINT;
  s: SHORTINT;
  ch: CHAR;
BEGIN
  s := 0;
  n := 0;
  F.Read(ch,1);
  WHILE ORD(ch) >= $80 DO BEGIN
    INC(n, _SYSTEM.ASH(ORD(ch) - $80, s) );
    INC(s, 7);
    F.Read(ch,1)
  END;
  x := n + _SYSTEM.ASH(ORD(ch) MOD 64 - ORD(ch) DIV 64 * 64, s);
END;

PROCEDURE ReadString(VAR F: TMemoryStream; VAR x: STRING);
VAR
  ch: CHAR;
BEGIN
  x:='';
  REPEAT
    F.Read(ch,1);
    IF ch<>#0 THEN x:=x+ch;
  UNTIL ch = #0;
END;

PROCEDURE ReadStringX(VAR Files: TMemoryStream; VAR strings: STRING);
VAR
  ch: CHAR;
BEGIN
  strings := '';
	REPEAT
    Files.Read(ch,1);
		IF ch = #0 THEN BEGIN
      strings := strings+'';
      EXIT;
		END ELSE IF ch > #$7F THEN BEGIN
      strings := strings+CHR(ORD(ch)-$80);
      EXIT;
		END ELSE BEGIN
      strings := strings+ch;
    END;
	UNTIL FALSE;
END;

FUNCTION FindModule(name: STRING): Module;
VAR
  t: Module;
BEGIN
	t := objectList;
	WHILE (t <> NIL) AND (t.name <> name) DO t := t.link;
	IF t <> NIL THEN BEGIN
    RESULT:=t;
    EXIT;
	END ELSE BEGIN
    RESULT:=NIL;
    EXIT;
	END;
END;

PROCEDURE ReadUnsigned (VAR Files: TMemoryStream; VAR u: LONGINT);
VAR
  low, high: CHAR;
BEGIN
	Files.Read (low,1);
  Files.Read (high,1);
	u := 256*ORD(high)+ORD(low);
END;

FUNCTION Align (num, boundary: LONGINT): LONGINT;
BEGIN
  RESULT:= num+_SYSTEM._MOD(-num,boundary);
END;

PROCEDURE AddAbsReloc(adr, val: LONGINT);
BEGIN
	AddReloc(adr-imageBase, val);
END;

FUNCTION FNAnd(x, y: LONGINT): LONGINT;
BEGIN
  RESULT:= x AND y;
END;

FUNCTION LSW (x: LONGINT): LONGINT;
BEGIN	(* least significant word (unsigned) *)
	RESULT:= x AND $0FFFF;
END;

FUNCTION MSW (x: LONGINT): LONGINT;
BEGIN
	RESULT:= _SYSTEM.LSH (x,-16);
END;

//====================================================================

PROCEDURE AddInitPoint (entryPoint: LONGINT; Vobject: Module);
VAR
	Vnew, p: InitPoint;
BEGIN
	NEW(Vnew);
	Vnew.entryPoint := entryPoint;
  Vnew.objects := Vobject;
	IF initPointList = NIL THEN BEGIN
		Vnew.next := initPointList;
    initPointList := Vnew;
	END ELSE BEGIN
		p := initPointList;
		WHILE p.next <> NIL DO p := p.next;
		Vnew.next := NIL;
    p.next := Vnew;
	END;
	INC (nofEntryPoints)
END;

PROCEDURE GetNum(refs: Bytes;  VAR i, num: LONGINT);
VAR
  n, s: LONGINT;
  x: CHAR;
BEGIN
	s := 0;
  n := 0;
  x := refs[i];
  INC(i);
	WHILE ORD(x) >= 128 DO BEGIN
		INC(n, _SYSTEM.ASH(ORD(x) - 128, s));
    INC(s, 7);
    x := refs[i];
    INC(i)
	END;
	num := n + _SYSTEM.ASH(ORD(x) MOD 64 - ORD(x) DIV 64 * 64, s)
END;

FUNCTION FindAdr(VAR Vmod: Module; VAR pat: Name; Vtype: LONGINT): LONGINT;	(* cf. System.FindProc *)
VAR
  i, j, m, t, ofs: LONGINT;
  ch: CHAR;
  found: BOOLEAN;
BEGIN
	i := 0;
  m := Vmod.refSize;
	ch := Vmod.refs[i];
  INC(i);
	WHILE (i < m) AND ((ch = #$0F8) OR (ch = #$0F9)) DO	BEGIN(* proc *)
		GetNum(Vmod.refs, i, ofs);	(* pofs *)
		IF ch = #$0F9 THEN BEGIN
			GetNum(Vmod.refs, i, t);	(* nofPars *)
			INC(i, 3)	(* RetType, procLev, slFlag *)
		END;
		found := TRUE;
    j := 1;
		REPEAT
      ch := Vmod.refs[i];
      found := found AND (ch = pat[j]);
      INC(j);
      INC(i);
    UNTIL ch = #0;	(* pname *)
		IF found AND (Vtype = ProcType) THEN BEGIN
      RESULT:= ofs;
      EXIT;
		END;
		IF i < m THEN BEGIN
			ch := Vmod.refs[i];
      INC(i);	(* 1X | 3X | 0F8X | 0F9X *)
			WHILE (i < m) AND (ch >= #1) AND (ch <= #3) DO BEGIN	(* var *)
				ch := Vmod.refs[i];
        INC(i);	(* type *)
				IF (ch >= #81) OR (ch = #16) OR (ch = #$1D) THEN GetNum(Vmod.refs, i, t);
				GetNum(Vmod.refs, i, ofs);	(* vofs *)
				found := TRUE;
        j := 1;
				REPEAT
          ch := Vmod.refs[i];
          found := found AND (ch = pat[j]);
          INC(j);
          INC(i)
        UNTIL ch = #0;	(* vname *)
				IF found AND (Vtype = VarType) THEN BEGIN
          RESULT:= ofs;
          EXIT;
				END;
				IF i < m THEN BEGIN ch := Vmod.refs[i]; INC(i) END;	(* 1X | 3X | 0F8X | 0F9X *)
			END;
		END;
	END;
	RESULT:= 0
	(* HALT(99)	name not found *)
END;

PROCEDURE Insert(o: Module);
VAR
  t: Module;
BEGIN	(* the modules must be sorted by the base address *)
	IF objectList = NIL THEN BEGIN
    objectList := o;
    o.link := NIL
	END ELSE BEGIN
		t := objectList;
		WHILE (t.link <> NIL) AND (t.link.base < o.base) DO t := t.link;
		o.link := t.link;
    t.link := o;
	END;
END;

PROCEDURE LoadModule (VAR Files: TMemoryStream; m: Module; VAR base: LONGINT);
TYPE
  TDesc = ^Desc;
  Desc = RECORD word: ARRAY [0..32000] OF LONGINT END;
VAR
  ch: CHAR;
  nofDataLinks, nofLinks: WORD;
  i, mno, j, symSize, e, t: LONGINT;
  imp: Module; mname: ModuleName;
  VlinkTab: LinkTab;
  dataLinks: DataLinkTab;

  PROCEDURE Expect (tag: CHAR);
	VAR
    ch: CHAR;
	BEGIN
    Files.Read (ch,1);
		IF (ch<>tag) THEN DebugPanel.AddMsg('corruptedObjFile '+m.name,4,NIL);
	END;

  PROCEDURE ReadImportList(m: Module);
  VAR
    Pos     : INTEGER;
    ImpPos  : INTEGER;
    LibName : STRING;
    mName   : STRING;
    iName   : STRING;
    i,n     : INTEGER;
	BEGIN
    Pos:=Files.Position;
    Files.Position:=Files.Size-4;
    ImpPos:=0;
    Files.Read(ImpPos,2);
    Files.Position:=ImpPos;
    Expect(#$8F);
    
    n:=0;
    Files.Read(n,1);

    WHILE n=$8E DO BEGIN
      LibName:='';
      ReadStringX(Files,LibName);

      LibName:=COPY(LibName,1,SYSTEM.Pos('.',LibName)-1);

      n:=0;
      Files.Read(n,2);

      FOR i:=0 TO n-1 DO BEGIN
        ReadStringX(Files,mName);
        ReadStringX(Files,iName);
        LNTi386.AddImport(m.name,mName,LibName,iName,0);
      END;

      n:=0;
      Files.Read(n,1);

    END;

    Files.Position:=Pos;
	END;

  PROCEDURE ReadHeader (VAR m: Module);
  VAR
    dummy: SMALLINT;
	BEGIN

    m.refSize:=0;
		m.nofEntries:=0;
		m.nofCmds:=0;
		m.nofPtrs:=0;
		m.nofTds:=0;
		m.nofImps:=0;
		nofDataLinks:=0;
		nofLinks:=0;
		m.dataSize:=0;
		m.conSize:=0;
		m.codeSize:=0;
		m.name:='';
    dummy:=0;

		Files.Read(m.refSize,4);
    Files.Read(dummy,2); m.nofEntries := dummy;
    Files.Read(dummy,2); m.nofCmds := dummy;
    Files.Read(dummy,2); m.nofPtrs := dummy;
    Files.Read(dummy,2); m.nofTds := dummy;
    Files.Read(dummy,2); m.nofImps := dummy;
		Files.Read(nofDataLinks,2);
		Files.Read(nofLinks,2);
		Files.Read(m.dataSize,4);
    Files.Read(dummy,2); m.conSize := dummy;
		ReadUnsigned(Files,m.codeSize);
		ReadString(Files,m.name);

	END;

  PROCEDURE ReadEntry (VAR m: Module);
	BEGIN
		IF m.nofEntries > 0 THEN SETLENGTH(m.entries,m.nofEntries);
		Expect(#$082);
    i := 0;
		WHILE i < m.nofEntries DO BEGIN
      ReadUnsigned(Files, m.entries[i]);
      INC(i);
    END;
	END;

  PROCEDURE ReadCmd(VAR m: Module);
	BEGIN
		IF m.nofCmds>0 THEN SETLENGTH(m.cmds, m.nofCmds);
		Expect (#$083);
    i := 0;
		WHILE i<m.nofCmds DO BEGIN
			ReadString(Files,m.cmds[i].name);
      ReadUnsigned(Files, m.cmds[i].adr);
			INC(i);
		END;
	END;

  PROCEDURE ReadPtr(VAR m: Module);
	BEGIN
		IF m.nofPtrs > 0 THEN SETLENGTH(m.ptrTab, m.nofPtrs);
		Expect(#$084);
		i := 0;
    WHILE i < m.nofPtrs DO BEGIN
			Files.Read(m.ptrTab[i],4);
			DEC(m.ptrTab[i], m.ptrTab[i] MOD 4);
			INC (i);
		END;
	END;

  PROCEDURE ReadImport (VAR m: Module);
	BEGIN
		IF m.nofImps > 0 THEN SETLENGTH(m.imports, m.nofImps);
		Expect (#$085);
    mno := 0;
		WHILE (mno < m.nofImps) AND (res = done) DO BEGIN
			ReadStringX(Files, mname);
			imp := FindModule (mname);
			IF imp = NIL THEN BEGIN
        WRITELN('Imported module not loaded');
        HALT(90);
      END;
			IF res = done THEN BEGIN
				INC (imp.refcnt);
				m.imports[mno] := imp;
				INC (mno)
			END
		END;
	END;

  PROCEDURE ReadDataLink(VAR m: Module);
	BEGIN
		Expect (#$08D);
    i := 0;
		WHILE i < nofDataLinks DO BEGIN
			Files.Read(dataLinks[i].Cmod,1);
			Files.Read(dataLinks[i].entry,2);
			Files.Read(dataLinks[i].nofFixups,2);
			IF dataLinks[i].nofFixups > 0 THEN BEGIN
				SETLENGTH(dataLinks[i].offset,dataLinks[i].nofFixups);
				j := 0;
				WHILE j<dataLinks[i].nofFixups DO BEGIN
					ReadUnsigned(Files, dataLinks[i].offset[j]);
					INC (j);
				END;
			END ELSE BEGIN
        SETLENGTH(dataLinks[i].offset,0);
        dataLinks[i].offset := NIL;
      END;
			INC(i);
		END;
	END;

  PROCEDURE ReadLink (VAR m: Module);
	BEGIN
		Expect(#$086);
    i := 0;
		WHILE i < nofLinks DO BEGIN
			Files.Read(VlinkTab[i].Cmod,1);
      Files.Read(VlinkTab[i].entry,1);
      ReadUnsigned(Files,VlinkTab[i].link);
			INC(i);
		END;
	END;

  PROCEDURE ReadDataConst (VAR m: Module);
	BEGIN
		m.dataSize := Align(m.dataSize,8);
		IF (m.dataSize+m.conSize)>0 THEN SETLENGTH(m.data, m.dataSize+m.conSize);
		i := 0;
		WHILE i < m.dataSize DO BEGIN
      m.data[i] := #0;
      INC(i);
    END;
		Expect (#$087);
    i := 0;
    t := m.dataSize;
		WHILE i < m.conSize DO BEGIN
      Files.Read (ch,1);
      m.data[t]:=ch;
      INC(t);
      INC(i);
    END;
	END;

  PROCEDURE ReadExport (VAR m: Module);
  VAR
    struct: ARRAY [0..1024] OF ExportPtr;
    nofStr: INTEGER;
		
		PROCEDURE LoadScope (VAR scope: ExportDesc; level: INTEGER; adr: LONGINT);
    VAR
      no, no2,jj: INTEGER;
      fp, off: LONGINT;
		BEGIN
			Files.Read(scope.nofExp, 2);
      no  := 0;
      no2 := 0;
			IF scope.nofExp <> 0 THEN BEGIN
				SETLENGTH(scope.dsc, scope.nofExp);
        FOR jj:=0 TO scope.nofExp-1 DO scope.dsc[jj]:=ExportDesc.Create;
				scope.dsc[0].adr := adr;
			END;
			IF level = EUrecScope THEN BEGIN
        INC (nofStr);
        struct[nofStr] := ExportPtr(_SYSTEM.ADR(scope));
      END;
			ReadNum(Files, fp);
			WHILE fp<>EUEnd DO BEGIN
				IF fp = EURecord THEN BEGIN
          ReadNum(Files, off);
					IF off < 0 THEN BEGIN
            scope.dsc[no2].nofExp := struct[-off].nofExp;
            scope.dsc[no2].dsc := struct[-off].dsc;
					END ELSE
            LoadScope (scope.dsc[no2], EUrecScope, off);
				END ELSE BEGIN
					IF level = EUobjScope THEN ReadNum(Files, scope.dsc[no].adr);
					scope.dsc[no].fp := fp;
          no2 := no;
          INC (no);
				END;
				ReadNum(Files, fp);
			END;
		END;
		
	BEGIN
		Expect (#$088);
		nofStr := 0;
    LoadScope (m.export, EUobjScope, 0);
	END;

  PROCEDURE ReadCode (VAR m: Module);
	BEGIN
		IF m.codeSize>0 THEN SETLENGTH(m.code, m.codeSize);
		Expect(#$089);
    Files.Read(m.code[0], m.codeSize);
	END;

  PROCEDURE AssignExportSizes (VAR exp: ExportDesc; m: Module);
  VAR
    size, i: LONGINT;
	BEGIN
		exp.done := TRUE;
    size := exp.nofExp;
		IF (size = 0) OR exp.dsc[0].done THEN EXIT;
		exp.dsc[0].cAdr := m.expAdr + m.expSize + 4;
		ASSERT (exp.dsc[0].cAdr MOD 32 = 0);
		INC (m.expSize, 32 * ((size*16 + TagSize+ArrayDescSize + 31) DIV 32));
		FOR i := 0 TO size-1 DO	AssignExportSizes (exp.dsc[i], m);
	END;

  PROCEDURE FixupCall (codeNow, codeImg, link, fixval: LONGINT);
	VAR
    instr, nextlink, jmp: LONGINT;
	BEGIN
		REPEAT
			_SYSTEM.GET(codeNow+link, instr);
			nextlink := MSW(instr);
			_SYSTEM.GET (codeNow+link-1, jmp);
			IF LSW (jmp) MOD $100 = $0E8 THEN BEGIN
				_SYSTEM.PUT (codeNow+link, fixval-(codeImg+link+4))
			END ELSE BEGIN
				AddAbsReloc(codeImg+link, fixval);
				_SYSTEM.PUT (codeNow+link, fixval)
			END;
			link := nextlink;
		UNTIL link = $0FFFF;
	END;
	
	PROCEDURE FixupVar (codeNow, codeImg, link, fixval: LONGINT);
	VAR
    nofFixups: INTEGER;
    i, val: LONGINT;
	BEGIN
		nofFixups := dataLinks[link].nofFixups;
		i := 0;
		WHILE i < nofFixups DO BEGIN
			_SYSTEM.GET (codeNow+dataLinks[link].offset[i], val);
			AddAbsReloc(codeImg+dataLinks[link].offset[i], val+fixval);
			_SYSTEM.PUT (codeNow+dataLinks[link].offset[i], val+fixval);
			INC (i);
		END;
	END;

  PROCEDURE CheckUseBlock (M: Module);
	VAR
    Vmod: Module;
    name: ModuleName;

		PROCEDURE Err;
		BEGIN
			IF res = 0 THEN BEGIN
        WRITELN(M.name,' imports incompatible version of ',Vmod.name,': ');
        res := incompImport
			END ELSE
        WRITE(#9);
		END;
		
		PROCEDURE CheckScope (scope: ExportDesc; level: INTEGER);
		VAR
      fp, link, i, tdadr: LONGINT;
      tadr: ExportPtr;
      tmpErr: BOOLEAN;
		BEGIN
      tmpErr := (level = EUerrScope);
			i := 0;
			ReadNum (Files, fp);
			WHILE fp <> EUEnd DO BEGIN
				IF fp = EURecord THEN BEGIN
          ReadNum(Files, link);
					IF tmpErr THEN
            CheckScope (scope.dsc[i], EUerrScope)
					ELSE BEGIN
						IF scope.dsc[i].dsc <> NIL THEN BEGIN
							IF link <> 0 THEN BEGIN
                tadr := ExportPtr(_SYSTEM.ADR(scope.dsc[i].dsc[0]));
								_SYSTEM.GET(_SYSTEM.ADR(Vmod.data[0])+Vmod.dataSize+tadr.adr, tdadr);
								AddAbsReloc(M.sb-link, tdadr);
								_SYSTEM.PUT(_SYSTEM.ADR(M.data[0])+M.dataSize-link, tdadr)	(* tdadr at tadr[0] *)
							END
						END;
						CheckScope (scope.dsc[i], EUrecScope)
					END;
				END ELSE BEGIN
					ReadStringX(Files, name);
					IF level >= EUobjScope THEN BEGIN
            tmpErr := FALSE;
						IF level = EUobjScope THEN ReadNum (Files, link);
						i := 0;
						WHILE (i < scope.nofExp) AND (scope.dsc[i].fp <> fp) DO INC (i);
						IF i >= scope.nofExp THEN BEGIN
              Err;
              tmpErr := TRUE;
							DEC (i);
						END ELSE IF (level = EUobjScope) AND (link <> 0) THEN BEGIN
							IF (link AND EUProcFlag) = 0 THEN
								FixupVar(_SYSTEM.ADR(M.code[0]),M.codeBase,link,Vmod.sb+scope.dsc[i].adr)
							ELSE
								FixupCall(_SYSTEM.ADR(M.code[0]),M.codeBase,link-INTEGER(EUProcFlag),scope.dsc[i].adr+Vmod.codeBase);
						END;
					END;
				END;
				ReadNum (Files, fp);
			END;
		END;
		
	BEGIN
    ReadStringX(Files, name);
		WHILE (name<>'') AND (res = 0) DO BEGIN
			Vmod := FindModule (name);
			IF Vmod = NIL THEN BEGIN
        WRITELN('HALT(99)');
        Load (Vmod, name, imageSize);
      END;
			IF res = 0 THEN CheckScope (Vmod.export, EUobjScope);
			ReadStringX(Files, name);
		END
	END;

  PROCEDURE ReadUse (VAR m: Module);
	BEGIN
		Expect(#$08A);
		CheckUseBlock(m);
	END;

  PROCEDURE ReadType (VAR m: Module);
  VAR
    dummy: SMALLINT;
    i, j: LONGINT;
	BEGIN
		IF m.nofTds>0 THEN SETLENGTH(m.tdescs, m.nofTds);
		Expect(#$08B);
    IF res <> done THEN EXIT;
		i := 0;
		WHILE i < m.nofTds DO BEGIN
			NEW(m.tdescs[i]);
			m.tdescs[i].initialized := FALSE;
      m.tdescs[i].module := m;

      m.tdescs[i].size:=0;
			m.tdescs[i].tdEntry:=0;
			m.tdescs[i].baseMod:=0;
			m.tdescs[i].baseEntry:=0;
			m.tdescs[i].nofMethods:=0;
			m.tdescs[i].nofNewMethods:=0;
			m.tdescs[i].nofPtrs:=0;
      dummy:=0;

			Files.Read(m.tdescs[i].size,4);
      Files.Read(dummy,2); m.tdescs[i].tdEntry := dummy;
      Files.Read(dummy,2); m.tdescs[i].baseMod := dummy;
			Files.Read(m.tdescs[i].baseEntry,4);
      Files.Read(dummy,2); m.tdescs[i].nofMethods := dummy;
			Files.Read(dummy,2);
      Files.Read(dummy,2); m.tdescs[i].nofNewMethods := dummy;
      Files.Read(dummy,2); m.tdescs[i].nofPtrs := dummy;
			ReadString(Files,m.tdescs[i].name);

			IF m.tdescs[i].nofNewMethods > 0 THEN
        SETLENGTH(m.tdescs[i].newMethods, m.tdescs[i].nofNewMethods);
			j := 0;
			WHILE j < m.tdescs[i].nofNewMethods DO BEGIN
				Files.Read(m.tdescs[i].newMethods[j].mthNo,2);
        Files.Read(m.tdescs[i].newMethods[j].entryNo,2);
				INC (j);
			END;
			IF m.tdescs[i].nofPtrs > 0 THEN
        SETLENGTH(m.tdescs[i].ptrOffset, m.tdescs[i].nofPtrs);
			j := 0;
			WHILE j < m.tdescs[i].nofPtrs DO BEGIN
        Files.Read(m.tdescs[i].ptrOffset[j],4);
        INC (j);
      END;
			
			(* td size *)
			ASSERT ((m.typeTableAdr+m.typeTableSize+4) MOD Boundary = 0);	(* type must aligned *)
			m.tdescs[i].tdSize := 13 (* fields tdsize..mdesc *) + m.tdescs[i].nofMethods + ExtTabWordSize + 1 (* tag *);
			INC (m.tdescs[i].tdSize, _SYSTEM._MOD((-m.tdescs[i].tdSize+2),4));	(* ADR(td.word[root]) MOD 16 = 8 *)
			m.tdescs[i].tdAdr := m.tdescs[i].tdSize*4 + m.typeTableAdr+m.typeTableSize+4;	(* <-- skip the td *)
			ASSERT (m.tdescs[i].tdAdr MOD 16 = 8);	(* type alignment *)
			m.tdescs[i].tdSize := (m.tdescs[i].tdSize + 1 (*recsize*) + m.tdescs[i].nofPtrs + 1 (*sentinel*)) * 4;
			
			(* write the td address in the constants section *)
			AddAbsReloc(m.sb+m.tdescs[i].tdEntry, m.tdescs[i].tdAdr);
			_SYSTEM.PUT (_SYSTEM.ADR (m.data[0])+m.dataSize+m.tdescs[i].tdEntry, m.tdescs[i].tdAdr);

			m.tdescs[i].padSize := _SYSTEM._MOD((-m.tdescs[i].tdSize-4(*tag*)),Boundary);
			INC(m.typeTableSize, 4(*tag*) + m.tdescs[i].tdSize + m.tdescs[i].padSize);

			ASSERT (m.typeTableSize MOD 32 = 0);
			IF (SysFix[modDescSF].module = m.name)
      AND (SysFix[modDescSF].command = m.tdescs[i].name) THEN
				SysFix[modDescSF].adr := m.tdescs[i].tdAdr;

			IF (SysFix[expDescSF].module = m.name)
      AND (SysFix[expDescSF].command = m.tdescs[i].name) THEN
				SysFix[expDescSF].adr := m.tdescs[i].tdAdr;
			INC(i);
		END;
		IF res <> done THEN EXIT;
	END;

  PROCEDURE ReadRef (VAR m: Module);
	BEGIN
		IF m.refSize > 0 THEN SETLENGTH(m.refs, m.refSize);
		Expect(#$08C);
    Files.Read(m.refs[0],m.refSize);
	END;

  PROCEDURE FixEntry (VAR m: Module);
	BEGIN
		e := m.codeBase;
		i := 0;
    WHILE i < m.nofEntries DO BEGIN
			INC (m.entries[i], e);
			INC (i);
		END;
	END;

  PROCEDURE FixCmd (VAR m: Module);
	BEGIN
		e := m.codeBase;
		i := 0;
    WHILE i < m.nofCmds DO BEGIN
			INC (m.cmds[i].adr, e);
			INC (i)
		END;
	END;

  PROCEDURE FixPtr (VAR m: Module);
	BEGIN
		e := m.sb;
    i := 0;
    WHILE i < m.nofPtrs DO BEGIN
			INC (m.ptrTab[i], e);
			INC (i)
		END;
	END;

  PROCEDURE FixDataLinks (VAR m: Module);
  VAR
    modNo, nofFixups: INTEGER;
    codebase, i, val,a: LONGINT;
	BEGIN
		codebase := _SYSTEM.ADR(m.code[0]);
		modNo := ORD(dataLinks[0].cmod);
    nofFixups := dataLinks[0].nofFixups;
		IF modNo = 0 THEN BEGIN
			i := 0;
			WHILE i < nofFixups DO BEGIN
        a:=codebase+dataLinks[0].offset[i];
				_SYSTEM.GET(a, val);
				AddAbsReloc(m.codeBase+dataLinks[0].offset[i], val+m.sb);
        a:=codebase+dataLinks[0].offset[i];
				_SYSTEM.PUT(a, val+m.sb);
				INC (i);
			END
		END;
	END;

  PROCEDURE Fixup (m: Module; VAR link: LinkTab);
	VAR
    i, val, offs: LONGINT;
		codebase, database: LONGINT;
	BEGIN
		codebase := _SYSTEM.ADR(m.code[0]);
		database := _SYSTEM.ADR(m.data[0]) + m.dataSize;
    i := 0;
		WHILE i < nofLinks DO BEGIN
			IF ORD(link[i].cmod) = 0 THEN BEGIN

				CASE ORD(link[i].entry) OF
						255:
                  BEGIN
                    offs := link[i].link;
                    WHILE offs <> $0FFFF DO BEGIN
                      _SYSTEM.GET(database+offs, val);
                      AddAbsReloc(m.sb+offs, m.codeBase+LSW (val));
                      _SYSTEM.PUT (database+offs, m.codeBase+LSW (val));
                      offs := MSW(val);
                    END;
                  END;
					  254:
                  BEGIN
                    offs := link[i].link;
                    WHILE offs <> $0FFFF DO BEGIN
                      _SYSTEM.GET (codebase+offs, val);
                      AddAbsReloc(m.codeBase+offs, m.entries[LSW (val)]);
                      _SYSTEM.PUT (codebase+offs, m.entries[LSW (val)]);
                      offs := MSW (val);
                    END;
                  END;
					  253:
                  BEGIN
                    ASSERT (SysFix[newSF].adr <> 0);
                    offs := link[i].link;
                    WHILE offs <> $0FFFF DO BEGIN
                      _SYSTEM.GET (codebase+offs, val);
                      _SYSTEM.PUT (codebase+offs, SysFix[newSF].adr-(m.codeBase+offs+4));
                      offs := MSW(val);
                    END;
                  END;
					  252:
                  BEGIN
                    ASSERT (SysFix[sysnewSF].adr <> 0);
                    offs := link[i].link;
                    WHILE offs <> $0FFFF DO BEGIN
                      _SYSTEM.GET (codebase+offs, val);
                      _SYSTEM.PUT (codebase+offs, SysFix[sysnewSF].adr-((*codebase*)m.codeBase+offs+4));
                      offs := MSW (val);
                    END;
                  END;
					  251:
                  BEGIN
                    ASSERT (SysFix[newarrSF].adr <> 0);
                    offs := link[i].link;
                    WHILE offs <> $0FFFF DO BEGIN
                      _SYSTEM.GET (codebase+offs, val);
                      _SYSTEM.PUT (codebase+offs, SysFix[newarrSF].adr-((*codebase*)m.codeBase+offs+4));
                      offs := MSW (val);
                    END;
                  END;
					  250:
                  BEGIN
                    ASSERT (SysFix[StartSF].adr <> 0);
                    offs := link[i].link;
                    WHILE offs <> $0FFFF DO BEGIN
                      _SYSTEM.GET (codebase+offs, val);
                      _SYSTEM.PUT (codebase+offs, SysFix[StartSF].adr - (m.codeBase + offs + 4));
                      offs := MSW(val);
                    END;
                  END;
					  249:
                  BEGIN
                    ASSERT (SysFix[PassivateSF].adr <> 0);
                    offs := link[i].link;
                    WHILE offs <> $0FFFF DO BEGIN
                      _SYSTEM.GET (codebase+offs, val);
                      _SYSTEM.PUT (codebase+offs, SysFix[PassivateSF].adr - (m.codeBase + offs + 4));
                      offs := MSW(val)
                    END;
                  END;
					  248:
                  BEGIN
                    ASSERT (SysFix[ActivateSF].adr <> 0);
                    offs := link[i].link;
                    WHILE offs <> $0FFFF DO BEGIN
                      _SYSTEM.GET (codebase+offs, val);
                      _SYSTEM.PUT (codebase+offs, SysFix[ActivateSF].adr - (m.codeBase + offs + 4));
                      offs := MSW(val)
                    END;
                  END;
					  247:
                  BEGIN
                    ASSERT (SysFix[LockSF].adr <> 0);
                    offs := link[i].link;
                    WHILE offs <> $0FFFF DO BEGIN
                      _SYSTEM.GET (codebase+offs, val);
                      _SYSTEM.PUT (codebase+offs, SysFix[LockSF].adr - (m.codeBase + offs + 4));
                      offs := MSW(val)
                    END;
                  END;
					  246:
                  BEGIN
                    ASSERT (SysFix[UnlockSF].adr <> 0);
                    offs := link[i].link;
                    WHILE offs <> $0FFFF DO BEGIN
                      _SYSTEM.GET (codebase+offs, val);
                      _SYSTEM.PUT (codebase+offs, SysFix[UnlockSF].adr - (m.codeBase + offs + 4));
                      offs := MSW(val)
                    END;
                  END;

(*
					  244:
                  BEGIN
                    ASSERT (SysFix[copyarraySF].adr # 0);
                    offs := link[i].link;
                    WHILE offs # 0FFFFH DO
                      S.GET (codebase+offs, val);
                      S.PUT (codebase+offs, SysFix[copyarraySF].adr - (m.codeBase + offs + 4));
                      offs := MSW(val)
                    END;
                  END;
					  243:
                  BEGIN
                    ASSERT (SysFix[newsysarrSF].adr # 0);
                    offs := link[i].link;
                    WHILE offs # 0FFFFH DO
                      S.GET (codebase+offs, val);
                      S.PUT (codebase+offs, SysFix[newsysarrSF].adr - (m.codeBase + offs + 4));
                      offs := MSW(val)
                    END;
                  END;
*)
					ELSE
						  WRITELN('Unknown fixup ',ORD(link[i].entry));
						  WRITELN('HALT (98)');
				  END;
			END ELSE
				WRITELN('HALT (98)');
			INC (i);
		END;
	END;

  PROCEDURE SetTDescAdr (M: Module; entry: LONGINT; t: TType);
  VAR
    i: INTEGER;
	BEGIN
		i := 0;
		WHILE (i < M.export.nofExp) AND ((M.export.dsc[i].dsc = NIL)
    OR ((M.export.dsc[i].dsc <> NIL) AND (M.export.dsc[i].dsc[0].adr <> entry))) DO
			INC (i);
		IF i < M.export.nofExp THEN M.export.dsc[i].dsc[0].types := t;
	END;
	
	PROCEDURE FindTDescAdr (M: Module; fp: LONGINT; VAR t: TType);
	VAR
    i: INTEGER;
	BEGIN
		i := 0;
		WHILE (i < M.export.nofExp) AND (M.export.dsc[i].fp <> fp) DO INC (i);
		IF i < M.export.nofExp THEN
      t := M.export.dsc[i].dsc[0].types
		ELSE
      WRITELN('HALT (99)');
	END;

  PROCEDURE InitType (i: LONGINT);
	VAR
    j, baseModNo: LONGINT;
    baseMod: Module;
	BEGIN
		ASSERT (m.tdescs <> NIL);
		ASSERT (i < m.nofTds);
		IF NOT m.tdescs[i].initialized THEN BEGIN
			SetTDescAdr (m, m.tdescs[i].tdEntry, m.tdescs[i]);
			baseModNo := m.tdescs[i].baseMod;
      m.tdescs[i].extlev := 0;
			IF baseModNo <>-1 THEN BEGIN	(* extended type *)
				IF baseModNo = 0 THEN	BEGIN(* base type in this module, initialize first! *)
					j := 0;
					WHILE (j < m.nofTds) AND (m.tdescs[j].tdEntry <> m.tdescs[i].baseEntry) DO INC (j);
					InitType(j);
					m.tdescs[i].baseType := m.tdescs[j];
				END ELSE IF baseModNo > 0 THEN BEGIN
					baseMod := m.imports[baseModNo-1];
					FindTDescAdr(baseMod, m.tdescs[i].baseEntry, m.tdescs[i].baseType);
					ASSERT(m.tdescs[i].baseType <> NIL);
				END ELSE
					WRITELN('HALT (99)');
				ASSERT (m.tdescs[i].baseType <> NIL);
				m.tdescs[i].extlev := m.tdescs[i].baseType.extlev + 1;
			END;
			m.tdescs[i].initialized := TRUE;
		END;
	END;

  PROCEDURE InitTypes;
	VAR
    i: LONGINT;
	BEGIN
    i := 0;
		WHILE i < m.nofTds DO BEGIN
      InitType(i);
      INC(i);
    END;
	END;

BEGIN
  m.init := FALSE;
  m.refcnt := 0;
  m.term := NIL;
	ReadNum(Files,symSize);
  Files.Position:=symSize+Files.Position;

  ReadHeader(m);
  ReadImportList(m);
  ReadEntry(m);
	ReadCmd(m);
	ReadPtr(m);
	ReadImport(m);		(* <- this changes the base !!! *)
	ReadDataLink(m);
	ReadLink(m);
	ReadDataConst(m);
	ReadExport(m);
	ReadCode(m);

  m.base := base;
  m.imageSize := 28; (* hidden td *)
	Insert(m);
	INC (m.imageSize, 16 * _SYSTEM._DIV((m.nofEntries*4 + ArrayDescSize +4 + 15),16) );
	INC (m.imageSize, 16 * _SYSTEM._DIV((m.nofCmds*36 + TagSize+ArrayDescSize + 15),16));
	INC (m.imageSize, 16 * _SYSTEM._DIV((m.nofPtrs*4 + TagSize+ArrayDescSize + 15),16));
	INC (m.imageSize, 16 * _SYSTEM._DIV((m.nofImps*4 + TagSize+ArrayDescSize + 15),16));
	m.sb := m.base + m.imageSize + ArrayDescSize + m.dataSize;
	INC (m.imageSize, _SYSTEM._DIV((TagSize+ArrayDescSize + m.dataSize + m.conSize + 15),16) * 16);
	m.codeBase := m.base + m.imageSize + ArrayDescSize;
	INC (m.imageSize, _SYSTEM._DIV((TagSize+ArrayDescSize + m.codeSize + 15),16) * 16);
	INC (m.imageSize, 16 * _SYSTEM._DIV((m.nofTds*4 + TagSize+ArrayDescSize + 15),16));
	m.refBase := m.base + m.imageSize + ArrayDescSize;
	INC (m.imageSize, _SYSTEM._DIV((m.refSize + TagSize+ArrayDescSize + 15),16) * 16);
	INC (m.imageSize, _SYSTEM._DIV((DefMaxImport*4 + TagSize+ArrayDescSize + 15),16) * 16);
	INC (m.imageSize, _SYSTEM._DIV((DefMaxStruct*4 + TagSize+ArrayDescSize + 15),16) * 16);
	INC (m.imageSize, DefMaxReimp*4 + TagSize+ArrayDescSize);
  m.expAdr := Align (m.base + m.imageSize + 4, Boundary) - 4;

  ASSERT (m.expAdr MOD 32 = 28);
	m.expPadding := m.expAdr - m.base - m.imageSize;
  m.expSize := 0;
	INC (m.imageSize, m.expPadding);
	AssignExportSizes(m.export, m);
	INC (m.imageSize, m.expSize);

	ReadUse(m);

  m.typeTableAdr := m.base + m.imageSize;
	m.typeTableSize := 0;
	ASSERT ((m.imageSize+m.base+4) MOD Boundary = 0);

	ReadType(m);

	INC (m.imageSize, m.typeTableSize);
	ASSERT ((m.imageSize+m.base+4) MOD Boundary = 0);

  ReadRef (m);

	ASSERT ((m.imageSize+m.base+4) MOD Boundary = 0);
	m.modDescAdr := m.imageSize + m.base + 4;
	INC (m.imageSize, moduleDescSize);

  IF res = done THEN BEGIN

		FOR i := newSF TO copyarraySF DO BEGIN
			IF SysFix[i].module = m.name THEN
        SysFix[i].adr := m.codeBase + FindAdr(m, SysFix[i].command, ProcType);
		END;

		IF SysFix[CurProcSF].module = m.name THEN BEGIN
			SysFix[CurProcSF].adr := m.sb + FindAdr (m, SysFix[CurProcSF].command, VarType);
		END;

		IF SysFix[listSF].module = m.name THEN BEGIN
			t := FindAdr (m, SysFix[listSF].command, VarType);
			SysFix[listSF].adr := _SYSTEM.ADR(m.data[m.dataSize + t]);
			AddAbsReloc(m.sb+t, objectList.modDescAdr);
			_SYSTEM.PUT(_SYSTEM.ADR(m.data[m.dataSize + t]), objectList.modDescAdr);
		END;

		FixEntry(m);
		FixCmd(m);
		FixPtr(m);
		FixDataLinks(m);
		Fixup(m,VlinkTab);

		AddAbsReloc(m.sb,m.modDescAdr);
		_SYSTEM.PUT (_SYSTEM.ADR(m.data[m.dataSize]),m.modDescAdr);

		InitTypes;
		m.init:=TRUE;
		AddInitPoint(m.codeBase,m);
		base:=base+m.imageSize;
	END ELSE
		WRITELN('HALT (99)');

END;

FUNCTION Load(VAR m: Module; name: STRING; VAR base: LONGINT): BOOLEAN;
VAR
  fname: STRING;
  tag: CHAR;
  Files: TMemoryStream;

BEGIN
  RESULT:=FALSE;
	m := FindModule(name);
	IF m = NIL THEN BEGIN
		fname:=name;
    fname:=fname+OPM.extension;

    Files:= TMemoryStream.Create;
		IF NOT FileExists(fname) THEN BEGIN
      DebugPanel.AddMsg('FileNotFound '+fname,4,NIL);
      FreeAndNil(Files);
			EXIT;
		END;
    Files.LoadFromFile(fname);
		Files.Position:=0;
    Files.Read (tag,SIZEOF(CHAR));
		IF (tag <> #$0BB) THEN BEGIN
      DebugPanel.AddMsg(fname+' is not an object file',4,NIL);
      FreeAndNil(Files);
			EXIT;
		END;

		NEW(m);
    m.export:= ExportDesc.Create;
    SETLENGTH(m.import, DefMaxImport);
    SETLENGTH(m.struct, DefMaxStruct);
		SETLENGTH(m.reimp, DefMaxReimp);
    m.nofImps :=-1;

		Files.Read (tag,SIZEOF(CHAR));
		IF tag <> OFVersion THEN BEGIN
      DebugPanel.AddMsg(fname+': wrong object file version',4,NIL);
      FreeAndNil(Files);
			EXIT;
		END;
		LoadModule (Files, m, base);

    FreeAndNil(Files);

	END ELSE BEGIN
		IF NOT m.init THEN BEGIN
      DebugPanel.AddMsg('Cyclic import not allowed',4,NIL);
			EXIT;
		END;
	END;
  RESULT:=TRUE;
END;  

PROCEDURE InitSysFix(i: LONGINT;  name, module, command: STRING);
BEGIN
	SysFix[i].name:=name;
	SysFix[i].module:=module;
	SysFix[i].command:=command;
	SysFix[i].adr := 0
END;

PROCEDURE Initialise;
VAR
  i: LONGINT;
BEGIN
	objectList := NIL;
  includeRefs := TRUE;
  protObj := NIL;
  res := done;
	FOR i := 0 TO MaxSF-1 DO BEGIN
    SysFix[i].module := '';
    SysFix[i].command := '';
    SysFix[i].adr := 0
  END;
	InitSysFix(newSF,       'new',      'Kernel', 'NewRec');
	InitSysFix(sysnewSF,    'sysnew',   'Kernel', 'NewSys');
	InitSysFix(newarrSF,    'newarr',   'Kernel', 'NewArr');
	InitSysFix(StartSF,     'start',    '',       '');
	InitSysFix(PassivateSF, 'passivate','',       '');
	InitSysFix(ActivateSF,  'activate', '',       '');
	InitSysFix(LockSF,      'lock',     '',       '');
	InitSysFix(UnlockSF,    'unlock',   '',       '');
	InitSysFix(CurProcSF,   '',         '',       '');
	InitSysFix(commandSF,   'command',  '',       '');
	InitSysFix(listSF,      'list',     'Kernel', 'modules');
	InitSysFix(modDescSF,   'mdesc',    'Kernel', 'ModuleDesc');
	InitSysFix(expDescSF,   'expdesc',  'Kernel', 'ExportDesc');
	InitSysFix(objectSF,    '',         '',       '');

	nofEntryPoints := 0;
	moduleDescSize := SIZEOF(DumpModuleDesc)+4; //!!!+++
	mDescPadSize := Align (moduleDescSize, Boundary)-moduleDescSize;
	INC(moduleDescSize, mDescPadSize);

	initPointList := NIL;
	FOR i := 0 TO PaddingSize-1 DO padding[i] := 0;
END;

PROCEDURE BeginLink(imgBase: LONGINT; VAR base: LONGINT);
BEGIN
	Initialise;
  imageSize := 0;
  imageBase := imgBase;
	base := imageBase+28+32	(* for Oberon GC aligment & header *)
END;

PROCEDURE BuildImage(VAR Files: TMemoryStream; entryPoint, imageBase: LONGINT; VAR initOffs: DWORD); FORWARD;

PROCEDURE EndLink(VAR Files: TMemoryStream; base: LONGINT; VAR initOffs: DWORD);
VAR
  Vobject: Module;
  i: LONGINT;
BEGIN
	IF SysFix[commandSF].module <> '' THEN BEGIN
		Vobject := FindModule(SysFix[commandSF].module);
		IF Vobject <> NIL THEN BEGIN
			i := 0;
			WHILE (i < Vobject.nofCmds) AND (Vobject.cmds[i].name <> SysFix[commandSF].command) DO INC (i);
			IF i < Vobject.nofCmds THEN
        AddInitPoint (Vobject.cmds[i].adr, NIL)
			ELSE BEGIN
        WRITELN('Procedure in ''command'' option not found');
        ABORT;
      END;
		END ELSE BEGIN
      WRITELN('Module in ''command'' option not included in image');
      ABORT;
    END;
	END;
	BuildImage(Files, base, imageBase, initOffs);
END;

PROCEDURE Close();
BEGIN
	objectList := NIL;
  initPointList := NIL	(* so that memory can be reclaimed by GC *)
END;

//#####################################################################

PROCEDURE PutDWord (VAR code: ARRAY OF CHAR; idx, value: LONGINT);
BEGIN
  code[idx]:= CHR(_SYSTEM._MOD(value,$100));                      INC(idx);
  code[idx]:= CHR(_SYSTEM._MOD(_SYSTEM._DIV(value,$100),$100));   INC(idx);
  code[idx]:= CHR(_SYSTEM._MOD(_SYSTEM._DIV(value,$10000),$100)); INC(idx);
  code[idx]:= CHR(_SYSTEM._DIV(value,$1000000));
END;

PROCEDURE InsertPad (VAR Files: TMemoryStream; VAR pos: LONGINT; size, alignTo: LONGINT);
VAR
  padSize: LONGINT;
BEGIN
	padSize := _SYSTEM._MOD(-size,alignTo);
	Files.Write(padding[0], padSize);
	INC(pos, size+padSize);
END;

PROCEDURE DumpPtrHeader (VAR Files: TMemoryStream; address, size: LONGINT; VAR adrPad, sizePad: LONGINT);
VAR
  header: ARRAY [0..7] OF LONGINT;	(* 28 bytes = 7 words *)
BEGIN
	adrPad := Align (address+4, Boundary)-4-address;
	sizePad := Align (size+28, Boundary)-size;
	header[0] := (address+adrPad)+4;	(* tag field *)
	header[1] := (size+sizePad)-4;	(* size *)
	header[2] :=-4;	(* sentinel *)
	header[3] := 0;
	header[4] := 0;
	header[5] := 0;
	header[6] := header[0];	(* tag *)
	IF adrPad > 0 THEN Files.Write(padding[0], adrPad);
	AddReloc(Files.Position + 0*4, header[0]);
	AddReloc(Files.Position + 6*4, header[6]);
	Files.Write(header, 28);
	DEC(sizePad, 28)	(* header already dumped - was included *)
END;

PROCEDURE DumpExport (VAR Files: TMemoryStream; VAR exp: ExportDesc; VAR next: LONGINT; tag: LONGINT);
VAR
  i, num: LONGINT;
  arrHdr: ArrayDesc;
  W     : WORD;
  B     : LONGINT;
BEGIN
	ASSERT (next MOD 32 = 0);
	num := exp.nofExp; (*arrHdr.len := num;*)
  exp.done := FALSE;
	IF (num = 0) OR NOT exp.dsc[0].done THEN EXIT;	(* don't traverse it *)
	
	(* write the array *)
	ASSERT (exp.dsc[0].cAdr = next);
	AddReloc(Files.Position, tag);
	Files.Write(tag,SIZEOF(LONGINT));	(* <- next already updated *)
	//IF _SYSTEM.TraceMore THEN LogStr (' --------          ExpArray at'); LogHex (exp.dsc[0].Adr); LogLn END;
	//ASSERT (Files.Length (Files.Base (rider)) + imageBase = exp.dsc[0].Adr);
	ASSERT (num > 0);
	arrHdr.a := exp.dsc[0].cAdr + (num-1)*16 + ArrayDescSize;		(* <- last element to mark *)
	arrHdr.b := 0; 														(* <- reserved *)
	arrHdr.c := exp.dsc[0].cAdr + ArrayDescSize; 							(* <- First Element *)
	arrHdr.len := num;																	(* <- dim *)
	(*arrHdr.lb := 0;
	arrHdr.empty := 0;*)

	AddReloc(Files.Position + 0, arrHdr.a);  (* .a *)
	AddReloc(Files.Position + 8, arrHdr.c); (* .c *)
	Files.Write(arrHdr, ArrayDescSize);
	ASSERT ((exp.dsc[0].cAdr + ArrayDescSize) MOD 8 = 0);

  W:=$00;
  B:=$00;
	FOR i := 0 TO num-1 DO BEGIN
		Files.Write(exp.dsc[i].fp,SIZEOF(LONGINT));
    Files.Write(exp.dsc[i].adr,SIZEOF(LONGINT)); //?????
		Files.Write(exp.dsc[i].nofExp,SIZEOF(WORD));
    Files.Write(W,SIZEOF(WORD));	(* <- padding *)
		//IF _SYSTEM.TraceMore & (exp.dsc[i].nofExp # 0) THEN	LogStr (' --------           -------- '); LogHex (i); LogStr (': '); LogHex (exp.dsc[i].dsc[0].Adr); LogLn; END;
		IF exp.dsc[i].nofExp = 0 THEN
      Files.Write(B,SIZEOF(LONGINT))
		ELSE BEGIN
			AddReloc(Files.Position, exp.dsc[i].dsc[0].cAdr);
			Files.Write(exp.dsc[i].dsc[0].cAdr,SIZEOF(LONGINT));
		END;
	END;
	InsertPad (Files, next, TagSize+ArrayDescSize+num*16, 32);
  
	FOR i := 0 TO num-1 DO
		DumpExport (Files, exp.dsc[i], next, tag);
END;

PROCEDURE DumpTypes (VAR Files: TMemoryStream; m: Module; VAR next: LONGINT);
VAR
  t: tType;
  i, j, tag,N: LONGINT;
	
		FUNCTION GetMethodAddress (n: LONGINT; t: tType): LONGINT;
    VAR
      i: LONGINT;
		BEGIN
			//IF t = NIL THEN LogClose;
			i := 0;
			WHILE (i < t.nofNewMethods) AND (t.newMethods[i].mthNo <> n) DO INC (i);
			IF i = t.nofNewMethods THEN BEGIN
        RESULT:= GetMethodAddress (n, t.baseType);
        EXIT;
			END ELSE BEGIN
        RESULT:= t.module.entries[t.newMethods[i].entryNo];
        EXIT;
			END
		END;
		
		PROCEDURE WriteTags (t: tType);
		BEGIN
			AddReloc(Files.Position, t.tdAdr);
			Files.Write(t.tdAdr,SIZEOF(LONGINT));
			IF t.extlev <> 0 THEN WriteTags (t.baseType);
		END;
		
BEGIN
	FOR i := 0 TO m.nofTds-1 DO BEGIN
		ASSERT (next MOD Boundary = 28);	(* <- already points to the next one *)
		t := m.tdescs[i];
    tag := next + 4;
		AddReloc(Files.Position, tag);
		Files.Write(tag,SIZEOF(LONGINT));	(* <-- tag of the TDesc *)
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = next + 4);
		Files.Write(t.tdSize,SIZEOF(LONGINT));
    N:=-4;
		Files.Write(N,SIZEOF(LONGINT));
		AddReloc(Files.Position, t.tdAdr);
		Files.Write(t.tdAdr,SIZEOF(LONGINT));
		Files.Write(t.extlev,SIZEOF(LONGINT));
    t.name:=t.name+StringOfChar(#0,32-LENGTH(t.name));
		Files.Write(t.name[1], 32);
		AddReloc(Files.Position, m.modDescAdr);
		Files.Write(m.modDescAdr (*+ 4*),SIZEOF(LONGINT));
		INC (next, 13*4 + 4 (*tag*));	(* tdsize .. mdesc *)
		INC (next, t.nofMethods*4 + 16*4(*tags*));
		
		(* padding for methods *)
		IF (-next-12) MOD 16 <> (-next+4) MOD 16 THEN HALT (99);
		j := _SYSTEM._MOD((-next - 12),16);
    INC (next, j);			(* next should point to the tag *)
		ASSERT (next MOD 16 = 4);
		
		j := j DIV 4;
    N:=0;
		WHILE j > 0 DO BEGIN
      Files.Write(N,SIZEOF(LONGINT));
      DEC (j);
    END;
		
		(* write methods *)
		j := t.nofMethods;
		WHILE j > 0 DO BEGIN
			DEC (j);
      N:=GetMethodAddress(j, t);
			Files.Write(N,SIZEOF(LONGINT));
		END;
		
		(* padding for tags *)
    N:=0;
		FOR j := t.extlev+1 TO 15 DO Files.Write(N,SIZEOF(LONGINT));
		
		(* tags *)
		WriteTags (t);
		
		ASSERT (next MOD 16 = 4);
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = next);
		AddReloc(Files.Position, tag);
		Files.Write(tag,SIZEOF(LONGINT));
		INC (next, 4);
		ASSERT (next MOD 16 = 8);
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = next);
		ASSERT (next = t.tdAdr);
		
		Files.Write(t.size,SIZEOF(LONGINT));
		j := 0;
		WHILE j < t.nofPtrs DO BEGIN
      Files.Write(t.ptrOffset[j],SIZEOF(LONGINT));
      INC (j)
    END;
    N:=-4 * (j+1);
		Files.Write(N,SIZEOF(LONGINT));
		Files.Write(padding[0], t.padSize);
		INC (next, 4 (*size*) + t.nofPtrs*4 + 4(*sentinel*) + t.padSize);
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = next);
	END
END;

PROCEDURE DumpInitCalls (VAR Files: TMemoryStream; VAR entry: LONGINT; VAR initOffs: DWORD);
VAR
	i: InitPoint;
	code: ARRAY [0..5] OF CHAR;
	relCallValue, adrPad, sizePad, base: LONGINT;
BEGIN
	ASSERT ((entry+4) MOD Boundary = 0);
	DumpPtrHeader (Files, entry, nofEntryPoints*5+4, adrPad, sizePad);
	INC (entry, 28);	(* skip header *)
	ASSERT (adrPad = 0);
	initOffs := Files.Position;
	i := initPointList;
  base := entry;
	code[0] := #$0E8;	(* CALL *)
	WHILE (i<>NIL) DO BEGIN
		relCallValue := (i.entryPoint-(base+5));
		PutDWord (code, 1, relCallValue);
		Files.Write(code[0], 5);	(* CALL <initcode> *)
		INC (base, 5);
		i := i.next
	END;
	code[0] := #$0C3; (* RET *)
	code[1] := #$090; (* NOP *)
	code[2] := #$090; (* NOP *)
	code[3] := #$090; (* NOP *)
	Files.Write(code[0], 4);
	IF sizePad > 0 THEN Files.Write(padding[0], sizePad);
END;

PROCEDURE DumpModules (VAR Files: TMemoryStream);
VAR
  img: DumpModule;
  size, next, num, tag: LONGINT;
	m: Module;
  i, j: LONGINT;
	arrHdr: ArrayDesc;
	prevmod: LONGINT;
  B: LONGINT;
  TT: TSET;
BEGIN
	i := 0;
  j := 0;
	arrHdr.a := 0;
  arrHdr.b := 0;
  arrHdr.c := 0; (*arrHdr.lb:=0; arrHdr.empty:=0;*)
	prevmod:=0;
	IF SysFix[listSF].adr <> 0 THEN BEGIN
		m := objectList;
		WHILE m.link<>NIL DO m:=m.link;
		_SYSTEM.PUT(SysFix[listSF].adr,m.modDescAdr);
	END;

	m := objectList;

	WHILE m <> NIL DO BEGIN
		NEW (img);
		img.link := prevmod;
    prevmod := m.modDescAdr;
    MOVE(m.name[1],img.name[0],LENGTH(m.name));
		img.init := m.init;
		img.refcnt := m.refcnt;
		IF img.refcnt = 0 THEN img.refcnt := 1;	(* disallow unloading of bootlinked modules, e.g. drivers *)
		img.sb := m.sb;
		img.term := NIL;

		next := m.base;
    tag := next+4;

		ASSERT ((next+4) MOD Boundary = 0);
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = m.base);
		size := m.imageSize-moduleDescSize-m.typeTableSize-m.expSize-28;

		DumpPtrHeader (Files, next, size, i, j);
		INC (next, 28);	(* sys block td size *)
		ASSERT ((i = 0) AND (j = 0));

		(* Entries *)
		//ASSERT (Files.Length (Files.Base (rider)) + imageBase = next);
		ASSERT (next MOD 16 = 8);
		img.entries := next;
    num := m.nofEntries;
    arrHdr.len := num;
		Files.Write(arrHdr, ArrayDescSize);	(* hidden header *)
		IF num > 0 THEN BEGIN
			i := 0;
      j := Files.Position;
			WHILE i < num DO BEGIN
				AddReloc(j, m.entries[i]);
        INC(i);
        INC(j, 4)
			END;
			Files.Write(m.entries[0], num*4);
		END;
		InsertPad (Files, next, TagSize+ArrayDescSize+num*4, 16);	(* <-- 4 bytes left out for the tag *)

		(* Cmds *)
		AddReloc(Files.Position, tag);
		Files.Write(tag,SIZEOF(LONGINT));
		ASSERT (next MOD 16 = 8);
		img.cmds := next;
    num := m.nofCmds;
    arrHdr.len := num;
		Files.Write(arrHdr, ArrayDescSize);
		IF num > 0 THEN BEGIN
			i := 0;
      j := Files.Position+32;
			WHILE i < num DO BEGIN
				AddReloc(j, m.cmds[i].adr);
        INC(i);
        INC(j, 36)
			END;
      i := 0;
			WHILE i < num DO BEGIN
        m.cmds[i].name:=m.cmds[i].name+StringOfChar(#0,32-LENGTH(m.cmds[i].name));
        Files.Write(m.cmds[i].name[1],32);
        Files.Write(m.cmds[i].adr,4);
        INC(i);
			END;
			//Files.Write(m.cmds, num*36);
		END;
		InsertPad (Files, next, TagSize+ArrayDescSize+num*36, 16);

		(* Ptrs *)
		AddReloc(Files.Position, tag);
		Files.Write(tag,SIZEOF(LONGINT));	(* <- next already upadted *)
		//ASSERT (Files.Length (Files.Base (rider)) + imageBase = next);
		ASSERT (next MOD 16 = 8);
		img.ptrTab := next;
    num := m.nofPtrs;
    arrHdr.len := num;
		Files.Write(arrHdr, ArrayDescSize);	(* hidden header *)
		IF num > 0 THEN BEGIN
			i := 0;
      j := Files.Position;
			WHILE i < num DO BEGIN
				AddReloc(j, m.ptrTab[i]);
        INC(i);
        INC(j, 4)
			END;
			Files.Write(m.ptrTab[0], num*4);
		END;
		InsertPad (Files, next, TagSize+ArrayDescSize+num*4, 16);

		(* Imports *)
		AddReloc(Files.Position, tag);
		Files.Write(tag,SIZEOF(LONGINT));	(* <- next already updated *)
		//ASSERT (Files.Length (Files.Base (rider)) + imageBase = next);
		ASSERT (next MOD 16 = 8);
		img.imports := next;
    num := m.nofImps;
    arrHdr.len := num;
		Files.Write(arrHdr, ArrayDescSize);	(* hidden header *)
		i := 0;
    WHILE i < m.nofImps DO BEGIN
			AddReloc(Files.Position, tag);
			Files.Write(m.imports[i].modDescAdr,SIZEOF(LONGINT));
      INC (i);
		END;
		InsertPad (Files, next, TagSize+ArrayDescSize+num*4, 16);

		(* Data *)(* Consts *)
		AddReloc(Files.Position, tag);
		Files.Write(tag,SIZEOF(LONGINT));	(* <- next already updated *)
		//ASSERT (Files.Length (Files.Base (rider)) + imageBase = next);
		ASSERT (next MOD 16 = 8);
		img.data := next;

		arrHdr.len := m.dataSize + m.conSize;
		Files.Write(arrHdr, ArrayDescSize);	(* hidden header *)

		ASSERT (m.sb = next + ArrayDescSize + m.dataSize);
		Files.Write(m.data[0], m.dataSize + m.conSize);
		InsertPad (Files, next, TagSize+ArrayDescSize + m.dataSize + m.conSize, 16);
		
		(* Code *)
		AddReloc(Files.Position, tag);
		Files.Write(tag,SIZEOF(LONGINT));	(* <- next already updated *)
		//ASSERT (Files.Length (Files.Base (rider)) + imageBase = next);
		ASSERT (next MOD 16 = 8);
		img.code := next;
    num := m.codeSize;
    arrHdr.len := num;
		ASSERT (m.codeBase = next + ArrayDescSize);
		Files.Write(arrHdr, ArrayDescSize);	(* hidden header *)
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = m.codeBase);
		Files.Write(m.code[0], num);
		InsertPad (Files, next, TagSize+ArrayDescSize+num, 16);
		
		(* TDescs *)
		AddReloc(Files.Position, tag);
		Files.Write(tag,SIZEOF(LONGINT));	(* <- next already updated *)
		//ASSERT (Files.Length (Files.Base (rider)) + imageBase = next);
		ASSERT (next MOD 16 = 8);
		img.tdescs := next;
    num := m.nofTds;
    arrHdr.len := num;
		Files.Write(arrHdr, ArrayDescSize);	(* hidden header *)
		i := 0;
		WHILE i < num DO BEGIN
			AddReloc(Files.Position, tag);
			Files.Write(m.tdescs[i].tdAdr,SIZEOF(LONGINT));
      INC (i)
		END;
		InsertPad (Files, next, TagSize+ArrayDescSize+num*4, 16);

		(* Refs *)
		AddReloc(Files.Position, tag);
		Files.Write(tag,SIZEOF(LONGINT));	(* <- next already updated *)
		//ASSERT (Files.Length (Files.Base (rider)) + imageBase = next);
		ASSERT (next MOD 16 = 8);
		img.refs := next;
    num := m.refSize;
    arrHdr.len := num;
		ASSERT (m.refBase = next + ArrayDescSize);
		Files.Write(arrHdr, ArrayDescSize);	(* hidden header *)
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = m.refBase);
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = next + ArrayDescSize);
		IF num > 0 THEN Files.Write(m.refs[0], num);
		InsertPad (Files, next, TagSize+ArrayDescSize+num, 16);
		
		(* import *)
		AddReloc(Files.Position, tag);
		Files.Write(tag,SIZEOF(LONGINT));
		ASSERT (next MOD 16 = 8);
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = next);
		img.import := next;
    num := DefMaxImport;
    arrHdr.len := num;
    img.nofimp := m.nofimp;
		Files.Write(arrHdr, ArrayDescSize);	(* hidden header *)
    B:=0;
		FOR i := 0 TO num-1 DO Files.Write(B,SIZEOF(LONGINT));
		InsertPad (Files, next, TagSize+ArrayDescSize+num*4, 16);

		(* struct *)
		AddReloc(Files.Position, tag);
		Files.Write(tag,SIZEOF(LONGINT));
		ASSERT (next MOD 16 = 8);
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = next);
		img.struct := next;
    num := DefMaxStruct;
    arrHdr.len := num;
    img.nofstrc := m.nofstrc;
		Files.Write(arrHdr, ArrayDescSize);	(* hidden header *)
		FOR i := 0 TO num-1 DO Files.Write(B,SIZEOF(LONGINT));
		InsertPad (Files, next, TagSize+ArrayDescSize+num*4, 16);

		(* reimp *)
		AddReloc(Files.Position, tag);
		Files.Write(tag,SIZEOF(LONGINT));
		ASSERT (next MOD 16 = 8);
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = next);
		img.reimp := next;
    num := DefMaxReimp;
    arrHdr.len := num;
    img.nofreimp := m.nofreimp;
		Files.Write(arrHdr, ArrayDescSize);	(* hidden header *)
		FOR i := 0 TO num-1 DO Files.Write(B,SIZEOF(LONGINT));
		(*InsertPad (rider, next, 20+num*4, 16);*)
		INC (next, TagSize+ArrayDescSize + num*4);

		(* export *)
		(* align to word boundary - 4 *)
		Files.Write(padding[0], m.expPadding+4);		(* <- +4 compensate the tag *)
		INC (next, m.expPadding+4);
		
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = next - 4);
		ASSERT (next MOD 32 = 0);
		IF _SYSTEM.CheckGC THEN ASSERT(SysFix[expDescSF].adr <> 0);
    TT:=TSET(SysFix[expDescSF].adr)+[1];
		DumpExport(Files, m.export, next,INTEGER(TT)); (* a bit *)
		img.export.fp := m.export.fp; img.export.adr := m.export.adr;
		img.export.nofExp := m.export.nofExp;
		IF img.export.nofExp = 0 THEN
      img.export.dsc := 0
    ELSE
      img.export.dsc := m.export.dsc[0].cAdr;

		(* align to word boundary - 4 *)
		(* tdescs *)
		DEC (next, 4);
		ASSERT (next MOD Boundary = 28);
		ASSERT (m.typeTableAdr = next);
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = next );
		DumpTypes (Files, m, next);

		(* Module Descr *)
		ASSERT (next MOD Boundary = 28);
		ASSERT (next = m.modDescAdr - 4);
		IF _SYSTEM.CheckGC THEN ASSERT(SysFix[modDescSF].adr <> 0);
		AddReloc(Files.Position, SysFix[modDescSF].adr);
		Files.Write(SysFix[modDescSF].adr,SIZEOF(LONGINT));
		//ASSERT (Files.Length(Files.Base(rider)) + imageBase = m.modDescAdr (*+4*));

		i := Files.Position;

    j := _SYSTEM.ADR(img^);
		AddReloc(i+_SYSTEM.ADR(img.link)-j, img.link);
		AddReloc(i+_SYSTEM.ADR(img.sb)-j, img.sb);
		AddReloc(i+_SYSTEM.ADR(img.entries)-j, img.entries);
		AddReloc(i+_SYSTEM.ADR(img.cmds)-j, img.cmds);
		AddReloc(i+_SYSTEM.ADR(img.ptrTab)-j, img.ptrTab);
		AddReloc(i+_SYSTEM.ADR(img.tdescs)-j, img.tdescs);
		AddReloc(i+_SYSTEM.ADR(img.imports)-j, img.imports);
		AddReloc(i+_SYSTEM.ADR(img.data)-j, img.data);
		AddReloc(i+_SYSTEM.ADR(img.code)-j, img.code);
		AddReloc(i+_SYSTEM.ADR(img.refs)-j, img.refs);
		AddReloc(i+_SYSTEM.ADR(img.import)-j, img.import);
		AddReloc(i+_SYSTEM.ADR(img.struct)-j, img.struct);
		AddReloc(i+_SYSTEM.ADR(img.reimp)-j, img.reimp);
		AddReloc(i+_SYSTEM.ADR(img.export.dsc)-j, img.export.dsc);

    Files.Write(img^, moduleDescSize - mDescPadSize - 4);
		Files.Write(padding[0], mDescPadSize);

		INC (next, moduleDescSize);
		ASSERT (m.imageSize +m.base = next);

		m := m.link;
	END;
  //DISPOSE(img);
END;

PROCEDURE PatchHeader (VAR Files: TMemoryStream; base, entry, size: LONGINT);
VAR
  C: CHAR;
  I: LONGINT;
BEGIN
	Files.Position:=0;
  C:=#$0E8;
  Files.Write (C,1);
  I:=entry-(base+5);
  Files.Write(I,SIZEOF(LONGINT));
	AddReloc(6, base);
  AddReloc(22, base+size);
	Files.Position:=6;
  Files.Write(base,SIZEOF(LONGINT));	(* LinkBase *)
	Files.Position:=22;
  I:=base+size;
  Files.Write(I,SIZEOF(LONGINT));	(* HeapStart *)
	Files.Position:=30;
  I:=0;
  Files.Write(I,SIZEOF(LONGINT));	(* PatchSize *)
END;

PROCEDURE BuildImage(VAR Files: TMemoryStream; entryPoint, imageBase: LONGINT; VAR initOffs: DWORD);
VAR
	OutputFile: TMemoryStream;
	B: DWORD;
BEGIN
	IF _SYSTEM.CheckGC THEN ASSERT (SysFix[modDescSF].adr <> 0);
	IF _SYSTEM.CheckGC THEN ASSERT (SysFix[listSF].adr <> 0);

  OutputFile:= TMemoryStream.Create;
	OutputFile.Position:=0;
	OutputFile.Write(padding[0], 28+32);	(* for Oberon GC aligment & header *)

  DumpModules (OutputFile);
  DumpInitCalls (OutputFile, entryPoint, initOffs);	(* entryPoint VAR *)
  PatchHeader (OutputFile, imageBase, entryPoint, OutputFile.Size);
	OutputFile.Position:=0;

  Files.Write(OutputFile.Memory^,OutputFile.Size);
  B:=$00; Files.Write(B,4);

  OutputFile.Free;
END;

END.
