UNIT LNKi386;

INTERFACE

USES
  Windows, Classes, SysUtils, OPM, LNTi386, LNBi386, LNRi386, _SYSTEM, DebugFrm,
  DIALOGS;

CONST
  SECTION_TEXT      = $00000020;{Contains code}
  SECTION_DATA      = $00000040;{Contains initialized data}
  SECTION_BSS       = $00000080;{Contains uninitialized data}
  SECTION_DISCARD   = $02000000;{Section Can be discarded}
  SECTION_EXEC      = $20000000;{Section is Executable}
  SECTION_READ      = $40000000;{Section is Readable}
  SECTION_WRITE     = $80000000;{Section is Writeable}
  SECTION_RSRC      = $50000040;

  MajorLinkVersion  = $00;
  MinorLinkVersion  = $5A;

  DOSCode: array[0..$D] of Char = (#$0E,#$1F,#$BA,#$0E,#$00,#$B4,#$09,
                                   #$CD,#$21,#$B8,#$01,#$4C,#$CD,#$21);
  MsgDOS: String    = 'This program cannot be run in DOS mode.'#$0D#$0D#$0A'$';
  Magic: DWORD      = IMAGE_NT_SIGNATURE;

  MaxRelocs         = 16*1024;

  EXEImageBase      = $0400000;
  DLLImageBase      = $010000000;

  ImageFileRelocsStripped     = $00001;
  ImageFileExecutableImage    = $00002;
  ImageFileLineNumsStripped   = $00004;
  ImageFileLocalSymsStripped  = $00008;
  ImageFile32BitMachine       = $00100;
  ImageFileDll                = $02000;

  DefaultLanguage             = $0400;

TYPE
  APP_TYPE = (CUI,GUI,NATIVE);

  TSection          = CLASS
  PRIVATE
    Head            : _IMAGE_SECTION_HEADER;
    Body            : TMemoryStream;
  PUBLIC
    CONSTRUCTOR Create;
    DESTRUCTOR  Destroy; OVERRIDE;
  END;

  TLinker = CLASS
    PROTECTED
      Done            : BOOLEAN;
      Files           : TMemoryStream;
      ExeHeader       : _IMAGE_DOS_HEADER;
      PeDirectory     : _IMAGE_DATA_DIRECTORY;
      PeHeader        : _IMAGE_FILE_HEADER;
      PeSection       : _IMAGE_SECTION_HEADER;
      ExportDirectory : _IMAGE_EXPORT_DIRECTORY;

      SectionIDATA    : TSection;
      SectionEDATA    : TSection;
      SectionRELOC    : TSection;
      SectionRSRC     : TSection;

      Sections        : DWORD;
      PROCEDURE       FileAlign(VAR Files: TMemoryStream; Valign: LONGINT; pad: CHAR);
      PROCEDURE       InitSection(VAR Section: TSection; Name: STRING; scnCnt, scnMem: DWORD);
      FUNCTION        GetObjAdr(objname: STRING; Vtype: LONGINT): LONGINT;
      PROCEDURE       AddFileReloc(val: LONGINT);
      PROCEDURE       WriteText();
      PROCEDURE       WriteIData();
      PROCEDURE       WriteEData();
      PROCEDURE       WriteReloc();
      PROCEDURE       WriteRSRC();
    PUBLIC
      Res             : TResource;
      Image           : STRING;
      Stub            : STRING;
      CONSTRUCTOR     Create;
      DESTRUCTOR      Destroy; OVERRIDE;
      FUNCTION        DoLink(aType: APP_TYPE; isDLL: BOOLEAN; LinkName, Name: STRING): INTEGER;
    END;

VAR
  isEXE               : BOOLEAN;
  OptHeader           : _IMAGE_OPTIONAL_HEADER;
  SectionTEXT         : TSection;
  nRelocs             : DWORD;
  relocs              : ARRAY [0..MaxRelocs] OF LONGINT;

  FUNCTION Align(Avalue, Aalign: LONGINT): DWORD;
  
IMPLEMENTATION

USES
  LNSi386;
  
PROCEDURE AddReloc(offs, val: DWORD);
BEGIN
  IF NOT isEXE AND (val <> 0) THEN BEGIN
    ASSERT(val >= OptHeader.ImageBase);
    IF nRelocs < MaxRelocs THEN BEGIN
      relocs[nRelocs] := SectionTEXT.Head.VirtualAddress+offs;
      INC(nRelocs);
    END ELSE BEGIN
      WRITELN('Too many relocations');
      ABORT;
    END;
  END;
END;

FUNCTION Align(Avalue, Aalign: LONGINT): DWORD;
BEGIN
  RESULT := Avalue + _SYSTEM._MOD((Aalign-_SYSTEM._MOD(Avalue,Aalign)),Aalign);
END;

FUNCTION TimeDateStamp(): DWORD;
BEGIN
  RESULT:=Round((NOW - 25569.0) * 86400);
END;

{###########################################################################}

CONSTRUCTOR TSection.Create;
BEGIN
  INHERITED;
  Body:=TMemoryStream.Create;
END;

DESTRUCTOR TSection.Destroy;
BEGIN
  FreeAndNil(Body);
  INHERITED;
END;

{###########################################################################}

CONSTRUCTOR TLinker.Create;
VAR
  i: INTEGER;
BEGIN
  INHERITED;

  Res   := TResource.Create;

  Files := TMemoryStream.Create;
  isEXE := TRUE;
  
  FillChar (ExeHeader, SizeOf(ExeHeader), 0);
  WITH ExeHeader DO BEGIN
      e_magic                   := $5A4D;
      e_cblp                    := $0090;
      e_cp                      := $0003;
      e_crlc                    := $0000;
      e_cparhdr                 := $0004;
      e_minalloc                := $0000;
      e_maxalloc                := $FFFF;
      e_ss                      := $0000;
      e_sp                      := $00B8;
      e_csum                    := $0000;
      e_ip                      := $0000;
      e_cs                      := $0000;
      e_lfarlc                  := $0040;
      _lfanew                   := $80;
  END;

  FillChar (PEheader, SizeOf(PEheader), 0);
  WITH PEheader DO BEGIN
    Machine                     := IMAGE_FILE_MACHINE_I386;
    NumberOfSections            := 0;
    TimeDateStamp               := Round((NOW - 25569.0) * 86400); //DateTimeToUnix
    PointerToSymbolTable        := $00;
    NumberOfSymbols             := $00;
    SizeOfOptionalHeader        := SizeOf(OptHeader);
    Characteristics             := $818E {or IMAGE_FILE_RELOCS_STRIPPED};
  END;

  FillChar(OptHeader, Sizeof(_IMAGE_OPTIONAL_HEADER), 0);
  WITH OptHeader DO BEGIN
    Magic                       := $010B;
    MajorLinkerVersion          := MajorLinkVersion;
    MinorLinkerVersion          := MinorLinkVersion;
    SizeOfCode                  := 0;{ Set Later }
    SizeOfInitializedData       := 0;
    SizeOfUninitializedData     := 0;
    AddressOfEntryPoint         := 0;
    BaseOfCode                  := 0;
    BaseOfData                  := 0;
    ImageBase                   := 0;
    SectionAlignment            := $1000;
    FileAlignment               := $200;
    MajorOperatingSystemVersion := 4;
    MinorOperatingSystemVersion := 0;
    MajorImageVersion           := 0;
    MajorImageVersion           := 0;
    MajorSubsystemVersion       := 4;
    Win32VersionValue           := 0;
    SizeOfImage                 := (1+PEheader.NumberOfSections)*OptHeader.SectionAlignment;
    SizeOfHeaders               := $200; {//FilePos(F) + SizeOf(ImageOptionalHeader)}
    CheckSum                    := 0;
    Subsystem                   := IMAGE_SUBSYSTEM_WINDOWS_CUI;
    DllCharacteristics          := 0;
    SizeOfStackReserve          := $100000;
    SizeOfStackCommit           := $1000;
    SizeOfHeapReserve           := $10000;
    SizeOfHeapCommit            := $1000;
    LoaderFlags                 := 0;
    NumberOfRvaAndSizes         := IMAGE_NUMBEROF_DIRECTORY_ENTRIES;

    FOR i := 0 TO IMAGE_NUMBEROF_DIRECTORY_ENTRIES-1 DO BEGIN
			DataDirectory[i].VirtualAddress := 0;
			DataDirectory[i].Size := 0
		END;

  END;

  InitSection(SectionTEXT , '.text' , SECTION_TEXT, SECTION_READ OR SECTION_WRITE OR SECTION_EXEC);
  InitSection(SectionIDATA, '.idata', SECTION_DATA, SECTION_READ OR SECTION_WRITE);
  InitSection(SectionRELOC, '.reloc', SECTION_DATA, SECTION_READ OR SECTION_DISCARD);
  InitSection(SectionEDATA, '.edata', SECTION_DATA, SECTION_READ);
  InitSection(SectionRSRC , '.rsrc' , SECTION_DATA, SECTION_READ);
END;

DESTRUCTOR TLinker.Destroy;
BEGIN
  SectionTEXT.Free;
  SectionIDATA.Free;
  SectionEDATA.Free;
  SectionRELOC.Free;
  SectionRSRC.Free;
  Res.Free;
  Files.Free;
  INHERITED;
END;

PROCEDURE TLinker.FileAlign(VAR Files: TMemoryStream; Valign: LONGINT; pad: CHAR);
VAR
  n: DWORD;
BEGIN
  n := Files.Position;
  n := Align(n, Valign)-n;
  WHILE n > 0 DO BEGIN
    Files.Write(pad,1);
    DEC(n);
  END
END;

PROCEDURE TLinker.InitSection(VAR Section: TSection; Name: STRING; scnCnt, scnMem: DWORD);
VAR
  i: LONGINT;
BEGIN
  Section:=TSection.Create;
  FOR i := 0 TO IMAGE_SIZEOF_SHORT_NAME-1 DO Section.Head.Name[i] := 0;
  MOVE(Name[1],Section.Head.Name[0],LENGTH(Name));
  WITH Section.Head DO BEGIN
    Misc.VirtualSize     := 0;
    VirtualAddress       := 0;
    SizeOfRawData        := 0;
    PointerToRawData     := 0;
    PointerToRelocations := 0;
    PointerToLinenumbers := 0;
    NumberOfRelocations  := 0;
    NumberOfLinenumbers  := 0;
    Characteristics      := scnCnt OR scnMem;
  END;
END;

FUNCTION TLinker.GetObjAdr(objname: STRING; Vtype: LONGINT): LONGINT;
VAR
  name: LNBi386.Name;
  Vmod: LNBi386.Module;
  adr, i: LONGINT;
BEGIN
  i := 1;
  name:='';

  WHILE (objname <> '') AND (objname[i] <> '.') DO BEGIN
    name := name+objname[i];
    INC(i);
  END;

  IF objname[i] = '.' THEN BEGIN
    Vmod := LNBi386.FindModule(name);
    IF Vmod=NIL THEN BEGIN
      RESULT:=0;
      SHOWMESSAGE(name+': module not found');
      EXIT;
    END;
    INC(i);
    name:='';
    WHILE objname[i] <> '' DO BEGIN
      name := name+objname[i];
      INC(i);
    END;
    adr := LNBi386.FindAdr(Vmod, name, Vtype);
    IF adr=0 THEN BEGIN
      RESULT:=0;
      WRITELN(name+': not found');
      EXIT;
    END;
    IF Vtype = LNBi386.VarType THEN BEGIN
      RESULT:= Vmod.sb+adr;
      EXIT;
    END ELSE IF Vtype = LNBi386.ProcType THEN BEGIN
      RESULT:= Vmod.codeBase+adr;
      EXIT;
    END ELSE
      WRITELN('invalid object type', '');
  END;
  WRITELN(objname, ': not a variable');
  RESULT:=0;
END;

PROCEDURE TLinker.AddFileReloc(val: LONGINT);
VAR
  offs: LONGINT;
BEGIN
  IF NOT isEXE AND (val <> 0) THEN BEGIN
    offs := Files.Position-SectionTEXT.Head.PointerToRawData;
    AddReloc(offs, val);
  END;
END;

PROCEDURE TLinker.WriteText();
VAR
  adr, pos, heapSize,
  unload, exit: DWORD;
  Imp   : ImportList;
  Ent   : EntryList;
  fBuff : LONGINT;

  base  : LONGINT;
  offs  : DWORD;

  List  : LNTi386.NameList;
  Mdl   : LNBi386.Module;

  PROCEDURE FilesWrite(Ch: ARRAY OF CHAR);
  BEGIN
    Files.Write(Ch[0],LENGTH(Ch));
  END;

BEGIN
  INC(Sections);
  unload:=0;
  exit:=0;

  SectionTEXT.Head.VirtualAddress := OptHeader.SectionAlignment;
  SectionTEXT.Head.PointerToRawData := Files.Position;
  base := OptHeader.ImageBase+OptHeader.SectionAlignment;
  
  LNBi386.AddReloc := AddReloc;
  LNBi386.BeginLink(base, base);
  List := LNTi386.modules;
  WHILE List <> NIL DO BEGIN
    IF NOT LNBi386.Load(Mdl, List.name, base) THEN BEGIN
      SELF.Done:=FALSE;
      BREAK;
    END;
    List := List.next;
  END;
  LNBi386.EndLink(Files, base, offs);
  
  heapSize := Files.Position-SectionTEXT.Head.PointerToRawData;

  OptHeader.AddressOfEntryPoint := SectionTEXT.Head.VirtualAddress+Files.Position-SectionTEXT.Head.PointerToRawData;

  FilesWrite([#$053,#$055,#$056,#$057]);
  
  IF NOT isEXE THEN BEGIN
    FilesWrite([#$08B,#$05C,#$024,#$018,#$083,#$0FB,#$000,
                #$00F,#$085,#$005,#$000,#$000,#$000,#$0E9]);
    unload := Files.Position;
    FilesWrite([#$000,#$000,#$000,#$000,#$083,#$0FB,#$001,#$00F,#$085]);
    exit := Files.Position;
    FilesWrite([#$000,#$000,#$000,#$000]);
  END;

  Imp := LNTi386.imports.next;
  WHILE Imp <> NIL DO BEGIN
    Ent:=Imp.entries;
    WHILE Ent <> NIL DO BEGIN
      adr := GetObjAdr(Ent.name, LNBi386.VarType);
      IF adr=0 THEN BEGIN
        SELF.Done:=FALSE;
        BREAK;
      END;
      FilesWrite([#$08B,#$01D]);
      Ent.fixup := Files.Position;
      AddFileReloc(OptHeader.ImageBase);
      FilesWrite([#$000,#$000,#$000,#$000,#$089,#$01D]);
      AddFileReloc(adr);
      Files.Write(adr,SIZEOF(LONGINT));
      Ent := Ent.next;
    END;
    IF NOT SELF.Done THEN BREAK;
    Imp := Imp.next;
  END;

  IF SELF.Done THEN BEGIN

    Ent := LNTi386.imports.entries;
    WHILE Ent <> NIL DO BEGIN
      adr := GetObjAdr(Ent.name, LNBi386.VarType);
      IF adr=0 THEN BEGIN
        SELF.Done:=FALSE;
        BREAK;
      END;
      IF Ent.value = 'HEAPSIZE' THEN BEGIN
        FilesWrite([#$0BB]); (* MOX EBX, HeapSize *)
        Files.Write(heapSize,SIZEOF(LONGINT));
      END ELSE IF Ent.value = 'HEAPADR' THEN BEGIN
        FilesWrite([#$0BB]); (* MOX EBX, HeapAdr *)
        AddFileReloc(OptHeader.ImageBase+SectionTEXT.Head.VirtualAddress);
        fBuff:=OptHeader.ImageBase+SectionTEXT.Head.VirtualAddress;
        Files.Write(fBuff,SIZEOF(LONGINT));
      END ELSE IF Ent.value = 'HINSTANCE' THEN BEGIN
        IF isEXE THEN
          FilesWrite([#$0BB,#$000,#$000,#$000,#$000])
        ELSE
          FilesWrite([#$08B,#$05C,#$024,#$014]);
      END ELSE
        DebugPanel.AddMsg(Ent.value+': unknown entry',4,NIL);
      FilesWrite([#$089,#$01D]);
      AddFileReloc(adr);
      Files.Write(adr,SIZEOF(LONGINT));
      Ent := Ent.next;
    END;

    IF Done THEN BEGIN
      pos := Files.Position;
      FilesWrite([#$0E8]); (* CALL *);
      fBuff:=SectionTEXT.Head.PointerToRawData+offs-pos-5;
      Files.Write(fBuff,SIZEOF(LONGINT));
      IF NOT isEXE THEN BEGIN
        FilesWrite([#$0E9,#$005,#$000,#$000,#$000]); (* JMP Exit *)
        pos := Files.Position;
        Files.Position:=unload;
        fBuff:=pos-unload-4;
        Files.Write(fBuff,SIZEOF(LONGINT));
        Files.Position:=SectionTEXT.Head.PointerToRawData+offs+1;
        Files.Read(adr,SIZEOF(LONGINT));
        Files.Position:=pos;
        FilesWrite([#$0E8]); (* CALL first module (Kernel32) *)
        fBuff:=SectionTEXT.Head.PointerToRawData+offs+adr-pos;
        Files.Write(fBuff,SIZEOF(LONGINT));
        pos := Files.Position;
        Files.Position:=exit;
        fBuff:=pos-exit-4;
        Files.Write(fBuff,SIZEOF(LONGINT));
        Files.Position:=pos;
      END;

      FilesWrite([#$05F,#$05E,#$05D,#$05B]);
      IF isEXE THEN
        FilesWrite([#$0B8,#$000,#$000,#$000,#$000,#$0C3])
      ELSE
        FilesWrite([#$0B8,#$0FF,#$0FF,#$0FF,#$0FF,#$0C2,#$00C,#$000]);//!!! -1
      SectionTEXT.Head.Misc.VirtualSize := Files.Position-SectionTEXT.Head.PointerToRawData;
      FileAlign(Files,OptHeader.FileAlignment,#$0CC); (* INT 3 *)
      SectionTEXT.Head.SizeOfRawData := Files.Position-SectionTEXT.Head.PointerToRawData;
    END;
  END;

END;

PROCEDURE TLinker.WriteIData();
VAR
  iPos, Pos, p, i, shift, adr: DWORD;
  fName : STRING;
  B     : BYTE;
  fBuff : LONGINT;
  name  : ImportList;
  Imps  : ImportList;
  Ent   : EntryList;
BEGIN
  INC(Sections);
  imps := imports.next;
  SectionIDATA.Head.PointerToRawData := Files.Position;
  SectionIDATA.Head.VirtualAddress := SectionTEXT.Head.VirtualAddress+Align(SectionTEXT.Head.SizeOfRawData,OptHeader.SectionAlignment);
  OptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress := SectionIDATA.Head.VirtualAddress;
  shift := SectionIDATA.Head.VirtualAddress-SectionIDATA.Head.PointerToRawData;
  iPos:=SectionIDATA.Head.PointerToRawData;

  name := imps;
  WHILE name <> NIL DO BEGIN
    Files.Write(name.impDesc, SIZEOF(ImageImportDescriptor));
    name := name.next;
  END;
  FOR i:=0 TO SIZEOF(ImageImportDescriptor)-1 DO BEGIN
    B:=0; Files.Write(B,1);
  END;

  OptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].Size := Files.Position-SectionIDATA.Head.PointerToRawData;

  name := imps;
  WHILE name <> NIL DO BEGIN
    name.impDesc.Characteristics := Files.Size;
    Files.Position:=name.impDesc.Characteristics;
    Ent := name.entries;
    WHILE Ent <> NIL DO BEGIN
      fBuff:=0;
      Files.Write(fBuff,SIZEOF(LONGINT));
      Ent := Ent.next;
    END;
    fBuff:=0;
    Files.Write(fBuff,SIZEOF(LONGINT));
    name := name.next;
  END;

  pos := Files.Position;
  OptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IAT].VirtualAddress := pos+shift;

  name := imps;
  WHILE name <> NIL DO BEGIN
    name.impDesc.FirstThunk := Files.Size;
    Files.Position:= name.impDesc.FirstThunk;
    Ent := name.entries;
    WHILE Ent <> NIL DO BEGIN
      fBuff:=0;
      Files.Write(fBuff,SIZEOF(LONGINT));
      Ent := Ent.next;
    END;
    fBuff:=0;
    Files.Write(fBuff,SIZEOF(LONGINT));
    name := name.next
  END;

  OptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IAT].Size := Files.Position-pos;
  //Files.Position:=idata.PointerToRawData;

  name := imps;
  WHILE name <> NIL DO BEGIN
    i := 0;
    B:=$00;
    fName:='';

    Ent := name.entries;
    WHILE Ent <> NIL DO BEGIN
      p := Files.Position;
      Files.Write(Ent.entry,SIZEOF(WORD));
      fName:=Ent.value+#0;
      Files.Write(fName[1],LENGTH(fName));
      FileAlign(Files,2,#0);
      p:=p+shift;

      Pos:=Files.Position;
      Files.Position:=name.impDesc.Characteristics+i*4;
      fBuff:=p;
      Files.Write(fBuff,SIZEOF(LONGINT));

      adr := name.impDesc.FirstThunk+i*4;
      Files.Position:=adr;
      fBuff:=p;
      Files.Write(fBuff,SIZEOF(LONGINT));
      //Text Info
      Files.Position:= Ent.fixup;
      fBuff:=OptHeader.ImageBase+adr+shift;
      Files.Write(fBuff,SIZEOF(LONGINT));
      Files.Position:=Pos;
      INC(i);
      Ent := Ent.next;
    END;

    name.impDesc.Name := Files.Size; //!!!
    Files.Position:=name.impDesc.Name;
    fName:=name.name+#0;
    Files.Write(fName[1],LENGTH(fName));
    FileAlign(Files,2,#0);
    Pos:=Files.Position;

    name.impDesc.Characteristics := name.impDesc.Characteristics+shift;
    name.impDesc.FirstThunk := name.impDesc.FirstThunk+shift;
    name.impDesc.Name := name.impDesc.Name+shift;
    Files.Position:= iPos;
    Files.Write(name.impDesc, SIZEOF(ImageImportDescriptor));
    iPos:=Files.Position;
    Files.Position:= Pos;

    name := name.next;
  END;

  SectionIDATA.Head.Misc.VirtualSize := Files.Position-SectionIDATA.Head.PointerToRawData;
  FileAlign(Files,OptHeader.FileAlignment, #0);
  SectionIDATA.Head.SizeOfRawData := Files.Position-SectionIDATA.Head.PointerToRawData;
END;

PROCEDURE TLinker.WriteEData();
TYPE
  ImageExportDirectory = RECORD
    Characteristics: DWORD;
    TimeDateStamp: DWORD;
    MajorVersion: WORD;
    MinorVersion: WORD;
    Name: DWORD;
    Base: DWORD;
    NumberOfFunctions: DWORD;
    NumberOfNames: DWORD;
    AddressOfFunctions: DWORD;
    AddressOfNames: DWORD;
    AddressOfNameOrdinals: DWORD;
  END;

VAR
  head: ImageExportDirectory;
  adr, i, n, pos, fix: DWORD;
  e: ExportEntryList;
  p: EntryList;
  fBuff: LONGINT;
  fName: STRING;
  W: WORD;

  FUNCTION AssignExportOrdinals(): LONGINT;
  VAR
    e, f, g, next, new: ExportEntryList;
    n: DWORD;
  BEGIN
    f := NIL;
    e := exportOrds;
    WHILE (e <> NIL) AND (e.entry <= 0) DO BEGIN
      f := e;
      e := e.nextOrd
    END;
    IF f <> NIL THEN f.nextOrd := NIL;
    new := e;
    e := exportOrds;
    WHILE e <> NIL DO BEGIN
      next := e.nextOrd;
      g := NIL;
      f := new;
      n := 1;
      WHILE (f <> NIL) AND (f.entry = n) DO BEGIN
        g := f;
        f := f.nextOrd;
        INC(n);
      END;
      e.entry := n;
      e.nextOrd := f;
      IF g <> NIL THEN
        g.nextOrd := e
      ELSE
        new := e;
      e := next
    END;
    e := new; n := 1;
    WHILE (e <> NIL) AND (e.entry = n) DO BEGIN
      e := e.nextOrd;
      INC(n);
    END;
    (* Assert(e = NIL, e.name, " exported ordinals must be unique") *)
    IF e <> NIL THEN BEGIN
      WRITELN('Error(e.name, " exported ordinals must be unique"');
    END;
    exportOrds := new;
    RESULT:= n-1
  END;

BEGIN
  INC(Sections);
  n := AssignExportOrdinals();
  SectionEDATA.Head.PointerToRawData := Files.Position;
  SectionEDATA.Head.VirtualAddress := SectionIDATA.Head.VirtualAddress+Align(SectionIDATA.Head.SizeOfRawData, OptHeader.SectionAlignment);
  OptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress := SectionEDATA.Head.VirtualAddress;

  head.Characteristics := 0;
  head.TimeDateStamp := TimeDateStamp();
  head.MajorVersion := 0;
  head.MinorVersion := 0;
  head.Base := 1;
  head.NumberOfFunctions := n;
  head.NumberOfNames := n;
  head.AddressOfFunctions := SectionEDATA.Head.VirtualAddress+SIZEOF(ImageExportDirectory);
  head.AddressOfNames := head.AddressOfFunctions+n*4;
  head.AddressOfNameOrdinals := head.AddressOfNames+n*4;
  head.Name := head.AddressOfNameOrdinals+n*2;
  Files.Write(head, SIZEOF(ImageExportDirectory));

  e := exportOrds;
  WHILE e <> NIL DO BEGIN
    adr := GetObjAdr(e.value, LNBi386.ProcType);
    IF adr=0 THEN BEGIN
      Done:=FALSE;
      EXIT;
    END;
    fBuff:=adr-OptHeader.ImageBase;
    Files.Write(fBuff,SIZEOF(LONGINT));
    e := e.nextOrd
  END;

  fix := Files.Position;
  FOR i := 1 TO n DO BEGIN
    fBuff:=0;
    Files.Write(fBuff,SIZEOF(LONGINT));
  END;

  p := iexports;
  WHILE p <> NIL DO BEGIN
    W:=p.entry-head.Base;
    Files.Write(W,SIZEOF(WORD));
    p := p.next;
  END;

  //fName:=Image+#0;
  Files.Write(fName[1],LENGTH(fName));
  FileAlign(Files, 2, #0);

  p := iexports;
  WHILE p <> NIL DO BEGIN
    pos := Files.Position;
    Files.Position:= fix;
    fBuff:=pos-SectionEDATA.Head.PointerToRawData+SectionEDATA.Head.VirtualAddress;
    Files.Write(fBuff,SIZEOF(LONGINT));
    Files.Position:= pos;
    INC(fix, 4);
    Files.Write(p.name[1],LENGTH(p.name));
    FileAlign(Files, 2, #0);
    p := p.next;
  END;

  SectionEDATA.Head.Misc.VirtualSize := Files.Position-SectionEDATA.Head.PointerToRawData;
  OptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].Size := SectionEDATA.Head.Misc.VirtualSize;
  FileAlign(Files,OptHeader.FileAlignment, #0);
  SectionEDATA.Head.SizeOfRawData := Files.Position-SectionEDATA.Head.PointerToRawData
END;

PROCEDURE TLinker.WriteReloc();
TYPE
  ImageBaseRelocation = RECORD
    VirtualAddress: DWORD;
    SizeOfBlock   : DWORD;
  END;

CONST
  ImageRelBasedAbsolute = 0;
  ImageRelBasedHigh = 1; 
  ImageRelBasedLow = 2; 
  ImageRelBasedHighLow = 3; 
  ImageRelBasedHighAdj = 4; 
  ImageRelBasedMipsJmpAddr = 5; 
  ImageRelBasedSection = 6; 
  ImageRelBasedRel32 = 7;

VAR
  head: ImageBaseRelocation;
  last: INTEGER;
  this: INTEGER;
  i, fix: DWORD;

  PROCEDURE WriteOffs(VAR Files: TMemoryStream; offs, Vtype: LONGINT);
  VAR
    W: WORD;
  BEGIN
    W:=WORD(_SYSTEM.ASH(Vtype, 12)+offs);
    Files.Write(W,2);
  END;

  PROCEDURE BeginHeader();
  BEGIN
    fix := Files.Position;
    head.VirtualAddress := this - (this MOD PageSize);
    head.SizeOfBlock := SIZEOF(ImageBaseRelocation);
    Files.Write(head, SIZEOF(ImageBaseRelocation))
  END;

  PROCEDURE EndHeader();
    VAR pos: LONGINT;
  BEGIN
    WriteOffs(Files, 0, ImageRelBasedAbsolute);
    INC(head.SizeOfBlock, 2);
    IF (head.SizeOfBlock MOD 4) <> 0 THEN BEGIN
      WriteOffs(Files, 0, ImageRelBasedAbsolute);
      INC(head.SizeOfBlock, 2);
    END;
    pos := Files.Position;
    Files.Position:=fix;
    Files.Write(head, SIZEOF(ImageBaseRelocation));
    Files.Position:=pos;
  END;

  PROCEDURE QuickSortRelocs(lo, hi: LONGINT);
  VAR
    i, j, t, rva: LONGINT;
  BEGIN
    IF lo < hi THEN BEGIN
      i := lo;
      j := hi;
      rva := relocs[_SYSTEM._DIV((lo + hi),2)];
      REPEAT
        WHILE relocs[i] < rva DO INC(i);
        WHILE rva < relocs[j] DO DEC(j);
        IF i <= j THEN BEGIN
          t := relocs[i];
          relocs[i] := relocs[j];
          relocs[j] := t;
          INC(i);
          DEC(j)
        END
      UNTIL i > j;
      QuickSortRelocs(lo, j);
      QuickSortRelocs(i, hi)
    END;
  END;

BEGIN
  INC(Sections);
  SectionRELOC.Head.PointerToRawData := Files.Position;

  IF iexports <> NIL THEN
    SectionRELOC.Head.VirtualAddress := SectionEDATA.Head.VirtualAddress+Align(SectionEDATA.Head.SizeOfRawData, OptHeader.SectionAlignment)
  ELSE
    SectionRELOC.Head.VirtualAddress := SectionIDATA.Head.VirtualAddress+Align(SectionIDATA.Head.SizeOfRawData, OptHeader.SectionAlignment);

  OptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].VirtualAddress := SectionRELOC.Head.VirtualAddress;

  QuickSortRelocs(0, nRelocs-1);

  i := 0;
  this := relocs[0];
  last := -1;

  BeginHeader();
  WHILE i < nRelocs DO BEGIN
    IF this <> last THEN BEGIN
      ASSERT(this > last);
      IF DWORD(this) >= (head.VirtualAddress+PageSize) THEN BEGIN
        EndHeader();
        BeginHeader()
      END;
      ASSERT((DWORD(this) >= SectionTEXT.Head.VirtualAddress) AND (DWORD(this) <= (SectionTEXT.Head.VirtualAddress+SectionTEXT.Head.Misc.VirtualSize)));
      WriteOffs(Files, DWORD(this)-head.VirtualAddress, ImageRelBasedHighLow);
      INC(head.SizeOfBlock, 2);
      last := this;
    END;
    INC(i);
    this := relocs[i];
  END;
  EndHeader();

  SectionRELOC.Head.Misc.VirtualSize := Files.Position-SectionRELOC.Head.PointerToRawData;
  OptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size := SectionRELOC.Head.Misc.VirtualSize;
  FileAlign(Files, OptHeader.FileAlignment, #0);
  SectionRELOC.Head.SizeOfRawData := Files.Position-SectionRELOC.Head.PointerToRawData;
END;

PROCEDURE TLinker.WriteRSRC();
TYPE
  NameU = ARRAY OF BYTE;
VAR
  Rsc           : NameList;
  Directory     : _IMAGE_RESOURCE_DIRECTORY;
  DirectoryEntry: _IMAGE_RESOURCE_DIRECTORY_ENTRY;
  i,j,c         : integer;

  SizeTypeDir   : integer;
  SizeNameDir   : integer;
  SizeDataDir   : integer;
  SizeNameData  : integer;
  SizeDataData  : integer;

  OffsetNameDir : DWORD;
  OffsetDataDir : DWORD;
  OffsetNameData: DWORD;
  OffsetDataData: DWORD;
  OffsetData    : DWORD;

  NameDirStream : TMemoryStream;
  DataDirStream : TMemoryStream;
  NameStream    : TMemoryStream;
  DataDataStream: TMemoryStream;
  DataData      : TMemoryStream;

  WideName      : WideString;

  FUNCTION WriteName(Name: WideString; Len: WORD): DWORD;
  BEGIN
    Result:=(OffsetNameData+NameStream.Position) or $80000000;
    NameStream.Write(Len,SizeOf(WORD));
    NameStream.Write(PWideChar(Name)^,Length(Name)*SizeOf(WideChar));
  END;

  FUNCTION WriteData(Item: TResourceItem): DWORD;
  VAR
    Data: _IMAGE_RESOURCE_DATA_ENTRY;
  BEGIN
    Result:=OffsetDataData+DataDataStream.Position;
    FillChar(Data,SizeOf(_IMAGE_RESOURCE_DATA_ENTRY),0);
    WITH Data DO BEGIN
      OffsetToData  := DWORD(SectionRSRC.Head.VirtualAddress)+OffsetData+DataData.Position;
      Size          := Item.Header.DataSize;
      CodePage      := 0;
      Reserved      := 0;
    END;
    DataDataStream.Write(Data,SizeOf(_IMAGE_RESOURCE_DATA_ENTRY));
    DataData.Write(Item.ResData[0],Length(Item.ResData));
  END;

  FUNCTION WriteDataDir(Item: TResourceItem): DWORD;
  VAR
    DirectoryData     : _IMAGE_RESOURCE_DIRECTORY;
    DirectoryDataEntry: _IMAGE_RESOURCE_DIRECTORY_ENTRY;
  BEGIN
    Result:=(OffsetDataDir+DataDirStream.Position) or $80000000;
    FillChar(DirectoryData,SizeOf(_IMAGE_RESOURCE_DIRECTORY),0);
    DirectoryData.TimeDateStamp := Round((NOW - 25569.0) * 86400); //DateTimeToUnix
    DirectoryData.NumberOfIdEntries:=1;
    DataDirStream.Write(DirectoryData,SizeOf(_IMAGE_RESOURCE_DIRECTORY));

    FillChar(DirectoryDataEntry,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY),0);
    DirectoryDataEntry.Name:=Item.Suffix.LanguageID;
    DirectoryDataEntry.OffsetToData:=WriteData(Item);
    DataDirStream.Write(DirectoryDataEntry,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY));
  END;

  FUNCTION WriteNameDir(n: integer): DWORD;
  VAR
    i                 : integer;
    DirectoryName     : _IMAGE_RESOURCE_DIRECTORY;
    DirectoryNameEntry: _IMAGE_RESOURCE_DIRECTORY_ENTRY;
  BEGIN
    Result:=(OffsetNameDir+NameDirStream.Position) or $80000000;
    FillChar(DirectoryName,SizeOf(_IMAGE_RESOURCE_DIRECTORY),0);
    DirectoryName.TimeDateStamp := Round((NOW - 25569.0) * 86400); //DateTimeToUnix
    FOR i:=0 TO (Res.Entries[n].SubEntries.Count-1) DO BEGIN
      IF Res.SubEntries[n,i].ResName.Named
        THEN INC(DirectoryName.NumberOfNamedEntries)
        ELSE INC(DirectoryName.NumberOfIdEntries);
    END;
    NameDirStream.Write(DirectoryName,SizeOf(_IMAGE_RESOURCE_DIRECTORY));

    FillChar(DirectoryNameEntry,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY),0);
    FOR i:=0 TO Res.Entries[n].SubEntries.Count-1 DO BEGIN
      IF Res.SubEntries[n,i].ResName.Named THEN
        DirectoryNameEntry.Name:=WriteName(Res.SubEntries[n,i].ResName.Name,Length(Res.SubEntries[n,i].ResName.Name))
      ELSE
        DirectoryNameEntry.Name:=Res.SubEntries[n,i].ResName.ResId;
      DirectoryNameEntry.OffsetToData:=WriteDataDir(Res.SubEntries[n,i]);
      NameDirStream.Write(DirectoryNameEntry,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY));
    END;
  END;

BEGIN

  INC(Sections);

  Rsc := LNTi386.resources;
  WHILE Rsc <> NIL DO BEGIN
    Res.LoadFromFile(Rsc.Name);
    Rsc := Rsc.next;
  END;

  SectionRSRC.Head.PointerToRawData := Files.Position;

  IF NOT isEXE THEN
    SectionRSRC.Head.VirtualAddress := SectionRELOC.Head.VirtualAddress+Align(SectionRELOC.Head.SizeOfRawData, OptHeader.SectionAlignment)
  ELSE IF iexports <> NIL THEN
    SectionRSRC.Head.VirtualAddress := SectionEDATA.Head.VirtualAddress+Align(SectionEDATA.Head.SizeOfRawData, OptHeader.SectionAlignment)
  ELSE
    SectionRSRC.Head.VirtualAddress := SectionIDATA.Head.VirtualAddress+Align(SectionIDATA.Head.SizeOfRawData, OptHeader.SectionAlignment);

  OptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_RESOURCE].VirtualAddress := SectionRSRC.Head.VirtualAddress;

  WITH SectionRSRC DO BEGIN

    NameDirStream   := TMemoryStream.Create;
    NameDirStream.Position  :=0;
    DataDirStream   := TMemoryStream.Create;
    DataDirStream.Position  :=0;
    DataDataStream  := TMemoryStream.Create;
    DataDataStream.Position :=0;
    NameStream      := TMemoryStream.Create;
    NameStream.Position     :=0;
    DataData        := TMemoryStream.Create;
    DataData.Position       :=0;

    FillChar(Directory,SizeOf(_IMAGE_RESOURCE_DIRECTORY),0);
    Directory.TimeDateStamp := Round((NOW - 25569.0) * 86400); //DateTimeToUnix

    SizeTypeDir:=Sizeof(_IMAGE_RESOURCE_DIRECTORY_ENTRY) * Res.EntryCount+Sizeof(_IMAGE_RESOURCE_DIRECTORY_ENTRY);
    SizeNameDir:=0;
    SizeDataDir:=0;
    SizeDataData:=0;
    SizeNameData:=0;
    FOR i:=0 TO (Res.EntryCount-1) DO BEGIN
      c:=Res.Entries[i].SubEntries.Count;
      INC(SizeNameDir, Sizeof(_IMAGE_RESOURCE_DIRECTORY)+Sizeof(_IMAGE_RESOURCE_DIRECTORY_ENTRY)*c);
      INC(SizeDataDir,(Sizeof(_IMAGE_RESOURCE_DIRECTORY)+SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY))*c);
      IF Res.Entries[i].ResType.Named THEN BEGIN
        INC(Directory.NumberOfNamedEntries);
        WideName:=Res.Entries[i].ResType.Name;
        INC(SizeNameData,SizeOf(WORD)+Length(WideName)*SizeOf(WideChar));
      END ELSE
        INC(Directory.NumberOfIdEntries);

        FOR j:=0 TO (Res.Entries[i].SubEntries.Count-1) DO BEGIN
          IF Res.SubEntries[i,j].ResName.Named THEN BEGIN
            WideName:=Res.SubEntries[i,j].ResName.Name;
            INC(SizeNameData,SizeOf(WORD)+Length(WideName)*SizeOf(WideChar));
          END;
          INC(SizeDataData,SizeOf(_IMAGE_RESOURCE_DATA_ENTRY));
        END;

    END;

    INC(SizeNameDir,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY));
    INC(SizeDataDir,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY));

    Body.Write(Directory,SizeOf(_IMAGE_RESOURCE_DIRECTORY));

    OffsetNameDir := Body.Position+SizeTypeDir;
    OffsetDataDir := Body.Position+SizeTypeDir+SizeNameDir;
    OffsetDataData:= Body.Position+SizeTypeDir+SizeNameDir+SizeDataDir;
    OffsetNameData:= Body.Position+SizeTypeDir+SizeNameDir+SizeDataDir+SizeDataData;
    OffsetData    := Body.Position+SizeTypeDir+SizeNameDir+SizeDataDir+SizeNameData+SizeDataData;

    FOR i:=0 TO (Res.EntryCount-1) DO BEGIN
      FillChar(DirectoryEntry,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY),0);
      IF Res.Entries[i].ResType.Named THEN
        DirectoryEntry.Name:=WriteName(Res.Entries[i].ResType.Name,Length(Res.Entries[i].ResType.Name))
      ELSE
        DirectoryEntry.Name:=Res.Entries[i].ResType.ResId;
      DirectoryEntry.OffsetToData:=WriteNameDir(i);
      Body.Write(DirectoryEntry,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY));
    END;
    FillChar(DirectoryEntry,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY),$00);
    Body.Write(DirectoryEntry,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY));

    NameDirStream.Write(DirectoryEntry,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY));
    DataDirStream.Write(DirectoryEntry,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY));

    Body.Write(NameDirStream.Memory^, NameDirStream.Size);
    Body.Write(DataDirStream.Memory^, DataDirStream.Size);
    Body.Write(DataDataStream.Memory^,DataDataStream.Size);
    Body.Write(NameStream.Memory^,    NameStream.Size);
    Body.Write(DataData.Memory^,      DataData.Size);

    NameDirStream.Free;
    DataDirStream.Free;
    DataDataStream.Free;
    NameStream.Free;
    DataData.Free;
  END;

  Files.Write(SectionRSRC.Body.Memory^, SectionRSRC.Body.Size);

  OptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_RESOURCE].Size:=SectionRSRC.Body.Size;
  SectionRSRC.Head.Misc.VirtualSize := Files.Position-SectionRSRC.Head.PointerToRawData;
  FileAlign(Files,OptHeader.FileAlignment,#0);
  SectionRSRC.Head.SizeOfRawData := Files.Position-SectionRSRC.Head.PointerToRawData;

END;

FUNCTION TLinker.DoLink(aType: APP_TYPE; isDLL: BOOLEAN; LinkName,Name: STRING): INTEGER;
VAR
  Pos   : LONGINT;
  VC    : INTEGER;
  LS    : TLinkScaner;
BEGIN
  SELF.Done := TRUE;

  //SUBSYSTEM GUI\CUI\NATIVE
  IF aType=GUI THEN
    OptHeader.Subsystem := IMAGE_SUBSYSTEM_WINDOWS_GUI
  ELSE IF aType=CUI THEN
    OptHeader.Subsystem := IMAGE_SUBSYSTEM_WINDOWS_CUI
  ELSE IF aType=NATIVE THEN
    OptHeader.Subsystem := IMAGE_SUBSYSTEM_NATIVE;
  //BASE PROGRAM\LIBRARY
  IF isDLL THEN BEGIN
    OptHeader.ImageBase := DLLImageBase;
    Name:=Name+OPM.dllext;
    isEXE := FALSE;
  END ELSE BEGIN
    OptHeader.ImageBase := EXEImageBase;
    Name:=Name+OPM.exeext;
    isEXE := TRUE;
  END;
  //HEAPSIZE
  OptHeader.SizeOfHeapReserve := Align(OptHeader.SizeOfHeapReserve, OptHeader.SectionAlignment);
  IF OptHeader.SizeOfHeapReserve > 0 THEN
    OptHeader.SizeOfHeapCommit := OptHeader.SectionAlignment
  ELSE
    OptHeader.SizeOfHeapCommit := 0;
  //STACKSIZE
  OptHeader.SizeOfStackReserve := Align(OptHeader.SizeOfStackReserve, OptHeader.SectionAlignment);
  IF OptHeader.SizeOfStackReserve > 0 THEN
    OptHeader.SizeOfStackCommit := OptHeader.SectionAlignment
  ELSE
    OptHeader.SizeOfStackCommit := 0;
  //IMGVERSION
  OptHeader.MajorImageVersion := 2;
  OptHeader.MinorImageVersion := 4;

  Image     := Name;
  
  IF (LinkName<>'') AND FileExists(LinkName) THEN BEGIN
    LS    := TLinkScaner.Create(LinkName,SELF);
    LS.Free;
  END;

  Sections := 0; (* .text, .idata, .rsrc *)

  Files.Write(ExeHeader,  SIZEOF(_IMAGE_DOS_HEADER));
  Files.Write(DOSCode,    SIZEOF(DOSCode));
  Files.Write(MsgDOS[1],Length(MsgDOS));
  FileAlign(Files,16, #0);
  Files.Write(Magic,      SIZEOF(Magic));
  Pos:=Files.Position;
  Files.Write(PEHeader,   SIZEOF(_IMAGE_FILE_HEADER));

  OptHeader.SizeOfHeaders := Files.Position;
  Files.Write(OptHeader,  SIZEOF(_IMAGE_OPTIONAL_HEADER));

  //#################################################
  Files.Write(SectionTEXT.Head, SIZEOF(_IMAGE_SECTION_HEADER));
  IF imports.next<>NIL   THEN Files.Write(SectionIDATA.Head,SIZEOF(_IMAGE_SECTION_HEADER));
  IF iexports<>NIL  THEN Files.Write(SectionEDATA.Head, SIZEOF(_IMAGE_SECTION_HEADER));
  IF NOT isEXE THEN      Files.Write(SectionRELOC.Head, SIZEOF(_IMAGE_SECTION_HEADER));
  IF resources<>NIL THEN Files.Write(SectionRSRC.Head , SIZEOF(_IMAGE_SECTION_HEADER));
  FileAlign(Files,OptHeader.FileAlignment, #0);
  //#################################################

  WriteText();
  IF imports.next<>NIL   THEN WriteIData();
  IF iexports<>NIL  THEN WriteEData();
  IF NOT isEXE      THEN WriteReloc();  
  IF resources<>NIL THEN WriteRSRC();

  OptHeader.SizeOfCode := SectionTEXT.Head.SizeOfRawData;
  OptHeader.SizeOfInitializedData   := SectionIDATA.Head.SizeOfRawData+SectionRSRC.Head.SizeOfRawData;
  OptHeader.SizeOfUninitializedData := 0;
  OptHeader.BaseOfCode := SectionTEXT.Head.VirtualAddress;
  OptHeader.BaseOfData := 0;
  OptHeader.SizeOfHeaders := OptHeader.SizeOfHeaders+SIZEOF(OptHeader)+Sections*SIZEOF(_IMAGE_SECTION_HEADER);
  OptHeader.SizeOfHeaders := Align(OptHeader.SizeOfHeaders, OptHeader.FileAlignment);
  OptHeader.SizeOfImage   := Align(OptHeader.SizeOfHeaders, OptHeader.SectionAlignment)+
                             Align(SectionTEXT.Head.SizeOfRawData, OptHeader.SectionAlignment)+
                             Align(SectionIDATA.Head.SizeOfRawData, OptHeader.SectionAlignment)+
                             Align(SectionEDATA.Head.SizeOfRawData, OptHeader.SectionAlignment)+
                             Align(SectionRELOC.Head.SizeOfRawData, OptHeader.SectionAlignment)+
                             Align(SectionRSRC.Head.SizeOfRawData, OptHeader.SectionAlignment);

  Files.Position:=Pos;
  WITH PEHeader DO BEGIN
    NumberOfSections            := Sections;
    TimeDateStamp               := Round((NOW - 25569.0) * 86400);
    SizeOfOptionalHeader        := SizeOf(OptHeader);
    VC := ImageFileExecutableImage OR ImageFile32BitMachine;
    IF isEXE THEN
      VC := VC OR ImageFileRelocsStripped
    ELSE
      VC := VC OR ImageFileDll;
    VC := VC OR ImageFileLineNumsStripped;
    VC := VC OR ImageFileLocalSymsStripped;
    Characteristics := WORD(VC);
  END;
  Files.Write(PEHeader, SIZEOF(_IMAGE_FILE_HEADER));

  Files.Write(OptHeader,         SIZEOF(_IMAGE_OPTIONAL_HEADER));

  //#################################################
  Files.Write(SectionTEXT.Head,  SIZEOF(_IMAGE_SECTION_HEADER));
  IF imports.next<>NIL   THEN Files.Write(SectionIDATA.Head, SIZEOF(_IMAGE_SECTION_HEADER));
  IF iexports<>NIL  THEN Files.Write(SectionEDATA.Head, SIZEOF(_IMAGE_SECTION_HEADER));
  IF NOT isEXE THEN      Files.Write(SectionRELOC.Head, SIZEOF(_IMAGE_SECTION_HEADER));
  IF resources<>NIL THEN Files.Write(SectionRSRC.Head,  SIZEOF(_IMAGE_SECTION_HEADER));
  //#################################################

  Files.SaveToFile(Image);
  RESULT:=Files.Size;
  FreeAndNil(Files);
END;

END.
