unit LNRi386;

interface

uses
  Windows, Classes, SysUtils, Contnrs; 

const
  RT_GIF        = MAKEINTRESOURCE(13);
  RT_JPEG       = MAKEINTRESOURCE(15);
  RT_PNG        = MAKEINTRESOURCE(18);
  RT_HTML       = MAKEINTRESOURCE(23);
  RT_MANIFEST   = MAKEINTRESOURCE(24);
  RT_AVI        = MAKEINTRESOURCE(128);
  RT_DLGINIT    = MAKEINTRESOURCE(240);
  RT_TOOLBAR    = MAKEINTRESOURCE(241);

  CResources: array[1..27] of record
    ID  : PAnsiChar;
    Name: ShortString;
  end = ((ID:RT_CURSOR;       Name:'CURSOR'),
         (ID:RT_BITMAP;       Name:'BITMAP'),
         (ID:RT_ICON;         Name:'ICON'),
         (ID:RT_MENU;         Name:'MENU'),
         (ID:RT_DIALOG;       Name:'DIALOG'),
         (ID:RT_STRING;       Name:'STRING'),
         (ID:RT_FONTDIR;      Name:'FONTDIR'),
         (ID:RT_FONT;         Name:'FONT'),
         (ID:RT_ACCELERATOR;  Name:'ACCELERATOR'),
         (ID:RT_RCDATA;       Name:'RCDATA'),
         (ID:RT_MESSAGETABLE; Name:'MESSAGETABLE'),
         (ID:RT_GROUP_CURSOR; Name:'GROUP_CURSOR'),
         (ID:RT_GIF;          Name:'GIF'),
         (ID:RT_GROUP_ICON;   Name:'GROUP_ICON'),
         (ID:RT_JPEG;         Name:'JPEG'),
         (ID:RT_VERSION;      Name:'VERSION'),
         (ID:RT_DLGINCLUDE;   Name:'DLGINCLUDE'),
         (ID:RT_PNG;          Name:'PNG'),
         (ID:RT_PLUGPLAY;     Name:'PLUGPLAY'),
         (ID:RT_VXD;          Name:'VXD'),
         (ID:RT_ANICURSOR;    Name:'ANICURSOR'),
         (ID:RT_ANIICON;      Name:'ANIICON'),
         (ID:RT_HTML;         Name:'HTML'),
         (ID:RT_MANIFEST;     Name:'MANIFEST'),
         (ID:RT_AVI;          Name:'AVI'),
         (ID:RT_DLGINIT;      Name:'DLGINIT'),
         (ID:RT_TOOLBAR;      Name:'TOOLBAR'));
type
  _IMAGE_RESOURCE_DIRECTORY =  packed record
    Characteristics       : DWORD;
    TimeDateStamp         : DWORD;
    MajorVersion          : WORD;
    MinorVersion          : WORD;
    NumberOfNamedEntries  : WORD;
    NumberOfIdEntries     : WORD;
  end;

  _IMAGE_RESOURCE_DIRECTORY_ENTRY = packed record
      case Integer of
        0:  (
              NameOffset        : DWORD;
              OffsetToData      : DWORD
            );
        1:  (
              Name              : DWORD;
              OffsetToDirectory : DWORD;
            );
        2:  (
              Id                : WORD;
            ); 
  end;

  _IMAGE_RESOURCE_DATA_ENTRY = packed record
    OffsetToData    : Cardinal;
    Size            : Cardinal;
    CodePage        : Cardinal;
    Reserved        : Cardinal;
  end;

  TFileResHeader = record
    DataSize        : DWORD;
    HeaderSize      : DWORD;
    ResType         : DWORD;
    ResId           : DWORD;
    DataVersion     : DWORD;
    MemoryFlags     : WORD;
    LanguageId      : WORD;
    Version         : DWORD;
    Characteristics : DWORD;
  end;

  TResHeader = packed record
    DataSize        : DWORD;
    HeaderSize      : DWORD;
  end;

  TResHeaderSuffix = packed record
    DataVersion     : DWORD;
    MemoryFlags     : WORD;
    LanguageID      : WORD;
    Version         : DWORD;
    Characteristics : DWORD;
  end;

  TResName   = packed record
    Named           : boolean;
    ResId           : DWORD;
    Name            : String;
  end;

  TResourceItem = class
    Header          : TResHeader;
    ResName         : TResName;
    ResType         : TResName;
    ResData         : array of Byte;
    Suffix          : TResHeaderSuffix;
    SubEntries      : TObjectList;
  end;

  TResource          = class
  private
    ExeHeader       : _IMAGE_DOS_HEADER;
    PeHeader        : _IMAGE_FILE_HEADER;
    OptHeader       : _IMAGE_OPTIONAL_HEADER;
    PeSection       : _IMAGE_SECTION_HEADER;
    FileMemory      : TMemoryStream;
    fEntries        : TObjectList;
    ResHeader       : TFileResHeader;
    function      GetEntryCount: Integer;
    function      GetEntry(Idx: Integer): TResourceItem;
    function      GetSubEntry(Idx,SubIdx: Integer): TResourceItem;
    procedure     AddEntry(Item: TResourceItem);
    procedure     LoadFromPeFile;
  public
    constructor   Create;
    destructor    Destroy; override;
    procedure     LoadFromFile(Name: String);
    procedure     SaveData(Name: String; Idx: Integer);
    property      EntryCount: Integer read GetEntryCount;
    property      Entries[Idx: Integer]: TResourceItem read GetEntry;
    property      SubEntries[Idx,SubIdx: Integer]: TResourceItem read GetSubEntry;
  end;

const
  IMAGE_RESOURCE_NAME_IS_STRING    = Cardinal($80000000);
  IMAGE_RESOURCE_DATA_IS_DIRECTORY = Cardinal($80000000);

implementation

function _ReadWideString(MS: TMemoryStream; Len: integer; Flag: boolean): string;
var
  c         : WORD;
  n         : Integer;
  Condition : boolean;
begin
  Result:='';
  repeat
    n:=MS.Read(c,SizeOf(word));
    Result:=Result+Copy(WideCharToString(PWideChar(@c)),1,1);
    if Flag then begin
      Dec(Len);
      Condition:=(Len=0);
    end else
      Condition:=(c=0) or (n=0);
  until Condition;
end;

{###########################################################################}

constructor TResource.Create;
begin
  FileMemory    := TMemoryStream.Create;
  fEntries      := TObjectList.Create(true);
end;

destructor TResource.Destroy;
var
  i,j: integer;
begin
  FileMemory.Free;
  for i:=0 to fEntries.Count-1 do begin
    for j:=0 to TResourceItem(fEntries[i]).SubEntries.Count-1 do
      SetLength(TResourceItem(TResourceItem(fEntries[i]).SubEntries[j]).ResData,0);
    TResourceItem(fEntries[i]).SubEntries.Free;
    SetLength(TResourceItem(fEntries[i]).ResData,0);
  end;
  fEntries.Free;
  inherited;
end;

function TResource.GetEntryCount: Integer;
begin
  Result := fEntries.Count;
end;

function TResource.GetEntry(Idx: Integer): TResourceItem;
begin
  Result := TResourceItem(fEntries[Idx]);
end;

function TResource.GetSubEntry(Idx,SubIdx: Integer): TResourceItem;
begin
  Result := TResourceItem(TResourceItem(fEntries[Idx]).SubEntries[SubIdx]);
end;

procedure TResource.AddEntry(Item: TResourceItem);
var
  i       : integer;
  f       : boolean;
  NewItem : TResourceItem;
begin
  i:=0;
  f:=true;
  while (i<fEntries.Count) and f do begin

    if Item.ResType.Named then begin
      if SameText(TResourceItem(fEntries[i]).ResType.Name,Item.ResType.Name) then f:=false;
    end else begin
      if TResourceItem(fEntries[i]).ResType.ResId=Item.ResType.ResId then f:=false;
    end;

    Inc(i);
  end;
  if f then begin
    NewItem           := TResourceItem.Create;
    NewItem.ResType   := Item.ResType;
    NewItem.SubEntries:= TObjectList.Create(true);
    NewItem.SubEntries.Add(Item);
    fEntries.Add(NewItem);
  end else
    TResourceItem(fEntries[i-1]).SubEntries.Add(Item);
end;

procedure TResource.SaveData(Name: String; Idx: Integer);
var
  OutData: TMemoryStream;
begin
  OutData:= TMemoryStream.Create;
  OutData.Write(TResourceItem(fEntries[Idx]).ResData[0],Length(TResourceItem(fEntries[Idx]).ResData));
  OutData.SaveToFile(Name);
  OutData.Free;
end;

procedure TResource.LoadFromPeFile;
var
  Magic         : DWORD;
  i             : integer;
  Name          : String;
  ImageRD       : _IMAGE_RESOURCE_DIRECTORY;
  ImageRDE      : _IMAGE_RESOURCE_DIRECTORY_ENTRY ;
  SectionOffset : DWORD;
  SectionVA     : DWORD;
  ResItem       : TResourceItem;
  IsName        : boolean;

  function CopyArr(Arr: Array of Byte): String;
  var
    i: integer;
  begin
    i:=0;
    Result:='';
    while Arr[i]<>0 do begin
      Result:=Result+CHR(Arr[i]);
      Inc(i);
    end;
  end;

  function ReadNameOffset(Offset: DWORD): String;
  var
    pos : integer;
    Len : WORD;
  begin
    pos:=FileMemory.Position;
    FileMemory.Position:=SectionOffset+Offset;
    FileMemory.Read(Len,SizeOf(WORD));
    Result:=_ReadWideString(FileMemory,Len,true);
    FileMemory.Position:=pos;
  end;

  procedure LoadData(var Data: array of BYTE; Syze: integer; Address: DWORD);
  var
    pos : Integer;
  begin
    pos:=FileMemory.Position;
    FileMemory.Position:=Address;
    FileMemory.Read(Data[0],Syze);
    FileMemory.Position:=pos;
  end;

  procedure ReadSubDir(AEntry: _IMAGE_RESOURCE_DIRECTORY_ENTRY; AType: Cardinal; ATypename: string; Named: boolean);
  var
    i,pos,depos : Integer;
    SubName     : String;
    SubImageRD  : _IMAGE_RESOURCE_DIRECTORY;
    SubImageRDE : _IMAGE_RESOURCE_DIRECTORY_ENTRY;
    SubImageDE  : _IMAGE_RESOURCE_DATA_ENTRY;
    SubNamed    : boolean;
    SubNameID   : DWORD;
    Lang        : WORD;

  begin
    pos:=FileMemory.Position;

    SubNamed:=false;
    SubNameID:=0;

    if AEntry.Name and IMAGE_RESOURCE_NAME_IS_STRING<>0 then begin
      SubName:=ReadNameOffset(AEntry.Name and $7FFFFFFF);
      SubNamed:=true;
    end else
      SubNameID:=AEntry.Name and $FFFF;

    FileMemory.Position:=SectionOffset+(AEntry.OffsetToDirectory and $7FFFFFFF);
    FileMemory.Read(SubImageRD,SizeOf(_IMAGE_RESOURCE_DIRECTORY));

    for i:=0 to SubImageRD.NumberOfNamedEntries+SubImageRD.NumberOfIdEntries-1 do begin
      FileMemory.Read(SubImageRDE,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY));

      if SubImageRDE.OffsetToData and IMAGE_RESOURCE_DATA_IS_DIRECTORY<>0 then
        ReadSubDir(SubImageRDE,AType,ATypename,Named)
      else begin
        depos:=FileMemory.Position;
        FileMemory.Position:=SectionOffset+(SubImageRDE.OffsetToData and $7FFFFFFF);
        FileMemory.Read(SubImageDE,SizeOf(_IMAGE_RESOURCE_DATA_ENTRY));

        if SubImageRDE.Name and IMAGE_RESOURCE_NAME_IS_STRING<>0 then begin
          Lang:=0;
        end else
          Lang:=SubImageRDE.Name and $FFFF;

        ResItem                     := TResourceItem.Create;
        ResItem.Header.DataSize     := SubImageDE.Size;
        ResItem.ResType.Named       := Named;
        ResItem.ResType.ResId       := AType;
        ResItem.ResType.Name        := ATypename;
        ResItem.ResName.Named       := SubNamed;
        ResItem.ResName.Name        := SubName;
        ResItem.ResName.ResId       := SubNameID;
        ResItem.Suffix.LanguageID   := Lang;

        SetLength(ResItem.ResData,SubImageDE.Size);
        LoadData(ResItem.ResData,SubImageDE.Size,SectionOffset+(SubImageDE.OffsetToData-SectionVA));

        AddEntry(ResItem);
        FileMemory.Position:=depos;
      end;

    end;

    FileMemory.Position:=pos;
  end;

begin
  FileMemory.Read(ExeHeader,SizeOf(_IMAGE_DOS_HEADER));
  FileMemory.Position     :=ExeHeader._lfanew;
  FileMemory.Write(Magic,   SizeOf(DWORD));
  FileMemory.Read(PeHeader, SizeOf(_IMAGE_FILE_HEADER));
  FileMemory.Read(OptHeader,SizeOf(_IMAGE_OPTIONAL_HEADER));
  if OptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_RESOURCE].Size=0 then Exit;

  for i:=0 to PeHeader.NumberOfSections-1 do begin
    FileMemory.Read(PeSection,SizeOf(_IMAGE_SECTION_HEADER)); 
    if OptHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_RESOURCE].VirtualAddress=PeSection.VirtualAddress then begin
      FileMemory.Position:=PeSection.PointerToRawData;
      SectionVA:=PeSection.VirtualAddress;
      Break;
    end;
  end;

  SectionOffset:=FileMemory.Position;
  FileMemory.Read(ImageRD,SizeOf(_IMAGE_RESOURCE_DIRECTORY));

  for i:=0 to ImageRD.NumberOfNamedEntries+ImageRD.NumberOfIdEntries-1 do begin
    FileMemory.Read(ImageRDE,SizeOf(_IMAGE_RESOURCE_DIRECTORY_ENTRY));
    IsName:=false;
    if ImageRDE.Name and IMAGE_RESOURCE_NAME_IS_STRING<>0 then begin
      Name:=ReadNameOffset(ImageRDE.Name and $7FFFFFFF);
      IsName:=true;
    end else
      Name:=CResources[ImageRDE.Name and $FFFF].Name;
    ReadSubDir(ImageRDE,ImageRDE.Name,Name,IsName);
  end;

  FileMemory.Clear;
end;

procedure TResource.LoadFromFile(Name: String);
var
  ResItem   : TResourceItem;
  Signature : Word;

  procedure ReadResID(var id: TResName);
  var
    w   : WORD;
  begin
    FileMemory.Read(w,SizeOf(WORD));
    if (w=$ffff) then begin
      FileMemory.Read(w,SizeOf(WORD));
      id.Named:=false;
      id.ResID:=w;
    end else begin
      id.Named:=true;
      id.Name:=CHR(w)+_ReadWideString(FileMemory,0,false);
    end;
  end;

  procedure AlignDWORD;
  var
    SkipBytes : Integer;
    Dummy     : DWORD;
  begin
    if FileMemory.Position mod SizeOf(DWORD) <> 0 then begin
      SkipBytes := SizeOf(DWORD) - FileMemory.Position mod SizeOf(DWORD);
      FileMemory.Read(Dummy, SkipBytes);
    end;
  end;

begin
  FileMemory.LoadFromFile(Name);
  FileMemory.Read(Signature,SizeOf(WORD));
  FileMemory.Position:=0;
  if Signature=$5A4D then begin
    LoadFromPeFile;
    Exit;
  end;
  FileMemory.Read(ResHeader,SizeOf(TFileResHeader));
  while FileMemory.Position<FileMemory.Size do begin
    ResItem   := TResourceItem.Create;
    FileMemory.Read(ResItem.Header,Sizeof(TResHeader));
    if (ResItem.Header.DataSize<>0) then begin
      ReadResID(ResItem.ResType);
      ReadResID(ResItem.ResName);
      AlignDWORD;
      FileMemory.Read(ResItem.Suffix,SizeOf(TResHeaderSuffix));
      SetLength(ResItem.ResData,ResItem.Header.DataSize);
      FileMemory.Read(ResItem.ResData[0],ResItem.Header.DataSize);
      AlignDWORD;
      AddEntry(ResItem);
    end else
      ResItem.Free;
  end;
  FileMemory.Clear;
end;

end.
