UNIT LNSi386;

INTERFACE

USES
  Windows, Classes, SysUtils, Dialogs, LNTi386, LNKi386, _SYSTEM;

TYPE
  TLinkToken= (
        _null,_number,_string,_ident,_term,_eof
  );
  TLNumTyp = ( _nil, _char,_integer,_hex,_real,_longreal);

  ScanProc = FUNCTION (VAR Sender: TObject; VAR T: TLinkToken): BOOLEAN;

CONST
  _CEOF   = #26;
  _EOT    = #0;
  MaxStrLen = 256;

TYPE
  TLinkScaner  = CLASS
    ch,OldCh    : CHAR;
    str         : String;
    intval      : integer;
    realval     : REAL;
    lrlval      : DOUBLE;
    numtyp      : TLNumTyp;
    len         : INTEGER;
    FileStream  : TMemoryStream;
    Tok         : TLinkToken;
    Done        : BOOLEAN;
    Linker      : TLinker;
    PROCEDURE   Number;
    PROCEDURE   GetStr(VAR Tok: TLinkToken);
    PROCEDURE   GetSymbolStr(VAR Tok: TLinkToken);
    PROCEDURE   Identifier(VAR Tok: TLinkToken);
    PROCEDURE   ScanLinkText;
  PUBLIC
    Name        : String;
    constructor Create(Path: String; ALinker: TLinker);
    destructor  Destroy; override;
    PROCEDURE   GetChar(VAR ch: CHAR);
    PROCEDURE   Scan(VAR Tok: TLinkToken);
  END;

VAR
  StrPos : INTEGER;
  CharPoS: INTEGER;

IMPLEMENTATION

CONST
  MaxHDig      = 8;  (* maximal hexadecimal longint length *)
  MaxRExp      = 38;  (* maximal real exponent *)
  MaxLExp      = 308;  (* maximal DOUBLE exponent *)

PROCEDURE err(n: INTEGER);
BEGIN
  SHOWMESSAGE(INTTOSTR(n));
END;

CONSTRUCTOR TLinkScaner.Create(Path: String; ALinker: TLinker);
BEGIN
  inherited Create;

  Linker:=ALinker;
  
  StrPos:=1;
  CharPoS:=1;

  ch      := ' ';
  OldCh   := ' ';
  Tok     := _null;

  FileStream  := TMemoryStream.Create;
  Done    := FALSE;

  IF FileExists(Path) THEN
    FileStream.LoadFromFile(Path)
  ELSE BEGIN
    ShowMessage('File "'+Path+'" not found!');
    EXIT;
  END;

  Done    := TRUE;

  ScanLinkText;
  
END;

DESTRUCTOR TLinkScaner.Destroy;
BEGIN
  FileStream.Free;
  inherited Destroy;
END;

PROCEDURE TLinkScaner.GetChar(VAR ch: CHAR);
BEGIN
  OldCh:=ch;
  IF FileStream.Position<FileStream.Size THEN BEGIN
    FileStream.Read(ch,1);
    IF (OldCh=#13) AND (ch=#10) THEN BEGIN
      Inc(StrPos);
      CharPoS:=1;
    END ELSE IF ch=#10 THEN BEGIN
      Inc(StrPos);
      CharPoS:=1;
    END ELSE
      Inc(CharPos);
  END ELSE BEGIN
    ch:=_CEOF;
  END;
END;

PROCEDURE TLinkScaner.Number;
VAR
  i, m, n, d, e: INTEGER;
  dig: ARRAY [0..24] OF CHAR;
  f: DOUBLE;
  expCh: CHAR;
  neg: BOOLEAN;

  FUNCTION Ten(e: INTEGER): DOUBLE;
  VAR
    x, p: DOUBLE;
  BEGIN
    x := 1;
    p := 10;
    WHILE e > 0 DO BEGIN
      IF ODD(e) THEN x := x*p;
      e := e DIV 2;
      IF e > 0 THEN p := p*p;
    END;
    RESULT:= x
  END;

  FUNCTION FOrd(ch: CHAR; hex: BOOLEAN): INTEGER;
  BEGIN
    IF ch <= '9' THEN RESULT:= ORD(ch) - ORD('0')
    ELSE IF hex THEN RESULT:= ORD(ch) - ORD('A') + 10
    ELSE BEGIN
      err(2);
      RESULT:= 0
    END
  END;

BEGIN
  i := 0;
  m := 0;
  n := 0;
  d := 0;
  len:=0;
  
  REPEAT
    IF ('0' <= ch) AND (ch <= '9') OR (d = 0) AND ('A' <= ch) AND (ch <= 'F') THEN BEGIN
      IF (m > 0) OR (ch <> '0') THEN BEGIN
        IF n < LENGTH(dig) THEN BEGIN
          dig[n] := ch;
          INC(n)
        END;
        INC(m);
      END;
      GetChar(ch);
      INC(i)
    END ELSE IF ch = '.' THEN BEGIN
      GetChar(ch);
      IF ch = '.' THEN BEGIN
        ch := #$7F;
        BREAK;
      END ELSE IF d = 0 THEN
        d := i
      ELSE
        err(2);
    END ELSE
      BREAK;
  UNTIL FALSE;

  IF d = 0 THEN BEGIN
    IF n = m THEN BEGIN
      intval := 0;
      i := 0;
      IF ch = 'X' THEN BEGIN
        GetChar(ch);
        numtyp := _char;
        IF n <= 2 THEN BEGIN
          WHILE i < n DO BEGIN
            intval := intval*$10 + FOrd(dig[i], TRUE);
            INC(i)
          END;
        END ELSE
          err(203);
      END ELSE IF ch = 'H' THEN BEGIN
        GetChar(ch);
        numtyp := _hex;
        IF n <= MaxHDig THEN BEGIN
          IF (n = MaxHDig) AND (dig[0] > '7') THEN intval := -1;
          WHILE i < n DO BEGIN
            intval := intval*$10 + FOrd(dig[i], TRUE);
            INC(i)
          END;
          len:=n;
        END ELSE
          err(203);
      END ELSE BEGIN
        numtyp := _integer;
        WHILE i < n DO BEGIN
          d := FOrd(dig[i], FALSE);
          INC(i);
          IF intval <= (_SYSTEM.MAX(_LONGINTS) - d) DIV 10 THEN
            intval := intval*10 + d
          ELSE
            err(203);
        END;
      END;
    END ELSE
      err(203);
  END ELSE BEGIN
    f := 0;
    e := 0;
    expCh := 'E';
    WHILE n > 0 DO BEGIN
      DEC(n);
      f := (FOrd(dig[n], FALSE) + f)/10;
    END;
    IF (ch = 'E') OR (ch = 'D') THEN BEGIN
      expCh := ch;
      GetChar(ch);
      neg := FALSE;
      IF ch = '-' THEN BEGIN
        neg := TRUE;
        GetChar(ch);
      END ELSE IF ch = '+' THEN
        GetChar(ch);
      IF ('0' <= ch) AND (ch <= '9') THEN BEGIN
        REPEAT
          n := FOrd(ch, FALSE);
          GetChar(ch);
          IF e <= (_SYSTEM.MAX(_INTEGERS) - n) DIV 10 THEN
            e := e*10 + n
          ELSE
            err(203);
        UNTIL (ch < '0') OR ('9' < ch);
        IF neg THEN e := -e;
      END ELSE
        err(2);
    END;

    DEC(e, i-d-m);

    IF expCh = 'E' THEN BEGIN
      numtyp := _real;
      IF (1-MaxRExp < e) AND (e <= MaxRExp) THEN BEGIN
        IF e < 0 THEN
          realval := f / Ten(-e)
        ELSE
          realval := f * Ten(e);
      END ELSE
        err(203);
    END ELSE BEGIN
      numtyp := _longreal;
      IF (1-MaxLExp < e) AND (e <= MaxLExp) THEN BEGIN
        IF e < 0 THEN
          lrlval := f / Ten(-e)
        ELSE
          lrlval := f * Ten(e);
      END ELSE
        err(203);
    END;
    
  END;
END;

PROCEDURE TLinkScaner.GetStr(VAR Tok: TLinkToken);
VAR
  i: INTEGER;
  och: CHAR;
BEGIN
  i := 0;
  och := ch;
  str:='';
  REPEAT
    GetChar(ch);
    IF ch = och THEN BREAK;
    IF ch < ' ' THEN BEGIN
      err(345);
      BREAK;
    END ;
    IF i = MaxStrLen-1 THEN BEGIN
      err(241);
      BREAK;
    END ;
    str := str+ch;
    inc(i);
  UNTIL False;
  GetChar(ch);
  intval := i + 1;
  IF intval = 2 THEN BEGIN
    Tok := _number;
    numtyp := _char;
    intval := ORD(str[1]);
  END ELSE
    Tok := _string;
END;

PROCEDURE TLinkScaner.GetSymbolStr(VAR Tok: TLinkToken);
BEGIN
  str:='';
  GetChar(ch);
  WHILE NOT (ch<' ') DO BEGIN
    str:=str+ch;
    GetChar(ch);
  END;
  Tok := _string;
END;

PROCEDURE TLinkScaner.Identifier(VAR Tok: TLinkToken);

  FUNCTION InNotRange: BOOLEAN;
  BEGIN
    RESULT:=(ch < '0') OR ('9' < ch) AND (CAP(ch) < 'A') OR ('Z' < CAP(ch)) AND (ch <> '_');
  END;

BEGIN
  name:='';
  WHILE ch='_' DO GetChar(ch);
  IF InNotRange THEN err(240);
  REPEAT
    name := name+ch;
    GetChar(ch);
  UNTIL InNotRange;
  IF Length(name)>=MaxStrLen THEN err(240);
  Tok := _ident;
END;

PROCEDURE TLinkScaner.Scan(VAR Tok: TLinkToken);
var
  s: TLinkToken;

  	PROCEDURE Comment;	(* do not read after END of file *)
		var
      dump: BOOLEAN;
		BEGIN
      GetChar(ch);
      dump:=FALSE;
			IF ch = '@' THEN dump := TRUE;
			REPEAT
				REPEAT
					WHILE ch = '(' DO BEGIN
            GetChar(ch);
						IF ch = '*' THEN Comment
            ELSE IF dump THEN ;//SHOWMESSAGE('(');
					END;
					IF ch = '*' THEN BEGIN
            GetChar(ch);
            BREAK;
          END;
					IF ch = _CEOF THEN EXIT;
					IF dump THEN ;//SHOWMESSAGE(ch);
					GetChar(ch)
				UNTIL False;
				if ch = ')' THEN BEGIN
          GetChar(ch);
          BREAK;
        END;
				IF dump THEN ;//SHOWMESSAGE('*');
				IF ch = _CEOF then BEGIN
          err(5);
          BREAK;
        END;
			UNTIL False;
		END;

BEGIN

  WHILE ch <= ' ' DO BEGIN
    IF ch=_CEOF then BEGIN
      Tok:=_eof;
      Exit;
    END ELSE
      GetChar(ch);
  END;

  name   := '';
  numtyp := _nil;

  CASE ch OF
      #$22, #$27  : GetStr(s);
			'0'..'9'    :
                    BEGIN
                      Number;
                      s := _number;
                    END;
      '('  :
            BEGIN
              GetChar(ch);
              IF ch = '*' then BEGIN
                Comment;
                Scan(Tok);
                Exit;
              END ELSE
                s := _term;
            END;
      '.','=',',',
      ')' :
                    BEGIN
                      s :=  _term;
                      name:=ch;
                      GetChar(ch);
                    END;
      ':','/',
      'A'..'Z',
      'a'..'z'    : Identifier(s);
  ELSE
    s :=  _null;
    GetChar(ch);
  END;

  Tok := s;
END;

PROCEDURE TLinkScaner.ScanLinkText;
VAR
  LinkTok: TLinkToken;
  keyword: STRING;
  tmp    : DWORD;

  FUNCTION ScanInt(VAR i: DWORD; Msg: STRING): BOOLEAN;
	BEGIN
    RESULT:=TRUE;
		Scan(LinkTok);
    IF numtyp IN [_hex,_integer] THEN
      i:=intval
    ELSE BEGIN
      ShowMessage('Invalid format '+Msg);
      Done:=FALSE;
      RESULT:=FALSE;
    END;
	END;

  FUNCTION ScanVersion(VAR Sender: TObject; VAR T: TLinkToken): BOOLEAN;
  VAR
    valname  : STRING;
    valvalue : STRING;
	BEGIN
    RESULT:=TRUE;

		IF T<>_ident THEN BEGIN
      SHOWMESSAGE('Error NAME in ScanVersion');
      TLinkScaner(Sender).Done:=FALSE;
      RESULT:=FALSE;
      EXIT;
    END;

    REPEAT

      TLinkScaner(Sender).Scan(T);
      IF T<>_ident THEN BEGIN
        SHOWMESSAGE('Error val.name in ScanVersion');
        TLinkScaner(Sender).Done:=FALSE;
        RESULT:=FALSE;
        EXIT;
      END;
      valname:=TLinkScaner(Sender).Name;

      TLinkScaner(Sender).Scan(T);
      IF TLinkScaner(Sender).Name<>'=' THEN BEGIN
        SHOWMESSAGE('Error = in ScanVersion');
        TLinkScaner(Sender).Done:=FALSE;
        RESULT:=FALSE;
        EXIT;
      END;

      TLinkScaner(Sender).Scan(T);
      IF T<>_string THEN BEGIN
        SHOWMESSAGE('Error STRING VALUE in ScanVersion');
        TLinkScaner(Sender).Done:=FALSE;
        RESULT:=FALSE;
        EXIT;
      END;
      valvalue:=TLinkScaner(Sender).Name;

      TLinkScaner(Sender).Scan(T);

    UNTIL TLinkScaner(Sender).Name<>',';

  END;

  FUNCTION ScanImport(VAR Sender: TObject; VAR T: TLinkToken): BOOLEAN;
  VAR
    impname : STRING;
    entname : STRING;
    libname : STRING;
    impvalue: STRING;
    impentry: INTEGER;
	BEGIN
    RESULT:=TRUE;

    IF T<>_ident THEN BEGIN
      SHOWMESSAGE('Error NAME in ScanImport');
      TLinkScaner(Sender).Done:=FALSE;
      RESULT:=FALSE;
      EXIT;
    END;

    impname:=TLinkScaner(Sender).Name;

    TLinkScaner(Sender).Scan(T);
    IF TLinkScaner(Sender).Name='.' THEN BEGIN
      TLinkScaner(Sender).Scan(T);
      entname:=TLinkScaner(Sender).Name;
      TLinkScaner(Sender).Scan(T);
    END;

		IF TLinkScaner(Sender).Name<>'=' THEN BEGIN
      SHOWMESSAGE('Error = in ScanImport');
      TLinkScaner(Sender).Done:=FALSE;
      RESULT:=FALSE;
      EXIT;
    END;

    TLinkScaner(Sender).Scan(T);
    IF T<>_ident THEN BEGIN
      SHOWMESSAGE('Error NAME in ScanImport');
      TLinkScaner(Sender).Done:=FALSE;
      RESULT:=FALSE;
      EXIT;
    END;
		libname:=TLinkScaner(Sender).Name;

    TLinkScaner(Sender).Scan(T);
    IF TLinkScaner(Sender).Name='.' THEN BEGIN
      TLinkScaner(Sender).Scan(T);
      impvalue:=TLinkScaner(Sender).Name;
      TLinkScaner(Sender).Scan(T);
    END ELSE BEGIN
      impvalue:=libname;
      libname := ''; (* pseudo lib *)
			impvalue:=UPPERCASE(impvalue);
    END;

		IF TLinkScaner(Sender).numtyp IN [_hex,_integer] THEN
			impentry:=TLinkScaner(Sender).intval
		ELSE
			impentry := 0;

    LNTi386.AddImport(impname,entname,libname,impvalue,impentry);

	END;

  FUNCTION ScanExport(VAR Sender: TObject; VAR T: TLinkToken): BOOLEAN;
  VAR
    expname : STRING;
    expentry: INTEGER;
		expvalue: STRING;
	BEGIN
    RESULT:=TRUE;

    IF T<>_ident THEN BEGIN
      SHOWMESSAGE('Error NAME in ScanExport');
      TLinkScaner(Sender).Done:=FALSE;
      RESULT:=FALSE;
      EXIT;
    END;
    expname:=TLinkScaner(Sender).Name;

    TLinkScaner(Sender).Scan(T);
		IF TLinkScaner(Sender).numtyp IN [_hex,_integer] THEN BEGIN
			expentry:=TLinkScaner(Sender).intval;
      TLinkScaner(Sender).Scan(T);
		END ELSE
			expentry := 0;

		IF TLinkScaner(Sender).Name<>'=' THEN BEGIN
      SHOWMESSAGE('Error = in ScanExport');
      TLinkScaner(Sender).Done:=FALSE;
      RESULT:=FALSE;
      EXIT;
    END;

    TLinkScaner(Sender).Scan(T);
    IF T<>_ident THEN BEGIN
      SHOWMESSAGE('Error NAME in ScanExport');
      TLinkScaner(Sender).Done:=FALSE;
      RESULT:=FALSE;
      EXIT;
    END;
		expvalue:=TLinkScaner(Sender).Name;

    TLinkScaner(Sender).Scan(T);
    IF TLinkScaner(Sender).Name='.' THEN BEGIN
      TLinkScaner(Sender).Scan(T);
      expvalue:=expvalue+'.'+TLinkScaner(Sender).Name;
      TLinkScaner(Sender).Scan(T);
    END;

	END;

  FUNCTION ScanModule(VAR Sender: TObject; VAR T: TLinkToken): BOOLEAN;
  VAR
    modname: STRING;
	BEGIN
    RESULT:=TRUE;
    IF T<>_ident THEN BEGIN
      SHOWMESSAGE('Error NAME in ScanModule');
      TLinkScaner(Sender).Done:=FALSE;
      RESULT:=FALSE;
      EXIT;
    END;
    modname:=TLinkScaner(Sender).Name;
    TLinkScaner(Sender).Scan(T);
    LNTi386.AddModule(modname);
	END;

  FUNCTION ScanIcon(VAR Sender: TObject; VAR T: TLinkToken): BOOLEAN;
  VAR
    iconame: STRING;
    icovalue: STRING;
	BEGIN
    RESULT:=TRUE;

    TLinkScaner(Sender).Scan(T);
    IF T<>_ident THEN BEGIN
      SHOWMESSAGE('Error NAME in ScanIcon');
      TLinkScaner(Sender).Done:=FALSE;
      RESULT:=FALSE;
      EXIT;
    END;
    iconame:=UPPERCASE(TLinkScaner(Sender).Name);

    TLinkScaner(Sender).Scan(T);
		IF TLinkScaner(Sender).Name<>'=' THEN BEGIN
      SHOWMESSAGE('Error = in ScanIcon');
      TLinkScaner(Sender).Done:=FALSE;
      RESULT:=FALSE;
      EXIT;
    END;
    TLinkScaner(Sender).GetSymbolStr(T);
    icovalue:=TLinkScaner(Sender).Str;
    TLinkScaner(Sender).Scan(T);
	END;

  FUNCTION ScanCursor(VAR Sender: TObject; VAR T: TLinkToken): BOOLEAN;
  VAR
    curname: STRING;
    curvalue: STRING;
	BEGIN
    RESULT:=TRUE;

    TLinkScaner(Sender).Scan(T);
    IF T<>_ident THEN BEGIN
      SHOWMESSAGE('Error NAME in ScanCursor');
      TLinkScaner(Sender).Done:=FALSE;
      RESULT:=FALSE;
      EXIT;
    END;
    curname:=UPPERCASE(TLinkScaner(Sender).Name);

    TLinkScaner(Sender).Scan(T);
		IF TLinkScaner(Sender).Name<>'=' THEN BEGIN
      SHOWMESSAGE('Error = in ScanCursor');
      TLinkScaner(Sender).Done:=FALSE;
      RESULT:=FALSE;
      EXIT;
    END;
    TLinkScaner(Sender).GetSymbolStr(T);
    curvalue:=TLinkScaner(Sender).Str;
    TLinkScaner(Sender).Scan(T);
	END;

  FUNCTION ScanList(Sender: TObject; VAR T: TLinkToken; entry: ScanProc): BOOLEAN;
	BEGIN
    RESULT:=TRUE;
    Done:=entry(Sender,T);
    IF NOT Done THEN BEGIN
      RESULT:=FALSE;
      EXIT;
    END;
    WHILE (T=_term) AND (name=',') DO BEGIN
      Scan(T);
      entry(Sender,T);
      IF NOT Done THEN BEGIN
        RESULT:=FALSE;
        EXIT;
      END;
		END;
	END;

BEGIN
  Scan(LinkTok);
  Done:=TRUE;

  WHILE LinkTok=_ident DO BEGIN
    keyword:=UpperCase(Name);
    IF keyword = 'PROGRAM' THEN BEGIN
      Scan(LinkTok);
      Linker.Image:=Name+'.exe';
      LNKi386.isEXE := TRUE;
      IF OptHeader.ImageBase = 0 THEN
        OptHeader.ImageBase := LNKi386.EXEImageBase;
      Scan(LinkTok);
    END ELSE IF keyword = 'LIBRARY' THEN BEGIN
      Scan(LinkTok);
      Linker.Image:=Name+'.dll';
      LNKi386.isEXE := FALSE;
      IF OptHeader.ImageBase = 0 THEN
        OptHeader.ImageBase := LNKi386.DLLImageBase;
      Scan(LinkTok);
    END ELSE IF keyword = 'STUB' THEN BEGIN
      Scan(LinkTok);
      IF LinkTok<>_string THEN BEGIN
        SHOWMESSAGE('Error STUB in project');
        Done:=FALSE;
        EXIT;
      END;
      Linker.Stub:=Str;
      Scan(LinkTok);
    END ELSE IF keyword = 'HEAPSIZE' THEN BEGIN
      IF NOT ScanInt(OptHeader.SizeOfHeapReserve,'HEAPSIZE') THEN BREAK;
      OptHeader.SizeOfHeapReserve := LNKi386.Align(OptHeader.SizeOfHeapReserve, OptHeader.SectionAlignment);
      IF OptHeader.SizeOfHeapReserve > 0 THEN
        OptHeader.SizeOfHeapCommit := OptHeader.SectionAlignment
      ELSE
        OptHeader.SizeOfHeapCommit := 0;
      Scan(LinkTok);
    END ELSE IF keyword = 'STACKSIZE' THEN BEGIN
      IF NOT ScanInt(OptHeader.SizeOfStackReserve,'STACKSIZE') THEN BREAK;
      OptHeader.SizeOfStackReserve := LNKi386.Align(OptHeader.SizeOfStackReserve, OptHeader.SectionAlignment);
      IF OptHeader.SizeOfStackReserve > 0 THEN
        OptHeader.SizeOfStackCommit := OptHeader.SectionAlignment
      ELSE
        OptHeader.SizeOfStackCommit := 0;
      Scan(LinkTok);
    END ELSE IF keyword = 'BASE' THEN BEGIN
      IF NOT ScanInt(OptHeader.ImageBase,'BASE') THEN BREAK;
      Scan(LinkTok);
    END ELSE IF keyword = 'IMGVERSION' THEN BEGIN
      Scan(LinkTok);
      IF numtyp IN [_real,_longreal] THEN BEGIN
        OptHeader.MajorImageVersion := TRUNC(realval);
        OptHeader.MinorImageVersion := ROUND(Frac(realval)*10);
      END ELSE BEGIN
        SHOWMESSAGE('IMGVERSION is incorrect');
        Done:=FALSE;
        BREAK;
      END;
      Scan(LinkTok);
    END ELSE IF keyword = 'VERSION' THEN BEGIN
      IF NOT ScanList(SELF,LinkTok,@ScanVersion) THEN EXIT;
    END ELSE IF keyword = 'SUBSYSTEM' THEN BEGIN
      Scan(LinkTok);
      keyword:=UpperCase(Name);
      IF keyword = 'GUI' THEN
        OptHeader.Subsystem := IMAGE_SUBSYSTEM_WINDOWS_GUI
      ELSE IF keyword = 'CUI' THEN
        OptHeader.Subsystem := IMAGE_SUBSYSTEM_WINDOWS_CUI
      ELSE IF keyword = 'NATIVE' THEN
        OptHeader.Subsystem := IMAGE_SUBSYSTEM_NATIVE
      ELSE BEGIN
        SHOWMESSAGE('unknown subsystem '+keyword);
        Done:=FALSE;
        BREAK;
      END;
      Scan(LinkTok);
    END ELSE IF keyword = 'IMPORTS' THEN BEGIN
      Scan(LinkTok);
      IF NOT ScanList(SELF,LinkTok,@ScanImport) THEN EXIT;
    END ELSE IF keyword = 'EXPORTS' THEN BEGIN
      Scan(LinkTok);
      IF NOT ScanList(SELF,LinkTok,@ScanExport) THEN EXIT;
    END ELSE IF keyword = 'MODULES' THEN BEGIN
      Scan(LinkTok);
      IF NOT ScanList(SELF,LinkTok,@ScanModule) THEN EXIT;
    END ELSE IF keyword = 'ICONS' THEN
      ScanList(SELF,LinkTok,@ScanIcon)
    ELSE IF keyword = 'CURSORS' THEN
      ScanList(SELF,LinkTok,@ScanCursor)
    ELSE BEGIN
      SHOWMESSAGE('unknown keyword '+keyword);
      Done:=FALSE;
      BREAK;
    END;
    
  END;
  
  IF FileStream.Position<FileStream.Size THEN BEGIN
    SHOWMESSAGE('Not current EOF file');
    Done:=FALSE;
  END;

END;

END.
