UNIT LNTi386;

INTERFACE

USES
  Windows, SysUtils, Classes, DebugFrm;
  
TYPE
  ImageImportDescriptor = RECORD
    Characteristics : DWORD;
    TimeDateStamp   : DWORD;
    ForwarderChain  : DWORD;
    Name            : DWORD;
    FirstThunk      : DWORD;
  END;

  EntryList = CLASS
    name  : STRING;
    value : STRING;
    entry : DWORD;
    fixup : DWORD;
    next  : EntryList;
  END;

  ImportList = CLASS
    name   : STRING;
    impDesc: ImageImportDescriptor;
    entries: EntryList;
    next   : ImportList;
  END;

  NameList = CLASS
    name: STRING;
    next: NameList;
  END;

  ExportEntryList = CLASS(EntryList)
    nextOrd: ExportEntryList;
  END;

VAR
  modules         : NameList;
  resources       : NameList;
  imports         : ImportList;
  iexports        : ExportEntryList;
  exportOrds      : ExportEntryList;
  
  PROCEDURE Init;
  PROCEDURE Done();
  PROCEDURE AddImport(Module,Name,LName,Value: STRING; Entry: DWORD);
  PROCEDURE AddModule(ModuleName: STRING);
  PROCEDURE AddResource(ResName: STRING);
  PROCEDURE AddExport(Name,MName,Value: STRING; Entry: DWORD);

IMPLEMENTATION

PROCEDURE Init();
BEGIN
  imports         := ImportList.Create;
  imports.name    := '';
  imports.next    := NIL;
  imports.entries := NIL;

  modules         := NIL;
  iexports        := NIL;    
  resources       := NIL;
END;

PROCEDURE InitImageImportDescriptor(VAR impDesc: ImageImportDescriptor);
BEGIN
  impDesc.Characteristics := 0;
  impDesc.TimeDateStamp   := 0;
  impDesc.ForwarderChain  := 0;
  impDesc.Name            := 0;
  impDesc.FirstThunk      := 0
END;

PROCEDURE FreeEntry(Ent: EntryList);
BEGIN
  IF Ent.next<>NIL THEN FreeEntry(EntryList(Ent.Next));
  FreeAndNil(Ent);
END;

PROCEDURE FreeImport(Imp: ImportList);
BEGIN
  IF Imp.next<>NIL THEN FreeImport(ImportList(Imp.Next));
  IF Imp.entries<>NIL THEN FreeEntry(Imp.entries);
  FreeAndNil(Imp);
END;

PROCEDURE FreeModule(Mdl: NameList);
BEGIN
  IF Mdl<>NIL THEN FreeModule(Mdl.next);
  FreeAndNil(Mdl);
END;

PROCEDURE FreeExport(Exp: ExportEntryList);
BEGIN
  IF Exp<>NIL THEN FreeExport(ExportEntryList(Exp.next));
  FreeAndNil(Exp);
END;

PROCEDURE Done();
BEGIN
  FreeImport(imports);
  FreeModule(modules);
  FreeExport(iexports);
END;

PROCEDURE AddImport(Module,Name,LName,Value: STRING; Entry: DWORD);
VAR
  imp,q,e   : EntryList;
  lib,p,i   : ImportList;

BEGIN
  imp       := EntryList.Create;
  imp.next  := NIL;
  imp.name  := Module+'.'+Name;
  imp.value := Value;
  imp.entry := Entry;

  if LName='' THEN
    imp.value := UPPERCASE(imp.value)
  ELSE
    LName:= LName+'.DLL';

  p := Imports;
  WHILE (p <> NIL) AND (p.name<>LName) DO p := p.next;

  IF p = NIL THEN BEGIN
    lib         := ImportList.Create;
    lib.name    := LName;
    lib.entries := NIL;
    InitImageImportDescriptor(lib.impDesc);
    i := Imports;
    p := Imports.next;
    WHILE (p <> NIL) AND (p.name < lib.name) DO BEGIN
      i := p;
      p := p.next;
    END;
    lib.next := p;
    i.next := lib;
  END ELSE
    lib := p;

  q := NIL;

  e := lib.entries;
  WHILE (e <> NIL) AND (e.value <> imp.value) DO e := e.next;
  IF e<>NIL THEN BEGIN
    FreeAndNil(imp);
    EXIT;
  END;
  
  e := lib.entries;
  WHILE (e <> NIL) AND (e.value < imp.value) DO BEGIN
    q := e;
    e := e.next;
  END;

  imp.next := e;

  IF q <> NIL THEN
    q.next := imp
  ELSE
    lib.entries := imp;

END;

PROCEDURE AddExport(Name,MName,Value: STRING; Entry: DWORD);
VAR
  exp, e : ExportEntryList;
  p, q   : EntryList;
BEGIN
  exp         :=ExportEntryList.Create;
  exp.next    := NIL;
  exp.nextOrd := NIL;

  exp.name  := Name;
  exp.entry := Entry;                
  exp.value := MName+'.'+Value;

  q := NIL;
  p := iexports;

  WHILE (p <> NIL) AND (p.name < exp.name) DO BEGIN
    q := p;
    p := p.next;
  END;

  IF NOT ((p = NIL) OR (p.name > exp.name)) THEN BEGIN
    //SHOWMESSAGE(exp.name+ ' exported names must be unique');
    EXIT;
  END;

  exp.next := p;
  IF q <> NIL THEN
    q.next := exp
  ELSE
    iexports := exp;

  IF exp.entry > 0 THEN BEGIN
    q := NIL;
    e := exportOrds;
    WHILE (e <> NIL) AND (e.entry < exp.entry) DO BEGIN
      q := e;
      e := e.nextOrd
    END;

    IF NOT ((e = NIL) OR (e.entry > exp.entry)) THEN BEGIN
      //SHOWMESSAGE(exp.name+ ' exported names must be unique');
      EXIT;
    END;

    exp.nextOrd := e;

    IF q <> NIL THEN
      ExportEntryList(q).nextOrd := exp
    ELSE
      exportOrds := exp;
  END ELSE BEGIN
    exp.nextOrd := exportOrds;
    exportOrds := exp;
  END;

END;

PROCEDURE AddModule(ModuleName: STRING);
VAR
  n, m: NameList;

  PROCEDURE Add(VAR a,md: NameList);
  BEGIN
    IF md<>NIL THEN
      Add(a,md.next)
    ELSE
      md:=a;
  END;

BEGIN
  n      := NameList.Create;
  n.name := ModuleName;
  n.next := NIL;
  IF modules <> NIL THEN BEGIN
    m := modules;
    WHILE (m <> NIL) AND (m.name<>ModuleName) DO m := m.next;
    IF m=NIL THEN
      Add(n,modules)
    ELSE
      FreeAndNil(n);
  END ELSE
    modules := n;
END;

PROCEDURE AddResource(ResName: STRING);
VAR
  n, m: NameList;

  PROCEDURE Add(VAR a,md: NameList);
  BEGIN
    IF md<>NIL THEN
      Add(a,md.next)
    ELSE
      md:=a;
  END;

BEGIN
  n      := NameList.Create;
  n.name := ResName;
  n.next := NIL;
  IF resources <> NIL THEN BEGIN
    m := resources;
    WHILE (m <> NIL) AND (m.name<>ResName) DO m := m.next;
    IF m=NIL THEN
      Add(n,resources)
    ELSE
      FreeAndNil(n);
  END ELSE
    resources := n;
END;

END.
