UNIT OPL;

INTERFACE

USES
  OPT, OPM, OPOi386, _SYSTEM, SYSUTILS;

CONST
	aNil	        = OPM.LANotAlloc;
  MaxExtensions	= 15;
  noEntry       = -1;
  BUG           = 41;

TYPE
  Instruction = RECORD
    scale, reg: SHORTINT; 
    op: INTEGER;
    dest, src1, src2,
    inx, pc, hint: LONGINT; (* pc: PC on the target machine, used for jumps *)
    link, used: INTEGER;
    node: Nodes;
    abs: BOOLEAN;	(* absolute address, should not be fixed. Def = FALSE. *)		(* GetPut *)
  END;
  InstructionTable = ARRAY OF Instruction;

	RecTable   = ARRAY OF Structs;
  EntryTable = ARRAY OF INTEGER;

VAR
  InstrTab:  InstructionTable; (* Pseudo Code *)
  recTab: RecTable;	(* record table *)
  entry: EntryTable;	(* entry table *)
  nofEntries, nofrecs: INTEGER;
  ConstErr: BOOLEAN;
(*
CONST
  OPOMaxConstLength	= 64*1024;
  OPOAbs = 16;
TYPE
  OPOConstArray = ARRAY OF CHAR;
VAR
  OPOdsize    : INTEGER;
  OPOcsize    : INTEGER;
  OPOconstant : OPOConstArray;
*)
  PROCEDURE AllocTypDesc (typ: Structs);
  PROCEDURE NewEntry(VAR entryNr: LONGINT);
  PROCEDURE Init;
  PROCEDURE OPLBegin;
  PROCEDURE Close;
  PROCEDURE AllocConst (VAR s: STRING; len, align: LONGINT; VAR adr: LONGINT);
  PROCEDURE AllocCaseTab (low, high: LONGINT; VAR tab: LONGINT);
  
CONST
  noHint = -1;
  useEAX = 0;
  useECX = 1;
  useEDX = 2;
  useEBX = 3;
  useESP = 4;
  useEBP = 5;
  useESI = 6;
  useEDI = 7;
  useST  = 8;
  tryEAX = 16;
  tryECX = 17;
  tryEDX = 18;
  tryEBX = 19;
  tryESI = 22;
  tryEDI = 23;

  WithTrap        = 1;
  CaseTrap        = 2;
  FuncTrap        = 3;
  EqualGuardTrap  = 5;
  GuardTrap       = 6;
  OverflowTrap    = 4;
  RangeTrap       = 7;
  DimTrap         = 9;
  nofTrapFixups   = 10;
  
IMPLEMENTATION

PROCEDURE Init;
BEGIN
  ConstErr    := FALSE;
  nofrecs     := 0;
  nofEntries  := 0;
  OPOi386.dsize := 0;
  OPOi386.pcc := 0;
  OPOi386.csize := 4;		(*module.SELF*)
END;

PROCEDURE NewEntry(VAR entryNr: LONGINT);
VAR
  table: EntryTable;
  i: LONGINT;
BEGIN
  IF nofEntries >= LENGTH(entry) THEN BEGIN
    SETLENGTH(table, LENGTH(entry)*2);
    FOR i:=0 TO LENGTH(entry)-1 DO table[i]:=entry[i];
    entry:=table
  END;
  entry [nofEntries] := aNil;
  entryNr := nofEntries;
  INC (nofEntries)
END;

PROCEDURE AllocConst (VAR S: STRING; len, align: LONGINT; VAR adr: LONGINT);
VAR
  fill,l: LONGINT;
  table: OPOi386.ConstArray;
BEGIN
  IF NOT ConstErr THEN BEGIN
    adr := OPOi386.csize;
    OPOi386.csize := OPOi386.csize + len;
    fill := _SYSTEM._MOD((-len),align);
    IF OPOi386.csize + fill > OPOi386.MaxConstLength THEN BEGIN
      OPM.err (230);
      ConstErr := TRUE
    END ELSE BEGIN
      IF OPOi386.csize + fill >= LENGTH(OPOi386.constant) THEN BEGIN
        SETLENGTH(table, LENGTH(OPOi386.constant)*2);
        l := LENGTH(OPOi386.constant)*2;
        WHILE OPOi386.csize >= l DO l:=l*2;
        SETLENGTH(table,l);
        MOVE(OPOi386.constant[0],table[0],adr);
        OPOi386.constant:=table
      END;
      WHILE fill > 0 DO BEGIN
        OPOi386.constant [OPOi386.csize]:=#0;
        INC (OPOi386.csize);
        DEC (fill);
      END;
      IF S='' THEN S:=#0;
      WHILE (LENGTH(S) MOD len)<>0 DO S:=S+#0;
      MOVE(S[1],OPOi386.constant[adr],len);
    END
  END ELSE
    adr := 0;
END;

PROCEDURE AllocCaseTab (low, high: LONGINT; VAR tab: LONGINT);
VAR
  len, i: LONGINT;
  table: OPOi386.ConstArray;
BEGIN
  IF NOT ConstErr THEN BEGIN
    len := 4 * (high - low + 1);
    tab := OPOi386.csize;
    OPOi386.csize := OPOi386.csize + len;
    IF OPOi386.csize > OPOi386.MaxConstLength THEN BEGIN
      OPM.err (230);
      ConstErr := TRUE
    END ELSE BEGIN
      IF OPOi386.csize >= LENGTH(OPOi386.constant) THEN BEGIN
        SETLENGTH(table, LENGTH(OPOi386.constant)*2);
        len := LENGTH(OPOi386.constant)*2;
        WHILE OPOi386.csize >= len DO len := len*2;
        SETLENGTH(table, len);
        MOVE(OPOi386.constant[0],table[0],tab);
        OPOi386.constant:=table
      END;
      i := tab;
      WHILE i < OPOi386.csize DO BEGIN
        OPOi386.constant[i]:=CHR(255); (* = -1 *)
        INC(i);
      END
    END
  END ELSE
    tab := 0;
END;

PROCEDURE AllocTypDesc (typ: Structs);
VAR
  vnil: STRING;
  table: RecTable;
  i: LONGINT;
BEGIN
  IF typ.comp = csRecord THEN BEGIN
    vnil := #0;
    AllocConst (vnil, 4, 4, typ.tdadr);
    IF typ.extlev > MaxExtensions THEN
      OPM.err (233)
    ELSE IF (typ.mno = 0) THEN BEGIN
      IF nofrecs>=LENGTH(recTab) THEN BEGIN
        SETLENGTH(table, LENGTH(recTab)*2);
        FOR i:=0 TO LENGTH(recTab)-1 DO  table[i]:=recTab[i];
        recTab:=table
      END;
      recTab [nofrecs] := typ;
      INC (nofrecs)
    END
  END
END;

PROCEDURE Close;
VAR
  i: LONGINT;
BEGIN
  FOR i := 0 TO LENGTH(InstrTab)-1 DO InstrTab[i].node:= NIL;
  i := 0;
  WHILE i < LENGTH(recTab) DO BEGIN
    recTab[i] := NIL;
    INC(i)
  END;
  SETLENGTH(recTab, 0);
  SETLENGTH(entry, 0);
	SETLENGTH(InstrTab, 0);
  SETLENGTH(OPOi386.constant, 0);
END;

//######################################################################

PROCEDURE OPLBegin;
BEGIN
  SETLENGTH(recTab, 64);
  SETLENGTH(entry, 128);
	SETLENGTH(InstrTab, 500);
  SETLENGTH(OPOi386.constant, 1024);
END;

END.