UNIT OPP;

INTERFACE

USES
  Classes, SysUtils, OPT, OPS, OPM, OPB, OER, OPA, _SYSTEM, LNTi386, DebugFrm,
  DIALOGS;

TYPE
  CaseTable = ARRAY [0..MaxCases] OF RECORD
    low, high: LONGINT
  END;

  FixPtr = CLASS
    mode: LONGINT;
    scope: Objects;
    level: SHORTINT;
    node: Nodes;
    stream: TokenStream;
    next: FixPtr;
  END;

  TypeFix = CLASS
    pos: TErrorPos;
    name: STRING;
    scope: Objects;
    level: SHORTINT;
  END;

  ObjFixDesc    = CLASS (TypeFix)  obj: OPT.Objects  END;
  StructFixDesc = CLASS (TypeFix)  typ: OPT.Structs  END;

CONST
  super = 1;
  HiddenSelf = '@SELF';

VAR
  SelfName            : STRING;
  sym                 : TToken;
  level               : SHORTINT;
  LoopLevel           : INTEGER;
  TDinit, lastTDinit  : Nodes;
  CondNodes           : Nodes;
  CondCount           : SHORTINT;
  UndefFlag, FixPhase : BOOLEAN;
  UndefObj            : OPT.Objects;
  scopeLock           : BOOLEAN;

  FixList, FixLast    : FixPtr;
  ExtObj              : Objects;
  RecInit, lastRecInit: Nodes;
  Oinc                : Objects;
  ptrToRec            : Structs;

  ObjCheck  :  ARRAY [0..1024] OF RECORD
                obj  : Objects;
                level: SHORTINT
               END;
  ObjFix    :  ARRAY [0..1024] OF ObjFixDesc;
  TypFix    :  ARRAY [0..1024] OF StructFixDesc;
  nofCheck, nofObjFix, nofTypFix: LONGINT;

  retcount: LONGINT;
  fincount: LONGINT;

  FUNCTION  GetSelf():OPT.Objects;
  PROCEDURE Resolve(VAR obj: OPT.Objects; name: STRING; VAR x, y: OPT.Nodes);
  PROCEDURE OwnScope(VAR obj: OPT.Objects);
  PROCEDURE Block(VAR Scaner: TScaner; VAR procdec, statseq: OPT.Nodes);
  PROCEDURE StatBlock(VAR Scaner: TScaner; VAR stseq: OPT.Nodes;  local: BOOLEAN);
  PROCEDURE ActualParameters(VAR Scaner: TScaner; VAR aparlist: OPT.Nodes; fpar: OPT.Objects; x: OPT.Nodes);
  PROCEDURE InlineProc(VAR x: OPT.Nodes; VAR apar: OPT.Nodes; fpar: OPT.Objects);
  PROCEDURE Await(VAR Scaner: TScaner; VAR x: OPT.Nodes; VAR n: INTEGER);
  PROCEDURE Module(VAR prog: OPT.Nodes; VAR modName: STRING; optn: TSET);

IMPLEMENTATION

USES
  OPE;

FUNCTION IsObjectScope ():BOOLEAN;
BEGIN
  RESULT:=OPT.topScope.link.mode = omTyp;
END;

PROCEDURE Resolve(VAR obj: OPT.Objects; name: STRING; VAR x, y: OPT.Nodes);
BEGIN
  obj:=OPT.Resolve(name, x, y);
  OPT.MarkObj(obj);
END;

PROCEDURE SetFlags (obj: OPT.Objects; flag: TSET);
BEGIN
  obj.conval^.setval := obj.conval^.setval + flag
END;

FUNCTION IsRecordScope(): BOOLEAN;
VAR
  owner: OPT.Objects;
BEGIN
  owner:=OPT.topScope.link;
  RESULT:= (owner.mode=omTyp) AND (owner.typ<>NIL) AND (owner.typ.comp=csRecord);
END;

PROCEDURE OwnScope(VAR obj: OPT.Objects);
BEGIN
  IF NOT (obj.mode IN [omCon, omTyp, omSProc]) THEN
    IF IsRecordScope() OR ((obj.mnolev=OPT.topScope.mnolev) AND (obj.link2=NIL) AND (obj.myscope=OPT.topScope)) THEN err(90);
END;

FUNCTION GetSelf():OPT.Objects;
VAR
  self: OPT.Objects;
BEGIN
  OPT.FindInScope(HiddenSelf, OPT.topScope, self);
  RESULT:= self;
END;

FUNCTION GetFlags (obj: OPT.Objects):TSET;
BEGIN
  RESULT:= obj.conval^.setval
END;

PROCEDURE CheckBool(VAR x: OPT.Nodes);
BEGIN
  IF (x.clas = Ntype) OR (x.clas = Nproc) THEN BEGIN
    err(126);
    x := OPB.NewBoolConst(FALSE)
  END ELSE IF x.typ = NIL THEN BEGIN
    err(120);
    x := OPB.NewBoolConst(FALSE)
  END ELSE IF x.typ.form <> sfBool THEN BEGIN
    err(120);
    x := OPB.NewBoolConst(FALSE)
  END ;
END;

FUNCTION IsModuleScope ():BOOLEAN;
BEGIN
  RESULT:= OPT.topScope.link.mode = omMod;
END;

PROCEDURE NewFixO(VAR f: ObjFixDesc; name: STRING);
BEGIN
  f:= ObjFixDesc.Create;
  f.pos := OPM.errpos;
  f.scope := OPT.topScope;
  f.name := name;
  f.level := level;
END;

PROCEDURE NewFixS(VAR f: StructFixDesc; name: STRING);
BEGIN
  f:= StructFixDesc.Create;
  f.pos := OPM.errpos;
  f.scope := OPT.topScope;
  f.name := name;
  f.level := level
END;

PROCEDURE NewCheck(obj: Objects);
BEGIN
  ObjCheck[nofCheck].obj := obj;
  ObjCheck[nofCheck].level := level;
  INC(nofCheck);
END;

PROCEDURE CheckMark(VAR Scaner: TScaner; VAR vis: SHORTINT; rdonlyAllowed: BOOLEAN);
BEGIN
  Scaner.Get(sym);
  IF (sym = _times) OR (sym = _minus) THEN BEGIN
    IF level > 0 THEN err(47);
    IF sym = _times THEN
      vis := external
    ELSE IF rdonlyAllowed  THEN
      vis := externalR
    ELSE
      err(47);
    Scaner.Get(sym)
  END ELSE
    vis := internal
END;

PROCEDURE CheckStruct(str: OPT.Structs); FORWARD;

PROCEDURE CheckObj(obj: OPT.Objects);
BEGIN
  IF (obj.mode IN [omXProc, omLProc]) AND (obj.typ = NIL) AND FixPhase THEN	BEGIN(*avoid NIL trap when no return type is defined*)
    obj.typ := OPT.undftyp;
    EXIT;
  END;
  CheckStruct(obj.typ);
  IF obj.typ.incomplete THEN EXIT;
  IF (obj.typ.comp = csSDynArr) AND (level = 0) THEN
    OPM.Mark(90, obj.errpos)
  ELSE BEGIN
    CASE obj.mode OF
      omVar, omVarPar:
                IF OPT.param IN obj.flag THEN BEGIN
                  IF obj.typ.comp = csDynArr THEN OPM.Mark(91, obj.errpos);
                END ELSE BEGIN
                  IF obj.typ.comp = csOpenArr THEN OPM.Mark(88, obj.errpos);
                END;
      omFld:	  IF obj.typ.comp IN [csOpenArr, csDynArr, csSDynArr] THEN OPM.Mark(88, obj.errpos);
      omXProc,
      omLProc:  IF (obj.typ = NIL) AND FixPhase THEN obj.typ := OPT.notyp;	(*avoid NIL trap when no return type is defined*)
    END;
  END;
END;

PROCEDURE CheckRecord(typ: OPT.Structs);
CONST
  stamp = $1234;
VAR
  t, base: OPT.Structs;

  PROCEDURE Traverse(p: OPT.Objects);
  VAR
    fwd: OPT.Objects;
  BEGIN
    IF p <> NIL THEN BEGIN
      Traverse(p.left);
      IF p.mode = omTProc THEN BEGIN
        OPT.FindField(p.name, typ.BaseTyp, fwd, TRUE);
        IF (fwd <> NIL) AND (fwd.mode = omTProc) THEN BEGIN
          OPM.errpos.CodePos := p.conval^.intval;
          OPB.CheckParameters (p.link.link, fwd.link.link, FALSE);
        END;
        p.conval^.intval := 0
      END ELSE IF p.mode = omFld THEN BEGIN
        OPT.FindField(p.name, typ.BaseTyp, fwd, TRUE);
        IF (fwd <> NIL) THEN OPM.Mark(1, p.conval^.errpos);
      END;
      Traverse(p.right)
    END;
  END;

BEGIN
  ASSERT(typ.comp = csRecord);
  IF typ.rectest = stamp THEN BEGIN
    err(58);
    typ.BaseTyp := NIL;
    EXIT;
  END;
  base := typ.BaseTyp;
  IF base <> NIL THEN BEGIN
    typ.rectest := stamp;
    CheckStruct(base);
    typ.rectest := 0;
    IF  base.incomplete THEN EXIT;
  END;
  IF base = NIL THEN
    (*ok*)
  ELSE IF base = OPT.undftyp THEN	BEGIN
    typ.BaseTyp := NIL;
    err(999);	(* wrong base type *)
  END ELSE BEGIN
    IF base.form = sfPointer THEN BEGIN
      base := base.BaseTyp;
      typ.BaseTyp := base;
      IF typ.ptr = NIL THEN OPM.Mark(249, typ.txtpos);	(*this record must also be a pointer to record*)
    END;
    OPT.MarkType(base);
    IF base.comp <> csRecord THEN BEGIN
      OPM.Mark(52, typ.txtpos2);
      EXIT;
    END;
    IF (base.sysflag = notag) OR (typ.sysflag = notag) THEN
      OPM.Mark(200, typ.txtpos2);
    typ.extlev := base.extlev+1;
    SetFlags(typ.strobj,GetFlags(base.strobj) * [protectedObj]);
    IF typ.strobj.link2 = NIL THEN
      typ.strobj.link2 := base.strobj.link2;
    t := base;
    WHILE (t <> NIL) AND NOT (hasBody IN GetFlags(t.strobj)) DO t:=t.BaseTyp;
    IF t <> NIL THEN SetFlags(typ.strobj, [isRedef]);
    IF checkOverwriting IN typ.strobj.conval^.setval THEN
      Traverse(typ.strobj.scope.right);
  END;
  typ.incomplete := FALSE;
END;

PROCEDURE CheckArray(typ: OPT.Structs);
VAR
  base: OPT.Structs;
BEGIN
  ASSERT(typ.comp IN ArraySet);
  CheckStruct(typ.BaseTyp);
  base := typ.BaseTyp;
  IF base.incomplete THEN EXIT;
  CASE typ.comp OF
     csStaticArr:       IF base.comp IN [csOpenArr, csDynArr, csSDynArr] THEN OPM.Mark(89, typ.txtpos2);
     csOpenArr, csDynArr:
                        IF base.comp IN ([csSDynArr, csDynArr, csOpenArr]-[typ.comp]) THEN
                          OPM.Mark(89, typ.txtpos2)
                        ELSE IF base.comp = typ.comp THEN
                          typ.n := base.n +1
                        ELSE
                          typ.n := 0;
     csSDynArr:
                        BEGIN
                          IF base.comp IN [csOpenArr, csDynArr] THEN OPM.Mark(89, typ.txtpos2);
                          typ.offset := 0;
                        END;
  END;
  typ.incomplete := FALSE;
END;

PROCEDURE CheckPointer(typ: OPT.Structs);
VAR
  base: OPT.Structs;
BEGIN
  ASSERT(typ.form = sfPointer);
  CheckStruct(typ.BaseTyp);
  base := typ.BaseTyp;
  IF base.incomplete THEN EXIT;
  IF NOT (base.comp IN [csRecord, csStaticArr, csDynArr, csOpenArr]) THEN BEGIN
    typ.BaseTyp := OPT.undftyp;
    OPM.Mark(57, typ.txtpos2);
  END;
  typ.incomplete := FALSE
END;
	
PROCEDURE CheckProcedure(typ: OPT.Structs);
VAR
  par: OPT.Objects;
  quit: BOOLEAN;
BEGIN
  IF typ.BaseTyp = NIL THEN EXIT;
  par := typ.link;
  quit := FALSE;
  WHILE (par<>NIL) AND NOT quit DO BEGIN
    IF par.typ.incomplete THEN BEGIN
      CheckObj(par);
      quit := par.typ.incomplete
    END;
    par := par.link
  END;
  CheckStruct(typ.BaseTyp);
  typ.incomplete := quit OR typ.BaseTyp.incomplete;
END;
	
PROCEDURE CheckStruct(str: OPT.Structs);
BEGIN
  IF NOT str.incomplete OR (str = OPT.ToBeFixed) THEN EXIT;
  IF str.comp = csRecord THEN CheckRecord(str)
  ELSE IF str.comp IN ArraySet THEN CheckArray(str)
  ELSE IF str.form = sfPointer THEN CheckPointer(str)
  ELSE IF str.form = sfProcTyp THEN CheckProcedure(str)
  ELSE OPM.err(200);
END;

PROCEDURE NewType(obj: OPT.Objects);	(*a new type is declared, do an early fix*)
VAR
  i: LONGINT;
  old: SHORTINT;
BEGIN
  i := 0;
  WHILE i < nofTypFix DO BEGIN
    IF (OPT.topScope = TypFix[i].scope) AND (obj.name = TypFix[i].name) THEN BEGIN
      TypFix[i].typ.BaseTyp := obj.typ;
      CheckStruct(TypFix[i].typ);
      INCL(obj.flag, OPT.used);
      IF NOT TypFix[i].typ.incomplete AND (TypFix[i].typ.strobj <> NIL) THEN BEGIN
        CheckObj(TypFix[i].typ.strobj);
        TypFix[i].typ.strobj.mode := ABS(TypFix[i].typ.strobj.mode);
      END;
      TypFix[i].scope := NIL;  TypFix[i].typ := NIL
    END;
    INC(i)
  END;
  WHILE i < nofObjFix DO BEGIN
    IF (OPT.topScope = ObjFix[i].scope) AND (obj.name = ObjFix[i].name) THEN BEGIN
      ObjFix[i].obj.typ := obj.typ;
      CheckObj(ObjFix[i].obj);
      INCL(obj.flag, OPT.used);
      IF NOT obj.typ.incomplete THEN ObjFix[i].obj.mode := ABS(ObjFix[i].obj.mode);
      ObjFix[i].scope := NIL;
      ObjFix[i].obj := NIL;
    END;
    INC(i);
  END;
  i := 0;
  WHILE i < nofCheck DO BEGIN
    IF (ObjCheck[i].obj <> NIL) AND (ObjCheck[i].obj.typ <> NIL) THEN BEGIN
      CheckStruct(ObjCheck[i].obj.typ);
      IF NOT ObjCheck[i].obj.typ.incomplete THEN BEGIN
        old := level;
        level := ObjCheck[i].level;
        ObjCheck[i].obj.mode := ABS(ObjCheck[i].obj.mode);
        CheckObj(ObjCheck[i].obj);
        ObjCheck[i].obj := NIL;
        level := old;
      END
    END;
    INC(i);
  END;
END;

PROCEDURE CheckSysFlag(VAR Scaner: TScaner; VAR sysflag: SHORTINT; default: SHORTINT; allowed: TSET);	(* ejz *)
VAR
  sf: LONGINT;
BEGIN
  sf := default;
  IF sym = _lbrak THEN BEGIN
    IF NOT OPT.SYSimported THEN err(135);
    Scaner.Get(sym);
    IF sym IN [_c,_winapi,_notag,_untraced] THEN BEGIN
      CASE sym OF
        _c        :  sf := cdecl;
        _winapi   :  sf := stdcall;
        _notag    :  sf := notag;
        _untraced :  sf := untraced;
      END;
      IF NOT (sf IN allowed) THEN BEGIN
        sf := default;
        err(200)
      END;
      Scaner.Get(sym)
    END ELSE
      err(115);
    Scaner.CheckSym(SYM,_rbrak);
  END;
  sysflag := sf;
END;

PROCEDURE CheckAndFix(typ: OPT.Structs);
VAR
  lev: SHORTINT;
BEGIN
  WHILE typ.comp = csSDynArr DO BEGIN
    lev := typ.link.mnolev;
    IF lev<>level THEN BEGIN
      typ.link.leaf:=FALSE;
      IF lev>0 THEN OPB.StaticLink(level-lev);
    END;
    typ:=typ.BaseTyp
  END;
END;

PROCEDURE AddTraceCode(owner: OPT.Objects;  VAR statseq: OPT.Nodes);
VAR
  curScope, obj, last: OPT.Objects;
  x: OPT.Nodes;
  name: STRING;
  m: SHORTINT;
BEGIN
  obj:=NIL;
  x:=NIL;
  curScope := OPT.topScope;
  OPT.topScope := OPT.Objects(OPT.modulesx[0]);
  name := 'trace'+owner.name;
  OPT.Insert(name, obj);
  obj.mode := omVar;
  obj.typ := OPT.linttyp;
  last := OPT.topScope.scope;
  IF last <> NIL THEN BEGIN
    WHILE last.link <> NIL DO last := last.link;
    last.link := obj
  END ELSE
    OPT.topScope.scope := obj;
  OPT.topScope := curScope;
  m := Oinc.adr;
  x := OPB.NewLeaf(obj);
  OPB.StPar0(x, m);
  OPB.StFct(x, m, 1);
  x.conval := OPT.NewConst();
  x.conval^.intval := OPM.errpos.CodePos;
  x.link := statseq;
  statseq := x
END;

PROCEDURE InsertFix(mode: LONGINT; stream: OPS.TokenStream; node: OPT.Nodes);
VAR
  fix: FixPtr;
BEGIN
  fix:=FixPtr.Create;
  fix.mode := mode;
  fix.stream := stream;
  fix.scope := OPT.topScope;
  fix.level := level;
  fix.node := node;
  fix.next := NIL;
  IF FixLast = NIL THEN FixList := fix	ELSE FixLast.next := fix;
  FixLast := fix;
  node.link := NIL
END;

PROCEDURE Await(VAR Scaner: TScaner; VAR x: OPT.Nodes; VAR n: INTEGER);
VAR
  proc, Aself: OPT.Objects;
  exp, cond: OPT.Nodes;
  name: STRING;
  ts: OPS.TokenStream;
BEGIN
  proc:=NIL;
  cond:=NIL;
  Aself := GetSelf();
  exp:=NIL;
  IF sym = _lparen THEN BEGIN
    n := 1;
    Scaner.Get(sym);
    IF OPT.topScope.link = NIL THEN
      name:='$$'
    ELSE
      name:=OPT.topScope.link.name;
    name := name+'.@';
    name := name+CHR((CondCount DIV 100) MOD 10 + ORD('0'));
    name := name+CHR((CondCount DIV 10) MOD 10 + ORD('0'));
    name := name+CHR(CondCount MOD 10 + ORD('0'));
    INC(CondCount);
    OPT.Insert(name, proc);
    proc.mode := omXProc;
    proc.link := NIL;
    proc.vis := internal;
    proc.adr := 0;
    proc.typ := OPT.booltyp;
    proc.conval := OPT.NewConst();
    proc.conval^.setval := [hasBody, slNeeded];
    proc.conval^.intval := 12;
    INC(level);
    OPT.OpenScope(level, proc);
    Expression(Scaner,x);
    IF UndefFlag THEN BEGIN
      exp := x;
      proc.typ := x.typ;
      OPB.Return(x, proc);
      proc.typ := OPT.booltyp
    END ELSE BEGIN
      OPB.StPar0(x, awaitfn);
      OPB.Return(x, proc);
    END;
    OPT.CloseScope;
    DEC(level);
    x.conval := OPT.NewConst();
    x.conval^.intval := OPM.errpos.CodePos;
    cond := NIL;
    OPB.Enter(cond, x, proc);
    cond.link := CondNodes;
    CondNodes := cond;
    cond.conval := OPT.NewConst();
    cond.conval^.intval := OPM.errpos.CodePos;
    x := OPB.NewLeaf(proc);
    x.link := OPB.NewLeaf(Aself);
    Scaner.CheckSym(SYM,_rparen);
  END ELSE
    err (65);
  IF AwaitIF THEN
    x.link.link := OPB.NewIntConst(0)
  ELSE
    x.link.link := OPB.NewIntConst(2);
  IF UndefFlag THEN BEGIN
    UndefFlag := FALSE;
    Scaner.StopRecording(ts);
    InsertFix(FPassExpr, ts, exp);
  END
END;

//#######################################################################

PROCEDURE FType(VAR Scaner: TScaner; VAR typ, banned: OPT.Structs); FORWARD;

PROCEDURE InlineProc(VAR x: OPT.Nodes; VAR apar: OPT.Nodes; fpar: OPT.Objects);			(* Inline *)
VAR
  inlines, par, list: OPT.Objects;
  glueCode, src, trg: OPT.Nodes;
  assembly: BOOLEAN;
  c: OPT.Consts;
		
  PROCEDURE CopyLocals(obj: OPT.Objects);
  VAR
    new: OPT.Objects;
    name: STRING;
    i, j: LONGINT;
  BEGIN
    IF obj <> NIL THEN BEGIN
      IF (obj.mode = omVar) OR (assembly AND (obj.mode = omVarPar)) THEN BEGIN
        i := 0;
        WHILE inlines.name[i] <> #0 DO BEGIN
          name[i] := inlines.name[i];
          INC(i)
        END;
        name[i] := '.';
        INC(i);
        j := 0;
        WHILE obj.name[j] <> #0 DO BEGIN
          name[i] := obj.name[j];
          INC(j);
          INC(i)
        END;
        name[i] := #0;
        OPT.Insert(name, new);
        new.flag := [OPT.used];
        par.link := new;
        par := new;
        new.leaf := FALSE;
        new.mode := obj.mode;
        new.vis := internal;
        new.typ := obj.typ;
        obj.link2 := new;
        IF assembly THEN new.link2 := obj;
      END;
      CopyLocals(obj.left);		(* pre-order traversing --> prevents a degenerated list *)
      CopyLocals(obj.right)
    END
  END;
			
BEGIN
  par:=NIL;
  trg:=NIL;

    inlines := x.obj;
    assembly := asmProc IN inlines.conval^.setval;
    IF NOT OPM.noerr THEN BEGIN
      IF assembly THEN BEGIN
        x.clas := Ncall;
        x.typ := x.obj.typ;
        x.subcl := 2;
      END;
      EXIT;
    END;
    IF NOT (hasBody IN inlines.conval^.setval) THEN BEGIN
      err(129);
      EXIT;
    END;

    IF NOT assembly AND (inlines.link2 <> OPT.topScope) THEN BEGIN
      par := OPT.topScope.scope;
      IF par <> NIL THEN BEGIN
        WHILE par.link <> NIL DO par := par.link;
        CopyLocals(inlines.scope.right)
      END ELSE BEGIN
        list:=Objects.Create;
        par := list;
        CopyLocals(inlines.scope.right);
        IF OPT.topScope.scope <> NIL THEN OPT.topScope.scope := list.link;
      END;
      inlines.link2 := OPT.topScope;
    END;

    IF assembly THEN BEGIN
      OPB.Call(x, apar, fpar);
      OPB.CopyAndSubst(x.left, inlines.code);
      x.subcl := 2;
    END ELSE BEGIN
      x := NIL;
      glueCode := NIL;
      c := OPT.NewConst();
      c^.intval := OPM.errpos.CodePos;
      WHILE fpar <> NIL DO BEGIN
        IF fpar.mode = omVarPar THEN BEGIN
          fpar.link2 := apar.obj;
          apar := apar.link
        END ELSE BEGIN
            src := apar;
            apar := apar.link;
            src.link := NIL;
            trg := OPB.NewLeaf(fpar.link2);
            OPB.Assign(trg, src);
            trg.conval^.intval := c^.intval;
            IF glueCode = NIL THEN x := trg ELSE glueCode.link := trg;
            glueCode := trg
        END;
        fpar := fpar.link
      END;
      IF glueCode = NIL THEN
        OPB.CopyAndSubst(x, inlines.code)
      ELSE
        OPB.CopyAndSubst(glueCode.link, inlines.code);
    END
END;

PROCEDURE CaseLabelList(VAR Scaner: TScaner; VAR lab: OPT.Nodes; LabelForm: INTEGER; VAR n: INTEGER; VAR tab: CaseTable);
VAR
  x, y, lastlab: OPT.Nodes;
  i, f: INTEGER;
  xval, yval: LONGINT;
BEGIN
  lab := NIL;
  lastlab := NIL;
  x:=NIL;
  y:=NIL;
  REPEAT
    ConstExpression(Scaner,x);
    f := x.typ.form;
    IF f IN intSet + [sfChar] THEN
      xval := x.conval^.intval
    ELSE BEGIN
      err(61);
      xval := 1
    END ;
    IF f IN intSet THEN BEGIN
      IF LabelForm < f THEN err(60);
    END ELSE IF LabelForm <> f THEN
      err(60);
    IF sym = _upto THEN BEGIN
      Scaner.Get(sym);
      ConstExpression(Scaner,y);
      yval := y.conval^.intval;
      IF (y.typ.form <> f) AND NOT ((f IN intSet) AND (y.typ.form IN intSet)) THEN err(60);
      IF yval < xval THEN BEGIN
        err(63);
        yval := xval
      END;
    END ELSE
      yval := xval;
    x.conval^.intval2 := yval;

    i := n;
    IF i < OPM.MaxCases THEN BEGIN
      REPEAT
        IF i = 0 THEN BREAK;
        IF tab[i-1].low <= yval THEN BEGIN
          IF tab[i-1].high >= xval THEN err(62);
          BREAK;
        END;
        tab[i] := tab[i-1];
        DEC(i)
      UNTIL FALSE;
      tab[i].low := xval;
      tab[i].high := yval;
      INC(n)
    END ELSE
      err(213);
    OPB.Link(lab, lastlab, x);
    IF sym = _comma THEN Scaner.Get(sym)
    ELSE IF (sym = _number) OR (sym = _ident) THEN err(ORD(_comma))
    ELSE BREAK;
  UNTIL FALSE;
END;

PROCEDURE ArrayType(VAR Scaner: TScaner; VAR typ, banned: OPT.Structs);
VAR
  x: OPT.Nodes;
  n: LONGINT;
  sysflag: SHORTINT;

  FUNCTION HiddenVar (name: STRING; exp: OPT.Nodes): OPT.Objects;
  VAR
    dim, owner: OPT.Objects;
    y: OPT.Nodes;
  BEGIN
    owner:=OPT.topScope;
    WHILE (OPT.topScope.left<>NIL) AND (OPT.topScope.link=OPT.topScope.left.link) DO
      OPT.topScope:=OPT.topScope.left;
    OPT.Insert('@@', dim);
    dim.name:=name;
    dim.link:=NIL;
    dim.mode:=omVar;
    dim.typ:=OPT.linttyp;
    dim.flag := [OPT.used];
    IF OPT.topScope.scope=NIL THEN
      OPT.topScope.scope:=dim
    ELSE BEGIN
      dim.link:=OPT.topScope.scope;
      OPT.topScope.scope:=dim;
    END;
    OPT.topScope:=owner;
    y:=OPB.NewLeaf(dim);
    OPB.Assign(y, exp);
    y.conval := OPT.NewConst();
    y.conval^.intval := OPM.errpos.CodePos;
    OPB.CodeInsert(y);
    RESULT:= dim;
  END;

BEGIN
  x:=NIL;
  CheckSysFlag(Scaner,sysflag, 0, [0, notag]);
  IF sym = _times THEN BEGIN
    typ := OPT.NewStr(sfComp, csDynArr);
    typ.sysflag := sysflag;
    Scaner.Get(sym);
    IF (OPM.oberon1 IN OPM.options) THEN err(271);
  END ELSE IF sym <> _of THEN BEGIN
    Expression(Scaner,x);
    IF x.clas = Nconst THEN BEGIN
      typ := OPT.NewStr(sfComp, csStaticArr);
      typ.sysflag := sysflag;
      IF x.typ.form IN intSet THEN BEGIN
        n := x.conval^.intval;
        IF (n <= 0) THEN BEGIN
          err(63);
          n := 1;
        END;
      END ELSE BEGIN
        err(51);
        n := 1;
      END;
      typ.n := n;
      typ.link:=NIL;
    END ELSE BEGIN
      typ := OPT.NewStr(sfComp, csSDynArr);
      typ.sysflag := sysflag;
      typ.n:=0;
      IF (x.typ = NIL) OR NOT (x.typ.form IN intSet) THEN BEGIN
        typ.comp := csStaticArr;
        err(68);
      END ELSE
        typ.link:=HiddenVar('@dim', x);
      IF (OPM.oberon1 IN OPM.options) THEN err(271);
    END
  END ELSE BEGIN
    typ := OPT.NewStr(sfComp, csOpenArr);
    typ.sysflag := sysflag;
  END;
  typ.incomplete := TRUE;
  IF sym = _comma THEN BEGIN
    Scaner.Get(sym);
    typ.errpos := OPM.errpos;
    ArrayType(Scaner,typ.BaseTyp, banned);
  END ELSE IF sym = _of THEN BEGIN
    Scaner.Get(sym);
    typ.errpos := OPM.errpos;
    FType(Scaner,typ.BaseTyp, banned);
    IF typ.BaseTyp = OPT.ToBeFixed THEN BEGIN
      NewFixS(TypFix[nofTypFix],Scaner.name);
      TypFix[nofTypFix].typ := typ;
      INC(nofTypFix);
    END
  END;
  CheckArray(typ);
END;

PROCEDURE RecordType(VAR Scaner: TScaner; VAR procdec: OPT.Nodes; VAR typ, banned: OPT.Structs; VAR owner: OPT.Objects);
VAR
  base: OPT.Objects;
  sysflag: SHORTINT;
  c: INTEGER;
  statseq: OPT.Nodes;
  lev: SHORTINT;
BEGIN
  base:=NIL;
  statseq:=NIL;
  typ := OPT.NewStr(sfComp, csRecord);
  typ.BaseTyp := NIL;
  IF ptrToRec <> NIL THEN BEGIN	(*POINTER TO RECORD*)
    typ.ptr := ptrToRec;
    ptrToRec := NIL;
  END;
  IF owner = NIL THEN	BEGIN
    owner := OPT.NewObj();
    owner.mode := omTyp;
    owner.name := '';
    owner.vis := internal;
    owner.typ := typ
  END;
  owner.conval := OPT.NewConst();
  typ.strobj := owner;
  CheckSysFlag(Scaner,sysflag, 0, [0, notag]);
  typ.incomplete := TRUE;
  IF sym = _lparen THEN BEGIN
    Scaner.Get(sym);
    IF sym = _ident THEN BEGIN
      typ.errpos := OPM.errpos;
      Qualident(Scaner,base);
      IF (base <> UndefObj) AND (base.mode = omTyp) THEN BEGIN
        IF base.typ = banned THEN
          err(58)
        ELSE
          typ.BaseTyp := base.typ;
      END ELSE IF base = UndefObj THEN BEGIN
        UndefFlag := FALSE;
        IF sym = _period THEN BEGIN
          OPT.FindInScope(Scaner.name, ExtObj, base);
          Scaner.Get(sym);
          IF (base <> NIL) AND (base.mode = omMod) THEN OPT.FindImport (Scaner.Name, base, base);
          Scaner.CheckSym(SYM,_ident);
          IF base = NIL THEN err (0)
          ELSE IF base.mode <> omTyp THEN  err(52)
          ELSE IF base.typ = banned THEN  err(58)
          ELSE  typ.BaseTyp := base.typ;
        END ELSE BEGIN
          typ.BaseTyp := OPT.ToBeFixed;
          NewFixS(TypFix[nofTypFix], Scaner.name);
          TypFix[nofTypFix].typ := typ;
          INC(nofTypFix);
        END;
      END ELSE
        err(52);
    END ELSE
      err(ORD(_ident));
    Scaner.CheckSym(SYM,_rparen);
  END;
  typ.sysflag := sysflag;
  CheckRecord(typ);
  OPT.OpenScope(0, owner);
  lev := level;
  level := 0;
  c := OPM.errpos.CodePos;
  IF sym = _semicolon THEN Scaner.Get(sym);
  Block(Scaner,procdec, statseq);
  IF typ.link = NIL THEN typ.link := OPT.topScope.right;
  ASSERT((statseq = NIL) OR (hasBody IN owner.conval^.setval));		(*stat#NIL  ==> hasBody*)
  IF (procdec <> NIL) OR (hasBody IN owner.conval^.setval) OR (statseq <> NIL) THEN BEGIN
    OPB.Enter(CondNodes, statseq, owner);
    CondNodes.link := procdec;
    procdec := CondNodes;
    procdec.conval := OPT.NewConst();
    procdec.conval^.intval := c;
  END;
  level := lev;
  OPT.CloseScope;
END;

PROCEDURE PointerType(VAR Scaner: TScaner; VAR typ, banned: OPT.Structs);
VAR
  id: OPT.Objects;
BEGIN
  id:=NIL;
  typ := OPT.NewStr(sfPointer, csBasic);
  CheckSysFlag(Scaner,typ.sysflag, 0, [0]);
  typ.incomplete := TRUE;
  Scaner.CheckSym(SYM,_to);
  IF sym = _ident THEN BEGIN
    OPT.Find(Scaner.Name,id);
    typ.errpos := OPM.errpos;
    IF id = NIL THEN BEGIN
      NewFixS(TypFix[nofTypFix], Scaner.name);
      TypFix[nofTypFix].typ := typ;
      INC(nofTypFix);
      typ.BaseTyp := OPT.ToBeFixed;
      Scaner.Get(sym)
    END ELSE BEGIN
      Qualident(Scaner,id);
      IF id.mode = omTyp THEN
        typ.BaseTyp := id.typ
      ELSE BEGIN
        typ.BaseTyp := OPT.undftyp;
        err(52);
      END;
    END;
  END ELSE BEGIN
    IF sym= _record THEN ptrToRec := typ;
    typ.errpos := OPM.errpos;
    FType(Scaner,typ.BaseTyp, banned);
    IF typ.BaseTyp = OPT.ToBeFixed THEN BEGIN
      NewFixS(TypFix[nofTypFix], Scaner.name);
      TypFix[nofTypFix].typ := typ;
      INC(nofTypFix)
    END;
    CheckPointer(typ);
  END;
END;

PROCEDURE ObjectType(VAR Scaner: TScaner; VAR typ, banned: OPT.Structs);
VAR
  x: OPT.Nodes;
  link: OPT.Objects;
BEGIN
  x:=NIL;
  link:=NIL;
  typ := OPT.NewStr(sfPointer, csBasic);
  typ.incomplete := TRUE;
  ptrToRec := typ;
  typ.errpos := OPM.errpos;
  RecordType(Scaner,x, typ.BaseTyp, banned, link);
  OPB.Inittd(TDinit, lastTDinit, typ);
  IF x <> NIL THEN OPB.Link(RecInit, lastRecInit, x);
  IF typ.BaseTyp = OPT.ToBeFixed THEN BEGIN
    NewFixS(TypFix[nofTypFix], Scaner.name);
    TypFix[nofTypFix].typ := typ;
    INC(nofTypFix);
  END;
  CheckPointer(typ);
END;

PROCEDURE ActualParameters(VAR Scaner: TScaner; VAR aparlist: OPT.Nodes; fpar: OPT.Objects; x: OPT.Nodes);
VAR
  apar, last, aparret, lastret: OPT.Nodes;
  b: BOOLEAN;
BEGIN
  apar:=NIL;
  aparlist := NIL;
  last := NIL;
  b := UndefFlag;
  lastret:=NIL;
  aparret:=NIL;
  IF sym <> _rparen THEN BEGIN
    REPEAT
      Expression(Scaner,apar);
      IF fpar <> NIL THEN BEGIN
        UndefFlag := UndefFlag OR (fpar.mode<0);
        IF NOT UndefFlag AND NOT b THEN BEGIN
          OPB.FNParam(apar, fpar);
          OPB.Link(aparlist, last, apar)
        END;
        fpar := fpar.link;
      END ELSE IF NOT UndefFlag AND NOT b THEN
        err(64);
      IF sym = _comma THEN
        Scaner.Get(sym)
      ELSE IF (_lparen <= sym) AND (sym <= _ident) THEN
        err(ORD(_comma))
      ELSE
        BREAK;
    UNTIL FALSE;
    IF aparret<>NIL THEN BEGIN
      lastret.link:=aparlist;
      aparlist:=aparret
    END;
  END;
  IF (fpar <> NIL) AND NOT UndefFlag AND NOT b THEN err(65);
  UndefFlag := UndefFlag OR b;
END;

PROCEDURE FormalParameters(VAR Scaner: TScaner; VAR firstPar: OPT.Objects; VAR resTyp: OPT.Structs;  sysflag: SHORTINT);
VAR
  mode: SHORTINT;
  par, first, last, res: OPT.Objects;
  typ, btyp: OPT.Structs;
  conPar: BOOLEAN; //!!!
BEGIN
  par := NIL;
  res := NIL;
  typ := NIL;
  first := NIL;
  last := firstPar;
  IF (sym = _ident) OR (sym = _var) OR (sym = _const) THEN BEGIN
    REPEAT
      conPar:=FALSE;
      IF (sym = _var) OR (sym = _const) THEN BEGIN
        conPar:= sym=_const;
        Scaner.Get(sym);
        mode := omVarPar;
      END ELSE
        mode := omVar;
      REPEAT
        IF sym = _ident THEN BEGIN
          OPT.Insert(Scaner.name, par);
          Scaner.Get(sym);
          IF conPar THEN BEGIN
            par.vis := internalR; //!!!!!
            mode := omVar;
          END;
          par.flag := [OPT.param];
          par.mode := mode;
          par.link := NIL;
          IF first = NIL THEN first := par;
          IF firstPar = NIL THEN firstPar := par ELSE last.link := par;
          last := par;
        END ELSE
          err(ORD(_ident));
        IF sym = _comma THEN Scaner.Get(sym)
        ELSE IF sym = _ident THEN err(ORD(_comma))
        ELSE IF sym = _var THEN BEGIN
          err(ORD(_comma));
          Scaner.Get(sym)
        END ELSE
          BREAK;
      UNTIL FALSE;
      Scaner.CheckSym(SYM,_colon);

      FType(Scaner,typ, OPT.notyp);
      
      IF NOT typ.incomplete AND (typ.strobj = NIL) AND (typ.comp <> csOpenArr) THEN err(-305);

      IF sysflag <> 0 THEN BEGIN
        IF (typ.comp IN [csOpenArr, csRecord]) AND (typ.sysflag <> notag) THEN err(200);
        IF typ.comp IN [csDynArr, csSDynArr] THEN err(200);
      END;

      IF mode = omVar THEN OPT.MarkType(typ);
      WHILE first<> NIL DO BEGIN
        IF typ = OPT.ToBeFixed THEN	BEGIN
          NewFixO(ObjFix[nofObjFix], Scaner.name);
          ObjFix[nofObjFix].obj := first;
          INC(nofObjFix);
        END;
        IF typ.incomplete THEN BEGIN
          NewCheck(first);
          first.mode := -first.mode;
        END;
        first.typ := typ;
        first := first.link
      END ;
      IF sym = _semicolon THEN Scaner.Get(sym)
      ELSE IF sym = _ident THEN err(ORD(_semicolon))
      ELSE
        BREAK;
    UNTIL FALSE;
  END ;
  Scaner.CheckSym(SYM,_rparen);
  IF sym = _colon THEN BEGIN
    Scaner.Get(sym);
    resTyp := OPT.undftyp;
    typ := NIL;
    WHILE sym = _array DO BEGIN
      Scaner.Get(sym);
      Scaner.CheckSym(SYM,_of);
      btyp := typ;
      typ := OPT.NewStr(sfComp, csOpenArr);
      typ.incomplete := TRUE;
      typ.BaseTyp := btyp;
      IF btyp = NIL THEN resTyp := typ;
    END;
    IF sym = _ident THEN BEGIN
      Qualident(Scaner,res);
      IF res = UndefObj THEN BEGIN
        IF typ <> NIL THEN BEGIN
          NewFixS(TypFix[nofTypFix], Scaner.name);
          TypFix[nofTypFix].typ := resTyp;
          INC(nofTypFix);
          resTyp := typ
        END ELSE
          resTyp := NIL;
        IF (sysflag <> 0) THEN err(200);
      END ELSE IF res.mode = omTyp THEN BEGIN
        IF typ<>NIL THEN  BEGIN
          resTyp.BaseTyp := res.typ;
          resTyp := typ;
          CheckStruct(typ);
        END ELSE
          resTyp := res.typ;
        IF (sysflag <> 0) AND (resTyp.form = sfComp) THEN err(200);
      END ELSE
        err(52);
    END ELSE
      err(ORD(_ident));
  END ELSE
    resTyp := OPT.notyp;
END;

PROCEDURE TypeDecl(VAR Scaner: TScaner; VAR x: OPT.Nodes; VAR typ, banned: OPT.Structs; VAR owner: OPT.Objects);
VAR
  id: OPT.Objects;
  name: STRING;
BEGIN
  typ := OPT.undftyp;
  x := NIL;
  id:=NIL;
  IF sym < _lparen THEN BEGIN
    err(12);
    REPEAT Scaner.Get(sym) UNTIL sym >= _lparen;
  END ;
  name:=Scaner.name;
  IF sym = _ident THEN BEGIN
    Qualident(Scaner,id);
    name:=Scaner.name;
    IF id = UndefObj THEN BEGIN
      typ := OPT.undftyp;
      UndefFlag := FALSE;
      IF (sym = _period) THEN BEGIN
        OPT.FindInScope(name, ExtObj, id);
        IF (id <> NIL) AND (id.mode = omMod) THEN BEGIN
          Scaner.Get(sym);
          OPT.FindImport(Scaner.Name, id, id);
          Scaner.Get(sym);
        END;
        IF id = NIL THEN BEGIN
          err (0);
          typ := OPT.undftyp;
        END ELSE IF id.mode = omTyp THEN BEGIN
          IF id.typ <> banned THEN
            typ := id.typ
          ELSE
            err(58);
        END ELSE BEGIN
          err (52);
          typ := OPT.undftyp;
        END;
      END ELSE
        typ := OPT.ToBeFixed;	(*will be fixed by the caller!!! Name is in OPS.name stored*)
    END ELSE IF id.mode = omTyp THEN BEGIN
      IF id.typ <> banned THEN
        typ := id.typ
      ELSE
        err(58);
    END ELSE
      err(52);
    IF id<>NIL THEN CheckAndFix(id.typ);
  END ELSE IF sym = _array THEN BEGIN
    Scaner.Get(sym);
    ArrayType(Scaner,typ, banned);
  END ELSE IF sym = _record THEN BEGIN
    Scaner.Get(sym);
    RecordType(Scaner, x, typ, banned, owner);
    OPB.Inittd(TDinit, lastTDinit, typ);
  END ELSE IF sym = _pointer THEN BEGIN
    Scaner.Get(sym);
    PointerType(Scaner,typ, banned);
  END ELSE IF sym = _object THEN BEGIN
    Scaner.Get(sym);
    ObjectType(Scaner,typ, banned);
  END ELSE IF sym = _procedure THEN BEGIN
    Scaner.Get(sym);
    typ := OPT.NewStr(sfProcTyp, csBasic);
    CheckSysFlag(Scaner,typ.sysflag, 0, [0, stdcall, cdecl]);
    IF sym = _lparen THEN BEGIN
      Scaner.Get(sym);
      OPT.OpenScope(level, NIL);
      FormalParameters(Scaner,typ.link, typ.BaseTyp, typ.sysflag);
      OPT.CloseScope;
      id := typ.link;
      WHILE (id<>NIL) AND NOT id.typ.incomplete DO id := id.link;
      typ.incomplete := (id<>NIL) AND id.typ.incomplete;
      IF typ.BaseTyp = NIL THEN BEGIN
        NewFixS(TypFix[nofTypFix], Scaner.name);
        TypFix[nofTypFix].typ := typ;
        INC(nofTypFix);
        typ.incomplete := TRUE;
      END;
      UndefFlag := FALSE
    END ELSE BEGIN
      typ.BaseTyp := OPT.notyp;
      typ.link := NIL
    END;
  END ELSE
    err(12);
  REPEAT
    IF (sym >= _semicolon) AND (sym <= _else)
    OR (sym = _rparen) OR (sym = _eof) OR (sym = _ident) THEN BREAK;
    err(15);
    IF sym = _ident THEN BREAK;
    Scaner.Get(sym)
  UNTIL FALSE;
END;

PROCEDURE FType(VAR Scaner: TScaner; VAR typ, banned: OPT.Structs);
VAR
  x: OPT.Nodes;
  link: OPT.Objects;
BEGIN
  x := NIL;
  link:=NIL;
  TypeDecl(Scaner,x, typ, banned, link);
  IF x <> NIL THEN OPB.Link(RecInit, lastRecInit, x);
END;

PROCEDURE ProcedureDeclaration(VAR Scaner: TScaner; VAR x: OPT.Nodes);
VAR
  proc, fwd: OPT.Objects;
  name: STRING;
  sysflag, mode, vis, prio: SHORTINT;
  forward, init: BOOLEAN;
  flags: TSET;
  scope: OPT.Objects;
  redef: BOOLEAN;
  selfname: STRING;
  selftype, selfrec: OPT.Structs;
  selfmode: SHORTINT;

  PROCEDURE GetParam;
  BEGIN
    proc.typ := OPT.notyp;
    proc.link := NIL;
    proc.sysflag := sysflag;
    IF mode = omTProc THEN BEGIN
      OPT.Insert (selfname, proc.link); 
      proc.link.mode := selfmode;
      proc.link.typ := selftype;
    END;
    IF sym = _lparen THEN BEGIN
      Scaner.Get(sym);
      IF mode <> omTProc THEN
        FormalParameters(Scaner,proc.link, proc.typ, sysflag)
      ELSE
        FormalParameters(Scaner,proc.link.link, proc.typ, sysflag);
      IF proc.typ = NIL THEN BEGIN
        NewFixO(ObjFix[nofObjFix], Scaner.name);
        ObjFix[nofObjFix].obj := proc;
        INC(nofObjFix);
      END;
      IF (proc.typ = NIL) OR (proc.typ.incomplete) THEN  NewCheck(proc);
    END ELSE IF prio=126 THEN
      err(30);
    IF prio=126 THEN BEGIN
      IF proc.name=':=' THEN BEGIN
        IF proc.typ<>OPT.notyp THEN err(147);
        IF (proc.link<>NIL) AND (proc.link.mode<>omVarPar) THEN err(148);
      END ELSE IF proc.typ=OPT.notyp THEN
        err(141);
      IF NOT forward THEN BEGIN
        OPT.Contextualize(proc);
        proc.prio:=125;
        OPT.topScope.link:=proc
      END ELSE
        err(145);
    END;
    IF (fwd <> NIL) AND (prio<126) THEN BEGIN
      IF (mode = omTProc) AND redef THEN BEGIN
        proc.flag:=proc.flag+[OPT.used];
        OPB.CheckParameters (proc.link.link, fwd.link.link, FALSE);
        IF proc.link.mode <> fwd.link.mode THEN err(115);
      END ELSE
        OPB.CheckParameters (proc.link, fwd.link, TRUE);
      IF proc.typ <> fwd.typ THEN err(117);
      IF NOT redef THEN BEGIN
        proc := fwd;
        OPT.topScope := fwd.scope
      END ELSE
        proc.conval^.setval:=proc.conval^.setval+[isRedef];
    END;
    IF init AND (proc.typ <> OPT.notyp) THEN err(134);
    IF init AND (OPT.topScope.left.link.mode = omMod) AND (proc.link <> NIL) THEN err(133);
    IF (OPM.oberon1 IN OPM.options) AND (proc.typ <> NIL) AND (proc.typ.form = sfComp) THEN err(270);
  END;

  PROCEDURE Body;
  VAR
    c: INTEGER;
    n: LONGINT;
    code    : ARRAY [0..OPS.MaxStrLen] OF CHAR;
    procdec, statseq: OPT.Nodes;
  BEGIN
    procdec:=NIL;
    statseq:=NIL;
    c := OPM.errpos.CodePos;
    proc.conval^.setval:=proc.conval^.setval+[hasBody];
    IF (sym = _number) THEN BEGIN
      IF NOT OPT.SYSimported THEN err(135);
      statseq := OPT.NewNode(Ncode);
      statseq.conval := OPT.NewConst();
      procdec := NIL;
      n := 0;
      FILLCHAR(code,LENGTH(code),#0);
      {
      statseq.conval^.ext := ext;
      OPB.Construct (Nassembler, statseq, NIL);
      statseq.conval := OPT.NewConst();
      statseq.conval^.errpos := errpos;
      statseq.obj := proc;
      proc.conval^.setval:=proc.conval^.setval+[asmProc];
      proc.conval^.ext := ext;
      }
      REPEAT
        IF sym = _number THEN BEGIN
          INC(n);
          IF (Scaner.intval < 0) OR (Scaner.intval > 255) OR (n = OPT.MaxConstLen) THEN BEGIN
            err(63);
            Scaner.intval := 1;
            n := 1
          END ;
          code[n]:=CHR(Scaner.intval);
          Scaner.Get(sym);
        END ;
        IF sym = _comma THEN Scaner.Get(sym)
        ELSE IF sym = _number THEN err(ORD(_comma))
        ELSE BEGIN
          code[0] := CHR(n);
          BREAK
        END;
      UNTIL FALSE;

      MOVE(code[0],statseq.conval^.code[0],LENGTH(code));
      OPB.Construct (Nassembler, statseq, NIL);
      statseq.conval := OPT.NewConst();
      statseq.conval^.intval := errpos.CodePos;
      statseq.obj := proc;
      proc.conval^.setval:=proc.conval^.setval+[asmProc];
      MOVE(code[0],proc.conval^.code[0],LENGTH(code));

    END ELSE BEGIN
      Scaner.CheckSym(SYM,_semicolon);
      Block(Scaner,procdec, statseq);
      IF (sym=_number) OR (sym=_string) THEN BEGIN
        Scaner.name:=Scaner.str;
        sym := _ident;
        proc.prio := prio
      END;
      IF sym = _ident THEN BEGIN
        IF Scaner.name <> proc.name THEN err(4);
        Scaner.Get(sym)
      END ELSE
        err(ORD(_ident));
    END;
    IF mode = omCProc THEN BEGIN
      IF NOT (asmProc IN proc.conval^.setval) AND ((proc.typ <> OPT.notyp) OR (procdec <> NIL)) THEN err(200);
      proc.code := statseq;
      procdec := NIL
    END ELSE BEGIN
      OPB.Enter(procdec, statseq, proc);
      procdec.conval := OPT.NewConst();
      procdec.conval^.intval := c;
    END;
    x := procdec;
  END;
		
  PROCEDURE Receiver;
  VAR
    obj: OPT.Objects;
    rec: OPT.Structs;
  BEGIN
    obj:=NIL;
    IF IsObjectScope() THEN BEGIN
      mode := omTProc;
      rec := OPT.topScope.link.typ;
      ASSERT(rec.comp = csRecord);
      IF OPM.OptimizeSelf AND (rec.ptr <> NIL) THEN BEGIN
        ASSERT(rec.ptr.form = sfPointer);
        selfmode := omVar;
        selfname := HiddenSelf;
        selftype := rec.ptr;
        selfrec := rec
      END ELSE BEGIN
        selfmode := omVarPar;
        selfname := HiddenSelf;
        selftype := rec;
        selfrec := rec;
      END;
      IF (OPM.oberon1 IN OPM.options) THEN err(261);
    END ELSE IF (sym = _lparen) AND (OPM.oberon2 IN OPM.options) THEN BEGIN
      mode := omTProc;
      Scaner.Get(sym);
      IF sym = _var THEN BEGIN
        selfmode := omVarPar;
        Scaner.Get(sym)
      END ELSE
        selfmode := omVar;
      selfname := Scaner.name;
      Scaner.CheckSym(SYM,_ident);
      Scaner.CheckSym(SYM,_colon);
      IF sym = _ident THEN BEGIN
        OPT.Find (Scaner.Name,obj);
        Scaner.Get(sym);
        IF obj = NIL THEN BEGIN
          err(0);
          selftype := OPT.undftyp;
        END ELSE IF obj.mode <> omTyp THEN BEGIN
          err(72);
          selftype := OPT.undftyp;
        END ELSE BEGIN
          selftype := obj.typ;
          selfrec := selftype;
          IF selftype.form = sfPointer THEN selfrec := selfrec.BaseTyp;
          IF NOT ((selfmode = omVar) AND (selftype.form = sfPointer) AND (selfrec.comp = csRecord) OR
            (selfmode = omVarPar) AND (selftype.comp = csRecord)) THEN err(70);
          IF (selfrec <> NIL) AND (selfrec.mno <> level) THEN err(72);
          IF selfrec.comp = csRecord THEN OPT.topScope := selfrec.strobj.scope;
        END
      END ELSE
        err(ORD(_ident));
      Scaner.CheckSym(sym,_rparen)
    END
  END;

BEGIN

  forward := FALSE;
  scope := OPT.topScope;
  proc := NIL;
  x := NIL;
  mode := omLProc;
  init := FALSE;
  flags := [];
  prio := 0;
  sysflag:=0;

  fwd:= NIL;
  selftype:= NIL;
  selfrec:= NIL;

  IF sym = _arrow THEN BEGIN
    forward := TRUE;
    Scaner.Get(sym)
  END;
  Receiver;
		
  (* special modes *)
  IF (sym <> _ident) AND (sym <> _string) AND (sym <> _number) THEN BEGIN
    IF sym = _lbrak THEN BEGIN
      IF mode = omTProc THEN
        CheckSysFlag(Scaner,sysflag, 0, [0])
      ELSE
        CheckSysFlag(Scaner,sysflag, 0, [0, stdcall, cdecl]);
    END;
    IF sym = _times THEN
      Scaner.Get (sym)
    ELSE IF sym = _plus THEN BEGIN
      IF mode = omTProc THEN err(47);
      mode := omIProc;
      IF NOT OPT.SYSimported THEN err(135);
      Scaner.Get (sym)
    END ELSE IF sym = _and THEN BEGIN
      init := TRUE;
      IF NOT IsObjectScope() THEN err(253);
      IF (OPT.topScope.link.link2 <> NIL) AND (OPT.topScope.link.link2.link.typ.strobj.scope = OPT.topScope) THEN err(144);
      IF (OPT.topScope.link.typ.ptr = NIL) THEN err(249);
      Scaner.Get (sym)
    END ELSE IF sym = _minus THEN BEGIN
      IF mode = omTProc THEN err(47);
      mode := omCProc;
      Scaner.Get(sym);
    END ELSE IF sysflag = 0 THEN
      err(ORD(_ident));
  END;

  IF (sym=_number) OR (sym=_string) THEN BEGIN
    Scaner.CheckOperator(prio);
    IF prio=126 THEN BEGIN
      sym:=_ident;
      IF IsObjectScope() THEN
        err(140)
      ELSE IF NOT forward AND (operatorFlag IN OPT.topScope.conval^.setval) THEN
        err(-212);
    END ELSE
      err(142);
  END ELSE IF NOT IsObjectScope() THEN
    OPT.topScope.conval^.setval:=OPT.topScope.conval^.setval+[operatorFlag];

  IF (sym = _ident) OR (sym=_number) OR (sym=_string) THEN BEGIN
    OPT.Find (Scaner.Name,fwd);
    name := Scaner.name;
    CheckMark(Scaner,vis, TRUE);
    IF init THEN vis := external;
    IF (vis <> internal) AND (mode = omLProc) THEN mode := omXProc;
			
    (* false forwards *)
    IF (fwd <> NIL) AND (NOT (fwd.mode IN [omTProc, omLProc, omXProc, omCProc, omIProc]) OR ((fwd.mnolev >= 0) AND (fwd.mnolev <> level)) OR ((mode=omTProc) AND (fwd.mode<>omTProc)) OR (prio=126)) THEN fwd := NIL;
    redef := (fwd <> NIL) AND (mode = omTProc) AND (fwd.mode = omTProc) AND (selfrec <> fwd.link.typ) AND (selfrec <> fwd.link.typ.BaseTyp) ;
    IF (fwd = NIL) OR redef THEN BEGIN
      IF NOT (OPM.oberon2 IN OPM.options) OR (mode <> omTProc) OR (selftype <> NIL) THEN
        OPT.Insert (name, proc)
      ELSE BEGIN
        proc := OPT.NewObj();
        proc.name:=name;
      END;
      (* special case for o2-methods in an empty record *)
      IF (scope <> OPT.topScope) AND (selfrec.link = NIL) THEN
        selfrec.link := OPT.topScope.right;
      proc.prio := prio
    END ELSE BEGIN
      IF vis <> fwd.vis THEN err(118);
      IF hasBody IN fwd.conval^.setval THEN err(1);
      proc := OPT.NewObj();
      proc.leaf := TRUE;
    END;

    proc.mode := mode;
    proc.vis := vis;
    proc.conval := OPT.NewConst();
    proc.conval^.setval := flags;

    IF IsObjectScope() AND (selfrec.BaseTyp <> NIL) AND selfrec.BaseTyp.incomplete THEN BEGIN
      SetFlags(OPT.topScope.link, [checkOverwriting]);
      proc.conval^.intval := OPM.errpos.CodePos
    END;

    IF (mode <> omLProc) AND (level > 0) THEN err(73);
    IF init THEN OPT.topScope.link.link2 := proc;

    OPT.topScope := scope;
    INC(level);
    OPT.OpenScope(level, proc);
    OPT.topScope.adr := OPM.errpos.CodePos;
    OPT.topScope.conval := OPT.NewConst();
    GetParam;
    IF NOT forward THEN Body;
    DEC(level);
    OPT.CloseScope;

  END ELSE
    err(ORD(_ident));
END;

PROCEDURE StatSeq(VAR Scaner: TScaner; VAR stat: OPT.Nodes; local: BOOLEAN);
VAR
  fpar, id, t, obj, owner, op: OPT.Objects;
  typ, idtyp: OPT.Structs;
  sysflag, e: BOOLEAN;
  x1, x2, s, x, y, z, apar, last, lastif: OPT.Nodes;
  pos: INTEGER;
  ts: OPS.TokenStream;

  PROCEDURE CasePart(VAR x: OPT.Nodes);
  VAR
    n: INTEGER;
    low, high: LONGINT;
    e: BOOLEAN;
    tab: CaseTable;
    cases, lab, y, lastcase: OPT.Nodes;
  BEGIN
    FixPhase := TRUE;
    Expression(Scaner,x);
    FixPhase := FALSE;
    pos := OPM.errpos.CodePos;
    IF (x.clas = Ntype) OR (x.clas = Nproc) THEN err(126)
    ELSE IF NOT (x.typ.form IN [sfChar..sfLInt]) THEN err(125);
    Scaner.CheckSym(SYM,_of);
    cases := NIL;
    lastcase := NIL;
    n := 0;

    REPEAT
      IF sym < _bar THEN BEGIN
        FixPhase := TRUE;
        CaseLabelList(Scaner,lab, x.typ.form, n, tab);
        FixPhase := FALSE;
        Scaner.CheckSym(SYM,_colon);
        StatSeq(Scaner,y,local);
        OPB.Construct(Ncasedo, lab, y);
        OPB.Link(cases, lastcase, lab)
      END ;
      IF sym = _bar THEN Scaner.Get(sym) ELSE BREAK;
    UNTIL FALSE;

    IF n > 0 THEN BEGIN
      low := tab[0].low;
      high := tab[n-1].high;
      IF high - low > OPM.MaxCaseRange THEN err(209);
    END ELSE BEGIN
      low := 1;
      high := 0;
      err(213);
    END;

    e := sym = _else;
    IF e THEN BEGIN
      Scaner.Get(sym);
      StatSeq(Scaner,y,local);
    END ELSE
      y := NIL;

    OPB.Construct(Ncaselse, cases, y);
    OPB.Construct(Ncase, x, cases);
    cases.conval := OPT.NewConst();
    cases.conval^.intval := low;
    cases.conval^.intval2 := high;
    IF e THEN
      cases.conval^.setval := [1]
    ELSE
      cases.conval^.setval := [];
  END;
		
  PROCEDURE SetPos(x: OPT.Nodes);
  BEGIN
    x.conval := OPT.NewConst();
    x.conval^.intval := pos;
  END;

BEGIN
  stat := NIL;
  last := NIL;
  UndefFlag := FALSE;
  owner:=NIL;
  id:=NIL;

  fpar:= NIL;
  t:= NIL;
  obj:= NIL;
  op:= NIL;
  x1:= NIL;
  x2:= NIL;
  s:= NIL;
  x:= NIL;
  y:= NIL;
  z:= NIL;
  apar:= NIL;
  lastif:= NIL;

  REPEAT

    x := NIL;

    IF sym < _ident THEN BEGIN
      err(14);
      REPEAT Scaner.Get(sym) UNTIL sym >= _ident
    END ;

    IF UndefFlag THEN BEGIN
      err (1001);
      UndefFlag := FALSE
    END;

    CASE sym OF
      _ident:
              BEGIN
                Scaner.StartRecording (sym);
                Qualident(Scaner,id);
                x := OPB.NewLeaf(id);
                Selector(Scaner,x);
                IF sym = _becomes THEN BEGIN
                  Scaner.Get(sym);
                  IF UndefFlag THEN BEGIN
                    UndefFlag := FALSE;
                    Expression (Scaner,x);
                    Scaner.StopRecording (ts);
                    IF UndefFlag THEN BEGIN
                      InsertFix(FAssign,ts,x);
                      UndefFlag := FALSE
                    END ELSE BEGIN
                      OPB.Construct(Ntrap, x, NIL);
                      InsertFix(FTarget,ts,x);
                    END;
                  END ELSE BEGIN
                    Scaner.StartRecording(sym);
                    Expression(Scaner,y);
                    Scaner.StopRecording(ts);
                    IF UndefFlag THEN BEGIN
                      InsertFix (FExpr, ts, x);
                      UndefFlag := FALSE
                    END ELSE BEGIN
                      Resolve(op,':=', x, y);
                      IF op=NIL THEN  OPB.Assign(x, y)  ELSE  OpCall(op, x, y);
                    END;
                  END
                END ELSE IF sym = _eql THEN BEGIN
                  err(ORD(_becomes));
                  Scaner.Get(sym);
                  Expression(Scaner,y);
                  OPB.Assign(x, y)
                END ELSE IF (x.clas = Nproc) AND (x.obj.mode = omSProc) THEN BEGIN
                  StandProcCall(Scaner,x);
                  Scaner.StopRecording(ts);
                  IF UndefFlag THEN BEGIN
                    UndefFlag := FALSE;
                    InsertFix (FCall, ts, x)
                  END ELSE IF (x <> NIL) AND (x.typ <> OPT.notyp) THEN
                    err(55);
                END ELSE BEGIN
                  IF NOT UndefFlag THEN OPB.PrepCall(x, fpar);
                  IF sym = _lparen THEN BEGIN
                    Scaner.Get(sym);
                    ActualParameters(Scaner,apar, fpar, x);
                    Scaner.CheckSym(SYM,_rparen)
                  END ELSE BEGIN
                    apar := NIL;
                    IF (fpar <> NIL) AND NOT UndefFlag THEN err(65);
                  END;
                  IF NOT UndefFlag THEN BEGIN
                    IF (x.clas=Nproc) OR (x.clas=Nassembler) THEN
                      sysflag := x.obj.sysflag <> 0
                    ELSE
                      sysflag := x.typ.sysflag <> 0;
                    IF (x.clas = Nproc) AND (x.obj.mode = omCProc) THEN
                      InlineProc (x, apar, fpar)
                    ELSE
                      OPB.Call(x, apar, fpar);
                    IF (x <> NIL) AND (x.typ <> OPT.notyp) THEN BEGIN
                      IF NOT sysflag THEN
                        err(55)
                      ELSE
                        x.typ := OPT.notyp;
                    END;
                  END;
                  Scaner.StopRecording(ts);
                  IF UndefFlag THEN BEGIN
                    UndefFlag := FALSE;
                    InsertFix (FCall, ts, x)
                  END;
                  IF level > 0 THEN OPT.topScope.link.leaf := FALSE;
                END;
                pos := OPM.errpos.CodePos;
              END;
      _if:
              BEGIN
                Scaner.Get(sym);
                Scaner.StartRecording (sym);
                Expression(Scaner,x);
                Scaner.StopRecording (ts);
                IF UndefFlag THEN BEGIN
                  InsertFix (FCond, ts, x);
                  s := x;
                  x.clas := 0;
                  UndefFlag := FALSE
                END ELSE BEGIN
                  CheckBool(x);
                  s := NIL;
                END;
                Scaner.CheckSym(SYM,_then);
                StatSeq(Scaner,y,local);
                OPB.Construct(Nif, x, y);
                SetPos(x);
                lastif := x;
                WHILE sym = _elsif DO BEGIN
                  Scaner.Get(sym);
                  Scaner.StartRecording (sym);
                  Expression(Scaner,y);
                  Scaner.StopRecording (ts);
                  IF UndefFlag THEN BEGIN
                    InsertFix (FCond, ts, y);
                    s := x;
                    x.clas := 0;
                    UndefFlag := FALSE
                  END ELSE
                    CheckBool(y);
                  Scaner.CheckSym(SYM,_then);
                  StatSeq(Scaner,z,local);
                  OPB.Construct(Nif, y, z);
                  SetPos(y);
                  OPB.Link(x, lastif, y);
                END ;
                IF sym = _else THEN BEGIN
                  Scaner.Get(sym);
                  StatSeq(Scaner,y,local)
                END ELSE
                  y := NIL;
                OPB.Construct(Nifelse, x, y);
                Scaner.CheckSym(SYM,_end);
                IF s = NIL THEN  OPB.OptIf(x);
                pos := OPM.errpos.CodePos;
              END;
      _case:
              BEGIN
                Scaner.Get(sym);
                CasePart(x);
                Scaner.CheckSym(SYM,_end)
              END;
      _while:
              BEGIN
                Scaner.Get(sym);
                Scaner.StartRecording (sym);
                Expression(Scaner,x);
                Scaner.StopRecording (ts);
                IF UndefFlag THEN BEGIN
                  InsertFix(FCond, ts, x);
                  UndefFlag := FALSE
                END ELSE
                  CheckBool(x);
                Scaner.CheckSym(SYM,_do);
                StatSeq(Scaner,y,local);
                OPB.Construct(Nwhile, x, y);
                Scaner.CheckSym(SYM,_end)
              END;
      _repeat:
              BEGIN
                Scaner.Get(sym);
                StatSeq(Scaner,x,local);
                IF sym = _until THEN BEGIN
                  Scaner.Get(sym);
                  pos := OPM.errpos.CodePos;
                  Scaner.StartRecording (sym);
                  Expression(Scaner,y);
                  Scaner.StopRecording (ts);
                  IF UndefFlag THEN BEGIN
                    InsertFix (FCond, ts, y);
                    UndefFlag := FALSE
                  END ELSE
                    CheckBool(y);
                END ELSE
                  err(ORD(_until));
                OPB.Construct(Nrepeat, x, y)
              END;
      _for:
              BEGIN
                Scaner.Get(sym);
                IF sym = _ident THEN BEGIN
                  FixPhase := TRUE;
                  Qualident(Scaner,id);
                  x := OPB.NewLeaf(id);
                  selector (Scaner,x);
                  OPB.CopyAndSubst(x1, x);
                  OPB.CopyAndSubst(x2, x);
                  IF NOT (x.typ.form IN intSet) THEN err(68);
                  Scaner.CheckSym(SYM,_becomes);
                  FixPhase := FALSE;
                  typ := x.typ;
                  Scaner.StartRecording(sym);
                  Expression(Scaner,y);
                  Scaner.StopRecording (ts);
                  pos := OPM.errpos.CodePos;
                  IF UndefFlag THEN BEGIN
                    InsertFix(FExpr, ts, x);
                    UndefFlag := FALSE
                  END ELSE
                    OPB.Assign(x, y);
                  SetPos(x);
                  Scaner.CheckSym(SYM,_to);
                  Scaner.StartRecording(sym);
                  Expression(Scaner,y);
                  Scaner.StopRecording (ts);
                  pos := OPM.errpos.CodePos;
                  IF (y.clas <> Nconst) OR UndefFlag THEN BEGIN
                    owner := OPT.topScope;
                    WHILE (OPT.topScope.left <> NIL) AND (OPT.topScope.link = OPT.topScope.left.link) DO
                      OPT.topScope := OPT.topScope.left;
                    OPT.Insert('@@', t);
                    t.name := '@for';
                    t.mode := omVar;
                    t.typ := typ;
                    t.flag := [OPT.used];
                    obj := OPT.topScope.scope;
                    IF obj = NIL THEN
                      OPT.topScope.scope := t
                    ELSE BEGIN
                      WHILE obj.link <> NIL DO obj := obj.link;
                      obj.link := t
                    END ;
                    OPT.topScope := owner;
                    z := OPB.NewLeaf(t);
                    IF UndefFlag THEN BEGIN
                      InsertFix (FExpr, ts, z);
                      UndefFlag := FALSE
                    END ELSE
                      OPB.Assign(z, y);
                    SetPos(z);
                    OPB.Link(stat, last, z);
                    y := OPB.NewLeaf(t)
                  END ELSE IF (y.typ.form < sfSInt) OR (y.typ.form > typ.form) THEN
                    err(113);
                  OPB.Link(stat, last, x);
                  IF sym = _by THEN BEGIN
                    Scaner.Get(sym);
                    ConstExpression(Scaner,z)
                  END ELSE
                    z := OPB.NewIntConst(1);
                  pos := OPM.errpos.CodePos;
                  x := x1;
                  IF z.conval^.intval > 0 THEN
                    OPB.Op(vleq, x, y)
                  ELSE IF z.conval^.intval < 0 THEN
                    OPB.Op(vgeq, x, y)
                  ELSE BEGIN
                    err(63);
                    OPB.Op(vgeq, x, y)
                  END ;
                  Scaner.CheckSym(SYM,_do);
                  StatSeq(Scaner,s,local);
                  y := x2;
                  OPB.StPar1(y, z, incfn);
                  SetPos(y);
                  IF s = NIL THEN
                    s := y
                  ELSE BEGIN
                    z := s;
                    WHILE z.link <> NIL DO z := z.link;
                    z.link := y
                  END ;
                  Scaner.CheckSym(SYM,_end);
                  OPB.Construct(Nwhile, x, s);
                END ELSE
                  err(ORD(_ident));
              END;
      _loop:
              BEGIN
                Scaner.Get(sym);
                INC(LoopLevel);
                StatSeq(Scaner,x,local);
                DEC(LoopLevel);
                OPB.Construct(Nloop, x, NIL);
                Scaner.CheckSym(SYM,_end);
                pos := OPM.errpos.CodePos;
              END;
      _with:
              BEGIN
                Scaner.Get(sym);
                idtyp := NIL;
                x := NIL;

                REPEAT
                  IF sym = _ident THEN BEGIN
                    Qualident(Scaner,id);
                    y := OPB.NewLeaf(id);
                    IF (id <> NIL) AND (id.typ.form = sfPointer) AND ((id.mode = omVarPar) OR NOT id.leaf) THEN
                      err(-302);
                    Scaner.CheckSym(SYM,_colon);
                    IF sym = _ident THEN BEGIN
                      Qualident(Scaner,t);
                      IF t.mode = omTyp THEN BEGIN
                        IF id <> NIL THEN BEGIN
                          idtyp := id.typ;
                          OPB.TypTest(y, t, FALSE);
                          IF id.mnolev < 0 THEN
                            id.typ := t.typ
                          ELSE BEGIN
                            owner := OPT.topScope.link;
                            OPT.OpenScope (level, owner);
                            OPT.Insert (id.name, obj);
                            obj.mode := id.mode;
                            obj.scope := id;
                            obj.typ := t.typ
                          END
                        END ELSE
                          err(130);
                      END ELSE
                        err(52);
                    END ELSE
                      err(ORD(_ident));
                  END ELSE
                    err(ORD(_ident));
                  pos := OPM.errpos.CodePos;
                  Scaner.CheckSym(SYM,_do);
                  StatSeq(Scaner,s,local);
                  OPB.Construct(Nif, y, s);
                  SetPos(y);
                  IF id.mnolev < 0 THEN
                    id.typ := idtyp
                  ELSE BEGIN
                    IF idtyp <> NIL THEN BEGIN
                      OPT.CloseScope;
                      owner.scope := OPT.topScope;
                      idtyp := NIL;
                    END;
                  END;
                  IF x = NIL THEN BEGIN
                    x := y;
                    lastif := x
                  END ELSE
                    OPB.Link(x, lastif, y);
                  IF (sym = _bar) AND (OPM.oberon2 IN OPM.options) THEN Scaner.Get(sym) ELSE BREAK;
                UNTIL FALSE;

                e := sym = _else;
                IF e AND (OPM.oberon2 IN OPM.options) THEN BEGIN
                  Scaner.Get(sym);
                  StatSeq(Scaner,s,local);
                END ELSE
                  s := NIL;
                OPB.Construct(Nwith, x, s);
                Scaner.CheckSym(SYM,_end);
                IF e THEN x.subcl:=1;
              END;
      _exit:
              BEGIN
                Scaner.Get(sym);
                IF LoopLevel = 0 THEN err(38);
                OPB.Construct(Nexit, x, NIL);
                pos := OPM.errpos.CodePos;
              END;
      _return:
              BEGIN
                Scaner.Get(sym);
                IF sym < _semicolon THEN BEGIN
                  Scaner.StartRecording(sym);
                  Expression(Scaner,x);
                  Scaner.StopRecording(ts)
                END;
                IF UndefFlag OR (OPT.topScope.link.typ = NIL) THEN BEGIN
                  InsertFix (FReturn, ts, x);
                  UndefFlag := FALSE;
                END ELSE BEGIN
                  IF level > 0 THEN
                    OPB.Return(x, OPT.topScope.link)
                  ELSE
                    OPB.Return(x, NIL);
                END;
                INC(retcount);
                pos := OPM.errpos.CodePos
              END;
      _begin:
              BEGIN
                Scaner.Get(sym);
                StatBlock(Scaner,x, TRUE);
              END;
      _finally:
              BEGIN
                IF local THEN
                  err(160)
                ELSE BEGIN
                  IF retcount > 0 THEN
                    err(161)
                  ELSE IF fincount > 0 THEN
                    err(162);
                  IF (retcount = 0) AND (fincount = 0) THEN	BEGIN
                    OPT.MarkType(OPT.linttyp);
                    OPB.Construct(Nfinally, x, NIL);
                    x.obj := OPT.topScope.link;
                    pos := OPM.errpos.CodePos;
                  END;
                  INC(fincount);
                END;
                Scaner.Get(sym);
              END;
    END;

    IF UndefFlag THEN	BEGIN
      err(1000);
      UndefFlag := FALSE
    END;

    IF x <> NIL THEN BEGIN
      SetPos(x);
      OPB.Link(stat, last, x)
    END;
    
    IF (x=NIL) OR ((x<>NIL) AND (x.clas <> Nfinally)) THEN BEGIN
      IF sym = _semicolon THEN
        Scaner.Get(sym)
      ELSE IF sym = _finally THEN
      ELSE IF (sym <= _ident) OR (_if <= sym) AND (sym <= _return) THEN
        err(ORD(_semicolon))
      ELSE
        BREAK;
    END;

  UNTIL FALSE;
END;

PROCEDURE BlockMode (VAR Scaner: TScaner; VAR flag: TSET;  local: BOOLEAN; VAR lock: BOOLEAN);
VAR
  x: OPT.Nodes;
  objscope: BOOLEAN;
	
  PROCEDURE ProtectScope();
  VAR
    p: OPT.Objects;
  BEGIN
    p := OPT.topScope;
    IF (p<>NIL) AND (p.link.mode IN [omXProc, omLProc]) AND (p.link.typ.comp IN [csOpenArr, csSDynArr]) THEN err(200);
    WHILE (p <> NIL) AND NOT (p.link.mode IN [omTyp, omMod]) DO p := p.left;
    IF p = NIL THEN err(1000)
    ELSE IF (p.link.mode = omTyp) AND (p.link.typ.ptr = NIL) THEN err(252);
    p.link.conval^.setval:=p.link.conval^.setval+[protectedObj];
  END;
	
BEGIN
  x:=NIL;
  lock := FALSE;
  objscope := IsObjectScope();
  IF sym = _lbrace THEN BEGIN
    Scaner.Get(sym);
    REPEAT
      IF sym = _ident THEN BEGIN
        IF Scaner.name = 'EXCLUSIVE' THEN BEGIN
          Scaner.Get (sym);
          ProtectScope();
          flag:=flag+[locked];
          lock := TRUE
        END ELSE IF NOT local THEN BEGIN
          IF Scaner.name = 'ACTIVE' THEN BEGIN
            IF NOT objscope THEN err(200);
            flag:=flag+[activeObj];
            Scaner.Get(sym)
          END ELSE IF Scaner.name = 'PRIORITY' THEN BEGIN
            IF NOT objscope  THEN err(200);
            IF NOT (activeObj IN flag) THEN err(272);
            Scaner.Get(sym);
            IF sym = _lparen THEN BEGIN
              ConstExpression(Scaner,x);
              IF x.typ.form IN [sfSInt] THEN OPT.topScope.link.prio := x.conval^.intval
              ELSE err(50);
            END ELSE
              x := OPB.NewIntConst(0);
          END ELSE IF Scaner.name = 'SAFE' THEN BEGIN
            IF NOT (activeObj IN flag) THEN err(272);
            flag:=[safe];
            Scaner.Get(sym)
          END ELSE BEGIN
            err(0);
            Scaner.Get(sym)
          END;
        END ELSE BEGIN
          err(0);
          Scaner.Get(sym)
        END;
      END ELSE IF sym <> _rbrace THEN BEGIN
        err(0);
        Scaner.Get(sym)
      END;
      IF sym = _rbrace THEN BREAK;
      Scaner.CheckSym(SYM,_comma);
    UNTIL FALSE;
    Scaner.CheckSym(SYM,_rbrace)
  END;
  IF NOT local AND objscope AND NOT (activeObj IN flag) THEN err(999);
END;

PROCEDURE StatBlock(VAR Scaner: TScaner; VAR stseq: OPT.Nodes;  local: BOOLEAN);
VAR
  lock, oldScopeLock: BOOLEAN;
  pos: INTEGER;
  owner: OPT.Objects;
BEGIN
  IF NOT local THEN BEGIN
    retcount := 0;
    fincount := 0;
  END;
  owner := OPT.topScope.link;
  pos := OPM.errpos.CodePos;
  BlockMode(Scaner,owner.conval^.setval, local, lock);
  IF scopeLock AND lock THEN  err(246);
  oldScopeLock := scopeLock;
  scopeLock := scopeLock OR lock;
  StatSeq(Scaner,stseq,local);
  Scaner.CheckSym(sym,_end);
  scopeLock := oldScopeLock;
  IF lock THEN BEGIN
    OPB.Lock(stseq, GetSelf(), TRUE);
    stseq.conval := OPT.NewConst();
    stseq.conval^.intval := pos
  END
END;

PROCEDURE Block(VAR Scaner: TScaner; VAR procdec, statseq: OPT.Nodes);
VAR
  typ: OPT.Structs;
  owner, obj, first, last, dmy: OPT.Objects;
  x, lastdec: OPT.Nodes;
  pos: integer;
  vars, proc: BOOLEAN;

BEGIN
  first := NIL;
  last := NIL;
  dmy:=NIL;
  obj:=NIL;
  typ:=NIL;
  procdec := NIL;
  lastdec := NIL;
  proc := FALSE;
  vars := FALSE;

  REPEAT

    IF (sym = _const) THEN BEGIN
      IF (OPM.oberon1 IN OPM.options) AND proc THEN err(262);
      Scaner.Get(sym);
      WHILE sym = _ident DO BEGIN
        OPT.Insert(Scaner.name, obj);
        CheckMark(Scaner,obj.vis, FALSE);
        IF IsObjectScope() AND (obj.vis = external) THEN err (47);		(* <- prevent export, implementation restriction *)
        obj.typ := OPT.sinttyp;
        obj.mode := omVar;	(* Var to avoid recursive definition *)
        IF sym = _eql THEN BEGIN
          Scaner.Get(sym);
          ConstExpression(Scaner,x);
        END ELSE IF sym = _becomes THEN BEGIN
          err(ORD(_eql));
          Scaner.Get(sym);
          ConstExpression(Scaner,x);
        END ELSE BEGIN
          err(ORD(_eql));
          x := OPB.NewIntConst(1);
        END;
        obj.mode := omCon;
        obj.typ := x.typ;
        obj.conval := x.conval; (* ConstDesc ist not copied *)
        Scaner.CheckSym(sym,_semicolon);
        WHILE sym=_semicolon DO Scaner.Get(sym);
      END
    END;

    IF (sym = _type) THEN BEGIN
      IF (OPM.oberon1 IN OPM.options) AND proc THEN err(262);
      Scaner.Get(sym);
      WHILE sym = _ident DO BEGIN
        OPT.Insert(Scaner.name, obj);
        obj.mode := omTyp;
        obj.typ := OPT.undftyp;
        x := NIL;
        CheckMark(Scaner,obj.vis, FALSE);
        IF sym = _eql THEN BEGIN
          Scaner.Get(sym);
          TypeDecl(Scaner,x, obj.typ, obj.typ, obj)
        END ELSE IF (sym = _becomes) OR (sym = _colon) THEN BEGIN
          err(ORD(_eql));
          Scaner.Get(sym);
          TypeDecl(Scaner,x, obj.typ, obj.typ, obj);
        END ELSE
          err(ORD(_eql));
        IF obj.typ = OPT.ToBeFixed THEN	BEGIN	(*incomplete structure, add fixup, hidden*)
          NewFixO(ObjFix[nofObjFix], Scaner.name);
          ObjFix[nofObjFix].obj := obj;
          INC(nofObjFix);
          obj.mode := -obj.mode;
        END ELSE
          NewType(obj);
        IF obj.typ.incomplete THEN
          NewCheck(obj)
        ELSE
          CheckObj(obj);
        IF x <> NIL THEN BEGIN	(* insert x into parse tree *)
          x.obj := obj;		(* link code to the type *)
          OPB.Link (RecInit, lastRecInit, x);
          IF level = 0 THEN BEGIN
            OPB.Link (procdec, lastdec, RecInit);
            RecInit := NIL;
          END;
        END;
        IF ((obj.typ.comp = csRecord) OR (obj.typ.form = sfPointer)
        AND (obj.typ.BaseTyp.comp = csRecord)) AND (sym = _ident) THEN BEGIN
          IF obj.name <> Scaner.name THEN err(4);
          Scaner.Get(sym)
        END;
        IF obj.typ.strobj = NIL THEN obj.typ.strobj := obj;
        IF obj.conval = NIL THEN obj.conval := OPT.NewConst();
        Scaner.CheckSym(sym,_semicolon);
      END;
      IF OPT.topScope.scope<>NIL THEN BEGIN
        first:=OPT.topScope.scope;
        WHILE first.link<>NIL DO first:=first.link;
        last:=first;
        first:=NIL
      END;
    END;
    
    IF (sym = _var) OR (IsObjectScope() AND (sym = _ident)) THEN BEGIN
      vars := TRUE;
      IF (OPM.oberon1 IN OPM.options) AND proc THEN err(262);
      IF sym = _var THEN Scaner.Get(sym);
      WHILE sym = _ident DO BEGIN
        REPEAT
          IF sym = _ident THEN BEGIN
            IF NOT IsObjectScope() THEN
            ELSE IF OPT.topScope.link.typ.incomplete THEN
              SetFlags(OPT.topScope.link, [checkOverwriting])
            ELSE BEGIN
              OPT.FindField(Scaner.name, OPT.topScope.link.typ, obj, FALSE);
              IF (obj <> NIL) AND (obj.mode IN [omFld, omTProc]) THEN err(1);
            END;

            OPT.Insert(Scaner.name, obj);
            CheckMark(Scaner,obj.vis, TRUE);
            
            IF level = 0 THEN
              CheckSysFlag(Scaner,obj.sysflag, 0, [0, untraced])
            ELSE
              CheckSysFlag(Scaner,obj.sysflag, 0, [0]);

            IF IsObjectScope() THEN BEGIN
              obj.mode := omFld;
              obj.conval := OPT.NewConst();
              obj.conval^.intval := OPM.errpos.CodePos;
              OPT.FindField (obj.name, OPT.topScope.link.typ.BaseTyp, dmy, TRUE);
            END ELSE BEGIN
              obj.mode := omVar;
              obj.leaf := obj.vis = internal
            END;
            obj.link := NIL;
            obj.typ := OPT.undftyp;
            obj.myscope:=OPT.topScope;
            IF first = NIL THEN first := obj;
            IF last = NIL THEN BEGIN
              IF IsObjectScope() THEN
                OPT.topScope.link.typ.link := obj
              ELSE
                OPT.topScope.scope := obj;
            END ELSE
              last.link := obj;
            last := obj;
          END ELSE
            err(ORD(_ident));
          IF sym = _comma THEN Scaner.Get(sym)
          ELSE IF sym = _ident THEN err(ORD(_comma))
          ELSE BREAK;
        UNTIL FALSE;

        Scaner.CheckSym(sym,_colon);

        FType(Scaner,typ, OPT.notyp);
        IF IsObjectScope() AND (OPT.topScope.link.typ = typ) THEN	err (58);
        OPT.MarkType(typ);

        WHILE first <> NIL DO BEGIN
          first.typ := typ; 
          IF typ.incomplete THEN BEGIN	(*hide*)
            IF typ = OPT.ToBeFixed THEN BEGIN
              NewFixO(ObjFix[nofObjFix], Scaner.name);
              ObjFix[nofObjFix].obj := first;
              INC(nofObjFix)
            END;
            NewCheck(first);
            first.mode := -first.mode;	(* mark variable *)
          END ELSE
            CheckObj(first);
          first := first.link;
        END;

        (************)
        //LNTi386.AddImport(SelfName,obj.name,typ.lib.libname,typ.lib.alias,0);

        IF NOT IsObjectScope() OR (sym <> _end) THEN Scaner.CheckSym(sym,_semicolon);
        WHILE sym=_semicolon DO Scaner.Get(sym);
      END;
    END;

    IF sym = _procedure THEN BEGIN
      proc := TRUE;
      Scaner.Get(sym);
      ProcedureDeclaration(Scaner,x);
      IF x <> NIL THEN BEGIN
        IF lastdec = NIL THEN
          procdec := x
        ELSE
          lastdec.link := x;
        lastdec := x;
      END;
      IF (RecInit <> NIL) AND (level = 0) THEN BEGIN
        OPB.Link (procdec, lastdec, RecInit);
        RecInit := NIL;
      END;
      Scaner.CheckSym(sym,_semicolon);
    END;

    IF (sym < _const) OR (sym > _procedure) THEN BREAK;
  UNTIL FALSE ;	(* LOOP *)

  OPT.topScope.adr := OPM.errpos.CodePos;

  IF IsObjectScope() AND (last = NIL) THEN
    OPT.topScope.link.typ.link := OPT.topScope.right;

  CondNodes := NIL;
  CondCount := 0;
  owner := OPT.topScope.link;

  IF sym = _begin THEN BEGIN
    IF (owner.mode = omCProc) AND (owner.vis <> internal) THEN err(250);
    IF IsObjectScope() THEN	BEGIN		(* object body, VAR self: TDesc *)
      typ := owner.typ;
      IF typ.ptr = NIL THEN err(249);
      ASSERT(typ <> NIL);
      OPT.Insert(HiddenSelf, obj);
      obj.mode := omVarPar;
      obj.typ := typ;
      obj.adr := OPM.errpos.CodePos;
      obj.mnolev := 1;
      obj.link := NIL;
      owner.link := obj;	(* set parameter list *)
      IF owner.typ.link = NIL THEN
        owner.typ.link := obj;
    END;

    Scaner.Get(sym);
    StatBlock(Scaner, statseq, FALSE);
    SetFlags (owner, [hasBody]);

    IF (OPM.trace IN OPM.options) AND (level = 1) THEN  AddTraceCode(owner, statseq);
    IF (CondNodes <> NIL) AND NOT IsObjectScope() THEN BEGIN
      IF lastdec = NIL THEN
        procdec := CondNodes
      ELSE
        lastdec.link := CondNodes;
      lastdec := CondNodes;
    END;
  END ELSE IF sym = _code THEN BEGIN		(* iOPA *)
    IF (owner.mode = omCProc) AND vars THEN err(247);
    pos := OPM.errpos.CodePos;
    SetFlags(owner, [asmProc]);
    OPA.Assemble(Scaner, statseq, owner.vis <> internal, owner.mode = omCProc);
    IF statseq <> NIL THEN BEGIN
      statseq.conval := OPT.NewConst();
      statseq.conval^.intval := pos;
      statseq.obj := owner;
    END;
    Scaner.Get(sym)
  END ELSE BEGIN
    statseq := NIL;
    Scaner.CheckSym(sym,_end)
  END;

  IF (level = 0) AND (TDinit <> NIL) AND IsModuleScope() THEN BEGIN
    lastTDinit.link := statseq;
    statseq := TDinit
  END ;

  IF (level = 0) AND (RecInit <> NIL) AND IsModuleScope() THEN
    OPB.Link (procdec, lastdec, RecInit);

END;

PROCEDURE DoFixes(VAR Scaner: TScaner; VAR prog: OPT.Nodes);
VAR
  id, obj, typ, fpar, topscope,op: OPT.Objects;
  x, y, z, link, apar: OPT.Nodes;
  conval: OPT.Consts;
  i: LONGINT;
  sysflag, done: BOOLEAN;

  PROCEDURE Insert(VAR Find,Dest,x: OPT.Nodes);
  BEGIN
    IF (Dest<>NIL) THEN Insert(Find,Dest.left,x);
    IF (Dest<>NIL) THEN Insert(Find,Dest.right,x);
    IF (Dest<>NIL) THEN Insert(Find,Dest.link,x);
    IF (Dest<>NIL) AND (Find=Dest) THEN Dest:=x;
  END;

BEGIN

  id:=NIL;
  typ:=NIL;
  fpar:=NIL;
  op:=NIL;
  y:=NIL;
  z:=NIL;
  apar:=NIL;

  topscope := OPT.topScope;

  REPEAT
    i := 0;
    done := FALSE;
    WHILE i < nofObjFix DO BEGIN
      IF  ObjFix[i].obj <> NIL THEN BEGIN
        OPM.errpos := ObjFix[i].pos;
        level := ObjFix[i].level;
        OPT.FindInScope(ObjFix[i].name, ObjFix[i].scope, typ);
        IF typ <> NIL THEN BEGIN
          ObjFix[i].obj.mode := ABS(ObjFix[i].obj.mode);
          ObjFix[i].obj.typ := typ.typ;
          ObjFix[i].obj := NIL;
          done := TRUE; 
        END
      END;
      INC(i)
    END;
  UNTIL NOT done;

  FixPhase := TRUE;
  i := 0;
  WHILE i < nofObjFix DO BEGIN
    IF ObjFix[i].obj <> NIL THEN BEGIN
      OPM.Mark(0, ObjFix[i].pos);
      ObjFix[i].obj := NIL;
      ObjFix[i].scope := NIL;
    END;
    INC(i)
  END;
  nofObjFix := 0;

  WHILE 0 < nofTypFix DO BEGIN
    DEC(nofTypFix);
    IF TypFix[nofTypFix].typ <> NIL THEN BEGIN
      OPM.errpos := TypFix[nofTypFix].pos;
      level := TypFix[nofObjFix].level;
      OPT.FindInScope(TypFix[nofTypFix].name, TypFix[nofTypFix].scope, typ);
      IF typ <> NIL THEN
        TypFix[nofTypFix].typ.BaseTyp := typ.typ
      ELSE IF TypFix[nofTypFix].typ.comp = csRecord THEN BEGIN
        err(0);
        TypFix[nofTypFix].typ.BaseTyp := NIL
      END ELSE BEGIN
        err(0);
        TypFix[nofTypFix].typ.BaseTyp := OPT.undftyp
      END;
      TypFix[nofTypFix].typ := NIL;
      TypFix[nofTypFix].scope := NIL
    END
  END;
  i := 0;

  WHILE i < nofCheck DO BEGIN
    IF ObjCheck[i].obj <> NIL THEN BEGIN
      level := ObjCheck[i].level;
      ObjCheck[i].obj.mode := ABS(ObjCheck[i].obj.mode);
      CheckObj(ObjCheck[i].obj);
      ObjCheck[i].obj := NIL;
    END;
    INC(i)
  END;

  WHILE FixList <> NIL DO BEGIN
    OPT.topScope := FixList.scope;
    level := FixList.level;
    Scaner.StartPlaying(FixList.stream);
    x := FixList.node;
    Scaner.Get(sym);
    link := x.link;
    x.link := NIL;
    conval := x.conval;
    x.conval:= NIL;

    CASE FixList.mode OF
      FTarget:
                BEGIN
                  y:=Nodes.Create;
                  y := x.left;
                  Qualident(Scaner,id);
                  z := OPB.NewLeaf (id);
                  Selector(Scaner,z);
                  Resolve(op,':=', z, y);
                  IF op=NIL THEN
                    OPB.Assign(z, y)
                  ELSE
                    OpCall(op, z, y);
                END;
      FExpr:
                BEGIN
                  z:=Nodes.Create;
                  z:=x;
                  Expression(Scaner,y);
                  Resolve(op,':=', z, y);
                  IF op=NIL THEN
                    OPB.Assign(z, y)
                  ELSE
                    OpCall(op, z, y);
                END;
      FAssign:
                BEGIN
                  Qualident(Scaner,id);
                  z := OPB.NewLeaf(id);
                  Selector(Scaner,z);
                  IF sym = _becomes THEN BEGIN
                    Scaner.Get(sym);
                    Expression(Scaner,y);
                    Resolve(op,':=', z, y);
                    IF op=NIL THEN
                      OPB.Assign(z, y)
                    ELSE
                      OpCall(op, z, y);
                  END ELSE
                    err (1002);
                END;
      FCall:
                BEGIN
                  Qualident(Scaner,id);
                  z := OPB.NewLeaf(id);
                  Selector(Scaner,z);
                  IF (z.clas = Nproc) AND (z.obj.mode = omSProc) THEN BEGIN
                    StandProcCall(Scaner,z);
                    IF (z <> NIL) AND (z.typ <> OPT.notyp) THEN err(55);
                  END ELSE BEGIN
                    OPB.PrepCall(z, fpar);
                    IF sym = _lparen THEN BEGIN
                      Scaner.Get(sym);
                      ActualParameters(Scaner,apar, fpar, z);
                      Scaner.CheckSym(SYM,_rparen)
                    END ELSE BEGIN
                      apar := NIL;
                      IF fpar <> NIL THEN err(65);
                    END ;
                    IF (z.clas=Nproc) OR (z.clas=Nassembler) THEN
                      sysflag := z.obj.sysflag <> 0
                    ELSE
                      sysflag := z.typ.sysflag <> 0;
                    IF (z.clas = Nproc) AND (z.obj.mode = omCProc) THEN
                      InlineProc(z,apar,fpar)
                    ELSE
                      OPB.Call(z, apar, fpar);
                    IF z.typ <> OPT.notyp THEN BEGIN
                      IF NOT sysflag THEN err(55) ELSE z.typ := OPT.notyp;
                    END;
                    IF level > 0 THEN OPT.topScope.link.leaf := FALSE;
                  END;
                END;
      FCond:
                BEGIN
                  Expression(Scaner,z);
                  CheckBool(z);
                  conval := z.conval;
                END;
      FParam:
                BEGIN
                  obj := x.obj;
                  Expression(Scaner,z);
                  OPB.FNParam(z, obj);
                END;
      FPassExpr:
                BEGIN
                  Scaner.Get(sym);
                  Scaner.Get(sym);
                  Expression(Scaner,z);
                  conval := z.conval;
                  OPB.StPar0(z,awaitfn);
                END;
      FReturn:
                BEGIN
                  Expression(Scaner,z);
                  OPB.Return(z,OPT.topScope.link);
                END;
    END;
    IF z = NIL THEN OPB.Construct (Ninittd, z, NIL);
    x:= z;
    x.link  := link;
    x.conval:= conval;
    Insert(FixList.node,prog,x); //!!!***
    FixList := FixList.next;
    //Scaner.FreeRecording(Scaner.FixList.stream);
  END;
  OPT.topScope := topscope;
END;

//#######################################################################

PROCEDURE Module(VAR prog: OPT.Nodes; VAR modName: STRING; optn: TSET);
VAR
  shortName, impName, aliasName, ext: STRING;
  procdec, statseq: OPT.Nodes;
  c: INTEGER;
  Amod: OPT.Objects;
  Scaner: TScaner;
BEGIN
  SelfName:='';
  procdec:=NIL;
  statseq:=NIL;

  OPM.modulename:=modName;

  modName:=ExtractFileName(modName);
  ext:=ExtractFileExt(modName);
  modName:=COPY(modName,1,Pos(ext,modName)-1);

  SelfName:=modName;

  ptrToRec := NIL;
  nofObjFix := 0;
  nofTypFix := 0;
  nofCheck := 0;
  UndefObj := OPT.NewObj();
  UndefObj.mode := omVar;
  UndefObj.typ := OPT.undftyp;
  UndefObj.adr := 0;
  FixList := NIL;
  FixPhase := FALSE;
  UndefFlag := FALSE;
  FixLast := NIL;
  LoopLevel := 0; level := 0;
  scopeLock := FALSE;

  Scaner:= TScaner.Create(modName+moduleext);
  Scaner.Get(sym);

  IF sym = _module THEN Scaner.Get(sym) ELSE err(16);
  IF sym = _ident THEN BEGIN
    //FullName(modName);
    shortName:=Scaner.name;
    OPT.Init(modName);
    Amod := OPT.NewObj();
    Amod.scope := OPT.topScope;
    OPT.topScope.link := Amod;
    OPT.topScope.conval:=OPT.NewConst();
    Amod.mode := omMod;
    Amod.conval := OPT.NewConst();
    Amod.typ := OPT.notyp;
    INCL(Amod.conval^.setval, protectedObj);	(* module is always protected *)

    ExtObj := OPT.topScope;
    Amod.name:=modName;

    SETLENGTH(modulesx[0].struct,DefMaxStruct); //!!!!
    SETLENGTH(modulesx[0].import,DefMaxImport); //!!!!
    SETLENGTH(modulesx[0].reimp,DefMaxReimp);   //!!!!

    Scaner.Get(sym);
    Scaner.CheckSym(sym,_semicolon);

    IF sym = _import THEN BEGIN
      Scaner.Get(sym);
      REPEAT
        IF sym = _ident THEN BEGIN
          aliasName:=Scaner.name;
          impName:=aliasName;
          Scaner.Get(sym);
          IF sym = _becomes THEN BEGIN
            Scaner.Get(sym);
            IF sym = _ident THEN BEGIN
              impName:=Scaner.name;
              Scaner.Get(sym);
            END ELSE
              err(ORD(_ident));
          END ELSE IF sym = _period THEN
            err(200);
          IF (OPM.systemchk IN OPM.options) AND (impName = 'SYSTEM') THEN
            err(152)
          ELSE
            OPT.Import(aliasName, impName, modName);
        END ELSE
          err(ORD(_ident));
        IF sym = _comma THEN
          Scaner.Get(sym)
        ELSE IF sym = _ident THEN
          err(ORD(_comma))
        ELSE
          BREAK;
      UNTIL FALSE;
      Scaner.CheckSym(sym,_semicolon);
    END ;

    IF OPM.noerr THEN BEGIN
      TDinit := NIL;
      lastTDinit := NIL;
      c := OPM.errpos.CodePos;
      RecInit := NIL;
      lastRecInit := NIL;
      procdec:=NIL;
      statseq:=NIL;
      Block(Scaner,procdec, statseq);
      OPB.Enter(procdec, statseq, NIL);
      prog := procdec;
      prog.conval := OPT.NewConst();
      prog.conval^.intval := c;
      IF sym = _ident THEN BEGIN
        IF Scaner.name <> shortName THEN err(4);
        Scaner.Get(sym);
      END ELSE
        err(ORD(_ident));
      IF sym <> _period THEN err(ORD(_period));
    END;
  END ELSE
    err(ORD(_ident));

  TDinit      := NIL;
  lastTDinit  := NIL;
  RecInit     := NIL;
  lastRecInit := NIL;

  IF OPM.noerr THEN BEGIN
    DoFixes(Scaner,prog);
  END ELSE BEGIN
    WHILE nofObjFix > 0 DO BEGIN DEC(nofObjFix);  ObjFix[nofObjFix].scope := NIL;  ObjFix[nofObjFix].obj := NIL; END;
    WHILE nofTypFix > 0 DO BEGIN DEC(nofTypFix);  TypFix[nofTypFix].scope := NIL;  TypFix[nofTypFix].typ := NIL; END;
    WHILE nofCheck  > 0 DO BEGIN DEC(nofCheck);   ObjCheck[nofCheck].obj  := NIL;                                END;
  END;
  IF OPM.noerr AND (OPM.warning IN OPM.options) THEN OPT.TraverseObjects(OPT.topScope.right);
END;

END.
