unit ObrHighliter;

interface   

uses
  Classes, SysUtils, Windows, Graphics, StdCtrls, SynEdit, StrUtils,
  Dialogs;

type
  TOBRHighliter = class(TSCustomHighlight)
  private
    fRichMemo       : TCustomSyntaxMemo;
    f_OBR_ListWord  : TStringList;

    c_OBR_Keyword,
    c_OBR_Types,
    c_OBR_Integer,
    c_OBR_String,
    c_OBR_Comment,
    c_OBR_Bold,
    c_OBR_Brown,
    c_OBR_Operator,
    c_OBR_Bracket,
    c_OBR_System,
    c_OBR_SysMod  : TSCharHighlight;

    procedure   Set_OBR_Keyword   (const Value: TSCharHighlight);
    procedure   Set_OBR_Types     (const Value: TSCharHighlight);
    procedure   Set_OBR_Integer   (const Value: TSCharHighlight);
    procedure   Set_OBR_String    (const Value: TSCharHighlight);
    procedure   Set_OBR_Comment   (const Value: TSCharHighlight);
    procedure   Set_OBR_Bold      (const Value: TSCharHighlight);
    procedure   Set_OBR_Brown     (const Value: TSCharHighlight);
    procedure   Set_OBR_Operator  (const Value: TSCharHighlight);
    procedure   Set_OBR_Bracket   (const Value: TSCharHighlight);
    procedure   Set_OBR_System    (const Value: TSCharHighlight);
    procedure   Set_OBR_SysMod    (const Value: TSCharHighlight);

  protected
    function    Parse(Sender: TSyntaxParser; Line: string; LineIndex: Integer; LastToken: TToken; PA: TSyntaxAttributes): TToken; override;
    procedure   Change(RichMemo : TCustomSyntaxMemo; ChangedItems: TChangedItems); override;
    procedure   GetHighlitedParam(RichMemo : TCustomSyntaxMemo); override;
  public
    constructor Create(const AOwner: TCustomSyntaxMemo); reintroduce;
    destructor  Destroy; override;
  published
    property    Font_OBR_Keyword  : TSCharHighlight read c_OBR_Keyword    write Set_OBR_Keyword;
    property    Font_OBR_Types    : TSCharHighlight read c_OBR_Types      write Set_OBR_Types;
    property    Font_OBR_Integer  : TSCharHighlight read c_OBR_Integer    write Set_OBR_Integer;
    property    Font_OBR_String   : TSCharHighlight read c_OBR_String     write Set_OBR_String;
    property    Font_OBR_Comment  : TSCharHighlight read c_OBR_Comment    write Set_OBR_Comment;
    property    Font_OBR_Bold     : TSCharHighlight read c_OBR_Bold       write Set_OBR_Bold;
    property    Font_OBR_Brown    : TSCharHighlight read c_OBR_Brown      write Set_OBR_Brown;
    property    Font_OBR_Operator : TSCharHighlight read c_OBR_Operator   write Set_OBR_Operator;
    property    Font_OBR_Bracket  : TSCharHighlight read c_OBR_Bracket    write Set_OBR_Bracket;
    property    Font_OBR_System   : TSCharHighlight read c_OBR_System     write Set_OBR_System;
    property    Font_OBR_SysMod   : TSCharHighlight read c_OBR_SysMod     write Set_OBR_SysMod;
  end;

const
  tokKeyword    = tokUser - 1;
  tokNumber     = tokUser - 2;
  tokString     = tokUser - 3;
  tokComment    = tokUser - 4;
  tokMulty      = tokUser - 5;
  tokTypes      = tokUser - 6;
  tokBold       = tokUser - 7;
  tokBrown      = tokUser - 8;
  tokOperator   = tokUser - 9;
  tokBracket    = tokUser - 10;
  tokSystem     = tokUser - 11;
  tokSysmod     = tokUser - 12;

procedure Register;

implementation
  
procedure Register;
begin
  RegisterComponents('Soft Universe', [TOBRHighliter]);
end;

const
  TokenKey : array [0..49] of string[10]=(
    'DIV','MOD','OR','IN','IS','OF','THEN','DO','TO','BY','NIL',
    'END','ELSE','ELSIF','UNTIL','IF','CASE','LEN','NEW',
    'WHILE','REPEAT','FOR','LOOP','WITH','ARRAY','ASSERT',
    'OBJECT','RECORD','POINTER','BEGIN','CONST','TYPE','VAR',
    'PROCEDURE','MODULE','CODE','INCL','EXCL','CAP','COPY','ASH',
    'INC','DEC','CHR','ORD','XOR','SHORT','LONG','MAX','MIN'
  );

  TypesKey : array [0..7] of string[10]=(
    'CHAR','INTEGER','BOOLEAN','LONGINT','REAL',
    'LONGREAL','SET','SHORTINT'
  );

  SystemKey : array [0..5] of string[10]=(
    'TRUE','FALSE','IMPORT','FINALLY','RETURN','EXIT'
  );

  SystemMod : array [0..21] of string[10]=(
    'SYSTEM','SIZE','ADR','BIT','CC','SELF',
    'LSH','ROT','VAL','GET','PTR','BYTE',
    'PUT','GETREG','PUTREG','MOVE',
    'i386','i486','Pentium','PentiumPro',
    'FPU','Privileged'
  );

constructor TOBRHighliter.Create(const AOwner: TCustomSyntaxMemo);
const
  clNum = $00996600; //$00ff6633;
  clTok = $00880000; //clNavy;
  clTyp = $00008800;
  clOpr = $00E08B45;
  clBrw = $00003060;
  clBrt = $00888800;
  clSys = $000000BB;
var
  i: integer;
begin
  fRichMemo :=AOwner;
  TokenBlock:=[tokComment,tokMulty];

  c_OBR_Types                 := TSCharHighlight.Create;
  c_OBR_Keyword               := TSCharHighlight.Create;
  c_OBR_Integer               := TSCharHighlight.Create;
  c_OBR_String                := TSCharHighlight.Create;
  c_OBR_Comment               := TSCharHighlight.Create;
  c_OBR_Bold                  := TSCharHighlight.Create;
  c_OBR_Brown                 := TSCharHighlight.Create;
  c_OBR_Operator              := TSCharHighlight.Create;
  c_OBR_Bracket               := TSCharHighlight.Create;
  c_OBR_System                := TSCharHighlight.Create;
  c_OBR_SysMod                := TSCharHighlight.Create;

  c_OBR_Keyword.FontColor     := clTok;
  c_OBR_Keyword.BackColor     := clNone;
  c_OBR_Keyword.FontStyle     := [fsBold];

  c_OBR_Types.FontColor       := clTyp;
  c_OBR_Types.BackColor       := clNone;
  c_OBR_Types.FontStyle       := [fsBold];

  c_OBR_Integer.FontColor     := clNum;
  c_OBR_Integer.BackColor     := clNone;
  c_OBR_Integer.FontStyle     := [fsBold];

  c_OBR_String.FontColor      := clMaroon;
  c_OBR_String.BackColor      := clNone;
  c_OBR_String.FontStyle      := [fsBold];

  c_OBR_Comment.FontColor     := clGray;
  c_OBR_Comment.BackColor     := clNone;
  c_OBR_Comment.FontStyle     := [fsItalic];

  c_OBR_Bold.FontColor        := clBlack;
  c_OBR_Bold.BackColor        := clNone;
  c_OBR_Bold.FontStyle        := [fsBold];

  c_OBR_Brown.FontColor       := clBrw;
  c_OBR_Brown.BackColor       := clNone;
  c_OBR_Brown.FontStyle       := [fsBold];

  c_OBR_Operator.FontColor    := clOpr;
  c_OBR_Operator.BackColor    := clNone;
  c_OBR_Operator.FontStyle    := [fsBold];

  c_OBR_Bracket.FontColor     := clBrt;
  c_OBR_Bracket.BackColor     := clNone;
  c_OBR_Bracket.FontStyle     := [fsBold];

  c_OBR_System.FontColor      := clSys;
  c_OBR_System.BackColor      := clNone;
  c_OBR_System.FontStyle      := [fsBold];

  c_OBR_SysMod.FontColor      := clBlack;
  c_OBR_SysMod.BackColor      := clNone;
  c_OBR_SysMod.FontStyle      := [fsBold];

  f_OBR_ListWord              := TStringList.Create;
  with f_OBR_ListWord do begin
    Sorted:=False;
      for i:=0 to length(TokenKey)-1      do AddObject(TokenKey[i],     TObject(tokKeyword));
      for i:=0 to length(TypesKey)-1      do AddObject(TypesKey[i],     TObject(tokTypes));
      for i:=0 to length(SystemKey)-1     do AddObject(SystemKey[i],    TObject(tokSystem));
      for i:=0 to length(SystemMod)-1     do AddObject(SystemMod[i],    TObject(tokSysmod));
    Sort;
    Sorted:=False;
  end;

end;

destructor TOBRHighliter.Destroy;
begin
  c_OBR_Integer.Free;
  c_OBR_Keyword.Free;
  c_OBR_Types.Free;
  c_OBR_Comment.Free;
  c_OBR_String.Free;
  f_OBR_ListWord.Free;
  c_OBR_Bold.Free;
  c_OBR_Brown.Free;
  c_OBR_Operator.Free;
  c_OBR_Bracket.Free;
  c_OBR_System.Free;
  c_OBR_SysMod.Free;
  inherited;
end;

procedure TOBRHighliter.GetHighlitedParam(RichMemo: TCustomSyntaxMemo);
begin
  inherited;
  SetMemoColor(RichMemo, tokKeyword,     c_OBR_Keyword);
  SetMemoColor(RichMemo, tokTypes,       c_OBR_Types);
  SetMemoColor(RichMemo, tokNumber,      c_OBR_Integer);
  SetMemoColor(RichMemo, tokString,      c_OBR_String);
  SetMemoColor(RichMemo, tokComment,     c_OBR_Comment);
  SetMemoColor(RichMemo, tokMulty,       c_OBR_Comment);
  SetMemoColor(RichMemo, tokBold,        c_OBR_Bold);
  SetMemoColor(RichMemo, tokBrown,       c_OBR_Brown);
  SetMemoColor(RichMemo, tokOperator,    c_OBR_Operator);
  SetMemoColor(RichMemo, tokBracket,     c_OBR_Bracket);
  SetMemoColor(RichMemo, tokSystem,      c_OBR_System);
  SetMemoColor(RichMemo, tokSysmod,      c_OBR_SysMod);
end;

procedure TOBRHighliter.Change(RichMemo: TCustomSyntaxMemo; ChangedItems: TChangedItems);
begin
  inherited;
end;

procedure TOBRHighliter.Set_OBR_Integer(const Value: TSCharHighlight);
begin
  c_OBR_Integer.Assign(Value);
end;

procedure TOBRHighliter.Set_OBR_Keyword(const Value: TSCharHighlight);
begin
  c_OBR_Keyword.Assign(Value);
end;

procedure TOBRHighliter.Set_OBR_Types(const Value: TSCharHighlight);
begin
  c_OBR_Types.Assign(Value);
end;

procedure TOBRHighliter.Set_OBR_String(const Value: TSCharHighlight);
begin
  c_OBR_String.Assign(Value);
end;

procedure TOBRHighliter.Set_OBR_Comment(const Value: TSCharHighlight);
begin
  c_OBR_Comment.Assign(Value);
end;

procedure TOBRHighliter.Set_OBR_Bold(const Value: TSCharHighlight);
begin
  c_OBR_Bold.Assign(Value);
end;

procedure TOBRHighliter.Set_OBR_Brown(const Value: TSCharHighlight);
begin
  c_OBR_Brown.Assign(Value);
end;

procedure TOBRHighliter.Set_OBR_Operator(const Value: TSCharHighlight);
begin
  c_OBR_Operator.Assign(Value);
end;

procedure TOBRHighliter.Set_OBR_Bracket(const Value: TSCharHighlight);
begin
  c_OBR_Bracket.Assign(Value);
end;

procedure TOBRHighliter.Set_OBR_System(const Value: TSCharHighlight);
begin
  c_OBR_System.Assign(Value);
end;

procedure TOBRHighliter.Set_OBR_SysMod(const Value: TSCharHighlight);
begin
  c_OBR_SysMod.Assign(Value);
end;

//#######################################################################

function TOBRHighliter.Parse(Sender: TSyntaxParser; Line: string; LineIndex: Integer; LastToken: TToken; PA: TSyntaxAttributes): TToken;
const
  IdentChars : Set of Char = ['_', '0'..'9', 'a'..'z', 'A'..'Z',''..'',''..'','',''];
var
  Pos       : integer;
  //Count,i,j : integer;
  Start,t   : integer;
  HexFlag   : boolean;
  Flag      : boolean;
  C         : char;
  S         : String;

  procedure ParseWord(Word: String; var Token: TToken);
  var
    n : integer;
  begin
    if f_OBR_ListWord.Find(Word, n) then begin
      if (Word=f_OBR_ListWord.Strings[n]) then
        Token:=TToken(f_OBR_ListWord.Objects[n])
      else
        Token:=tokError;
    end;
  end;

  procedure CommentProc(const F: boolean; Tok: TToken);
  begin
    Start  :=Pos;
    if not F then Inc(Pos);
    while Pos<=Length(Line) do
      case Line[Pos] of
        '''' :
                begin
                  Inc(Pos);
                  while (Pos<=Length(Line)) and not (Line[Pos] in ['''']) do Inc(Pos);
                  if Line[Pos] in [''''] then Inc(Pos);
                end;
        ']'  :
                begin
                  Inc(Pos);
                  Break;
                end;
      else
        Inc(Pos);
      end;
    Sender.AddToken(Start,Pos-Start,Tok);
  end;

begin

  Sender.Clear;
  Result := LastToken;
  if (Length(Line) = 0) or (PA = nil) then Exit;
  if Line[Length(Line)] >= ' ' then Line := Line + ' ';

  Pos:=1;
  Flag:=false;

  case Result of
      tokComment:
                  begin
                    Result := tokComment;
                    CommentProc(true,Result);
                  end;

      tokMulty  :
                  begin
                    Result := tokMulty;
                    Start:=Pos;
                    while Line[Pos] <> #0 do
                      case Line[Pos] of
                        '*':
                              if Line[Pos + 1] = ')' then begin
                                Inc(Pos, 2);
                                Break;
                                (*
                                //!!!->TODO Optimize
                                Count:=0;
                                i:=0;
                                while i<=LineIndex do begin
                                  j:=length(fRichMemo.Lines[i]);
                                  if i=LineIndex then j:=Pos-2;
                                  while j>0 do begin
                                    if (fRichMemo.Lines[i][j]='*') and (fRichMemo.Lines[i][j+1]=')') then DEC(Count);
                                    if (fRichMemo.Lines[i][j]='(') and (fRichMemo.Lines[i][j+1]='*') then INC(Count);
                                    DEC(j);
                                  end;
                                  INC(i);
                                end;

                                if Count<=0 then Break;
                                *)
                              end else
                                Inc(Pos);
                        #10,#13:  Break;
                      else
                        Inc(Pos);
                      end;
                    Sender.AddToken(Start,Pos-Start,Result);
                  end;

  end;

  while Pos<=Length(Line) do begin
    case Line[Pos] of
      #0..#9,
      #11, #12,
      #14..#32  :
                  begin
                    Start:=Pos;
                    Inc(Pos);
                    Result := tokBlank;
                    while (Pos<=Length(Line)) and (Line[Pos] in [#0..#9, #11, #12, #14..#32]) do Inc(Pos);
                    Sender.AddToken(Start,Pos-Start,Result);
                  end;
      #160        :
                  begin
                    Start:=Pos;
                    Inc(Pos);
                    Result := tokBlank;
                    while (Pos<=Length(Line)) and (Line[Pos] in [#160]) do Inc(Pos);
                    Sender.AddToken(Start,Pos-Start,Result);
                  end;
      '0'..'9'  :
                  begin
                    Start:=Pos;
                    Inc(Pos);
                    Result := tokNumber;
                    HexFlag:= False;
                    while Line[Pos] in IdentChars do begin
                      case Line[Pos] of
                        '.': if Line[Pos+1] = '.'                            then Break;
                        'H':
                              begin
                                HexFlag:=true;
                                if not (Line[Pos-1] in ['0'..'9','A'..'F'])  then begin
                                  Result := tokText;
                                  HexFlag:=false;
                                end;
                              end;
                        'X':  if (Line[Pos-1] in ['0'..'9','A'..'F'])        then Result := tokString;
                        'a'..'z','A'..'G','I'..'W','Y','Z',
                        ''..'',''..'','','':                           begin
                                                                                  HexFlag:=false;
                                                                                  Result := tokText;
                                                                             end;
                      end;
                      Inc(Pos);
                    end;
                    if HexFlag then Result := tokNumber;
                    Sender.AddToken(Start,Pos-Start,Result);
                  end;
      'A'..'Z',
      'a'..'z',
      ''..'',
      ''..'',
      '', '',
      '_'       :
                  begin
                    Start:=Pos;
                    Result := tokText;
                    Inc(Pos);
                    while Line[Pos] in IdentChars do Inc(Pos);
                    if Line[Pos] in ['*','-'] then begin
                      Result := tokBold;
                      if Line[Pos] in ['-'] then Result := tokBrown;
                      if Sender.Count>1 then begin
                        case Sender.Tokens[Sender.Count-1].stToken of
                          tokBlank:
                                    begin
                                      with Sender.Tokens[Sender.Count-2] do
                                        case stToken of
                                          tokKeyword : if Copy(Line, stStart + 1, stLength)<>'PROCEDURE' then Result := tokText;
                                          tokOperator: if not((Copy(Line, stStart + 1, stLength)=',') or (Copy(Line, stStart + 1, stLength)=';')) then Result := tokText;
                                        else
                                          Result := tokText;
                                        end;
                                    end;
                          tokOperator:
                                    begin
                                      with Sender.Tokens[Sender.Count-1] do
                                        if not((Copy(Line, stStart + 1, stLength)=',') or (Copy(Line, stStart + 1, stLength)=';')) then Result := tokText;
                                    end;
                        end;
                      end;
                      Sender.AddToken(Start,Pos-Start,Result);
                      Inc(Pos);
                      Result := tokOperator;
                      Sender.AddToken(Pos-1,1,Result);
                    end else begin
                      ParseWord(Copy(Line,Start,Pos-Start),Result);
                      Sender.AddToken(Start,Pos-Start,Result);
                    end;
                  end;
      ''''      :
                  begin
                    Start  :=Pos;
                    Result := tokString;
                    if (Line[Pos + 1] = '''') and (Line[Pos + 2] = '''') then Inc(Pos, 2);
                    repeat
                      case Line[Pos] of
                        #0, #10, #13: Break;
                        ''''        : Inc(Pos);
                      end;
                      Inc(Pos);
                    until (Line[Pos] = '''') and (Line[Pos+1]<>'''');
                    if Line[Pos] <> #0 then Inc(Pos);
                    Sender.AddToken(Start,Pos-Start,Result);
                  end;
      '"'      :
                  begin
                    Start  :=Pos;
                    Result := tokString;
                    repeat
                      case Line[Pos] of
                        #0, #10, #13: Break;
                      end;
                      Inc(Pos);
                    until (Line[Pos] = '"');
                    if Line[Pos] <> #0 then Inc(Pos);
                    Sender.AddToken(Start,Pos-Start,Result);
                  end;
      '['       :
                  begin
                    Start  :=Pos;
                    Result := tokOperator;
                    repeat
                      case Line[Pos] of
                        #0, #10, #13: Break;
                      end;
                      Inc(Pos);
                    until (Line[Pos]>#32);
                    t:=Pos;
                    S:='';
                    if Line[Pos] in ['''','"'] then begin
                      Flag:=true;
                      while Line[Pos]<>']' do Inc(Pos);
                    end else
                      while (Line[Pos] in IdentChars) do begin
                        S:=S+Line[Pos];
                        Inc(Pos);
                      end;
                    if (S='UNTRACED')
                    or (S='WINAPI')
                    or (S='NOTAG')
                    or (S='CDECL')
                    or (Flag) then begin
                      Result := tokBracket;
                      Flag:=false;
                      while not (Line[Pos] in [']',#0, #10, #13]) do Inc(Pos);
                      if Line[Pos]=']' then Inc(Pos);
                    end else
                      Pos:=t;
                    Sender.AddToken(Start,Pos-Start,Result);
                  end;
      '('       :
                  begin
                    Start  :=Pos;
                    Inc(Pos);
                    case Line[Pos] of
                      '*':
                            begin
                              Result := tokMulty;
                              Inc(Pos);
                              while Line[Pos] <> #0 do
                                case Line[Pos] of
                                  '*':
                                        if Line[Pos + 1] = ')' then begin
                                          Inc(Pos, 2);
                                          Break;
                                        end else
                                          Inc(Pos);
                                  #10,#13:  Break;
                                else
                                  Inc(Pos);
                                end;
                            end;
                    else
                      Result := tokOperator;
                    end;
                    Sender.AddToken(Start,Pos-Start,Result);
                  end;
      '<','>',
      '=',':':
                  begin
                    Start  :=Pos;
                    Result := tokOperator;
                    C:=Line[Pos];
                    Inc(Pos);
                    case C of
                      '<':  if Line[Pos] in ['=','>'] then Inc(Pos);
                      '>':  if Line[Pos] in ['='] then Inc(Pos);
                      ':':  if Line[Pos] in ['='] then Inc(Pos);
                      '=':  if Line[Pos] in ['<','>','='] then begin
                              Result:=TokError;
                              Inc(Pos);
                            end;
                    end;
                    Sender.AddToken(Start,Pos-Start,Result);
                  end;

      '.',',','+',
      '-','*','@',
      '^','/',')',
      ';','{','}',
      '|','#','~',
      '&','!',']':
                  begin
                    Result := tokOperator;
                    Inc(Pos);
                    Sender.AddToken(Pos-1,1,Result);
                  end;
    else
      Result := tokBlank;
      Inc(Pos);
      Sender.AddToken(Pos-1,1,Result);
    end;
  end;

end;

end.
