UNIT OBRScaner;

INTERFACE

USES
  Classes, SysUtils, Dialogs;

TYPE
  TToken= (
        _null,      _times,     _slash,   _div,     _mod,
        _and,       _plus,      _minus,   _or,      _eql,
        _neq,       _lss,       _leq,     _gtr,     _geq,
        _in,        _is,        _arrow,   _period,  _comma,
        _colon,     _upto,      _rparen,  _rbrak,   _rbrace,
        _of,        _then,      _do,      _to,      _by,
        _lparen,    _lbrak,     _lbrace,  _not,     _becomes,
        _number,    _nil,       _true,    _false,   _string,
        _ident,     _semicolon, _bar,     _end,     _else,
        _elsif,     _until,     _if,      _case,    _while,
        _repeat,    _for,       _loop,    _with,    _exit,
        _return,    _array,     _object,  _record,  _pointer,
        _begin,     _const,     _type,    _var,     _procedure,
        _import,    _module,    _code,    _finally, _eof,

        _c,         _winapi,    _notag,   _untraced, 

        ScUndef,ScIdent,ScSemicolon,ScAt,ScNumber,ScComma,ScLBrak,ScRBrak,
        ScCR,ScMult,ScEnd,ScLabel,ScColon,ScPlus,ScMinus,ScDiv,ScLPar,
        ScRPar,ScString,ScPoint,ScLBrace,ScRBrace
  );

  (*symbols:
	    |  0          1          2          3          4
	 ---|--------------------------------------------------------
	  0 |  null       *          /          DIV        MOD
	  5 |  &          +          -          OR         =
	 10 |  #          <          <=         >          >=
	 15 |  IN         IS         ^          .          ,
	 20 |  :          ..         )          ]          }
	 25 |  OF         THEN       DO         TO         BY
	 30 |  (          [          {          ~          :=
	 35 |  number     NIL        TRUE       FALSE      string
	 40 |  ident      ;          |          END        ELSE
	 45 |  ELSIF      UNTIL      IF         CASE       WHILE
	 50 |  REPEAT     FOR        LOOP       WITH       EXIT
	 55 |  RETURN     ARRAY      OBJECT     RECORD     POINTER
	 60 |  BEGIN      CONST      TYPE       VAR        PROCEDURE
	 65 |  IMPORT     MODULE     CODE       eof        **
	 *)

  TNumTyp = (_char,_integer,_real,_longreal);

CONST
  _CEOF   = #26;
  _EOT    = #0;
  MaxStrLen = 256;

  MaxHDig      = 8;  (* maximal hexadecimal longint length *)
  MaxRExp      = 38;  (* maximal real exponent *)
  MaxLExp      = 308;  (* maximal DOUBLE exponent *)

  MaxInt        = $7FFF;
  MaxLInt       = $7FFFFFFF;

TYPE
  TScaner  = CLASS
    ch,OldCh    : CHAR;
    str         : String;
    ScanToken   : TToken;
    intval      : integer;
    realval     : SINGLE;
    lrlval      : DOUBLE;
    numtyp      : TNumTyp;
    FileStream  : TMemoryStream;
    Tok         : TToken;
    PROCEDURE   Number;
    PROCEDURE   GetStr(VAR Tok: TToken);
    PROCEDURE   Identifier(VAR Tok: TToken);
  PUBLIC
    Name        : String;
    constructor Create(Stream: TMemoryStream);
    destructor  Destroy; override;
    PROCEDURE   GetChar(VAR ch: CHAR);
    PROCEDURE   Get(VAR Tok: TToken);
    PROCEDURE   CheckSym(VAR Tok: TToken; s: TToken);
    PROCEDURE   CheckOperator(VAR id: SHORTINT);
  END;

IMPLEMENTATION

PROCEDURE TScaner.CheckOperator(VAR id: SHORTINT);
VAR
  ch0, ch1, ch2: CHAR;
BEGIN
  id:=126;
  name:=str;
  ch0:=str[1];
  ch1:=str[2];
  ch2:=str[3];
  CASE ch0 OF
    '=', '#', '&': IF ch1=#0 THEN EXIT;
    '<', '>': IF (ch1=#0) OR ((ch1='=') AND (ch2=#0)) THEN EXIT;
    'I': IF (ch1='N') AND (ch2=#0) THEN EXIT;
    'D': IF str='DIV' THEN EXIT;
    'M': IF str='MOD' THEN EXIT;
    'O': IF str='OR' THEN EXIT;
    '+', '-', '*', '/' : IF (ch1=#0) OR ((ch2=#0) AND (ch1='*')) THEN EXIT;
    '~': IF (ch1=#0) THEN EXIT;
    ':': IF str=':=' THEN EXIT;
  ELSE
  END;
  id:=0
END;

CONSTRUCTOR TScaner.Create(Stream: TMemoryStream);
BEGIN
  inherited Create;
  ch      := ' ';
  OldCh   := ' ';
  Tok     := _null;
  FileStream  := TMemoryStream.Create;
  FileStream.LoadFromStream(Stream);
END;

DESTRUCTOR TScaner.Destroy;
BEGIN
  FileStream.Free;
  inherited Destroy;
END;

PROCEDURE TScaner.GetChar(VAR ch: CHAR);
BEGIN
  OldCh:=ch;
  IF FileStream.Position<FileStream.Size THEN BEGIN
    FileStream.Read(ch,1);
  END ELSE BEGIN
    name:=ch;
    ch:=_CEOF;
  END;
END;

PROCEDURE TScaner.Number;
VAR
  i, m, n, d, e: INTEGER;
  dig: ARRAY [0..24] OF CHAR;
  f: DOUBLE;
  expCh: CHAR;
  neg: BOOLEAN;

  FUNCTION Ten(e: INTEGER): DOUBLE;
  VAR
    x, p: DOUBLE;
  BEGIN
    x := 1;
    p := 10;
    WHILE e > 0 DO BEGIN
      IF ODD(e) THEN x := x*p;
      e := e DIV 2;
      IF e > 0 THEN p := p*p;
    END;
    RESULT:= x
  END;

  FUNCTION FOrd(ch: CHAR; hex: BOOLEAN): INTEGER;
  BEGIN
    IF ch <= '9' THEN RESULT:= ORD(ch) - ORD('0')
    ELSE IF hex THEN RESULT:= ORD(ch) - ORD('A') + 10
    ELSE BEGIN
      //err(2);
      RESULT:= 0
    END
  END;

BEGIN
  i := 0;
  m := 0;
  n := 0;
  d := 0;

  REPEAT
    IF ('0' <= ch) AND (ch <= '9') OR (d = 0) AND ('A' <= ch) AND (ch <= 'F') THEN BEGIN
      IF (m > 0) OR (ch <> '0') THEN BEGIN
        IF n < LENGTH(dig) THEN BEGIN
          dig[n] := ch;
          INC(n)
        END;
        INC(m);
      END;
      GetChar(ch);
      INC(i)
    END ELSE IF ch = '.' THEN BEGIN
      GetChar(ch);
      IF ch = '.' THEN BEGIN
        ch := #$7F;
        BREAK;
      END ELSE IF d = 0 THEN
        d := i
      ELSE
        ;//err(2);
    END ELSE
      BREAK;
  UNTIL FALSE;

  IF d = 0 THEN BEGIN
    IF n = m THEN BEGIN
      intval := 0;
      i := 0;
      IF ch = 'X' THEN BEGIN
        GetChar(ch);
        numtyp := _char;
        IF n <= 2 THEN BEGIN
          WHILE i < n DO BEGIN
            intval := intval*$10 + FOrd(dig[i], TRUE);
            INC(i)
          END;
        END ELSE
          ;//err(203);
      END ELSE IF ch = 'H' THEN BEGIN
        GetChar(ch);
        numtyp := _integer;
        IF n <= MaxHDig THEN BEGIN
          IF (n = MaxHDig) AND (dig[0] > '7') THEN intval := -1;
          WHILE i < n DO BEGIN
            intval := intval*$10 + FOrd(dig[i], TRUE);
            INC(i)
          END;
        END ELSE
          ;//err(203);
      END ELSE BEGIN
        numtyp := _integer;
        WHILE i < n DO BEGIN
          d := FOrd(dig[i], FALSE);
          INC(i);
          IF intval <= MaxLInt - d DIV 10 THEN
            intval := intval*10 + d
          ELSE
            ;//err(203);
        END;
      END;
    END ELSE
      ;//err(203);
  END ELSE BEGIN
    f := 0;
    e := 0;
    expCh := 'E';
    WHILE n > 0 DO BEGIN
      DEC(n);
      f := (FOrd(dig[n], FALSE) + f)/10;
    END;
    IF (ch = 'E') OR (ch = 'D') THEN BEGIN
      expCh := ch;
      GetChar(ch);
      neg := FALSE;
      IF ch = '-' THEN BEGIN
        neg := TRUE;
        GetChar(ch);
      END ELSE IF ch = '+' THEN
        GetChar(ch);
      IF ('0' <= ch) AND (ch <= '9') THEN BEGIN
        REPEAT
          n := FOrd(ch, FALSE);
          GetChar(ch);
          IF e <= MaxInt - n DIV 10 THEN
            e := e*10 + n
          ELSE
            ;//err(203);
        UNTIL (ch < '0') OR ('9' < ch);
        IF neg THEN e := -e;
      END ELSE
        ;//err(2);
    END;
    DEC(e, i-d-m);
    IF expCh = 'E' THEN BEGIN
      numtyp := _real;
      IF (1-MaxRExp < e) AND (e <= MaxRExp) THEN BEGIN
        IF e < 0 THEN
          realval := f / Ten(-e)
        ELSE
          realval := f * Ten(e);
      END ELSE
        ;//err(203);
    END ELSE BEGIN
      numtyp := _longreal;
      IF (1-MaxLExp < e) AND (e <= MaxLExp) THEN BEGIN
        IF e < 0 THEN
          lrlval := f / Ten(-e)
        ELSE
          lrlval := f * Ten(e);
      END ELSE
        ;//err(203);
    END;
  END;
END;

PROCEDURE TScaner.GetStr(VAR Tok: TToken);
VAR
  i: INTEGER;
  och: CHAR;
BEGIN
  i := 0;
  och := ch;
  str:='';
  REPEAT
    GetChar(ch);
    IF ch = och THEN BREAK;
    IF ch < ' ' THEN BEGIN
      //err(345);
      BREAK;
    END ;
    IF i = MaxStrLen-1 THEN BEGIN
      //err(241);
      BREAK;
    END ;
    str := str+ch;
    inc(i);
  UNTIL False;
  GetChar(ch);
  intval := i + 1;
  IF intval = 2 THEN BEGIN
    Tok := _number;
    numtyp := _char;
    intval := ORD(str[1]);
  END ELSE
    Tok := _string;
END;

PROCEDURE TScaner.Identifier(VAR Tok: TToken);
BEGIN
  name:='';
  REPEAT
    name := name+ch;
    GetChar(ch);
  UNTIL (ch < '0') OR ('9' < ch) AND (UPCASE(ch) < 'A') OR ('Z' < UPCASE(ch)) AND (ch <> '_');
  IF Length(name)>=MaxStrLen THEN ;//err(240);
  Tok := _ident;
END;

PROCEDURE TScaner.Get(VAR Tok: TToken);
var
  s: TToken;

  	PROCEDURE Comment;	(* do not read after END of file *)
		var
      dump: BOOLEAN;
		BEGIN
      GetChar(ch);
      dump:=False;
			if ch = '@' then dump := TRUE;
			repeat
				repeat
					while ch = '(' do BEGIN
            GetChar(ch);
						if ch = '*' then
              Comment
            else if dump then ;//OPM.LogW ("(") END
					END;
					if ch = '*' then BEGIN
            GetChar(ch);
            BREAK;
          END;
					if ch = _CEOF then EXIT;
					if dump then ;//OPM.LogW (ch);
					GetChar(ch)
				until False;
				if ch = ')' then BEGIN
          GetChar(ch);
          BREAK;
        END;
				if dump then ;//OPM.LogW ('*');
				if ch = _CEOF then BEGIN
          //err(5);
          BREAK;
        END;
			until False;
		END;

BEGIN

  WHILE ch <= ' ' DO BEGIN
    IF ch=_CEOF then BEGIN
      Tok:=_eof;
      ScanToken:=_eof;
      Exit;
    END ELSE
      GetChar(ch);
  END;

  case ch of
      #$22, #$27  : GetStr(s);
		  '#' :
            BEGIN
              s := _neq;
              name:='#';
              GetChar(ch);
            END;
			'&'  :
            BEGIN
              s :=  _and;
              name:='&';
              GetChar(ch);
            END;
			'('  :
            BEGIN
              GetChar(ch);
              IF ch = '*' then BEGIN
                Comment;
                Get(Tok);
                Exit;
              END else
                s := _lparen;
            END;
			')'  :
            BEGIN
              s := _rparen;
              GetChar(ch);
            END;
			'*'  :
            BEGIN
              name:='*';
              s := _times;
              GetChar(ch);
            END;
			'+'  :
            BEGIN
              name:='+';
              s := _plus;
              GetChar(ch);
            END;
			','  :
            BEGIN
              s := _comma;
              GetChar(ch);
            END;
			'-'  :
            BEGIN
              name:='-';
              s := _minus;
              GetChar(ch);
            END;
			'.'  :
            BEGIN
              GetChar(ch);
              if ch = '.' then BEGIN
                GetChar(ch);
                s := _upto;
              END else
                s := _period;
            END;
			'/'  :
            BEGIN
              name:='/';
              GetChar(ch);
              s :=  _slash;
            END;
			'0'..'9':
            BEGIN
              Number;
              s := _number;
            END;
			':'  :
            BEGIN
              GetChar(ch);
              name := ':';
							if ch = '=' then BEGIN
                GetChar(ch);
                s := _becomes;
                name:=':=';
              END else
                s := _colon;
            END;
			';'  :
            BEGIN
              s := _semicolon;
              GetChar(ch);
            END;
			'<'  :
            BEGIN
              GetChar(ch);
              if ch = '=' then BEGIN
                GetChar(ch);
                s := _leq;
                name:='<=';
              END else BEGIN
                s := _lss;
                name:='<';
              END;
            END;
			'='  :
            BEGIN
              s := _eql;
              name:='=';
              GetChar(ch);
            END;
			'>'  :
            BEGIN
              GetChar(ch);
							if ch = '=' then BEGIN
                GetChar(ch);
                s := _geq;
                name:='>=';
              END else BEGIN
                s := _gtr;
                name:='>';
              END;
            END;
			'A':
            BEGIN
              Identifier(s);
              IF      name = 'ARRAY'  THEN s := _array;
            END;
			'B':
            BEGIN
              Identifier(s);
						  IF      name = 'BEGIN'  THEN s := _BEGIN
						  ELSE IF name = 'BY'     THEN s := _by;
            END;
			'C':
            BEGIN
              Identifier(s);
						  IF      name = 'CASE'   THEN s := _case
						  ELSE IF name = 'CODE'   THEN s := _code
              ELSE IF name = 'C'      THEN s := _c
						  ELSE IF name = 'CONST'  THEN s := _const;
            END;
			'D':
            BEGIN
              Identifier(s);
						  IF      name = 'DO'     THEN s := _do
						  ELSE IF name = 'DIV'    THEN s := _div;
						END;
			'E':
            BEGIN
              Identifier(s);
						  IF      name = 'END'    THEN s := _end
						  ELSE IF name = 'ELSE'   THEN s := _else
						  ELSE IF name = 'ELSIF'  THEN s := _elsif
						  ELSE IF name = 'EXIT'   THEN s := _exit
						END;
			'F':
            BEGIN
              Identifier(s);
						  IF      name = 'FALSE'  THEN s := _false
              ELSE IF name = 'FINALLY' THEN s := _finally
						  ELSE IF name = 'FOR'    THEN s := _for;
						END;
			'I':
            BEGIN
              Identifier(s);
						  IF      name = 'IF'     THEN s := _if
						  ELSE IF name = 'IN'     THEN s := _in
						  ELSE IF name = 'IS'     THEN s := _is
						  ELSE IF name = 'IMPORT' THEN s := _import
						END;
			'L':
            BEGIN
              Identifier(s);
              IF      name = 'LOOP' THEN s := _loop;
            END;
			'M':
            BEGIN
              Identifier(s);
						  IF      name = 'MOD'    THEN s := _mod
						  ELSE IF name = 'MODULE' THEN s := _module
						END;
			'N':
            BEGIN
              Identifier(s);
              IF      name = 'NIL'    THEN s := _nil
              ELSE IF name = 'NOTAG'  THEN s := _notag
            END;
			'O':
            BEGIN
              Identifier(s);
						  IF      name = 'OR'     THEN s := _or
						  ELSE IF name = 'OF'     THEN s := _of
						  ELSE IF name = 'OBJECT' THEN s := _object
						END;
			'P':
            BEGIN
              Identifier(s);
						  IF      name = 'PROCEDURE'  THEN s := _procedure
						  ELSE IF name = 'POINTER'    THEN s := _pointer
						END;
			'R':
            BEGIN
              Identifier(s);
						  IF      name = 'RECORD' THEN s := _record
						  ELSE IF name = 'REPEAT' THEN s := _repeat
						  ELSE IF name = 'RETURN' THEN s := _return
						END;
			'T':
            BEGIN
              Identifier(s);
						  IF      name = 'THEN'   THEN s := _then
						  ELSE IF name = 'TO'     THEN s := _to
						  ELSE IF name = 'TRUE'   THEN s := _true
						  ELSE IF name = 'TYPE'   THEN s := _type
						END;
			'U':
            BEGIN
              Identifier(s);
              IF      name = 'UNTIL'    THEN s := _until
              ELSE IF name = 'UNTRACED' THEN s := _untraced
            END;
			'V':
            BEGIN
              Identifier(s);
              IF      name = 'VAR'    THEN s := _var
            END;
			'W':
            BEGIN
              Identifier(s);
						  IF      name = 'WHILE'  THEN s := _while
						  ELSE IF name = 'WITH'   THEN s := _with
              ELSE IF name = 'WINAPI' THEN s := _winapi
						END;
			'G'..'H', 'J', 'K', 'Q', 'S', 'X'..'Z': Identifier(s);
      'a'..'z': Identifier(s);
			'[' : BEGIN s := _lbrak;  GetChar(ch); END;
			']' : BEGIN s := _rbrak;  GetChar(ch); END;
			'^' : BEGIN s := _arrow;  GetChar(ch); END;
			'{' : BEGIN s := _lbrace; GetChar(ch); END;
			'|' : BEGIN s := _bar;    GetChar(ch); END;
			'}' : BEGIN s := _rbrace; GetChar(ch); END;
			'~' : BEGIN s := _not;    name:='~'; GetChar(ch); END;
			#$7F: BEGIN s := _upto;   GetChar(ch); END;
  else
    s :=  _null;
    GetChar(ch);
  END;
  Tok := s;
  ScanToken:=Tok;
END;

PROCEDURE TScaner.CheckSym(VAR Tok: TToken; s: TToken);
BEGIN
  IF ScanToken=s THEN Get(Tok) ELSE ;//err(ORD(s));
END;

END.
